/* Copyright (c) 2019 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes

#include "AssignmentPage.h"

#include "shared/SharedObjects.h"
#include "shared/TournamentSettings.h"
#include "shared/TableDelegate.h"
#include "shared/SearchWidget.h"

// Qt includes
#include <QGroupBox>
#include <QVBoxLayout>
#include <QHBoxLayout>
#include <QPushButton>
#include <QApplication>
#include <QStyle>
#include <QTableWidget>
#include <QHeaderView>
#include <QLabel>
#include <QDebug>

AssignmentPage::AssignmentPage(QWidget *parent, SharedObjects *sharedObjects)
    : QWidget(parent),
      m_tournamentSettings(sharedObjects->tournamentSettings()),
      m_searchEngine(sharedObjects->searchEngine())
{
    QVBoxLayout *mainLayout = new QVBoxLayout(this);
    QGroupBox *box = new QGroupBox;
    box->setProperty("notitle", true);
    QHBoxLayout *layout = new QHBoxLayout(box);
    mainLayout->addWidget(box);

    QVBoxLayout *displayLayout = new QVBoxLayout;
    layout->addLayout(displayLayout);

    m_noAssignment = new QLabel(tr("<i>Es wurde noch keine Auslosung eingegeben</i>"));
    m_noAssignment->setAlignment(Qt::AlignCenter);
    displayLayout->addWidget(m_noAssignment);

    m_isSearch = new QLabel;
    m_isSearch->hide();
    displayLayout->addWidget(m_isSearch);

    m_assignment = new QTableWidget;
    m_assignment->setEditTriggers(QAbstractItemView::NoEditTriggers);
    m_assignment->setItemDelegate(new TableDelegate(this, TableDelegate::HighlightMode::EvenRows));
    m_assignment->verticalHeader()->hide();
    m_assignment->setShowGrid(false);
    m_assignment->setWordWrap(false);
    m_assignment->setSelectionMode(QTableWidget::NoSelection);
    m_assignment->setFocusPolicy(Qt::NoFocus);
    m_assignment->setColumnCount(3);
    m_assignment->setHorizontalHeaderLabels({ tr("Tisch"), QString(), tr("Namen") });
    displayLayout->addWidget(m_assignment);

    m_searchWidget = new SearchWidget;
    connect(m_searchWidget, &SearchWidget::searchChanged, this, &AssignmentPage::updateSearch);
    displayLayout->addWidget(m_searchWidget);

    QVBoxLayout *closeLayout = new QVBoxLayout;
    layout->addLayout(closeLayout);

    QPushButton *closeButton = new QPushButton(tr("Schließen"));
    closeButton->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);
    closeButton->setIcon(QApplication::style()->standardIcon(QStyle::SP_DialogCloseButton));
    connect(closeButton, &QPushButton::clicked, this, &AssignmentPage::closePage);
    closeLayout->addWidget(closeButton);
    closeLayout->addStretch();
}

void AssignmentPage::closePage()
{
    emit pageClosed();
    parent()->deleteLater();
}

void AssignmentPage::updateAssignment(const QHash<int, QHash<int, QString>> &assignment,
                                      const QHash<int, QHash<int, bool>> &pairComplete,
                                      const QString &namesSeparator, int lastTable)
{
    const int rows = assignment.count();
    if (! m_tournamentSettings->isPairMode()) {
        m_searchEngine->clearSearchCache(SearchEngine::SearchType::DrawnPairsPreAssignment);
    }

    if (rows == 0) {
        m_noAssignment->show();
        m_assignment->hide();
        m_searchWidget->clear();
        m_searchWidget->hide();
        return;
    } else {
        m_noAssignment->hide();
        m_assignment->show();
        m_assignment->clearSpans();
        m_assignment->setRowCount(lastTable * 2);
        m_searchWidget->show();
    }

    m_excludeFromSearch.clear();
    int row = 0;
    for (int table = 1; table <= lastTable; table++) {
        QTableWidgetItem *item = new QTableWidgetItem(QString::number(table));
        if (! assignment.contains(table)) {
            item->setForeground(Qt::red);
            m_assignment->setItem(row, 0, item);
            m_assignment->setSpan(row, 0, 2, 1);
            item = new QTableWidgetItem(tr("Noch keine Paare zugelost"));
            item->setForeground(Qt::red);
            m_assignment->setItem(row, 2, item);
            m_assignment->setSpan(row, 2, 2, 1);
            m_excludeFromSearch.insert(row, true);
            m_excludeFromSearch.insert(row + 1, true);
            row += 2;

        } else {
            m_assignment->setItem(row, 0, item);
            m_assignment->setSpan(row, 0, 2, 1);

            item = new QTableWidgetItem(tr("Paar 1:"));
            m_assignment->setItem(row,     1, item);
            item = new QTableWidgetItem(tr("Paar 2:"));
            m_assignment->setItem(row + 1, 1, item);

            for (int pair = 1; pair <= 2; pair++) {
                if (assignment.value(table).contains(pair)) {
                    m_excludeFromSearch.insert(row, false);
                    item = new QTableWidgetItem;
                    QString displayText = assignment.value(table).value(pair);
                    if (! m_tournamentSettings->isPairMode()) {
                        m_searchEngine->updateSearchCache(displayText,
                            SearchEngine::SearchType::DrawnPairsPreAssignment);

                        if (! pairComplete.value(table).value(pair)) {
                            item->setData(EntryData::ActualText, displayText);
                            displayText += namesSeparator
                                        + tr("(noch kein zweiter Spieler zugelost)");
                            item->setForeground(Qt::red);
                        }
                    }
                    item->setText(displayText);
                } else {
                    m_excludeFromSearch.insert(row, true);
                    item = new QTableWidgetItem(tr("Fehlt noch"));
                    item->setForeground(Qt::red);
                }
                m_assignment->setItem(row, 2, item);
                row++;
            }
        }
    }
    m_assignment->resizeColumnsToContents();
}

void AssignmentPage::updateSearch(const QString &searchText)
{
    if (searchText.isEmpty()) {
        for (int row = 0; row < m_assignment->rowCount(); row++) {
            m_assignment->showRow(row);
        }
        m_assignment->resizeColumnsToContents();
        m_isSearch->hide();
        return;
    }

    m_searchEngine->setSearchTerm(searchText);
    const SearchEngine::SearchType searchType
        = m_tournamentSettings->isPairMode() ? SearchEngine::SearchType::Default
                                             : SearchEngine::SearchType::DrawnPairsPreAssignment;
    int matches = 0;
    for (int row = 0; row < m_assignment->rowCount(); row += 2) {
        const bool rowMatches =    checkRowMatch(row,     searchType)
                                || checkRowMatch(row + 1, searchType);
        m_assignment->setRowHidden(row,     ! rowMatches);
        m_assignment->setRowHidden(row + 1, ! rowMatches);
        matches += rowMatches;
    }

    m_searchWidget->updateMatches(matches);
    m_assignment->resizeColumnsToContents();
    m_isSearch->setText(tr("Suchergebnis für „%1“").arg(searchText));
    m_isSearch->show();
}

bool AssignmentPage::checkRowMatch(int row, SearchEngine::SearchType searchType) const
{
    if (m_excludeFromSearch.value(row)) {
        return false;
    }

    QTableWidgetItem *item = m_assignment->item(row, 2);
    if (item->foreground() != Qt::red) {
        return m_searchEngine->checkMatch(item->text(), searchType);
    } else {
        return m_searchEngine->checkMatch(item->data(EntryData::ActualText).toString(), searchType);
    }
}

void AssignmentPage::resetSearch()
{
    m_searchWidget->clear();
}
