/* Copyright (c) 2018-2020 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes

#include "Client.h"
#include "RankingDialog.h"

#include "PlayersPage/PlayersPage.h"
#include "PlayersPage/MarkersWidget.h"

#include "shared/SharedObjects.h"
#include "shared/Database.h"
#include "shared/TournamentSettings.h"
#include "shared/StringTools.h"
#include "shared/JsonHelper.h"

// Qt includes
#include <QDebug>
#include <QHostAddress>
#include <QTcpSocket>
#include <QTimer>
#include <QJsonObject>
#include <QJsonArray>
#include <QColor>
#include <QApplication>
#include <QLocale>

using namespace ServerProtocol;

Client::Client(QObject *parent, int connectTimeout, int requestTimeout,
               SharedObjects *sharedObjects, PlayersPage *playersPage)
    : AbstractNetworkEngine(parent),
      m_connectTimeout(connectTimeout),
      m_db(sharedObjects->database()),
      m_tournamentSettings(sharedObjects->tournamentSettings()),
      m_stringTools(sharedObjects->stringTools()),
      m_playersPage(playersPage),
      m_markersWidget(playersPage->markersWidget()),
      m_requestHandlers {
          // Registration changes

          { ACTION_PLAYERS_REGISTER,           &Client::processRegisterPlayers },
          { ACTION_PLAYERS_RENAME,             &Client::processRenamePlayers },
          { ACTION_PLAYERS_ASSEMBLE_PAIR,      &Client::processAssemblePair },
          { ACTION_PLAYERS_DELETE,             &Client::processDeletePlayers },
          { ACTION_PLAYERS_UPDATE_MARKER,      &Client::processUpdatePlayersMarker },

          { ACTION_PLAYERS_DELETE_MARKED,      &Client::processDeleteMarkedPlayers },
          { ACTION_PLAYERS_DELETE_UNMARKED,    &Client::processDeleteUnmarkedPlayers },
          { ACTION_PLAYERS_DELETE_ALL,         &Client::processDeleteAllPlayers },
          { ACTION_PLAYERS_LIST_UPDATE_MARKER, &Client::processUpdateListMarker },
          { ACTION_PLAYERS_LIST_REMOVE_MARKER, &Client::processRemoveListMarker },

          // Preassignment changes
          { ACTION_UPDATE_PREASSIGNMENT,       &Client::processUpdatePreAssignment },
          { ACTION_DELETE_ALL_PREASSIGNMENTS,  &Client::processDeleteAllPreAssignments },

          // Markers definition changes
          { ACTION_MARKERS_ADD,    &Client::processAddMarker },
          { ACTION_MARKERS_DELETE, &Client::processDeleteMarker },
          { ACTION_MARKERS_MOVE,   &Client::processMoveMarker },
          { ACTION_MARKERS_EDIT,   &Client::processEditMarker },

          // Special markers changes
          { ACTION_SET_ABSENT_MARKER,         &Client::processSetAbsentMarker },
          { ACTION_CHANGE_SINGLES_MANAGEMENT, &Client::processChangeSinglesManagement },

          // Finishing the registration
          { ACTION_FINISH_REGISTRATION, &Client::processFinishRegistration },

          // Ranking comparison
          { ACTION_RANKING_REQUEST, &Client::processRankingRequest }
      }
{
    connect(socket(), QOverload<QAbstractSocket::SocketError>::of(&QAbstractSocket::error),
            this, &Client::error);
    connect(socket(), &QAbstractSocket::connected, this, &Client::socketConnected);

    m_connectionTimer = new QTimer(this);
    m_connectionTimer->setInterval(connectTimeout);
    m_connectionTimer->setSingleShot(true);
    connect(m_connectionTimer, &QTimer::timeout, this, &Client::connectionTimeout);

    m_requestTimer = new QTimer(this);
    m_requestTimer->setInterval(requestTimeout);
    m_requestTimer->setSingleShot(true);
    connect(m_requestTimer, &QTimer::timeout, this, &Client::requestTimeout);
}

void Client::connectToServer(const QHostAddress &address, quint16 port)
{
    emit message(tr("Verbinde mit Server %1").arg(address.toString()));
    m_connectionPending = true;
    m_connectionTimer->start();
    emit message(tr("Stelle Socketverbindung her …"));
    socket()->connectToHost(address, port);
}

QString Client::serverIp() const
{
    return socket()->peerAddress().toString();
}

int Client::serverPort() const
{
    return socket()->peerPort();
}

void Client::socketConnected()
{
    emit message(tr("Socketverbindung hergestellt, initialisiere Handshake …"));
    m_connectionPending = false;
    m_handshakePending = true;
}

void Client::abortHandshake()
{
    if (! m_handshakePending) {
        return;
    }

    closeConnection();
    m_connectionTimer->stop();
    m_handshakePending = false;
    m_stringTools->resetLocale();
}

void Client::error(QAbstractSocket::SocketError)
{
    closeConnection();

    if (m_connectionPending || m_handshakePending) {
        m_connectionTimer->stop();
        m_handshakePending = false;
        m_connectionPending = false;
        emit connectResult(false, socket()->errorString());
        return;
    }

    emit connectionError(socket()->errorString());
}

void Client::connectionTimeout()
{
    if (m_connectionPending) {
        emit connectionError(tr("Konnte keine Socketverbindung herstellen."));
        return;
    }

    closeConnection();
    QLocale locale;
    emit message(tr("Timeout: Keine Antwort nach %1 Sekunden").arg(
                 locale.toString(float(m_connectTimeout) / 1000.0)));
    emit connectResult(false, tr("Der Server hat zu lange keine Antwort geschickt."));
}

void Client::closeConnection()
{
    socket()->abort();
    m_connected = false;
}

void Client::sendRequest(QJsonObject &jsonObject, ProcessingTarget target,
                         const QString &request, const QString &scrollToTarget)
{
    emit message(tr("Frage %1 an").arg(request));
    emit statusUpdate(tr("Frage %1 beim Server an …").arg(request));

    switch (target) {
    case ProcessingTarget::Players:
        m_playersPage->setProcessingRequest(true, scrollToTarget);
        break;
    case ProcessingTarget::Markers:
        m_markersWidget->setProcessingRequest(true);
        break;
    case ProcessingTarget::Ranking:
        break;
    }

    jsonObject.insert(KEY_CHECKSUM,  m_playersPage->checksum());
    jsonObject.insert(KEY_CLIENT_ID, m_id);
    sendData(jsonObject);

    m_processingTarget.append(target);
    m_requestTimer->start();
}

int Client::id() const
{
    return m_id;
}

void Client::requestTimeout()
{
    m_playersPage->setProcessingRequest(false);
    m_markersWidget->setProcessingRequest(false);
    emit connectionError(tr("Der Server hat zu lange nicht auf eine Anfrage reagiert. Bitte die "
                            "Netzwerkverbindung überprüfen und neu verbinden."));
    closeConnection();
}

bool Client::connected() const
{
    return m_connected;
}

void Client::stopProcessingRequest()
{
    m_requestTimer->stop();
    switch (m_processingTarget.takeFirst()) {
    case Players:
        m_playersPage->setProcessingRequest(false);
        break;
    case Markers:
        m_markersWidget->setProcessingRequest(false);
        break;
    case Ranking:
        break;
    }
}

void Client::verifyChecksum(const QString &checksum)
{
    if (checksum != m_playersPage->checksum()) {
        // This should not happen. But who knows?!
        emit connectionError(tr("Die Synchronisation mit dem Server ist fehlgeschlagen, die Daten "
                                "weichen ab. Die Verbindung wird beendet. Bitte neu verbinden, um "
                                "den Datenbestand wieder mit dem Server abzugleichen!"));
    }
}

// ====================
// Handshake processing
// ====================

void Client::processHandshake(const QJsonObject &jsonObject)
{
    const QJsonValue phaseValue = jsonObject.value(HANDSHAKE_KEY_PHASE);
    if (phaseValue.type() != QJsonValue::String) {
        emit connectResult(false, tr("Der Server hat eine ungültige Begrüßung geschickt"));
        return;
    }

    const QString handshakePhase = phaseValue.toString();

    // Handshake Initialization
    // ========================

    if (handshakePhase == HANDSHAKE_PHASE_INITIALIZATION) {

        const QJsonValue serverSignatureValue = jsonObject.value(HANDSHAKE_KEY_SERVER_SIGNATURE);
        if (serverSignatureValue.type() != QJsonValue::String
            || serverSignatureValue.toString() != SERVER_SIGNATURE) {

            emit message(tr("Der Server hat eine ungültige Begrüßung geschickt."));
            emit connectResult(false, tr("Der Server hat eine ungültige Begrüßung geschickt."));
            return;
        }

        const QJsonValue protocolVersionValue = jsonObject.value(HANDSHAKE_KEY_PROTOCOL_VERSION);
        if (protocolVersionValue.type() != QJsonValue::Double) {
            emit message(tr("Der Server hat eine ungültige Begrüßung geschickt."));
            emit connectResult(false, tr("Der Server hat eine ungültige Begrüßung geschickt."));
            return;
        } else {
            const int protocolVersion = protocolVersionValue.toInt();
            if (protocolVersion != PROTOCOL_VERSION) {
                emit message(tr("Falsche Protokollrevision (ist %1, müsste %2 sein)").arg(
                                QString::number(protocolVersion),
                                QString::number(PROTOCOL_VERSION)));
                emit connectResult(false, tr(
                    "Der Server verwendet eine andere Protokollrevision (ist %1, müsste %2 sein). "
                    "Es kann keine Verbindung hergestellt werden.").arg(
                    QString::number(protocolVersion), QString::number(PROTOCOL_VERSION)));
                return;
            }
        }

        // The counterpart seems to be a Muckturnier server and we speak the same protocol revision
        emit message(tr("Begrüßung empfangen, frage Synchronisation an …"));
        sendData(HANDSHAKE_MESSAGE_SYNCHRONIZE);

    // Data Synchronization
    // ===================

    } else if (handshakePhase == HANDSHAKE_PHASE_SYNCHRONIZE) {

        // Check the settings and state
        // ----------------------------

        const QJsonObject settings = jsonObject.value(HANDSHAKE_KEY_SETTINGS).toObject();
        Muckturnier::TournamentMode tournamentMode
            = static_cast<Muckturnier::TournamentMode>(settings.value(
                  HANDSHAKE_KEY_TOURNAMENT_MODE).toInt());
        const int boogerScore = settings.value(HANDSHAKE_KEY_BOOGER_SCORE).toInt();
        const int boogersPerRound = settings.value(HANDSHAKE_KEY_BOOGERS_PER_ROUND).toInt();
        const bool networkTournamentStarted
                       = settings.value(HANDSHAKE_KEY_TOURNAMENT_STARTED).toBool();

        // Tournament settings
        if (m_tournamentSettings->tournamentMode() != tournamentMode
            || m_tournamentSettings->boogerScore() != boogerScore
            || m_tournamentSettings->boogersPerRound() != boogersPerRound) {

            emit message(tr("Die Datenbank des Servers hat andere Einstellungen als die lokale"));
            emit connectResult(false, tr(
                "Die Datenbank des Servers hat andere Einstellungen als die lokale:</p>"
                "<table>"
                "<tr>"
                    "<td></td>"
                    "<td>Server</td>"
                    "<td>Hier</td>"
                "</tr>"
                "<tr>"
                    "<td>Turniermodus:</td>"
                    "<td><i>%1</i>&nbsp;&nbsp;</td>"
                    "<td><i>%2</i></td>"
                "</tr>"
                "<tr>"
                    "<td>Punkte pro Bobbl:&nbsp;&nbsp;</td>"
                    "<td><i>%3</i></td>"
                    "<td><i>%4</i></td>"
                "</tr>"
                "<tr>"
                    "<td>Bobbl pro Runde:</td>"
                    "<td><i>%5</i></td>"
                    "<td><i>%6</i></td>"
                "</tr>"
                "</table>"
                "<p>Es kann keine Verbindung hergestellt werden.").arg(
                tournamentMode == Muckturnier::TournamentMode::FixedPairs ? tr("Feste Paare")
                                                                          : tr("Einzelne Spieler"),
                m_tournamentSettings->isPairMode() ? tr("Feste Paare") : tr("Einzelne Spieler"),
                QString::number(boogerScore),
                QString::number(m_tournamentSettings->boogerScore()),
                QString::number(boogersPerRound),
                QString::number(m_tournamentSettings->boogersPerRound())));
            return;
        }

        // Registration phase state
        if (m_db->localTournamentStarted() && ! networkTournamentStarted) {
            emit message(tr("Es wurden bereits Ergebnisse eingegeben, aber nicht auf dem Server"));
            emit connectResult(false,
                     tr("Es wurden bereits Ergebnisse eingegeben, aber der Server hat die "
                        "Anmeldung noch nicht beendet.</p>"
                        "<p>Bitte nach Beendigung der Anmeldung erneut verbinden, oder die bisher "
                        "eingegebenen Ergebnisse löschen und erneut verbinden!"));
            return;
        }

        // Process the Server's data
        // -------------------------

        // Extract the markers
        const QJsonArray markersArray(jsonObject.value(HANDSHAKE_KEY_MARKERS).toArray());
        QVector<Muckturnier::Marker> markers;
        markers.reserve(markersArray.count());
        for (const QJsonValue &value : markersArray) {
            const QJsonObject entry = value.toObject();
            Muckturnier::Marker marker;
            marker.id = entry.value(HANDSHAKE_KEY_MARKER_ID).toInt();
            marker.name = entry.value(HANDSHAKE_KEY_MARKER_NAME).toString();
            marker.color = colorFromValue(entry.value(HANDSHAKE_KEY_MARKER_COLOR));
            marker.sorting = static_cast<Muckturnier::MarkerSorting>(
                                entry.value(HANDSHAKE_KEY_MARKER_SORTING).toInt());
            markers.append(marker);
        }

        // Extract the players data
        QHash<QString, Database::PlayersData> playersData;
        for (const QJsonValue &value : jsonObject.value(HANDSHAKE_KEY_PLAYERS).toArray()) {
            const QJsonObject entry = value.toObject();
            const QString name = entry.keys().at(0);
            const QJsonObject jsonData = entry.value(name).toObject();
            Database::PlayersData data;
            data.marker = jsonData.value(PLAYERS_KEY_MARKER_ID).toInt();
            data.assignedTable = jsonData.value(PLAYERS_KEY_ASSIGNED_TABLE).toInt();
            data.assignedPairNumber = jsonData.value(PLAYERS_KEY_ASSIGNED_PAIR_NUMBER).toInt();
            playersData[name] = data;
        }
        QStringList players(playersData.keys());

        // Adopt the data
        // --------------

        // Adopt the locale
        const QString serverLocale = jsonObject.value(HANDSHAKE_KEY_LOCALE).toString();
        if (m_stringTools->locale() != serverLocale) {
            emit message(tr("Übernehme abweichendes Server-Gebietsschema „%1“ (die lokale "
                            "Einstellung ist „%2“)").arg(serverLocale, m_stringTools->locale()));
            m_stringTools->setLocale(serverLocale);
        }

        // Adopt the registration phase state
        m_db->setNetworkTournamentStarted(networkTournamentStarted);

        // If we don't have results yet, we can simply adopt all data
        if (! m_db->localTournamentStarted()) {
            if (! m_db->adoptAllData(playersData, markers)) {
                return;
            }
            m_playersPage->reload();

        // Otherwise, the local players list has to be matched and probably updated
        } else {
            bool success;
            QStringList localPlayers(m_db->players(&success));
            if (! success) {
                return;
            }

            if (localPlayers.count() != players.count()) {
                emit message(tr("Es wurden bereits Ergebnisse eingetragen, und es sind "
                                "unterschiedlich viele %1 angemeldet").arg(
                                m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler")));
                emit connectResult(false, tr(
                    "Es wurden bereits Ergebnisse eingetragen, und es sind unterschiedlich viele "
                    "%1 angemeldet (am Server %2, hier %3). Es kann keine Verbindung hergestellt "
                    "werden!").arg(m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler"),
                                   QString::number(players.count()),
                                   QString::number(localPlayers.count())));
                return;
            }

            m_stringTools->sortList(localPlayers);
            m_stringTools->sortList(players);
            if (localPlayers == players) {
                // The lists match, we can leave it as-is and adopt the markers
                if (! m_db->adoptMarkersAndPreAssignment(playersData, markers)) {
                    return;
                }
                m_playersPage->reload();
            } else {
                // The names differ, so we have to match the lists first
                emit namesDiffer(localPlayers, players);
                emit connectResult(false, tr(
                    "Es wurden bereits Ergebnisse eingetragen, und die %1 des Servers "
                    "unterscheidet sich von der lokalen. Bevor eine Verbindung hergestellt werden "
                    "kann, müssen alle Namen übereinstimmen.</p>"
                    "<p>Bitte die Liste überprüfen, die Namen übernehmen bzw. manuell korrigieren "
                    "und erneut verbinden!").arg(
                    m_tournamentSettings->isPairMode() ? tr("Paarliste") : tr("Spielerliste")));
                return;
            }
        }

        // Adopt the absent marker
        const int absentMarkerId = jsonObject.value(HANDSHAKE_KEY_ABSENT_MARKER).toInt();
        m_markersWidget->setAbsentMarker(absentMarkerId);
        m_tournamentSettings->setAbsentMarker(absentMarkerId);

        // Adopt the singles management settings
        const Muckturnier::SinglesManagement singlesManagement
            = JsonHelper::singlesManagementFromJson(jsonObject.value(
                  HANDSHAKE_KEY_SINGLES_MANAGEMENT).toObject());
        m_markersWidget->setSinglesManagement(singlesManagement);
        m_tournamentSettings->setSinglesManagement(singlesManagement);

        // Update the players list to show the adopted settings
        m_playersPage->updatePlayersList();

        // Here, the data should be in sync
        if (jsonObject.value(KEY_CHECKSUM).toString() == m_playersPage->checksum()) {
            // We are in sync with the server and everything seems to be fine, so we want an ID
            emit message(tr("Synchronisation war erfolgreich, frage ID an …"));
            sendData(HANDSHAKE_MESSAGE_ALLOCATE_ID);
        } else {
            // This should not happen, but who knows?!
            emit message(tr("Die Übernahme der Daten des Servers ist fehlgeschlagen!"));
            emit connectResult(false, tr("Die Übernahme der Daten des Servers ist "
                                           "fehlgeschlagen!"));
            return;
        }

    // Finish the handshake
    // ====================

    } else if (handshakePhase == HANDSHAKE_PHASE_ALLOCATE_ID) {
        m_connectionTimer->stop();
        m_handshakePending = false;
        m_id = jsonObject.value(KEY_CLIENT_ID).toInt();
        emit message(tr("Verbindung mit %1 Port %2 hergestellt, Client-ID: %3").arg(
                        serverIp(), QString::number(serverPort()), QString::number(m_id)));
        emit connectResult(true);
        m_connected = true;
    }
}

void Client::statusMessage(const QString &message, int clientId)
{
    if (clientId == m_id) {
        emit statusUpdate(message);
    } else {
        emit statusUpdate(tr("Netzwerkänderung: ") + message);
    }
}

// ======================
// Requests to the server
// ======================

// Registration changes
// ====================

void Client::requestRegisterPlayers(const QString &name, int marker)
{
    QJsonObject message { { KEY_ACTION,            ACTION_PLAYERS_REGISTER },
                          { PLAYERS_KEY_NAME,      name },
                          { PLAYERS_KEY_MARKER_ID, marker } };
    sendRequest(message, ProcessingTarget::Players,
                tr("Anmelden von „%1“").arg(name),
                name);
}

void Client::requestRenamePlayers(const QString &name, const QString &newName)
{
    QJsonObject message { { KEY_ACTION,           ACTION_PLAYERS_RENAME },
                          { PLAYERS_KEY_NAME,     name },
                          { PLAYERS_KEY_NEW_NAME, newName } };
    sendRequest(message, ProcessingTarget::Players,
                tr("Umbenennen von „%1“ zu „%2“").arg(name, newName),
                newName);
}

void Client::requestAssemblePair(const QString &player1, const QString &player2,
                                 const QString &separator, int assignedMarker)
{
    QJsonObject message { { KEY_ACTION,            ACTION_PLAYERS_ASSEMBLE_PAIR },
                          { PLAYERS_KEY_NAME,      QJsonArray { player1, player2, separator } },
                          { PLAYERS_KEY_MARKER_ID, assignedMarker } };
    sendRequest(message, ProcessingTarget::Players,
                tr("Zusammenfügen von „%1“ und „%2“ zu einem Paar").arg(player1, player2),
                player1 + separator + player2);
}

void Client::requestDeletePlayers(const QString &name)
{
    QJsonObject message { { KEY_ACTION,       ACTION_PLAYERS_DELETE },
                          { PLAYERS_KEY_NAME, name } };
    sendRequest(message, ProcessingTarget::Players,
                tr("Löschen von „%1“").arg(name));
}

void Client::requestUpdateMarker(const QString &name, const QString &markerName, int markerId)
{
    QJsonObject message { { KEY_ACTION,              ACTION_PLAYERS_UPDATE_MARKER },
                          { PLAYERS_KEY_NAME,        name },
                          { PLAYERS_KEY_MARKER_NAME, markerName },
                          { PLAYERS_KEY_MARKER_ID,   markerId } };
    if (markerId != 0) {
        sendRequest(message, ProcessingTarget::Players,
                    tr("Markieren von „%1“ als „%2“ an").arg(name, markerName),
                    name);
    } else {
        sendRequest(message, ProcessingTarget::Players,
                    tr("Entfernen der Markierung von „%1“ an").arg(name),
                    name);
    }
}

void Client::requestDeleteMarkedPlayers(const QString &markerName, int markerId)
{
    QJsonObject message { { KEY_ACTION,              ACTION_PLAYERS_DELETE_MARKED },
                          { PLAYERS_KEY_MARKER_NAME, markerName },
                          { PLAYERS_KEY_MARKER_ID,   markerId } };
    sendRequest(message, ProcessingTarget::Players,
                tr("Löschen aller mit „%1“ markierten %2").arg(
                   markerName, m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler")));
}

void Client::requestDeleteUnmarkedPlayers()
{
    QJsonObject message { { KEY_ACTION, ACTION_PLAYERS_DELETE_UNMARKED } };
    sendRequest(message, ProcessingTarget::Players,
                tr("Frage Löschen aller unmarkierten %1").arg(
                   m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler")));
}

void Client::requestDeleteAllPlayers()
{
    QJsonObject message { { KEY_ACTION, ACTION_PLAYERS_DELETE_ALL } };
    sendRequest(message, ProcessingTarget::Players,
                tr("Löschen aller %1 an").arg(
                   m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler")));
}

void Client::requestListUpdateMarker(const QString &markerName, int markerId)
{
    QJsonObject message { { KEY_ACTION,              ACTION_PLAYERS_LIST_UPDATE_MARKER },
                          { PLAYERS_KEY_MARKER_NAME, markerName },
                          { PLAYERS_KEY_MARKER_ID,   markerId } };

    if (markerId != 0) {
        sendRequest(message, ProcessingTarget::Players,
                    tr("Markieren aller %1 als „%2“ an").arg(
                       m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler"),
                       markerName));
    } else {
        sendRequest(message, ProcessingTarget::Players, tr("Entfernen aller Markierungen"));
    }
}

void Client::requestListRemoveMarker(const QString &markerName, int markerId)
{
    QJsonObject message { { KEY_ACTION,              ACTION_PLAYERS_LIST_REMOVE_MARKER },
                          { PLAYERS_KEY_MARKER_NAME, markerName },
                          { PLAYERS_KEY_MARKER_ID,   markerId } };
    sendRequest(message, ProcessingTarget::Players,
                tr("Entfernen aller „%1“-Markierungen").arg(markerName));
}

// Preassignment changes
// =====================

void Client::requestUpdatePreAssignment(const QString &name, int table, int pair, bool removeMarker)
{
    QJsonObject message { { KEY_ACTION,                       ACTION_UPDATE_PREASSIGNMENT },
                          { PLAYERS_KEY_NAME,                 name },
                          { PLAYERS_KEY_ASSIGNED_TABLE,       table },
                          { PLAYERS_KEY_ASSIGNED_PAIR_NUMBER, pair },
                          { PLAYERS_KEY_REMOVE_MARKER,        removeMarker } };
    const QString text = table == 0 ? tr("Löschen der Auslosung für „%1“").arg(name)
                                    : tr("Speichern der Auslosung für „%1“").arg(name);
    sendRequest(message, ProcessingTarget::Players, text, name);
}

void Client::requestDeleteAllPreAssignments()
{
    QJsonObject message { { KEY_ACTION, ACTION_DELETE_ALL_PREASSIGNMENTS } };
    sendRequest(message, ProcessingTarget::Players, tr("Löschen aller Auslosungen"));
}

// Markers definition changes
// ==========================

void Client::requestAddMarker(const QString &name, const QColor &color,
                              Muckturnier::MarkerSorting sorting)
{
    QJsonObject message { { KEY_ACTION,          ACTION_MARKERS_ADD },
                          { MARKERS_KEY_NAME,    name },
                          { MARKERS_KEY_COLOR,   colorToString(color) },
                          { MARKERS_KEY_SORTING, sorting } };
    sendRequest(message, ProcessingTarget::Markers,
                tr("Hinzufügen der Markierung „%1“").arg(name));
}

void Client::requestDeleteMarker(const QString &name, int id, int sequence)
{
    QJsonObject message { { KEY_ACTION,           ACTION_MARKERS_DELETE },
                          { MARKERS_KEY_NAME,     name },
                          { MARKERS_KEY_ID,       id },
                          { MARKERS_KEY_SEQUENCE, sequence } };
    sendRequest(message, ProcessingTarget::Markers,
                tr("Frage Löschen der Markierung „%1“").arg(name));
}

void Client::requestMoveMarker(const QString &name, int id, int sequence, int direction)
{
    QJsonObject message { { KEY_ACTION,            ACTION_MARKERS_MOVE },
                          { MARKERS_KEY_NAME,      name },
                          { MARKERS_KEY_ID,        id },
                          { MARKERS_KEY_SEQUENCE,  sequence },
                          { MARKERS_KEY_DIRECTION, direction } };
    sendRequest(message, ProcessingTarget::Markers,
                tr("Verschieben der Markierung „%1“ %2").arg(
                   name, direction == 1 ? tr("nach unten") : tr("nach oben")));
}

void Client::requestEditMarker(const QString &name, int id, const QString &newName,
                               const QColor &color, Muckturnier::MarkerSorting sorting)
{
    QJsonObject message { { KEY_ACTION,           ACTION_MARKERS_EDIT },
                          { MARKERS_KEY_NAME,     name },
                          { MARKERS_KEY_ID,       id },
                          { MARKERS_KEY_NEW_NAME, newName },
                          { MARKERS_KEY_COLOR,    colorToString(color) },
                          { MARKERS_KEY_SORTING,  sorting } };
    sendRequest(message, ProcessingTarget::Markers,
                tr("Bearbeiten der Markierung „%1“").arg(name));
}

// Special markers changes
// =======================

void Client::requestSetAbsentMarker(const QString &name, int id)
{
    QJsonObject message { { KEY_ACTION,       ACTION_SET_ABSENT_MARKER },
                          { MARKERS_KEY_NAME, name },
                          { MARKERS_KEY_ID,   id } };
    sendRequest(message, ProcessingTarget::Markers,
                id == -1 ? tr("Deaktivierung der Markierung für abwesende Anmeldungen")
                         : tr("Setzen der Markierung für abwesende Anmeldungen auf „%1“").arg(
                              name));
}

void Client::requestSinglesManagementChange(const Muckturnier::SinglesManagement &data)
{
    QJsonObject message { { KEY_ACTION, ACTION_CHANGE_SINGLES_MANAGEMENT },
                          { KEY_DATA,   JsonHelper::singlesManagementToJson(data) } };
    sendRequest(message, ProcessingTarget::Markers,
                tr("Änderung der Einstellungen für das Behandlen allein gekommener Spieler"));
}

// Finishing the registration
// ==========================

void Client::requestFinishRegistration()
{
    QJsonObject message { { KEY_ACTION, ACTION_FINISH_REGISTRATION } };
    sendRequest(message, ProcessingTarget::Players,
                tr("Beenden der Anmeldung"));
}

// Ranking comparison
// ==================

void Client::requestRanking(RankingDialog *dialog)
{
    m_rankingDialog = dialog;
    QJsonObject message { { KEY_ACTION, ACTION_RANKING_REQUEST } };
    sendRequest(message, ProcessingTarget::Ranking,
                tr("Rangliste"));
}

// ====================================
// Change announcements from the server
// ====================================

void Client::dataReceived(const QString &action, const QJsonObject &jsonObject)
{
    if (action == ACTION_HANDSHAKE) {
        if (! m_handshakePending) {
            return;
        }
        processHandshake(jsonObject);
        return;
    }

    const int clientId = jsonObject.value(KEY_CLIENT_ID).toInt();
    if (clientId == m_id) {
        stopProcessingRequest();
    } else {
        QApplication::setOverrideCursor(Qt::BusyCursor);
    }

    // Check if the request has been accepted
    if (action == ACTION_REJECTED) {
        QApplication::restoreOverrideCursor();
        emit statusUpdate(tr("Anfrage vom Server zurückgewiesen"));
        emit requestRejected(static_cast<RejectReason>(
                                 jsonObject.value(KEY_REJECT_REASON).toInt()));
        return;
    }

    // Pass the data to the respective handler function
    if (m_requestHandlers.contains(action)) {
        (this->*m_requestHandlers.value(action))(jsonObject, clientId);
    } else {
        // This should not happen
        emit connectionError(tr("Die Antwort des Servers konnte nicht verarbeitet werden."));
    }

    QApplication::restoreOverrideCursor();
    if (action != ACTION_RANKING_REQUEST) {
        verifyChecksum(jsonObject.value(KEY_CHECKSUM).toString());
    }
}

// Registration changes
// ====================

void Client::processRegisterPlayers(const QJsonObject &jsonObject, int clientId)
{
    const QString name = jsonObject.value(PLAYERS_KEY_NAME).toString();
    const int marker = jsonObject.value(PLAYERS_KEY_MARKER_ID).toInt();
    emit message(tr("Melde „%1“ an").arg(name));

    if (! m_db->registerPlayers(name, marker)) {
       return;
    }
    statusMessage(tr("%1 „%2“ angemeldet").arg(
                     m_tournamentSettings->isPairMode() ? tr("Paar") : tr("Spieler"), name),
                  clientId);

    m_playersPage->processNetworkChange(clientId);
}

void Client::processRenamePlayers(const QJsonObject &jsonObject, int clientId)
{
    const QString name = jsonObject.value(PLAYERS_KEY_NAME).toString();
    const QString newName = jsonObject.value(PLAYERS_KEY_NEW_NAME).toString();
    emit message(tr("Benenne „%1“ zu „%2“ um").arg(name, newName));

    if (! m_db->renamePlayers(name, newName)) {
        return;
    }

    statusMessage(tr("%1 „%2“ zu „%3“ umbenannt").arg(
                     m_tournamentSettings->isPairMode() ? tr("Paar") : tr("Spieler"),
                     name, newName),
                  clientId);

    m_playersPage->processNetworkChange(clientId);
    m_playersPage->checkUpdateAssignmentPage();
    emit nameEdited();
}

void Client::processAssemblePair(const QJsonObject &jsonObject, int clientId)
{
    const QJsonArray namesData = jsonObject.value(PLAYERS_KEY_NAME).toArray();
    const QString player1 = namesData.at(0).toString();
    const QString player2 = namesData.at(1).toString();
    const QString separator = namesData.at(2).toString();
    const int marker = jsonObject.value(PLAYERS_KEY_MARKER_ID).toInt();
    emit message(tr("Füge „%1“ und „%2“ zu einem Paar zusammen").arg(player1, player2));

    if (! m_db->assemblePair(player1, player2, separator, marker)) {
        return;
    }

    statusMessage(tr("„%1“ und „%2“ zu einem Paar zusammengefügt").arg(player1, player2),
                  clientId);

    m_playersPage->processNetworkChange(clientId);
    m_playersPage->checkUpdateAssignmentPage();
}

void Client::processDeletePlayers(const QJsonObject &jsonObject, int clientId)
{
    const QString name = jsonObject.value(PLAYERS_KEY_NAME).toString();
    emit message(tr("Lösche „%1“").arg(name));

    if (! m_db->deletePlayers(name)) {
        return;
    }

    statusMessage(tr("%1 „%2“ gelöscht").arg(
                     m_tournamentSettings->isPairMode() ? tr("Paar") : tr("Spieler"), name),
                  clientId);

    m_playersPage->processNetworkChange(clientId);
    m_playersPage->checkUpdateAssignmentPage();
}

void Client::processUpdatePlayersMarker(const QJsonObject &jsonObject, int clientId)
{
    const QString name = jsonObject.value(PLAYERS_KEY_NAME).toString();
    const QString markerName = jsonObject.value(PLAYERS_KEY_MARKER_NAME).toString();
    const int markerId = jsonObject.value(PLAYERS_KEY_MARKER_ID).toInt();

    if (markerId != 0) {
        emit message(tr("Markiere „%1“ als „%2“").arg(name, markerName));
    } else {
        emit message(tr("Entferne Markierung von „%1“").arg(name));
    }

    if (! m_db->updateMarker(name, markerId)) {
        return;
    }

    if (markerId != 0) {
        statusMessage(tr("%1 „%2“ als „%3“ markiert").arg(
                         m_tournamentSettings->isPairMode() ? tr("Paar") : tr("Spieler"),
                         name, markerName),
                      clientId);
    } else {
        statusMessage(tr("Markierung von %1 „%2“ entfernt").arg(
                         m_tournamentSettings->isPairMode() ? tr("Paar") : tr("Spieler"), name),
                      clientId);
    }

    m_playersPage->processNetworkChange(clientId);
}

void Client::processDeleteMarkedPlayers(const QJsonObject &jsonObject, int clientId)
{
    const QString markerName = jsonObject.value(PLAYERS_KEY_MARKER_NAME).toString();
    const int markerId = jsonObject.value(PLAYERS_KEY_MARKER_ID).toInt();
    emit message(tr("Lösche alle als „%1“ markierten %2").arg(
                    markerName, m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler")));

    if (! m_db->deleteMarkedPlayers(markerId)) {
        return;
    }

    statusMessage(tr("Alle als „%1“ markierten %2 gelöscht").arg(
                     markerName, m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler")),
                  clientId);

    m_playersPage->processNetworkChange(clientId);
    m_playersPage->checkUpdateAssignmentPage();
}

void Client::processDeleteUnmarkedPlayers(const QJsonObject &, int clientId)
{
    emit message(tr("Lösche alle unmarkierten %1").arg(
                    m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler")));

    if (! m_db->deleteUnmarkedPlayers()) {
        return;
    }

    statusMessage(tr("Alle unmarkierten %1 gelöscht").arg(
                     m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler")),
                  clientId);

    m_playersPage->processNetworkChange(clientId);
    m_playersPage->checkUpdateAssignmentPage();
}

void Client::processDeleteAllPlayers(const QJsonObject &, int clientId)
{
    emit message(tr("Lösche alle %1").arg(m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler")));

    if (! m_db->deleteAllPlayers()) {
        return;
    }

    statusMessage(tr("Alle %1 gelöscht").arg(
                     m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler")),
                  clientId);

    m_playersPage->processNetworkChange(clientId);
    m_playersPage->checkUpdateAssignmentPage();
}

void Client::processUpdateListMarker(const QJsonObject &jsonObject, int clientId)
{
    const QString markerName = jsonObject.value(PLAYERS_KEY_MARKER_NAME).toString();
    const int markerId = jsonObject.value(PLAYERS_KEY_MARKER_ID).toInt();

    if (markerId != 0) {
        emit message(tr("Markiere alle %1 als „%2“").arg(
                        m_tournamentSettings->isPairMode() ? tr("Paar") : tr("Spieler"),
                        markerName));
    } else {
        emit message(tr("Entferne alle Markierungen"));
    }

    if (! m_db->listUpdateMarker(markerId)) {
        return;
    }

    if (markerId != 0) {
        statusMessage(tr("Alle %1 als „%2“ markiert").arg(
                         m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler"),
                         markerName),
                      clientId);
    } else {
        statusMessage(tr("Alle Markierungen entfernt"), clientId);
    }

    m_playersPage->processNetworkChange(clientId);
}

void Client::processRemoveListMarker(const QJsonObject &jsonObject, int clientId)
{
    const QString markerName = jsonObject.value(PLAYERS_KEY_MARKER_NAME).toString();
    const int markerId = jsonObject.value(PLAYERS_KEY_MARKER_ID).toInt();
    emit message(tr("Entferne alle „%1“-Markierungen").arg(markerName));

    if (! m_db->listRemoveMarker(markerId)) {
        return;
    }

    statusMessage(tr("Alle „%1“-Markierungen entfernt").arg(markerName), clientId);

    m_playersPage->processNetworkChange(clientId);
}

// Preassignment changes
// =====================

void Client::processUpdatePreAssignment(const QJsonObject &jsonObject, int clientId)
{
    const QString name = jsonObject.value(PLAYERS_KEY_NAME).toString();
    const int table = jsonObject.value(PLAYERS_KEY_ASSIGNED_TABLE).toInt();
    const int pair = jsonObject.value(PLAYERS_KEY_ASSIGNED_PAIR_NUMBER).toInt();
    if (table == 0) {
        emit message(tr("Lösche Auslosung für „%1“").arg(name));
    } else {
        emit message(tr("Speichere Auslosung für „%1“").arg(name));
    }

    if (! m_db->updatePreAssignment(name, table, pair,
                                    jsonObject.value(PLAYERS_KEY_REMOVE_MARKER).toBool())) {
        return;
    }

    if (table == 0) {
        statusMessage(tr("Auslosung für „%1“ gelöscht").arg(name), clientId);
    } else {
        statusMessage(tr("Auslosung für „%1“ gespeichert (Tisch %2 Paar %3)").arg(
                         name, QString::number(table), QString::number(pair)), clientId);
    }

    m_playersPage->processNetworkChange(clientId);
    m_playersPage->checkUpdateAssignmentPage();
}

void Client::processDeleteAllPreAssignments(const QJsonObject &, int clientId)
{
    emit message(tr("Lösche alle Auslosungen"));

    if (! m_db->deleteAllPreAssignments()) {
        return;
    }

    statusMessage(tr("Alle Auslosungen gelöscht"), clientId);

    m_playersPage->processNetworkChange(clientId);
    m_playersPage->checkUpdateAssignmentPage();
}

// Marker definition changes
// =========================

void Client::processAddMarker(const QJsonObject &jsonObject, int clientId)
{
    const QString name = jsonObject.value(MARKERS_KEY_NAME).toString();
    emit message(tr("Füge Markierung „%1“ hinzu").arg(name));

    if (! m_db->addMarker(name, colorFromValue(jsonObject.value(MARKERS_KEY_COLOR)),
                          static_cast<Muckturnier::MarkerSorting>(
                              jsonObject.value(MARKERS_KEY_SORTING).toInt()))) {
        return;
    }

    statusMessage(tr("Markierung „%1“ hinzugefügt").arg(name), clientId);

    m_markersWidget->processNetworkChange();
}

void Client::processDeleteMarker(const QJsonObject &jsonObject, int clientId)
{
    const QString name = jsonObject.value(MARKERS_KEY_NAME).toString();
    emit message(tr("Lösche Markierung „%1“").arg(name));

    if (! m_db->deleteMarker(jsonObject.value(MARKERS_KEY_ID).toInt(),
                             jsonObject.value(MARKERS_KEY_SEQUENCE).toInt())) {
        return;
    }

    statusMessage(tr("Markierung „%1“ gelöscht").arg(name), clientId);

    m_markersWidget->processNetworkChange();
}

void Client::processMoveMarker(const QJsonObject &jsonObject, int clientId)
{
    const QString name = jsonObject.value(MARKERS_KEY_NAME).toString();
    const int direction = jsonObject.value(MARKERS_KEY_DIRECTION).toInt();
    emit message(tr("Verschiebe Markierung „%1“ %2").arg(
                    name, direction == 1 ? tr("nach unten") : tr("nach oben")));

    if (! m_db->moveMarker(jsonObject.value(MARKERS_KEY_ID).toInt(),
                           jsonObject.value(MARKERS_KEY_SEQUENCE).toInt(),
                           direction)) {
        return;
    }

    statusMessage(tr("Markierung „%1“ %2 verschoben").arg(
                     name, direction == 1 ? tr("nach unten") : tr("nach oben")),
                  clientId);

    m_markersWidget->processNetworkChange();
}

void Client::processEditMarker(const QJsonObject &jsonObject, int clientId)
{
    const QString name = jsonObject.value(MARKERS_KEY_NAME).toString();
    emit message(tr("Speichere Änderungen an Markierung „%1“").arg(name));

    if (! m_db->editMarker(jsonObject.value(MARKERS_KEY_ID).toInt(),
                           jsonObject.value(MARKERS_KEY_NEW_NAME).toString(),
                           colorFromValue(jsonObject.value(MARKERS_KEY_COLOR)),
                           static_cast<Muckturnier::MarkerSorting>(
                               jsonObject.value(MARKERS_KEY_SORTING).toInt()))) {
        return;
    }

    statusMessage(tr("Änderungen an Markierung „%1“ gespeichert").arg(name), clientId);

    m_markersWidget->processNetworkChange();
}

// Special markers changes
// =======================

void Client::processSetAbsentMarker(const QJsonObject &jsonObject, int clientId)
{
    const QString name = jsonObject.value(MARKERS_KEY_NAME).toString();
    const int markerId = jsonObject.value(MARKERS_KEY_ID).toInt();

    emit message(markerId == -1 ? tr("Deaktiviere Markierung für abwesende Anmeldungen")
                                : tr("Setze Markierung für abwesende Anmeldungen auf „%1“").arg(
                                     name));

    m_markersWidget->setAbsentMarker(markerId);
    m_tournamentSettings->setAbsentMarker(markerId);

    statusMessage(markerId == -1 ? tr("Markierung für abwesende Anmeldungen deaktiviert")
                                 : tr("Markierung für abwesende Anmeldungen auf „%1“ gesetzt").arg(
                                      name),
                  clientId);

    m_markersWidget->processNetworkChange();
}

void Client::processChangeSinglesManagement(const QJsonObject &jsonObject, int clientId)
{
    emit message(tr("Übernehme Änderungen an den Einstellungen für das Behandlen allein gekommener "
                    "Spieler"));

    const Muckturnier::SinglesManagement data = JsonHelper::singlesManagementFromJson(
                                               jsonObject.value(KEY_DATA).toObject());
    m_markersWidget->setSinglesManagement(data);
    m_tournamentSettings->setSinglesManagement(data);
    statusMessage(tr("Änderungen an den Einstellungen für das Behandeln allein gekommener Spieler "
                     "übernommen"),
                  clientId);

    m_markersWidget->processNetworkChange();
}

// Finishing the registration
// ==========================

void Client::processFinishRegistration(const QJsonObject &, int clientId)
{
    emit message(tr("Beende die Anmeldung"));
    m_playersPage->networkFinishRegistration(clientId);
    statusMessage(tr("Anmeldung beendet"), clientId);
}

// Ranking comparison
// ==================

void Client::processRankingRequest(const QJsonObject &jsonObject, int)
{
    if (! jsonObject.value(RANKING_KEY_AVAILABLE).toBool()) {
        emit statusUpdate(tr("Antwort der Ranglistenanfrage empfangen"));
        emit message(tr("Es ist noch keine Rangliste verfügbar"));
        m_rankingDialog->noRanking();
        return;
    }

    const QJsonArray names = jsonObject.value(RANKING_KEY_NAMES).toArray();
    const QJsonArray boogers = jsonObject.value(RANKING_KEY_BOOGERS).toArray();
    const QJsonArray goals = jsonObject.value(RANKING_KEY_GOALS).toArray();
    const QJsonArray opponentGoals = jsonObject.value(RANKING_KEY_OPPONENT_GOALS).toArray();
    const QJsonArray playedBoogers = jsonObject.value(RANKING_KEY_PLAYED_BOOGERS).toArray();
    QVector<Database::RankData> ranking;
    for (int i = 0; i < names.count(); i++) {
        Database::RankData rank;
        rank.name = names.at(i).toString();
        rank.boogers = boogers.at(i).toInt();
        rank.goals = goals.at(i).toInt();
        rank.opponentGoals = opponentGoals.at(i).toInt();
        rank.playedBoogers = playedBoogers.at(i).toInt();
        ranking.append(rank);
    }

    emit message(tr("Rangliste empfangen"));
    emit statusUpdate(tr("Rangliste empfangen"));
    m_rankingDialog->compareRanking(ranking);
}
