/* Copyright (c) 2018-2020 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef CLIENT_H
#define CLIENT_H

// Local includes
#include "shared/AbstractNetworkEngine.h"
#include "shared/ServerProtocol.h"
#include "shared/Muckturnier.h"

// Qt includes
#include <QAbstractSocket>

// Local classes
class SharedObjects;
class Database;
class TournamentSettings;
class StringTools;
class PlayersPage;
class RankingDialog;
class MarkersWidget;

// Qt classes
class QHostAddress;
class QTimer;
class QJsonObject;

class Client : public AbstractNetworkEngine
{
    Q_OBJECT

public:
    enum ProcessingTarget {
        Players,
        Markers,
        Ranking
    };

    explicit Client(QObject *parent, int connectTimeout, int requestTimeout,
                    SharedObjects *sharedObjects, PlayersPage *playersPage);
    void connectToServer(const QHostAddress &address, quint16 port);
    void closeConnection();
    QString serverIp() const;
    int serverPort() const;
    int id() const;
    bool connected() const;
    void abortHandshake();
    void checkAbortHandshake();

    // Change request functions
    // ========================

    // Registration changes

    void requestRegisterPlayers(const QString &name, int marker);
    void requestRenamePlayers(const QString &name, const QString &newName);
    void requestAssemblePair(const QString &player1, const QString &player2,
                             const QString &separator, int assignedMarker);
    void requestDeletePlayers(const QString &name);
    void requestUpdateMarker(const QString &name, const QString &markerName, int markerId);

    void requestDeleteMarkedPlayers(const QString &markerName, int markerId);
    void requestDeleteUnmarkedPlayers();
    void requestDeleteAllPlayers();
    void requestListUpdateMarker(const QString &markerName, int markerId);
    void requestListRemoveMarker(const QString &markerName, int markerId);

    // Preassignment changes
    void requestUpdatePreAssignment(const QString &name, int table, int pair, bool removeMarker);
    void requestDeleteAllPreAssignments();

    // Markers definition changes
    void requestAddMarker(const QString &name, const QColor &color,
                          Muckturnier::MarkerSorting sorting);
    void requestDeleteMarker(const QString &name, int id, int sequence);
    void requestMoveMarker(const QString &name, int id, int sequence, int direction);
    void requestEditMarker(const QString &name, int id, const QString &newName,
                           const QColor &color, Muckturnier::MarkerSorting sorting);

    // Special markers changes
    void requestSetAbsentMarker(const QString &name, int id);
    void requestSinglesManagementChange(const Muckturnier::SinglesManagement &data);

    // Finishing the registration
    void requestFinishRegistration();

    // Ranking comparison
    void requestRanking(RankingDialog *dialog);

signals:
    void message(const QString &text);
    void connectResult(bool success, const QString &error = QString());
    void connectionError(const QString &text);
    void requestRejected(ServerProtocol::RejectReason reason);
    void statusUpdate(const QString &text);
    void namesDiffer(QStringList &localNames, QStringList &remoteNames);
    void nameEdited();

private slots:
    void socketConnected();
    void error(QAbstractSocket::SocketError);
    void connectionTimeout();
    void requestTimeout();

private: // Functions
    void statusMessage(const QString &message, int clientId);
    void sendRequest(QJsonObject &jsonObject, ProcessingTarget target,
                     const QString &request, const QString &scrollToTarget = QString());
    void processHandshake(const QJsonObject &jsonObject);
    void verifyChecksum(const QString &checksum);
    void stopProcessingRequest();

    virtual void dataReceived(const QString &action, const QJsonObject &jsonObject) override;

    // Server change announcement handlers
    // ===================================

    // Helper macro to keep the header a bit more tidy
    #define CHANGE_HANDLER(name) void process##name(const QJsonObject &jsonObject, int clientId)

    // Registration changes

    CHANGE_HANDLER(RegisterPlayers);
    CHANGE_HANDLER(RenamePlayers);
    CHANGE_HANDLER(AssemblePair);
    CHANGE_HANDLER(DeletePlayers);
    CHANGE_HANDLER(UpdatePlayersMarker);

    CHANGE_HANDLER(DeleteMarkedPlayers);
    CHANGE_HANDLER(DeleteUnmarkedPlayers);
    CHANGE_HANDLER(DeleteAllPlayers);
    CHANGE_HANDLER(UpdateListMarker);
    CHANGE_HANDLER(RemoveListMarker);

    // Preassignment changes
    CHANGE_HANDLER(UpdatePreAssignment);
    CHANGE_HANDLER(DeleteAllPreAssignments);

    // Markers definition changes
    CHANGE_HANDLER(AddMarker);
    CHANGE_HANDLER(DeleteMarker);
    CHANGE_HANDLER(MoveMarker);
    CHANGE_HANDLER(EditMarker);

    // Special markers changes
    CHANGE_HANDLER(SetAbsentMarker);
    CHANGE_HANDLER(ChangeSinglesManagement);

    // Finishing the registration
    CHANGE_HANDLER(FinishRegistration);

    // Ranking comparison
    CHANGE_HANDLER(RankingRequest);

    #undef CHANGE_HANDLER

private: // Variables
    const int m_connectTimeout;
    Database *m_db;
    TournamentSettings *m_tournamentSettings;
    StringTools *m_stringTools;
    PlayersPage *m_playersPage;
    MarkersWidget *m_markersWidget;
    const QHash<QString, void(Client::*)(const QJsonObject &, int)> m_requestHandlers;

    int m_id;
    bool m_connected = false;
    bool m_connectionPending = false;
    bool m_handshakePending = false;
    QTimer *m_connectionTimer;
    QTimer *m_requestTimer;
    RankingDialog *m_rankingDialog;
    QVector<ProcessingTarget> m_processingTarget;

};

#endif // CLIENT_H
