/* Copyright (c) 2018-2020 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes

#include "ClientPage.h"
#include "Client.h"
#include "DifferencesDelegate.h"
#include "DiscoverClient.h"

#include "shared/SocketAddressWidget.h"
#include "shared/LogDisplay.h"
#include "shared/SharedStyles.h"
#include "shared/SharedObjects.h"
#include "shared/StringTools.h"
#include "shared/Database.h"
#include "shared/Settings.h"

// Qt includes
#include <QDebug>
#include <QVBoxLayout>
#include <QHBoxLayout>
#include <QLabel>
#include <QPushButton>
#include <QMessageBox>
#include <QApplication>
#include <QGroupBox>
#include <QTableWidget>
#include <QHeaderView>
#include <QTimer>
#include <QScrollBar>
#include <QUdpSocket>
#include <QNetworkInterface>

// C++ includes
#include <functional>

ClientPage::ClientPage(QWidget *parent, SharedObjects *sharedObjects, PlayersPage *playersPage,
                       const QString &ip, int port)
    : QWidget(parent),
      m_stringTools(sharedObjects->stringTools()),
      m_db(sharedObjects->database())
{
    Settings *settings = sharedObjects->settings();
    m_discoverAttempts = settings->discoverAttempts();
    m_connecting = false;

    m_discoverClient = new DiscoverClient(this, settings->discoverPort(),
                                          settings->discoverTimeout());
    connect(m_discoverClient, &DiscoverClient::discoverResult,
            this, &ClientPage::processDiscoverResult);

    m_client = new Client(this, settings->connectTimeout(), settings->requestTimeout(),
                          sharedObjects, playersPage);
    connect(m_client, &Client::protocolError, this, &ClientPage::connectionError);
    connect(m_client, &Client::connectResult, this, &ClientPage::checkConnectResult);
    connect(m_client, &Client::connectionError, this, &ClientPage::connectionError);
    connect(m_client, &Client::requestRejected, this, &ClientPage::requestRejected);
    connect(m_client, &Client::statusUpdate, this, &ClientPage::statusUpdate);
    connect(m_client, &Client::namesDiffer, this, &ClientPage::showDifferingNames);

    QVBoxLayout *layout = new QVBoxLayout(this);

    // Settings box

    m_settingsBox = new QGroupBox(tr("Socket-Adresse des Servers"));
    QVBoxLayout *mainSettingsLayout = new QVBoxLayout(m_settingsBox);
    QHBoxLayout *settingsLayout = new QHBoxLayout;
    mainSettingsLayout->addLayout(settingsLayout);
    layout->addWidget(m_settingsBox);

    m_notConnectedMessage = tr("<span style=\"%1\">Nicht<br/>verbunden</span>").arg(
                               SharedStyles::STYLE_RED_TEXT);
    m_connectedLabel = new QLabel(m_notConnectedMessage);
    settingsLayout->addWidget(m_connectedLabel);

    m_searchServerButton = new QPushButton(tr("Server suchen"));
    connect(m_searchServerButton, &QPushButton::clicked, this, &ClientPage::searchServer);
    settingsLayout->addWidget(m_searchServerButton);

    m_socketAddress = new SocketAddressWidget(SocketAddressWidget::Mode::Client,
                                              settings->serverPort(),
                                              settings->preferIpv6());
    connect(m_socketAddress, &SocketAddressWidget::returnPressed,
            this, &ClientPage::connectToServer);
    settingsLayout->addWidget(m_socketAddress);

    m_connectButton = new QPushButton(tr("Verbinden"));
    m_connectButton->setIcon(QApplication::style()->standardIcon(QStyle::SP_DialogYesButton));
    settingsLayout->addWidget(m_connectButton);
    connect(m_socketAddress, &SocketAddressWidget::settingsOkay,
            m_connectButton, &QPushButton::setEnabled);
    connect(m_connectButton, &QPushButton::clicked, this, &ClientPage::connectToServer);
    if (! ip.isEmpty()) {
        // This will also enable the connect button
        m_socketAddress->setSocketAddress(ip, port);
    }

    m_disconnectButton = new QPushButton(tr("Verbindung beenden"));
    m_disconnectButton->setIcon(QApplication::style()->standardIcon(QStyle::SP_DialogCancelButton));
    settingsLayout->addWidget(m_disconnectButton);
    m_disconnectButton->hide();
    connect(m_disconnectButton, &QPushButton::clicked, this, &ClientPage::closeConnection);

    settingsLayout->addStretch();

    m_closeButton = new QPushButton(tr("Schließen"));
    m_closeButton->setIcon(QApplication::style()->standardIcon(QStyle::SP_DialogCloseButton));
    settingsLayout->addWidget(m_closeButton);
    connect(m_closeButton, &QPushButton::clicked, this, &ClientPage::closeClient);

    m_warningLabel = new QLabel(tr(
        "<span style=\"%1 %2\">Achtung!</span> Wenn eine Verbindung hergestellt wird, werden "
        "<span style=\"%2\">alle lokalen Anmeldungen, Auslosungen und Markierungen "
        "verworfen</span> und der Datenbestand des Servers übernommen!").arg(
        SharedStyles::STYLE_BOLD_TEXT, SharedStyles::STYLE_RED_TEXT));
    m_warningLabel->setWordWrap(true);
    mainSettingsLayout->addWidget(m_warningLabel);

    // Differences box

    m_differencesBox = new QGroupBox(tr("Unterschiede der Anmeldungsliste"));
    QVBoxLayout *differencesLayout = new QVBoxLayout(m_differencesBox);
    layout->addWidget(m_differencesBox);

    m_differences = new QTableWidget;
    m_differences->setEditTriggers(QAbstractItemView::NoEditTriggers);
    m_differences->setItemDelegate(new DifferencesDelegate(this));
    m_differences->verticalHeader()->hide();
    m_differences->setShowGrid(false);
    m_differences->setWordWrap(false);
    m_differences->setSelectionMode(QTableWidget::NoSelection);
    m_differences->setFocusPolicy(Qt::NoFocus);
    m_differences->setContextMenuPolicy(Qt::NoContextMenu);
    m_differences->setColumnCount(2);
    m_differences->setHorizontalHeaderLabels({ tr("Name hier"), tr("Name Server") });
    differencesLayout->addWidget(m_differences);

    QHBoxLayout *differencesButtonsLayout = new QHBoxLayout;
    differencesLayout->addLayout(differencesButtonsLayout);

    differencesButtonsLayout->addStretch();

    m_adoptListButton = new QPushButton(tr("Namen vom Server übernehmen und neu verbinden"));
    m_adoptListButton->setIcon(QApplication::style()->standardIcon(QStyle::SP_DialogYesButton));
    differencesButtonsLayout->addWidget(m_adoptListButton);
    connect(m_adoptListButton, &QPushButton::clicked, this, &ClientPage::adoptList);

    m_cancelButton = new QPushButton(tr("Abbrechen"));
    m_cancelButton->setIcon(QApplication::style()->standardIcon(QStyle::SP_DialogCancelButton));
    differencesButtonsLayout->addWidget(m_cancelButton);
    connect(m_cancelButton, &QPushButton::clicked, this, &ClientPage::cancelAdoption);

    m_differencesBox->hide();

    // Display box

    m_displayBox = new QGroupBox(tr("Netzwerkinformationen"));
    QVBoxLayout *connectedLayout = new QVBoxLayout(m_displayBox);
    layout->addWidget(m_displayBox);

    m_display = new LogDisplay;
    connect(m_client, &Client::message, m_display, &LogDisplay::displayMessage);
    connect(m_client, &AbstractNetworkEngine::protocolError,
            m_display, &LogDisplay::displayProtocolError);
    connectedLayout->addWidget(m_display);

    m_disconnectButton->hide();

    // We have to do this via QTimer call so that the page shows up before we continue
    QTimer::singleShot(0, this, std::bind(&ClientPage::finishConstruction, this, ! ip.isEmpty()));
}

void ClientPage::finishConstruction(bool ipGiven)
{
    if (ipGiven) {
        if (QMessageBox::question(this,
            tr("Mit Server verbinden"),
            tr("Soll die letzte benutzte Verbindung zu dem Muckturnier-Server unter IP %1 Port %2 "
               "wiederhergestellt werden?").arg(m_socketAddress->ip(),
                                                QString::number(m_socketAddress->port())),
            QMessageBox::Yes | QMessageBox::No, QMessageBox::Yes) == QMessageBox::Yes) {

            connectToServer();
        } else {
            m_connectButton->setFocus();
        }
    } else {
        searchServer();
    }
}

void ClientPage::connectToServer()
{
    if (! m_connectButton->isEnabled()) {
        return;
    }

    m_display->clear();
    emit connectionRequested(m_socketAddress->ip(), m_socketAddress->port());
    m_connecting = true;
    QApplication::setOverrideCursor(Qt::WaitCursor);
    m_settingsBox->setEnabled(false);
    m_searchServerButton->setEnabled(false);
    m_connectButton->setEnabled(false);
    m_closeButton->setEnabled(false);
    m_client->connectToServer(QHostAddress(m_socketAddress->ip()), m_socketAddress->port());
}

void ClientPage::resetInterface()
{
    QApplication::restoreOverrideCursor();
    m_connecting = false;
    m_settingsBox->setEnabled(true);
    m_searchServerButton->setEnabled(true);
    m_connectButton->setEnabled(true);
    m_closeButton->setEnabled(true);
}

void ClientPage::checkConnectResult(bool success, const QString &error)
{
    if (success) {
        resetInterface();
        m_warningLabel->hide();
        m_socketAddress->hide();
        m_connectedLabel->setText(tr("<span style=\"%1\">Verbindung hergestellt</span><br/>"
                                     "Server %2 Port %3<br/>"
                                     "Client-ID: %4").arg(SharedStyles::STYLE_GREEN_TEXT,
                                                          m_client->serverIp(),
                                                          QString::number(m_client->serverPort()),
                                                          QString::number(m_client->id())));
        m_searchServerButton->hide();
        m_connectButton->hide();
        m_disconnectButton->show();
        m_closeButton->show();
        m_closeButton->clearFocus();
        emit connected(m_client);
        emit statusUpdate(tr("Verbindung mit Server unter IP %1 Port %2 hergestellt").arg(
                             m_client->serverIp(), QString::number(m_client->serverPort())));
        if (m_stringTools->localeChanged()) {
            QTimer::singleShot(0, this, [this] {
                QMessageBox::information(this,
                    tr("Verbindung hergestellt"),
                    tr("<p><b>Abweichendes Gebietsschema</b></p>"
                       "<p>Der Server benutzt eine anderes Gebietsschema („%1“) als die hier "
                       "ausgeführte Instanz von Muckturnier („%2“). Für die Zeit der Verbindung "
                       "wird das Server-Gebietsschema übernommen. Das führt evtl. zu einer anderen "
                       "Sortierung von Namen; alle Teilnehmer müssen gleich sortieren.</p>"
                       "<p>Nach dem Beenden der Verbindung wird die lokale Einstellung für das "
                       "Gebietsschema wiederhergestellt.</p>").arg(m_stringTools->locale(),
                                                                   m_stringTools->defaultLocale()));
            });
        }
    } else {
        m_client->abortHandshake();
        m_display->displayMessage(tr("Herstellen der Verbindung ist fehlgeschlagen"));
        resetInterface();
        QMessageBox::warning(this, tr("Mit Server verbinden"),
            tr("<p><b>Es kann keine Verbindung mit dem Server %1 hergestellt werden:</p>"
               "<p>%2</p>").arg(m_socketAddress->ip(), error));
    }
}

void ClientPage::connectionError(const QString &text)
{
    emit raiseMe();
    m_client->closeConnection();
    closeConnection();
    resetInterface();
    m_display->displayMessage(tr("Verbindungsfehler: %1").arg(text));
    QMessageBox::warning(this, tr("Verbindungsfehler"),
                         tr("<p>Es ist ein Fehler in der Serververbindung aufgetreten:</p>"
                            "<p>%1</p>").arg(text));
}

void ClientPage::closeConnection()
{
    if (! checkConnected()) {
        return;
    }

    if (m_client->connected()) {
        m_client->closeConnection();
    }

    m_connectedLabel->setText(m_notConnectedMessage);
    m_searchServerButton->show();
    m_connectButton->show();
    m_disconnectButton->hide();
    m_socketAddress->show();
    m_connectButton->setFocus();

    emit connectionClosed();
    emit statusUpdate(tr("Serververbindung beendet"));
}

void ClientPage::requestRejected(ServerProtocol::RejectReason reason)
{
    m_display->displayMessage(tr("Anfrage vom Server zurückgewiesen"));
    QString message;
    switch (reason) {
    case ServerProtocol::RejectReason::OutdatedRequest:
        message = tr("<p>Der Server hat die letzte Anfrage zurückgewiesen, da der Datenbestand "
                     "sich während der Anfrage verändert hat. Die Anfrage wird ignoriert.</p>"
                     "<p>Bitte die Änderung erneut anfragen!</p>");
        break;
    case ServerProtocol::RejectReason::UnknownAction:
        message = tr("<p>Die Anfrage wurde vom Server nicht verstanden. Das sollte eigentlich "
                     "nicht passieren, bitte mit möglichst vielen Informationen (wie kann man das "
                     "reproduzieren?) als Fehler melden!</p>");
        break;
    case ServerProtocol::RejectReason::Unknown:
        message = tr("<p>Die Anfrage wurde vom Server nicht verstanden.</p>");
        break;
    }
    QMessageBox::warning(qobject_cast<QWidget *>(parent()),
                         tr("Anfrage vom Server zurückgewiesen"), message);
}

bool ClientPage::checkConnected()
{
    if (m_client->connected()) {
        if (QMessageBox::question(this,
                tr("Serververbindung schließen"),
                tr("<p>Wenn die Verbindung getrennt wird, dann werden keine Änderungen mehr an die "
                   "anderen Netzwerkteilnehmer gesendet oder von ihnen empfangen!</p>"
                   "<p>Trotzdem schließen?</p>"),
                QMessageBox::Ok | QMessageBox::Cancel,
                QMessageBox::Cancel) == QMessageBox::Cancel) {

                return false;
        }
    }
    return true;
}

void ClientPage::closeClient()
{
    if (! checkConnected()) {
        return;
    }
    closePage();
}

void ClientPage::closePage()
{
    // This will only do something if a handshake is currently being processed
    client()->abortHandshake();

    // Inform the MainWindow
    emit pageClosed();

    // Close the DockWidget
    parent()->deleteLater();
}

void ClientPage::showDifferingNames(QStringList &localNames, QStringList &remoteNames)
{
    emit namesDifferencesDisplayed(true);

    m_stringTools->removeIntersection(localNames, remoteNames);
    QHash<int, int> assignment = m_stringTools->matchLists(localNames, remoteNames);

    m_settingsBox->hide();
    m_differencesBox->show();
    m_displayBox->hide();

    m_differences->clearContents();
    m_differences->setRowCount(localNames.count());
    for (int i = 0; i < localNames.count(); i++) {
        const QString longestCommonSubsequence = m_stringTools->longestCommonSubsequence(
            localNames.at(i), remoteNames.at(assignment[i]));
        QTableWidgetItem *localName = new QTableWidgetItem(localNames.at(i));
        QTableWidgetItem *remoteName = new QTableWidgetItem(remoteNames.at(assignment[i]));
        localName->setData(Qt::UserRole, QVariant(m_stringTools->diffList(
                                                      localNames.at(i), longestCommonSubsequence)));
        remoteName->setData(Qt::UserRole, QVariant(m_stringTools->diffList(
                                                      remoteNames.at(assignment[i]),
                                                      longestCommonSubsequence)));
        m_differences->setItem(i, 0, localName);
        m_differences->setItem(i, 1, remoteName);
    }
    m_differences->resizeColumnsToContents();
}

void ClientPage::adoptList()
{
    QStringList oldNames;
    QStringList newNames;
    for (int i = 0; i < m_differences->rowCount(); i++) {
        oldNames.append(m_differences->item(i, 0)->text());
        newNames.append(m_differences->item(i, 1)->text());
    }

    if (! m_db->renamePlayers(oldNames, newNames)) {
        return;
    }

    emit nameEdited();
    cancelAdoption();
    connectToServer();
}

void ClientPage::cancelAdoption()
{
    emit namesDifferencesDisplayed(false);
    m_settingsBox->show();
    m_differencesBox->hide();
    m_displayBox->show();
    m_display->clear();
}

Client * ClientPage::client() const
{
    return m_client;
}

void ClientPage::searchServer()
{
    if (m_currentDiscoverAttempts == 0) {
        m_currentDiscoverAttempts = m_discoverAttempts;
    }

    if (! m_discoverClient->discoverServer()) {
        QMessageBox::warning(this,
            tr("Muckturnier-Server suchen"),
            tr("<p><b>Der Discover-Client konnte nicht gestartet werden</b></p>"
               "<p>Die Fehlermeldung war: %1</p>"
               "<p>Es kann nicht automatisch nach einem Muckturnier-Server gesucht werden. Bitte "
               "die Socket-Adresse (IP-Adresse und Port) des Servers manuell eingeben.</p>").arg(
               m_discoverClient->errorString()));
        return;
    }

    m_settingsBox->setEnabled(false);
    m_display->displayMessage(tr("Suche nach einem Muckturnier-Server via UDP-Port %1 (%n "
                                 "Versuch(e) verbleiben) …", "", m_currentDiscoverAttempts).arg(
                                 m_discoverClient->port()));
    QApplication::setOverrideCursor(Qt::WaitCursor);
}

void ClientPage::processDiscoverResult(const QString &ip, int port)
{
    m_settingsBox->setEnabled(true);
    QApplication::restoreOverrideCursor();

    if (ip.isEmpty()) {
        m_display->displayMessage(tr("Es wurde kein Muckturnier-Server gefunden"));
        if (--m_currentDiscoverAttempts > 0) {
            searchServer();
        } else {
            m_socketAddress->focusIp();
        }

    } else {
        m_currentDiscoverAttempts = 0;
        m_display->displayMessage(tr("Muckturnier-Server mit IP %1 auf Port %2 gefunden").arg(
                                     ip, QString::number(port)));

        bool scopeIdChanged = false;
        QString scopeId;
        QHostAddress address(ip);
        if (address.protocol() == QAbstractSocket::IPv6Protocol && ! address.scopeId().isEmpty()) {
            // Get the possibly correct scope id for an IPv6 link-local address
            for (const QNetworkInterface &interface : QNetworkInterface::allInterfaces()) {
                const QNetworkInterface::InterfaceFlags flags = interface.flags();
                if (flags & QNetworkInterface::IsUp
                    && flags & QNetworkInterface::IsRunning
                    && flags & QNetworkInterface::CanBroadcast) {

                    scopeId = interface.name();
                    break;
                }
            }
            address.setScopeId(scopeId);
            scopeIdChanged = true;
        }

        m_socketAddress->setSocketAddress(address.toString(), port);
        QString message = tr("<p>Es wurde ein Muckturnier-Server mit der <b>IP %1</b> und "
                             "<b>Port %2</b> gefunden.").arg(ip, QString::number(port));
        if (scopeIdChanged) {
            message += tr("</p><p>Diese Adresse ist eine <b>IPv6-Link-Local-Adresse</b>, die die "
                          "passende <b>Scope-ID</b> enthalten muss. Diese wurde auf <b>„%1“</b> "
                          "gesetzt. Sollte das nicht korrekt sein, dann bitte korrigieren und "
                          "manuell verbinden.</p><p>").arg(scopeId);
        }
        message += tr("Soll jetzt eine Verbindung mit diesem Server hergestellt werden?</p>");

        if (QMessageBox::question(this, tr("Mit Server verbinden"), message,
            QMessageBox::Yes | QMessageBox::No, QMessageBox::Yes) == QMessageBox::Yes) {

            connectToServer();
        } else {
            QTimer::singleShot(0, this, [this] { m_connectButton->setFocus(); });
        }
    }
}
