/* Copyright (c) 2011-2020 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes

#include "MainWindow.h"
#include "AboutDialog.h"
#include "RankingPage.h"
#include "ExportTournamentDialog.h"
#include "TournamentInfoDialog.h"
#include "ExportPlayersDialog.h"
#include "ChecksumsDialog.h"
#include "ImportPlayersDialog.h"
#include "RestoreBackupDialog.h"
#include "RankingDialog.h"
#include "NewTournamentDialog.h"
#include "AssignmentPage.h"

#include "PlayersPage/PlayersPage.h"
#include "PlayersPage/MarkersWidget.h"
#include "ScorePage/ScorePage.h"
#include "StopWatchDialog/StopWatchDialog.h"
#include "ClientPage/Client.h"
#include "ClientPage/ClientPage.h"
#include "ServerPage/Server.h"
#include "ServerPage/ServerPage.h"
#include "SettingsDialog/SettingsDialog.h"

#include "shared/SharedStyles.h"
#include "shared/SharedObjects.h"
#include "shared/ResourceFinder.h"
#include "shared/BackupEngine.h"
#include "shared/Settings.h"
#include "shared/StringTools.h"
#include "shared/Muckturnier.h"

// Qt includes
#include <QLabel>
#include <QHBoxLayout>
#include <QVBoxLayout>
#include <QTabWidget>
#include <QMenuBar>
#include <QAction>
#include <QApplication>
#include <QStatusBar>
#include <QMessageBox>
#include <QUrl>
#include <QDesktopServices>
#include <QFileInfo>
#include <QDebug>
#include <QFileDialog>
#include <QDir>
#include <QDockWidget>
#include <QCloseEvent>
#include <QToolButton>
#include <QCheckBox>
#include <QRegularExpression>
#include <QScreen>
#include <QWhatsThis>
#include <QShortcut>
#include <QTimer>

// C++ includes
#include <functional>

MainWindow::MainWindow(SharedObjects *sharedObjects)
    : QMainWindow(),
      m_sharedObjects(sharedObjects),
      m_resourceFinder(sharedObjects->resourceFinder()),
      m_db(sharedObjects->database()),
      m_settings(sharedObjects->settings()),
      m_tournamentSettings(sharedObjects->tournamentSettings())
{
    setWindowIcon(QIcon(m_resourceFinder->find(QStringLiteral("muckturnier.svg"))));
    setWindowTitle(tr("Muckturnier"));

    connect(m_db, &Database::sqlError, this, &MainWindow::sqlError);

    // Splash Screen
    // =============

    m_centralWidget = new QWidget(this);
    setCentralWidget(m_centralWidget);
    QVBoxLayout *layout = new QVBoxLayout(m_centralWidget);

    // Banner

    QLabel *banner = new QLabel;
    banner->setPixmap(QIcon(m_resourceFinder->find(
                                QStringLiteral("banner.svg"))).pixmap(QSize(320, 80)));
    banner->setAlignment(Qt::AlignCenter);
    banner->setStyleSheet(SharedStyles::STYLE_LOGO_HEADER);
    banner->setSizePolicy(QSizePolicy::Minimum, QSizePolicy::Maximum);
    layout->addWidget(banner);
    layout->addStretch();

    // Startup buttons

    QHBoxLayout *buttonLayout = new QHBoxLayout;
    layout->addLayout(buttonLayout);

    QToolButton *newTournamentButton = new QToolButton;
    newTournamentButton->setAutoRaise(true);
    newTournamentButton->setStyleSheet(SharedStyles::STYLE_NO_BORDER);
    newTournamentButton->setToolButtonStyle(Qt::ToolButtonTextUnderIcon);
    newTournamentButton->setIcon(QIcon(m_resourceFinder->find(QStringLiteral("new.svg"))));
    newTournamentButton->setIconSize(QSize(100, 100));
    newTournamentButton->setText(tr("Ein neues Turnier starten"));
    buttonLayout->addWidget(newTournamentButton);

    QToolButton *openTournamentButton = new QToolButton;
    openTournamentButton->setAutoRaise(true);
    openTournamentButton->setStyleSheet(SharedStyles::STYLE_NO_BORDER);
    openTournamentButton->setToolButtonStyle(Qt::ToolButtonTextUnderIcon);
    openTournamentButton->setIcon(QIcon(m_resourceFinder->find(QStringLiteral("open.svg"))));
    openTournamentButton->setIconSize(QSize(100, 100));
    openTournamentButton->setText(tr("Ein Turnier öffnen"));
    buttonLayout->addWidget(openTournamentButton);

    // Footer

    QLabel *footer = new QLabel(tr("Das Programm für die Turnierleitung"));
    footer->setAlignment(Qt::AlignCenter);
    footer->setStyleSheet(SharedStyles::STYLE_LOGO_FOOTER);
    layout->addStretch();
    layout->addWidget(footer);

    // Pages and DockWidgets
    // =====================

    // PlayersPage
    // -----------

    m_playersPage = new PlayersPage(this, m_sharedObjects);
    connect(m_playersPage, &PlayersPage::statusUpdate, this, &MainWindow::updateStatus);
    connect(m_playersPage, &PlayersPage::registrationsAvailable,
            this, &MainWindow::registrationsAvailable);
    connect(m_playersPage, &PlayersPage::playersCountChanged,
            this, &MainWindow::playersCountChanged);
    connect(m_playersPage, &PlayersPage::nameEdited, this, &MainWindow::playersNameEdited);
    connect(this, &MainWindow::resetAllSearches, m_playersPage, &PlayersPage::resetSearch);

    // ScorePage
    // ---------

    m_scorePage = new ScorePage(this, m_sharedObjects);
    connect(m_scorePage, &ScorePage::statusUpdate, this, &MainWindow::updateStatus);
    connect(m_scorePage, &ScorePage::tournamentStarted,  // tournamentStarted is also emitted
            this, &MainWindow::tournamentStarted);       // if the last score has been deleted
    connect(m_scorePage, &ScorePage::scoreChanged, m_playersPage, &PlayersPage::scoreChanged);
    connect(m_playersPage, &PlayersPage::networkTournamentStarted,
            this, &MainWindow::networkTournamentStarted);
    connect(this, &MainWindow::resetAllSearches, m_scorePage, &ScorePage::resetSearch);
    connect(m_playersPage, &PlayersPage::assignmentChanged,
            m_scorePage, std::bind(&ScorePage::updateAssignment, m_scorePage, -1));

    // RankingPage
    // -----------

    m_rankingPage = new RankingPage(this, m_sharedObjects);
    connect(m_rankingPage, &RankingPage::statusUpdate, this, &MainWindow::updateStatus);
    connect(m_scorePage, &ScorePage::scoreChanged, m_rankingPage, &RankingPage::dataChanged);
    connect(this, &MainWindow::resetAllSearches, m_rankingPage, &RankingPage::resetSearch);

    // DockWidgets
    // -----------

    setDockNestingEnabled(true);

    m_playersDock = new QDockWidget(tr("Anmeldung"), this);
    m_playersDock->setWidget(m_playersPage);
    formatDockWidget(m_playersDock);
    m_playersDock->hide();

    m_scoreDock = new QDockWidget(tr("Ergebnisse"), this);
    m_scoreDock->setWidget(m_scorePage);
    formatDockWidget(m_scoreDock);
    tabifyDockWidget(m_playersDock, m_scoreDock);
    m_scoreDock->hide();

    m_rankingDock = new QDockWidget(tr("Rangliste"), this);
    m_rankingDock->setWidget(m_rankingPage);
    formatDockWidget(m_rankingDock);
    tabifyDockWidget(m_scoreDock, m_rankingDock);
    m_rankingDock->hide();

    m_allDocks = { m_playersDock, m_scoreDock, m_rankingDock };

    // Main Menu
    // =========

    // "File"
    // ------

    QMenu *fileMenu = menuBar()->addMenu(tr("Datei"));

    m_newTournamentAction = fileMenu->addAction(tr("Neues Turnier starten"));
    m_newTournamentAction->setShortcut(QKeySequence(tr("Ctrl+N")));
    connect(m_newTournamentAction, &QAction::triggered, this, &MainWindow::startNewTournament);
    connect(newTournamentButton, &QToolButton::clicked, m_newTournamentAction, &QAction::triggered);

    m_openTournamentAction = fileMenu->addAction(tr("Turnier öffnen"));
    m_openTournamentAction->setShortcut(QKeySequence(tr("Ctrl+O")));
    connect(m_openTournamentAction, &QAction::triggered,
            this, std::bind(&MainWindow::openTournament, this, QString()));
    connect(openTournamentButton, &QToolButton::clicked,
            m_openTournamentAction, &QAction::triggered);

    fileMenu->addSeparator();

    m_tournamentInfoAction = fileMenu->addAction(tr("Turnierinformationen"));
    connect(m_tournamentInfoAction, &QAction::triggered, this, &MainWindow::showTournamentInfo);
    m_tournamentInfoAction->setEnabled(false);

    m_imExportMenu = fileMenu->addMenu(tr("Import und Export"));
    m_imExportMenu->menuAction()->setEnabled(false);

    m_importPlayersListAction = m_imExportMenu->addAction(tr("Paarliste importieren"));
    connect(m_importPlayersListAction, &QAction::triggered, this, &MainWindow::importPlayersList);
    m_importPlayersListAction->setEnabled(false);

    m_exportPlayersListAction = m_imExportMenu->addAction(tr("Paarliste exportieren"));
    connect(m_exportPlayersListAction, &QAction::triggered, this, &MainWindow::exportPlayersList);
    m_exportPlayersListAction->setEnabled(false);

    m_imExportMenu->addSeparator();

    m_exportTournamentAction = m_imExportMenu->addAction(tr("Turnierdaten exportieren"));
    connect(m_rankingPage, &RankingPage::enableExport,
            m_exportTournamentAction, &QAction::setEnabled);
    connect(m_exportTournamentAction, &QAction::triggered,
            this, &MainWindow::showExportTournamentDialog);
    m_exportTournamentAction->setEnabled(false);

    fileMenu->addSeparator();

    m_closeTournamentAction = fileMenu->addAction(tr("Turnier schließen"));
    m_closeTournamentAction->setEnabled(false);
    connect(m_closeTournamentAction, &QAction::triggered, this, &MainWindow::closeTournament);

    QAction *quitAction = fileMenu->addAction(tr("Beenden"));
#ifdef Q_OS_MACOS
    quitAction->setMenuRole(QAction::QuitRole);
#endif
    quitAction->setShortcut(QKeySequence(tr("Ctrl+Q")));
    connect(quitAction, &QAction::triggered, this, &QWidget::close);

    // "View"
    // ------

    QMenu *viewMenu = menuBar()->addMenu(tr("Ansicht"));

    m_hideWindowFrameAction = viewMenu->addAction(tr("Fenstertitel und -rahmen ausblenden"));
    m_hideWindowFrameAction->setCheckable(true);
    m_hideWindowFrameAction->setShortcut(QKeySequence(tr("F11")));
    connect(m_hideWindowFrameAction, &QAction::triggered, this, &MainWindow::hideWindowFrame);

    m_hideMenubarAction = viewMenu->addAction(tr("Menüleiste ausblenden"));
    m_hideMenubarAction->setCheckable(true);
    m_hideMenubarAction->setShortcut(QKeySequence(tr("Ctrl+M")));
    connect(m_hideMenubarAction, &QAction::triggered, this, &MainWindow::hideMenuBar);

    // All menu shortcuts are disabled as soon as the menu is hidden,
    // so we also need a global shortcut to re-show it again
    m_viewMenubarShortcut = new QShortcut(QKeySequence(tr("Ctrl+M")), this);
    m_viewMenubarShortcut->setContext(Qt::ApplicationShortcut);
    m_viewMenubarShortcut->setEnabled(false);
    connect(m_viewMenubarShortcut, &QShortcut::activated, m_hideMenubarAction, &QAction::toggle);
    connect(m_viewMenubarShortcut, &QShortcut::activated, this, &MainWindow::hideMenuBar);

    viewMenu->addSeparator();

    QMenu *pagesMenu = viewMenu->addMenu(tr("Seiten arrangieren"));

    QAction *arrangeTabifiedAction = pagesMenu->addAction(tr("Als Tabs"));
    arrangeTabifiedAction->setShortcut(QKeySequence(tr("Ctrl+T")));
    connect(arrangeTabifiedAction, &QAction::triggered, this, &MainWindow::arrangeTabified);

    QAction *arrangeHorizontallyAction = pagesMenu->addAction(tr("Nebeneinander"));
    connect(arrangeHorizontallyAction, &QAction::triggered, this, &MainWindow::arrangeHorizontally);

    QAction *arrangeVerticallyAction = pagesMenu->addAction(tr("Übereinander"));
    connect(arrangeVerticallyAction, &QAction::triggered, this, &MainWindow::arrangeVertically);

    QAction *arrangeOwnWindowAction = pagesMenu->addAction(tr("Als einzelne Fenster"));
    connect(arrangeOwnWindowAction, &QAction::triggered, this, &MainWindow::arrangeOwnWindow);

    // "Network"
    // ---------

    QMenu *networkMenu = menuBar()->addMenu(tr("Netzwerk"));

    m_showServerPageAction = networkMenu->addAction(tr("Server starten"));
    connect(m_showServerPageAction, &QAction::triggered, this, &MainWindow::showServerPage);
    m_showServerPageAction->setEnabled(false);

    m_showClientPageAction = networkMenu->addAction(tr("Mit Server verbinden"));
    connect(m_showClientPageAction, &QAction::triggered, this, &MainWindow::showClientPage);
    m_showClientPageAction->setEnabled(false);

    networkMenu->addSeparator();

    m_finishRegistrationAction = networkMenu->addAction(tr("Anmeldung beenden"));
    connect(m_finishRegistrationAction, &QAction::triggered,
            m_playersPage, &PlayersPage::finishRegistration);
    m_finishRegistrationAction->setEnabled(false);

    networkMenu->addSeparator();

    m_showRankingDialogAction = networkMenu->addAction(tr("Ranglistenvergleich"));
    m_showRankingDialogAction->setShortcut(QKeySequence(tr("Ctrl+R")));
    connect(m_showRankingDialogAction, &QAction::triggered, this, &MainWindow::showRankingDialog);
    m_showRankingDialogAction->setEnabled(false);

    // "Extras"
    // --------

    QMenu *extrasMenu = menuBar()->addMenu(tr("Extras"));

    QAction *showSettingsDialogAction = extrasMenu->addAction(tr("Einstellungen"));
#ifdef Q_OS_MACOS
    showSettingsDialogAction->setMenuRole(QAction::PreferencesRole);
#endif
    connect(showSettingsDialogAction, &QAction::triggered, this, &MainWindow::showSettingsDialog);

    extrasMenu->addSeparator();

    QAction *showStopWatchDialogAction = extrasMenu->addAction(tr("Stoppuhr"));
    connect(showStopWatchDialogAction, &QAction::triggered, this, &MainWindow::showStopWatchDialog);

    extrasMenu->addSeparator();

    m_showAssignmentPageAction = extrasMenu->addAction(tr("Übersicht Auslosung"));
    connect(m_showAssignmentPageAction, &QAction::triggered, this, &MainWindow::showAssignmentPage);
    m_showAssignmentPageAction->setEnabled(false);

    extrasMenu->addSeparator();

    m_createDatabaseBackupAction = extrasMenu->addAction(tr("Backup der Datenbank anlegen"));
    m_createDatabaseBackupAction->setShortcut(QKeySequence(tr("Ctrl+B")));
    connect(m_createDatabaseBackupAction, &QAction::triggered,
            this, &MainWindow::createDatabaseBackup);
    m_createDatabaseBackupAction->setEnabled(false);

    m_showRestoreBackupDialogAction = extrasMenu->addAction(tr("Backup wiederherstellen"));
    m_showRestoreBackupDialogAction->setShortcut(QKeySequence(tr("Ctrl+W")));
    connect(m_showRestoreBackupDialogAction, &QAction::triggered,
            this, &MainWindow::triggerShowRestoreBackupDialog);
    m_showRestoreBackupDialogAction->setEnabled(false);

    extrasMenu->addSeparator();

    m_resetDatabaseAction = extrasMenu->addAction(tr("Datenbank zurücksetzen"));
    connect(m_resetDatabaseAction, &QAction::triggered, this, &MainWindow::resetDatabase);
    m_resetDatabaseAction->setEnabled(false);

    extrasMenu->addSeparator();

    m_showChecksumsDialogAction = extrasMenu->addAction(tr("Prüfsummen"));
    m_showChecksumsDialogAction->setShortcut(QKeySequence(tr("Ctrl+P")));
    connect(m_showChecksumsDialogAction, &QAction::triggered,
            this, &MainWindow::showChecksumsDialog);
    m_showChecksumsDialogAction->setEnabled(false);

    // "Help"
    // ------

    QMenu *helpMenu = menuBar()->addMenu(tr("Hilfe"));

    QAction *handbookAction = helpMenu->addAction(tr("Handbuch zu Muckturnier"));
    handbookAction->setShortcut(QKeySequence(tr("F1")));
    connect(handbookAction, &QAction::triggered, this, &MainWindow::showHandbook);

    QAction *whatsThisAction = helpMenu->addAction(tr("Was ist das?"));
    whatsThisAction->setShortcut(QKeySequence(tr("Shift+F1")));
    connect(whatsThisAction, &QAction::triggered, [] { QWhatsThis::enterWhatsThisMode(); });

    helpMenu->addSeparator();

    QAction *aboutAction = helpMenu->addAction(tr("Über Muckturnier"));
#ifdef Q_OS_MACOS
    aboutAction->setMenuRole(QAction::AboutRole);
#endif
    connect(aboutAction, &QAction::triggered, this, &MainWindow::showAboutDialog);

    // Restore or set the MainWindow geometry
    // ======================================

    const bool geometryRestored = m_settings->saveMainWindowGeometryEnabled()
                                      ? restoreGeometry(m_settings->mainWindowGeometry())
                                      : false;

    if (! geometryRestored) {
        const QRect availableGeometry = QGuiApplication::primaryScreen()->availableGeometry();
        if (availableGeometry.width() > 900 && availableGeometry.height() > 700) {
            resize(QSize(900, 700));
        } else {
            setWindowState(Qt::WindowMaximized);
        }
    }

    // Done :-)
    // ========

    updateStatus(tr("Bitte ein neues Turnier starten oder ein bestehendes öffnen!"));
}

void MainWindow::playersCountChanged(int count)
{
    const bool tournamentCanBeStarted = m_db->playersDividable(count)
                                        && ! m_playersPage->singlesPresent()
                                        && ! m_playersPage->absentsPresent();
    m_scoreDock->setVisible(tournamentCanBeStarted);
    m_scorePage->reload();
    m_finishRegistrationAction->setEnabled((m_isServer || m_isClient)
                                           && ! m_db->networkTournamentStarted()
                                           && tournamentCanBeStarted);
}

void MainWindow::playersNameEdited()
{
    if (m_scoreDock->isVisible()) {
        m_rankingPage->dataChanged();
    }
    if (m_scoreDock->isVisible()) {
        m_scorePage->playersNameEdited();
    }
}

void MainWindow::startNewTournament()
{
    NewTournamentDialog *dialog = new NewTournamentDialog(this, m_sharedObjects);
    connect(dialog, &NewTournamentDialog::updateStatus, this, &MainWindow::updateStatus);
    connect(dialog, &NewTournamentDialog::reopenLastOpenedDb,
            this, &MainWindow::reopenLastOpenedDb);
    if (! dialog->exec()) {
        return;
    }

    processDbReset();
    databaseChanged();
}

void MainWindow::openTournament(QString dbFileName)
{
    const QString lastOpenedDb = m_db->dbFile();

    if (dbFileName.isEmpty()) {
        QString startPath = m_db->dbPath();
        if (startPath.isEmpty()) {
            startPath = m_settings->lastDbPath();
        }
        dbFileName = QFileDialog::getOpenFileName(this,
            tr("Turnier öffnen"),
            startPath,
            tr("Muckturnier-Datenbanken") + QLatin1String(" (*.mtdb);;")
            + tr("Alle Dateien") + QLatin1String(" (*)"));
        if (dbFileName.isEmpty()) {
            return;
        }
    }

    if (dbFileName == lastOpenedDb) {
        QMessageBox::information(this,
            tr("Turnier öffnen"),
            tr("Die Turnierdatenbank „%1“ ist bereits geöffnet!").arg(
               QDir::toNativeSeparators(dbFileName)));
        return;
    }

    updateStatus(tr("Öffne Turnierdatenbank …"));
    QApplication::setOverrideCursor(Qt::WaitCursor);
    // Close the current database connection without touching the GUI. If no database is currently
    // opened, nothing happens. Otherwise, the GUI stays the same for now (in case the requested
    // database can't be opened and we re-open the last one).
    m_db->closeDatabaseConnection();
    Database::Error openResult = m_db->openDatabase(dbFileName);
    QApplication::restoreOverrideCursor();

    switch (openResult) {
    case Database::Error::NoError:
        // All tests passed correctly, everything is okay
        break;
    case Database::Error::NoErrorReadOnly:
        // Everything is okay, but we opened a read-only database
        QMessageBox::warning(this,
            tr("Turnier öffnen"),
            tr("<p><b>Turnierdatenbank ist nicht beschreibbar</b></p>"
               "<p>Die Datei „%1“ ist nicht beschreibbar. Die Turnierdatenbank kann nur angezeigt, "
               "aber nicht verändert werden.</p>").arg(
               QDir::toNativeSeparators(dbFileName)),
            QMessageBox::Ok);
        break;
    case Database::Error::DbVersion3:
    case Database::Error::DbVersion4:
    case Database::Error::DbVersion5:
    case Database::Error::DbVersion6:
    case Database::Error::DbVersion7:
        // We have to do a database revision update before we can use the database
        updateDatabase(lastOpenedDb, dbFileName, openResult);
        return;
    case Database::Error::CouldNotOpen:
        // Something went wrong. We can't use the requested file.
        QMessageBox::warning(this,
            tr("Turnier öffnen"),
            tr("<p><b>Das Öffnen der Turnierdatenbank ist nicht möglich:</b></p>")
            + m_db->errorMessage(),
            QMessageBox::Ok);
        if (! lastOpenedDb.isEmpty()) {
            // Re-open the last opened database. The GUI still remains unchanged, so the user
            // won't notice that the currently opened database has been closed and re-opened.
            m_db->openDatabase(lastOpenedDb);
        }
        return;
    }

    databaseChanged();

    updateStatus(tr("Turnierdatenbank %1 geöffnet").arg(QDir::toNativeSeparators(dbFileName)));
    QApplication::restoreOverrideCursor();

    // Did we just open a backup?
    if (! m_openingAfterRestore) {
        BackupEngine backupEngine(dbFileName);
        if (! backupEngine.originalBaseName().isEmpty() && askRestoreOriginalDb(backupEngine)) {
            backupEngine.setFileName(backupEngine.originalFileName());
            showRestoreBackupDialog(backupEngine, backupEngine.originalFileName(), true);
        }
    }
}

void MainWindow::updateDatabase(const QString &lastOpenedDb, const QString &dbFileName,
                                Database::Error versionError)
{
    if (! m_db->isWritable()) {
        updateStatus(tr("Datenbankupdate nicht möglich"));
        QMessageBox::warning(this,
            tr("Turnier öffnen"),
            tr("<p><b>Das Datenbankformat von Muckturnier hat sich geändert.</b></p>"
               "<p>Um die Turnierdatenbank „%1“ benutzen zu können, müsste ein Update der "
               "Datenbank durchgeführt werden, aber die Datei ist nicht beschreibbar.</p>"
               "<p>Bitte die Zugriffsrechte überprüfen, und die Datei erneut laden!</p>").arg(
               QDir::toNativeSeparators(dbFileName)));
        reopenLastOpenedDb(lastOpenedDb);
        return;
    }

    QMessageBox messageBox(QMessageBox::Question,
        tr("Update des Datenbankformats"),
        tr("<p><b>Das Datenbankformat von Muckturnier hat sich geändert.</b></p>"
           "<p>Soll die Turnierdatenbank „%1“ aktualisiert werden? Es wird dabei ein Backup der "
           "Original-Datenbank im selben Ordner angelegt.</p>").arg(
           QDir::toNativeSeparators(dbFileName)),
        QMessageBox::Cancel, this);
    QAbstractButton *yesButton = messageBox.addButton(tr("Aktualisieren"), QMessageBox::YesRole);
    messageBox.exec();
    if (messageBox.clickedButton() != yesButton) {
        updateStatus(tr("Datenbankupdate abgebrochen!"));
        QMessageBox::warning(this,
            tr("Update des Datenbankformats"),
            tr("<p><b>Das Update des Datenbankformats wurde abgebrochen.</b></p>"
               "<p>Die Turnierdatenbank „%1“ kann nicht geöffnet werden.</p>").arg(
               QDir::toNativeSeparators(dbFileName)));
        reopenLastOpenedDb(lastOpenedDb);
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    updateStatus(tr("Datenbankupdate läuft …"));

    int originalVersion = 0;
    switch (versionError) {
    case Database::Error::NoError:
    case Database::Error::NoErrorReadOnly:
    case Database::Error::CouldNotOpen:
        Q_UNREACHABLE();
        break;
    case Database::Error::DbVersion3:
        originalVersion = 3;
        break;
    case Database::Error::DbVersion4:
        originalVersion = 4;
        break;
    case Database::Error::DbVersion5:
        originalVersion = 5;
        break;
    case Database::Error::DbVersion6:
        originalVersion = 6;
        break;
    case Database::Error::DbVersion7:
        originalVersion = 7;
        break;
    }

    BackupEngine backupEngine(dbFileName);
    if (! backupEngine.createBackup(originalVersion)) {
        updateStatus(tr("Datenbankupdate abgebrochen!"));
        QApplication::restoreOverrideCursor();
        QMessageBox::critical(this,
            tr("Update des Datenbankformats"),
            tr("<p><b>Die Backup-Datei konnte nicht angelegt werden.</b></p>"
               "<p>Das Update wird abgebrochen. Bitte die Zugriffsrechte (des Verzeichnisses „%2“) "
               "überprüfen!</p>").arg(QDir::toNativeSeparators(backupEngine.path())));
        reopenLastOpenedDb(lastOpenedDb);
        return;
    }

    if (! m_db->updateDatabase(dbFileName, versionError)) {
        updateStatus(tr("Datenbankupdate fehlgeschlagen!"));
        QApplication::restoreOverrideCursor();
        QMessageBox::critical(this,
            tr("Update des Datenbankformats"),
            tr("<p><b>Das Update der Datenbank „%1“ ist fehlgeschlagen!</b></p>"
               "<p>Das sollte nicht passieren. Entweder ist die Datei keine gültige "
               "Muckturnier-Turnierdatenbank, oder sie wurde während des Updates von einem anderen "
               "Prozess verändert.</p>%2").arg(
               QDir::toNativeSeparators(dbFileName), m_db->errorMessage()));
        reopenLastOpenedDb(lastOpenedDb);
        return;
    }

    updateStatus(tr("Datenbankupdate beendet!"));
    QApplication::restoreOverrideCursor();

    QMessageBox::information(this,
        tr("Update des Datenbankformats"),
        tr("<p><b>Die Datenbank „%1“ wurde erfolgreich aktualisiert!</b></p>"
           "<p>Ein Backup der Originaldatei wurde mit dem Namen „%2“ erstellt.</p>").arg(
           QDir::toNativeSeparators(dbFileName),
           QDir::toNativeSeparators(backupEngine.backupFile())),
        QMessageBox::Ok);

    openTournament(dbFileName);
}

void MainWindow::reopenLastOpenedDb(const QString &lastOpenedDb)
{
    closeTournament();
    if (! lastOpenedDb.isEmpty()) {
        openTournament(lastOpenedDb);
    }
}

void MainWindow::showHandbook()
{
    const QString readmePath = m_resourceFinder->find(QStringLiteral("Readme.htm"));

    if (readmePath.isEmpty()) {
        QMessageBox::warning(this, tr("Handbuch anzeigen"),
        tr("<p>Das Handbuch kann nicht angezeigt werden: Die Datei „Readme.htm“ wurde nicht "
           "gefunden.</p>"
           "<p>Online ist unter <a href=\"http://muckturnier.org/readme/\">"
           "http://muckturnier.org/readme/</a> das aktuelle Handbuch zu finden!</p>"));
        return;
    }

    if (! QDesktopServices::openUrl(QUrl::fromLocalFile(readmePath))) {
        QMessageBox::warning(this, tr("Handbuch anzeigen"),
        tr("<p>Das Handbuch kann nicht angezeigt werden: Die Datei „%1“ konnte nicht mit einem "
           "Browser geöffnet werden.<p>"
           "<p>Bitte die Readme-Datei manuell öffnen oder unter "
           "<a href=\"http://muckturnier.org/readme/\">"
           "http://muckturnier.org/readme/</a> online lesen!</p>").arg(readmePath));
    }
}

void MainWindow::showAboutDialog()
{
    QDialog *aboutDialog = new AboutDialog(this, m_sharedObjects);
    aboutDialog->show();
}

void MainWindow::updateStatus(const QString &message)
{
    statusBar()->showMessage(message, 5000);
}

void MainWindow::databaseChanged()
{
    setWindowTitle(tr("Muckturnier: %1").arg(QFileInfo(m_db->dbFile()).fileName()));
    m_centralWidget->hide();
    m_playersDock->show();

    if (m_settings->saveLastDbPathEnabled()) {
        m_settings->saveLastDbPath(m_db->dbPath());
    }

    // Setup the ScorePage, hide/show it as needed
    m_scorePage->setupGui();
    if (m_db->tournamentStarted()) {
        m_scoreDock->show();
        m_scoreDock->raise();
    } else {
        m_scoreDock->setVisible(m_db->playersDividable()
                                && ! m_playersPage->absentsPresent()
                                && ! m_playersPage->singlesPresent());
        m_playersDock->raise();
    }

    // Setup the MarkersWidget
    MarkersWidget *markersWidget = m_playersPage->markersWidget();
    // First, we block the markersChanged() signal so that it won't trigger
    // PlayersPage::updatePlayersList here (first, we have to load the singles management settings)
    markersWidget->blockSignals(true);
    // Then we reload the MarkersWidget so that we have the current markers definitions
    markersWidget->reload();
    // Here, we reset the visibility settings and load the singles management settings
    markersWidget->reset();
    markersWidget->blockSignals(false);

    // Update the PlayersPage
    m_playersPage->setColumnsVisible(m_tournamentSettings->playersPageColumns());
    // Block signals so that the playersCountChanged signal is not emitted yet, as we want to run
    // MainWindow::playersCountChanged in each case (also if the count is the same, because another
    // database could have the same number of players like the currently opened one, and in this
    // case the signal would not be emitted, and the ScorePage update would not be triggered).
    m_playersPage->blockSignals(true);
    // Reload, but suppress the MarkersWidget reload (as this has been already done)
    m_playersPage->reload(false);
    m_playersPage->blockSignals(false);
    playersCountChanged(m_db->pairsOrPlayersCount()); // This also reloads the ScorePage

    // Enable/Disable menu actions
    registrationsAvailable(m_db->pairsOrPlayersCount() > 0);
    m_exportTournamentAction->setEnabled(false);
    m_closeTournamentAction->setEnabled(true);
    m_tournamentInfoAction->setEnabled(true);
    m_createDatabaseBackupAction->setEnabled(true);
    m_showRestoreBackupDialogAction->setEnabled(m_db->isWritable());
    m_showAssignmentPageAction->setEnabled(true);
    m_showServerPageAction->setEnabled(m_db->isWritable()
                                       && m_serverDock == nullptr && m_clientDock == nullptr);
    m_showClientPageAction->setEnabled(m_db->isWritable()
                                       && m_serverDock == nullptr && m_clientDock == nullptr);

    if (m_serverDock != nullptr && ! m_db->isWritable()) {
        QMessageBox::warning(this, tr("Netzwerk"),
            tr("<p><b>Turnierdatenbank ist nicht beschreibbar</b></p>"
               "<p>Die Netzwerkfunktionen können nur mit einer beschreibbaren Datenbank genutzt "
               "werden. Die Server-Seite wird geschlossen.</p>"));
        m_serverDock->deleteLater();
        serverPageClosed();
    } else if (m_clientDock != nullptr && ! m_db->isWritable()) {
        QMessageBox::warning(this, tr("Netzwerk"),
            tr("<p><b>Turnierdatenbank ist nicht beschreibbar</b></p>"
               "<p>Die Netzwerkfunktionen können nur mit einer beschreibbaren Datenbank genutzt "
               "werden. Die Client-Seite wird geschlossen.</p>"));
        m_clientDock->deleteLater();
        clientPageClosed();
    }

    m_imExportMenu->menuAction()->setEnabled(true);

    checkEnableRankingPage();
    if (m_db->tournamentStarted()) {
        m_rankingPage->dataChanged();
        m_importPlayersListAction->setEnabled(false);
    } else {
        m_importPlayersListAction->setEnabled(m_db->isWritable());
    }

    switch (m_tournamentSettings->tournamentMode()) {
    case Muckturnier::TournamentMode::FixedPairs:
        m_exportPlayersListAction->setText(tr("Paarliste exportieren"));
        m_importPlayersListAction->setText(tr("Paarliste importieren"));
        break;
    case Muckturnier::TournamentMode::SinglePlayers:
        m_exportPlayersListAction->setText(tr("Spielerliste exportieren"));
        m_importPlayersListAction->setText(tr("Spielerliste importieren"));
        break;
    }

    m_resetDatabaseAction->setEnabled(m_db->isWritable());

    // Restore the state of the options we save in the database
    m_scorePage->setNoTableNumbers(m_tournamentSettings->noTableNumbers());
    m_scorePage->setAutoSelectPairs(m_tournamentSettings->autoSelectPairs());
    m_rankingPage->setIncludeOpponentGoals(m_tournamentSettings->includeOpponentGoals());

    // Clear all possibly entered searches
    emit resetAllSearches();

    // Warn about special markers
    if (m_db->isWritable()
        && (m_tournamentSettings->absentMarker() != -1
            || m_tournamentSettings->singlesManagement().enabled)) {

        markersWidget->warnAboutSpecialMarkers();
    }
}

void MainWindow::settingsSaved()
{
    m_playersDock->show();
    m_playersPage->reload();
    m_playersDock->raise();
    m_tournamentInfoAction->setEnabled(true);
    m_createDatabaseBackupAction->setEnabled(true);
    m_showRestoreBackupDialogAction->setEnabled(m_db->isWritable());
}

void MainWindow::tournamentStarted()
{
    m_playersPage->reload();
    checkEnableRankingPage();
    m_importPlayersListAction->setEnabled(! m_db->tournamentStarted());
}

void MainWindow::checkEnableRankingPage()
{
    m_rankingDock->setVisible(m_db->localTournamentStarted());
}

void MainWindow::resizeEvent(QResizeEvent *event)
{
    QMainWindow::resizeEvent(event);
    if (m_scoreDock->isVisible()) {
        m_scorePage->setScrollAreaHeight();
    }
}

void MainWindow::showExportTournamentDialog()
{
    ExportTournamentDialog *dialog= new ExportTournamentDialog(this, m_sharedObjects);
    connect(dialog, &ExportTournamentDialog::updateStatus, this, &MainWindow::updateStatus);
    dialog->exec();
}

void MainWindow::showStopWatchDialog()
{
    if (findChildren<StopWatchDialog *>().count() > 0
        && QMessageBox::question(this,
               tr("Stoppuhr anzeigen"),
               tr("Es ist bereits mindestens eine Stoppuhr geöffnet. Soll wirklich eine weitere "
                  "angezeigt werden?"),
               QMessageBox::Yes | QMessageBox::Cancel, QMessageBox::Cancel)
           == QMessageBox::Cancel) {

        return;
    }

    StopWatchDialog *dialog = new StopWatchDialog(this);
    dialog->show();
}

void MainWindow::showTournamentInfo()
{
    TournamentInfoDialog *dialog = new TournamentInfoDialog(this, m_sharedObjects);
    dialog->exec();
}

void MainWindow::formatDockWidget(QDockWidget *dock)
{
    dock->setContextMenuPolicy(Qt::PreventContextMenu);
    dock->setFeatures(QDockWidget::DockWidgetMovable | QDockWidget::DockWidgetFloatable);
    dock->setStyleSheet(QStringLiteral("QDockWidget { font-weight: bold; font-size: %1pt; }").arg(
                                       (int) double(dock->font().pointSize()) * 1.2));
    addDockWidget(Qt::TopDockWidgetArea, dock);
}

void MainWindow::closeTournament()
{
    if (! m_db->saveSettings()) {
        return;
    }
    m_db->closeDatabaseConnection();

    m_playersDock->hide();
    m_scoreDock->hide();
    m_rankingDock->hide();
    m_centralWidget->show();

    m_closeTournamentAction->setEnabled(false);
    m_tournamentInfoAction->setEnabled(false);
    m_createDatabaseBackupAction->setEnabled(false);
    m_showRestoreBackupDialogAction->setEnabled(false);
    m_resetDatabaseAction->setEnabled(false);
    m_imExportMenu->menuAction()->setEnabled(false);
    m_exportPlayersListAction->setEnabled(false);
    m_exportTournamentAction->setEnabled(false);
    m_importPlayersListAction->setEnabled(false);
    m_showServerPageAction->setEnabled(false);
    m_showClientPageAction->setEnabled(false);
    m_showRankingDialogAction->setEnabled(false);
    m_showChecksumsDialogAction->setEnabled(false);
    m_showAssignmentPageAction->setEnabled(true);

    if (m_serverDock != nullptr) {
        m_serverDock->deleteLater();
    }
    if (m_clientDock != nullptr) {
        m_clientDock->deleteLater();
    }

    setWindowTitle(tr("Muckturnier"));
    updateStatus(tr("Bitte ein neues Turnier starten oder ein bestehendes öffnen!"));
}

void MainWindow::exportPlayersList()
{
    if (m_isClient || m_isServer) {
        QString message = tr("<p><b>Netzwerk läuft</b></p>"
                             "<p>Während das Netzwerk läuft, ist kein Listenexport möglich. ");
        if (m_isClient) {
            message.append(tr("Bitte die Verbindung zum Server beenden und den Export erneut "
                              "anfragen.</p>"));
        } else {
            message.append(tr("Bitte den Server stoppen und den Export erneut anfragen.</p>"));
        }

        QMessageBox::information(this, m_exportPlayersListAction->text(), message);
        return;
    }

    ExportPlayersDialog *dialog = new ExportPlayersDialog(this, m_sharedObjects);
    connect(dialog, &ExportPlayersDialog::updateStatus, this, &MainWindow::updateStatus);
    dialog->exec();
    return;
}

void MainWindow::hideWindowFrame()
{
    if (m_hideWindowFrameAction->isChecked()) {
        setWindowFlags(Qt::FramelessWindowHint);
    } else {
        setWindowFlags(windowFlags() & ~Qt::FramelessWindowHint);
    }
    show();
}

void MainWindow::hideMenuBar()
{
    const bool visible = ! m_hideMenubarAction->isChecked();

    menuBar()->setVisible(visible);
    m_viewMenubarShortcut->setEnabled(! visible);

    if (! visible) {
        QMessageBox::information(this, tr("Menüleiste ausgeblendet"),
            tr("<p>Die Menüleiste kann mit <b>Strg+M</b> wieder eingeblendet werden.</p>"
               "<p>Solang die Menüleiste ausgeblendet ist, sind alle Menü-Tastaturkürzel "
               "deaktiviert.</p>"));
    }
}

void MainWindow::attachDocks(Qt::DockWidgetArea area)
{
    for (QDockWidget *dock : m_allDocks) {
        dock->setFloating(false);
        addDockWidget(area, dock);
    }
}

void MainWindow::arrangeTabified()
{
    attachDocks(Qt::TopDockWidgetArea);
    for (int i = 1; i < m_allDocks.count(); i++) {
        tabifyDockWidget(m_allDocks.at(i - 1), m_allDocks.at(i));
    }
}

void MainWindow::arrangeHorizontally()
{
    attachDocks(Qt::TopDockWidgetArea);
    for (int i = 1; i < m_allDocks.count(); i++) {
        splitDockWidget(m_allDocks.at(i - 1), m_allDocks.at(i), Qt::Horizontal);
    }
}

void MainWindow::arrangeVertically()
{
    attachDocks(Qt::LeftDockWidgetArea);
    for (int i = 1; i < m_allDocks.count(); i++) {
        splitDockWidget(m_allDocks.at(i - 1), m_allDocks.at(i), Qt::Vertical);
    }
}

void MainWindow::arrangeOwnWindow()
{
    for (QDockWidget *dock : m_allDocks) {
        dock->setFloating(true);
    }
}

void MainWindow::processDbReset()
{
    m_playersDock->hide();
    m_playersPage->markersWidget()->reset();
    m_playersPage->reset();
    m_scoreDock->hide();
    m_rankingDock->hide();
    m_importPlayersListAction->setEnabled(true);
    m_exportPlayersListAction->setEnabled(false);
    m_exportTournamentAction->setEnabled(false);
}

void MainWindow::resetDatabase()
{
    QCheckBox *createBackup = new QCheckBox(tr("Vorher ein Backup der Datenbank anlegen"));
    createBackup->setChecked(true);
    QMessageBox messageBox(this);
    messageBox.setIcon(QMessageBox::Warning);
    messageBox.setWindowTitle(tr("Datenbank zurücksetzen"));
    messageBox.setText(tr("Sollen wirklich alle angemeldeten %1 und alle eingegebenen "
                          "Ergebnisse gelöscht werden?").arg(
                          m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler")));
    messageBox.setStandardButtons(QMessageBox::Yes | QMessageBox::Cancel);
    messageBox.setDefaultButton(QMessageBox::Cancel);
    messageBox.setCheckBox(createBackup);

    if (messageBox.exec() == QMessageBox::Cancel) {
        return;
    }

    if (createBackup->isChecked()) {
        if (! createDatabaseBackup()) {
            QMessageBox::warning(this,
                tr("Datenbank zurücksetzen"),
                tr("Das angefragte Datenbankbackup konnte nicht angelegt werden. Das Zurücksetzen "
                   "der Datenbank wird abgebrochen."));
            return;
        }
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    updateStatus(tr("Setzte Turnierdatenbank zurück …"));

    if (! m_db->reset()) {
        return;
    }

    if (! m_db->addMarkers(m_settings->markersTemplate())) {
        return;
    }

    m_tournamentSettings->setDefaultMarker(m_settings->defaultMarker());

    const int absentMarker = m_settings->absentMarker();
    if (absentMarker == -1) {
        // We need to explicitely set the absent marker if it's disabled, so that "(unmarked)" will
        // be selected in the disabled group box instead of a probably other setting being chosen
        // before.
        m_playersPage->markersWidget()->setAbsentMarker(0);
    }
    m_tournamentSettings->setAbsentMarker(absentMarker);

    m_tournamentSettings->setSinglesManagement(m_settings->singlesManagementTemplate());

    processDbReset();
    settingsSaved();

    QApplication::restoreOverrideCursor();
    updateStatus(tr("Turnierdatenbank zurückgesetzt."));

    QMessageBox::information(this, tr("Datenbank zurücksetzen"),
                             tr("Die Datenbank wurde zurückgesetzt."));
}

void MainWindow::importPlayersList()
{
    if (m_isClient || m_isServer) {
        QString message = tr("<p><b>Netzwerk läuft</b></p>"
                             "<p>Während eine Netzwerkverbindung bzw. der Server läuft, ist kein "
                             "Listenimport möglich.</p>"
                             "<p>Bitte den Server stoppen, die Liste am Server importieren und "
                             "danach alle Clients neu verbinden (die übernehmen dann die neuen "
                             "Anmeldungen automatisch).</p>");
        if (m_isClient) {
            message.append(tr("<p><b>Vorsicht:</b> Dieser Rechner läuft als Client. Alle "
                              "Änderungen an der Anmeldungsliste, die hier „offline“ durchgeführt "
                              "werden, werden beim nächsten Verbinden mit dem Server verworfen!"
                              "</p>"));
        }
        QMessageBox::information(this, m_importPlayersListAction->text(), message);
        return;
    }

    ImportPlayersDialog *dialog = new ImportPlayersDialog(this, m_sharedObjects);
    connect(dialog, &ImportPlayersDialog::updateStatus, this, &MainWindow::updateStatus);

    if (! dialog->exec()) {
        updateStatus(tr("Import abgebrochen"));
    } else {
        m_playersPage->updatePlayersList();
        updateStatus(tr("Import beendet"));
    }
}

void MainWindow::sqlError(const QString &text)
{
    // This should not happen

    while (QApplication::overrideCursor() != nullptr) {
        QApplication::restoreOverrideCursor();
    }

    QTimer::singleShot(0, this, [this, text] {
        // Close the server connection if we're connected to a server
        if (m_clientDock != nullptr) {
            qobject_cast<ClientPage *>(m_clientDock->widget())->closePage();
        }

        // Close the server if are a server
        if (m_serverDock != nullptr) {
            qobject_cast<ServerPage *>(m_serverDock->widget())->closePage();
        }

        QMessageBox::critical(this, tr("Fehler in der Datenbankkommunikation"),
            tr("<p><b>Fehler in der Datenbankkommunikation</b></p>"
            "<p>Es ist ein unerwarteter Fehler während einer Datenbankabfrage aufgetreten. Das "
            "sollte nicht passieren!</p>"
            "<p>Um einen Datenverlust zu vermeiden, wird versucht, den letzten konsistenten "
            "Zustand der Datenbank wiederherzustellen und die Datenbank geschlossen.</p>")
            + text);

        closeTournament();
    });
}

void MainWindow::closeEvent(QCloseEvent *event)
{
    // Do we have a running server?
    if (m_serverDock != nullptr
        && ! qobject_cast<ServerPage *>(m_serverDock->widget())->checkServerRunning()) {

        event->ignore();
        return;
    }

    // Do we have running stopwatches?
    bool stopWatchRunning = false;
    for (StopWatchDialog *stopWatchDialog : findChildren<StopWatchDialog *>()) {
        if (stopWatchDialog->isRunning()) {
            stopWatchRunning = true;
            break;
        }
    }
    if (stopWatchRunning && QMessageBox::warning(this,
            tr("Muckturnier beenden"),
            tr("<p>Es läuft mindestens eine Stoppuhr! Alle Stoppuhren werden beim Schließen "
               "ebenfalls geschlossen, und die Zeitinformation geht verloren!</p>"
               "<p>Soll Muckturnier trotzdem geschlossen werden?</p>"),
            QMessageBox::Yes | QMessageBox::Cancel, QMessageBox::Cancel)  == QMessageBox::Cancel) {

        event->ignore();
        return;
    }

    // Save the window position and state
    if (m_settings->saveMainWindowGeometryEnabled()) {
        m_settings->saveMainWindowGeometry(saveGeometry());
    }

    // If a database is currently opened, save the current settings and close it
    if (m_db->isOpen()) {
        m_db->saveSettings();
        m_db->closeDatabaseConnection();
    }

    // Accept the close
    event->accept();
}

bool MainWindow::createDatabaseBackup()
{
    const QString openedDb = m_db->dbFile();
    BackupEngine backupEngine(openedDb);

    if (! backupEngine.originalBaseName().isEmpty()
        && QMessageBox::question(this,
            tr("Backup der Datenbank anlegen"),
            tr("<p>Es sieht so aus, als wäre die geöffnete Turnierdatenbank<br/>"
               "<b>%1.mtdb</b><br/>"
               "ein Backup von<br/>"
               "<b>%2.mtdb</b><br/>"
               "das %3 erstellt wurde. Soll wirklich ein Backup von der Backup-Datei erstellt "
               "werden?</p>").arg(
               backupEngine.baseName(),
               backupEngine.originalBaseName(),
               backupEngine.messageDate()),
            QMessageBox::Yes | QMessageBox::No) == QMessageBox::No) {

        return false;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    updateStatus(tr("Backup wird angelegt …"));

    m_db->closeDatabaseConnection();
    bool success = backupEngine.createBackup();
    m_db->openDatabase(openedDb);

    QApplication::restoreOverrideCursor();

    if (! success) {
        updateStatus(tr("Backup konnte nicht erstellt werden!"));
        QMessageBox::critical(this,
            tr("Backup der Datenbank anlegen"),
            tr("<p><b>Das Anlegen eines Backups ist fehlgeschlagen!</b></p>"
               "<p>Die Datei konnte nicht angelegt werden. Bitte die Zugriffsrechte von „%1“ bzw. "
               "dem Verzeichnis „%2“) überprüfen!</p>").arg(
               QDir::toNativeSeparators(backupEngine.backupFile()),
               QDir::toNativeSeparators(backupEngine.path())),
            QMessageBox::Ok);
        return false;
    }

    updateStatus(tr("Backup erfolgreich erstellt!"));
    QMessageBox::information(this,
        tr("Backup der Datenbank anlegen"),
        tr("Es wurde erfolgreich ein Backup der Datenbank mit dem Dateinamen „%1“ erstellt.").arg(
           QDir::toNativeSeparators(backupEngine.backupFile())),
        QMessageBox::Ok);

    return true;
}

bool MainWindow::askRestoreOriginalDb(BackupEngine &backupEngine)
{
    if (backupEngine.originalBaseName().isEmpty()) {
        return false;
    }

    if (! m_askRestoreOriginalDb) {
        return m_restoreOriginalDb;
    }

    QMessageBox messageBox(QMessageBox::Question,
        tr("Backup wiederherstellen"),
        tr("<p><b>Soll ein Backup wiederhergestellt werden?</b></p>"
           "<p>Es sieht so aus, als wäre die soeben geöffnete Turnierdatenbank<br/>"
           "<b>%1.mtdb</b><br/>"
           "ein Backup von<br/>"
           "<b>%2.mtdb</b><br/>"
           "das %3 erstellt wurde. Soll das Backup wiederhergestellt werden?</p>").arg(
           backupEngine.baseName(),
           backupEngine.originalBaseName(),
           backupEngine.messageDate()),
        QMessageBox::Yes | QMessageBox::No, this);
    QCheckBox *dontAskAgain = new QCheckBox(tr("Nicht mehr nachfragen"));
    messageBox.setCheckBox(dontAskAgain);

    m_restoreOriginalDb = messageBox.exec() == QMessageBox::Yes;
    m_askRestoreOriginalDb = ! dontAskAgain->isChecked();

    return m_restoreOriginalDb;
}

void MainWindow::triggerShowRestoreBackupDialog()
{
    BackupEngine backupEngine(m_db->dbFile());
    if (askRestoreOriginalDb(backupEngine)) {
        backupEngine.setFileName(backupEngine.originalFileName());
        showRestoreBackupDialog(backupEngine, backupEngine.originalFileName(), false);
    } else {
        showRestoreBackupDialog(backupEngine, m_db->dbFile(), false);
    }
}

void MainWindow::showRestoreBackupDialog(BackupEngine &backupEngine, const QString &dbToOverwrite,
                                         bool autoAccept)
{
    RestoreBackupDialog *dialog = new RestoreBackupDialog(this, m_db, backupEngine, dbToOverwrite,
                                                          autoAccept);
    connect(dialog, &RestoreBackupDialog::statusUpdate, this, &MainWindow::updateStatus);
    connect(dialog, &RestoreBackupDialog::closeTournament, this, &MainWindow::closeTournament);
    connect(dialog, &RestoreBackupDialog::openTournament, this, &MainWindow::openRestoredDb);
    dialog->exec();
}

void MainWindow::openRestoredDb(const QString &dbFile)
{
    m_openingAfterRestore = true;
    openTournament(dbFile);
    m_openingAfterRestore = false;
}

void MainWindow::registrationsAvailable(bool state)
{
    m_exportPlayersListAction->setEnabled(state);
    m_showChecksumsDialogAction->setEnabled(state);
}

void MainWindow::showChecksumsDialog()
{
    ChecksumsDialog *dialog = new ChecksumsDialog(this, m_sharedObjects);
    dialog->exec();
}

void MainWindow::showSettingsDialog()
{
    const bool resetSingleConditionString = m_playersPage->markersWidget()->singleConditionString()
                                            == m_settings->namesSeparator().simplified();

    SettingsDialog *dialog = new SettingsDialog(this, m_sharedObjects);
    if (! dialog->exec()) {
        return;
    }

    m_db->setNamesSeparator(m_settings->namesSeparator());
    m_sharedObjects->stringTools()->setBoogerSymbol(m_settings->boogerSymbol());
    if (resetSingleConditionString) {
        m_playersPage->markersWidget()->resetSingleConditionString();
    }
    if (m_scoreDock->isVisible()) {
        // This will reload the score page and preserve possibly entered data
        m_scorePage->playersNameEdited();
    }
}

void MainWindow::showServerPage()
{
    m_showServerPageAction->setEnabled(false);
    m_showClientPageAction->setEnabled(false);

    ServerPage *serverPage = new ServerPage(this, m_sharedObjects, m_playersPage);
    connect(serverPage, &ServerPage::serverStarted, this, &MainWindow::serverStarted);
    connect(serverPage, &ServerPage::serverStopped, this, &MainWindow::serverStopped);
    connect(serverPage, &ServerPage::statusUpdate, this, &MainWindow::updateStatus);
    connect(serverPage, &ServerPage::pageClosed, this, &MainWindow::serverPageClosed);

    m_serverDock = new QDockWidget(tr("Netzwerk (Server)"), this);
    m_allDocks.append(m_serverDock);
    m_serverDock->setWidget(serverPage);
    formatDockWidget(m_serverDock);
    tabifyDockWidget(m_playersDock, m_serverDock);
    m_serverDock->show();
    m_serverDock->raise();
}

void MainWindow::checkDisplayFinishRegistrationInfo()
{
    if (m_db->networkTournamentStarted() || m_networkFinishRegistrationWarningAlreadyShown) {
        return;
    }

    QMessageBox::information(this,
        tr("Eingabe von Ergebnissen im Netzwerk"),
        tr("<p>Im Netzwerk-Modus muss die Anmeldephase explizit per per <i>„Netzwerk“ → „Anmeldung "
           "beenden“</i> beendet werden, damit Ergebnisse eingegeben werden können.</p>"
           "<p>Vorher ist die Ergebnisse-Seite bei allen Netzwerkteilnehmern deaktiviert.</p>"));

    m_networkFinishRegistrationWarningAlreadyShown = true;
}

void MainWindow::serverStarted(Server *server)
{
    m_db->setNetworkTournamentStarted(m_db->tournamentStarted());
    m_playersPage->setServer(server);
    setNetworkActionsEnabled(false);
    m_isServer = true;
    m_finishRegistrationAction->setEnabled(! m_db->networkTournamentStarted()
                                           && m_db->playersDividable()
                                           && ! m_playersPage->singlesPresent()
                                           && ! m_playersPage->absentsPresent());
    m_scorePage->setEnabled(m_db->networkTournamentStarted());
    m_showRankingDialogAction->setEnabled(m_db->networkTournamentStarted());
    checkDisplayFinishRegistrationInfo();
}

void MainWindow::serverStopped()
{
    m_playersPage->serverStopped();
    setNetworkActionsEnabled(true);
    m_isServer = false;
    m_finishRegistrationAction->setEnabled(false);
    m_db->setNetworkTournamentStarted(false);
    m_playersPage->reload();
    m_scorePage->setEnabled(true);
    m_showRankingDialogAction->setEnabled(false);
}

void MainWindow::showClientPage()
{
    m_showServerPageAction->setEnabled(false);
    m_showClientPageAction->setEnabled(false);

    ClientPage *clientPage = new ClientPage(this, m_sharedObjects, m_playersPage,
                                            m_lastServerIp, m_lastServerPort);
    connect(clientPage, &ClientPage::nameEdited, m_playersPage, &PlayersPage::nameEdited);
    connect(clientPage, &ClientPage::namesDifferencesDisplayed,
            this, &MainWindow::setNameAdoptionMode);
    connect(clientPage, &ClientPage::connectionRequested,
            this, &MainWindow::serverConnectionRequested);
    connect(clientPage, &ClientPage::connected, this, &MainWindow::clientConnected);
    connect(clientPage, &ClientPage::connectionClosed, this, &MainWindow::clientConnectionClosed);
    connect(clientPage, &ClientPage::statusUpdate, this, &MainWindow::updateStatus);
    connect(clientPage, &ClientPage::pageClosed, this, &MainWindow::clientPageClosed);

    m_clientDock = new QDockWidget(tr("Netzwerk (Client)"), this);
    connect(clientPage, &ClientPage::raiseMe, m_clientDock, &QDockWidget::raise);

    m_allDocks.append(m_clientDock);
    m_clientDock->setWidget(clientPage);
    formatDockWidget(m_clientDock);
    tabifyDockWidget(m_playersDock, m_clientDock);
    m_clientDock->show();
    m_clientDock->raise();
}

void MainWindow::serverConnectionRequested(const QString &ip, int port)
{
    m_lastServerIp = ip;
    m_lastServerPort = port;
}

void MainWindow::clientConnected(Client *client)
{
    m_playersPage->setClient(client);
    m_scorePage->reload();
    setNetworkActionsEnabled(false);
    m_isClient = true;
    m_finishRegistrationAction->setEnabled(! m_db->networkTournamentStarted()
                                           && m_db->playersDividable()
                                           && ! m_playersPage->singlesPresent()
                                           && ! m_playersPage->absentsPresent());
    m_scorePage->setEnabled(m_db->networkTournamentStarted());
    m_showRankingDialogAction->setEnabled(m_db->networkTournamentStarted());
    checkDisplayFinishRegistrationInfo();
}

void MainWindow::clientConnectionClosed()
{
    m_playersPage->clientDisconnected();
    setNetworkActionsEnabled(true);
    m_isClient = false;
    m_db->setNetworkTournamentStarted(false);
    m_sharedObjects->stringTools()->resetLocale();
    m_playersPage->reload();
    m_finishRegistrationAction->setEnabled(false);
    m_scorePage->setEnabled(true);
    m_showRankingDialogAction->setEnabled(false);
}

void MainWindow::serverPageClosed()
{
    m_allDocks.removeOne(m_serverDock);
    m_serverDock = nullptr;
    networkPageClosed();
}

void MainWindow::clientPageClosed()
{
    m_allDocks.removeOne(m_clientDock);
    m_clientDock = nullptr;
    m_sharedObjects->stringTools()->resetLocale();
    networkPageClosed();
}

void MainWindow::networkPageClosed()
{
    m_showServerPageAction->setEnabled(true);
    m_showClientPageAction->setEnabled(true);
    setNetworkActionsEnabled(true);
    m_isServer = false;
    m_isClient = false;
    m_finishRegistrationAction->setEnabled(false);
    m_db->setNetworkTournamentStarted(false);
    m_playersPage->reload();
    m_scorePage->setEnabled(true);
}

void MainWindow::setNameAdoptionMode(bool state)
{
    menuBar()->setEnabled(! state);
    for (QDockWidget *dock : m_allDocks) {
        if (dock != m_clientDock) {
            dock->setEnabled(! state);
        }
    }
}

void MainWindow::setNetworkActionsEnabled(bool state)
{
    m_resetDatabaseAction->setEnabled(state);
    m_showRestoreBackupDialogAction->setEnabled(state);
    m_newTournamentAction->setEnabled(state);
    m_openTournamentAction->setEnabled(state);
    m_closeTournamentAction->setEnabled(state);
}

void MainWindow::showRankingDialog()
{
    if (m_isServer) {
        QMessageBox::information(this,
                          tr("Ranglistenvergleich"),
                          tr("Diese Instanz von Muckturnier läuft als Server. Bitte den "
                             "Ranglistenvergleich von einem Client aus anfragen!"));
        return;
    }

    QDialog *existingDialog = findChild<RankingDialog *>();
    if (existingDialog != nullptr) {
        existingDialog->setWindowState((existingDialog->windowState()
                                       & ~Qt::WindowMinimized) | Qt::WindowActive);
        return;
    }

    ClientPage *clientPage = qobject_cast<ClientPage *>(m_clientDock->widget());
    RankingDialog *dialog = new RankingDialog(this, m_sharedObjects, clientPage->client());

    connect(clientPage, &ClientPage::connectionClosed, dialog, &RankingDialog::clientDisconnected);
    connect(clientPage, &ClientPage::pageClosed, dialog, &RankingDialog::clientDisconnected);
    connect(m_playersPage, &PlayersPage::nameEdited, dialog, &RankingDialog::dataChanged);
    connect(m_scorePage, &ScorePage::scoreChanged, dialog, &RankingDialog::dataChanged);

    dialog->show();
}

void MainWindow::networkTournamentStarted()
{
    m_scorePage->setEnabled(true);
    m_finishRegistrationAction->setEnabled(false);
    m_showRankingDialogAction->setEnabled(true);
}

void MainWindow::showAssignmentPage()
{
    m_showAssignmentPageAction->setEnabled(false);

    AssignmentPage *assignmentPage = new AssignmentPage(this, m_sharedObjects);
    connect(assignmentPage, &AssignmentPage::pageClosed, this, &MainWindow::assignmentPageClosed);
    connect(this, &MainWindow::resetAllSearches, assignmentPage, &AssignmentPage::resetSearch);

    m_assignmentDock = new QDockWidget(tr("Übersicht Auslosung"), this);
    m_allDocks.append(m_assignmentDock);
    m_assignmentDock->setWidget(assignmentPage);
    m_playersPage->setAssignmentPage(assignmentPage);
    m_playersPage->checkUpdateAssignmentPage();

    formatDockWidget(m_assignmentDock);
    tabifyDockWidget(m_playersDock, m_assignmentDock);
    m_assignmentDock->show();
    m_assignmentDock->raise();
}

void MainWindow::assignmentPageClosed()
{
    m_allDocks.removeOne(m_assignmentDock);
    m_assignmentDock = nullptr;
    m_playersPage->setAssignmentPage(nullptr);
    m_showAssignmentPageAction->setEnabled(true);
}
