/* Copyright (c) 2011-2020 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef MAINWINDOW_H
#define MAINWINDOW_H

// Local includes
#include "shared/Database.h"

// Qt includes
#include <QMainWindow>

// Local classes
class SharedObjects;
class ResourceFinder;
class PlayersPage;
class ScorePage;
class RankingPage;
class Server;
class Client;
class Settings;
class BackupEngine;
class TournamentSettings;

// Qt classes
class QAction;
class QDockWidget;
class QWidget;
class QShortcut;

class MainWindow : public QMainWindow
{
    Q_OBJECT

public:
    explicit MainWindow(SharedObjects *sharedObjects);

public slots:
    void updateStatus(const QString &message);
    void settingsSaved();
    void tournamentStarted();
    void openTournament(QString dbFileName);
    void closeTournament();
    void reopenLastOpenedDb(const QString &lastOpenedDb);

signals:
    void resetAllSearches();

protected:
    virtual void resizeEvent(QResizeEvent *event) override;
    virtual void closeEvent(QCloseEvent *event) override;

private: // Functions
    void checkEnableRankingPage();
    void formatDockWidget(QDockWidget *dock);
    void attachDocks(Qt::DockWidgetArea area);
    void updateDatabase(const QString &lastOpenedDb, const QString &dbFileName,
                        Database::Error versionError);
    void processDbReset();
    bool askRestoreOriginalDb(BackupEngine &backupEngine);
    void showRestoreBackupDialog(BackupEngine &backupEngine, const QString &dbToOverwrite,
                                 bool autoAccept);
    void checkDisplayFinishRegistrationInfo();

private slots:
    void playersCountChanged(int count);
    void registrationsAvailable(bool state);
    void playersNameEdited();
    void databaseChanged();
    void startNewTournament();
    void showHandbook();
    void showAboutDialog();
    void showExportTournamentDialog();
    void showStopWatchDialog();
    void showTournamentInfo();
    void exportPlayersList();
    void hideWindowFrame();
    void hideMenuBar();
    void arrangeTabified();
    void arrangeHorizontally();
    void arrangeVertically();
    void arrangeOwnWindow();
    void resetDatabase();
    void importPlayersList();
    void sqlError(const QString &text);
    void triggerShowRestoreBackupDialog();
    bool createDatabaseBackup();
    void showChecksumsDialog();
    void showSettingsDialog();
    void showServerPage();
    void showClientPage();
    void setNetworkActionsEnabled(bool state);
    void clientConnected(Client *client);
    void showRankingDialog();
    void clientConnectionClosed();
    void serverStopped();
    void serverStarted(Server *server);
    void serverConnectionRequested(const QString &ip, int port);
    void serverPageClosed();
    void clientPageClosed();
    void networkPageClosed();
    void setNameAdoptionMode(bool state);
    void openRestoredDb(const QString &dbFile);
    void networkTournamentStarted();
    void showAssignmentPage();
    void assignmentPageClosed();

private:
    // Shared objects
    SharedObjects *m_sharedObjects;
    ResourceFinder *m_resourceFinder;
    Database *m_db;
    Settings *m_settings;
    TournamentSettings *m_tournamentSettings;

    // GUI obects
    QWidget *m_centralWidget;
    QDockWidget *m_playersDock;
    PlayersPage *m_playersPage;
    QDockWidget *m_scoreDock;
    ScorePage *m_scorePage;
    QDockWidget *m_rankingDock;
    RankingPage *m_rankingPage;
    QDockWidget *m_serverDock = nullptr;
    QDockWidget *m_clientDock = nullptr;
    QDockWidget *m_assignmentDock = nullptr;

    // Menus
    QMenu *m_imExportMenu;

    // Menu actions
    QAction *m_newTournamentAction;
    QAction *m_openTournamentAction;
    QAction *m_tournamentInfoAction;
    QAction *m_closeTournamentAction;
    QAction *m_resetDatabaseAction;
    QAction *m_importPlayersListAction;
    QAction *m_exportPlayersListAction;
    QAction *m_exportTournamentAction;
    QAction *m_hideWindowFrameAction;
    QAction *m_hideMenubarAction;
    QAction *m_createDatabaseBackupAction;
    QAction *m_showRestoreBackupDialogAction;
    QAction *m_showChecksumsDialogAction;
    QAction *m_showServerPageAction;
    QAction *m_finishRegistrationAction;
    QAction *m_showClientPageAction;
    QAction *m_showRankingDialogAction;
    QAction *m_showAssignmentPageAction;

    // Global shortcuts
    QShortcut *m_viewMenubarShortcut;

    // Cache variables
    QString m_lastServerIp;
    int m_lastServerPort;
    bool m_askRestoreOriginalDb = true;
    bool m_restoreOriginalDb = false;
    bool m_openingAfterRestore = false;
    bool m_isServer = false;
    bool m_isClient = false;
    QVector<QDockWidget *> m_allDocks;
    bool m_networkFinishRegistrationWarningAlreadyShown = false;

};

#endif // MAINWINDOW_H
