/* Copyright (c) 2011-2020 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes

#include "NewTournamentDialog.h"

#include "shared/defaultValues.h"
#include "shared/SharedObjects.h"
#include "shared/Database.h"
#include "shared/Settings.h"
#include "shared/TournamentSettings.h"
#include "shared/Muckturnier.h"

// Qt includes
#include <QGroupBox>
#include <QLabel>
#include <QApplication>
#include <QDebug>
#include <QDir>
#include <QRadioButton>
#include <QGridLayout>
#include <QSpinBox>
#include <QMessageBox>
#include <QStandardPaths>
#include <QSettings>
#include <QDateTime>
#include <QCheckBox>

NewTournamentDialog::NewTournamentDialog(QWidget *parent, SharedObjects *sharedObjects)
    : FileDialog(parent, sharedObjects->database()),
      m_settings(sharedObjects->settings()),
      m_tournamentSettings(sharedObjects->tournamentSettings())
{
    setTitle(tr("Neues Turnier starten"));
    setOkButtonText(tr("Turnier starten"));
    setTargetFileBaseName(QDateTime::currentDateTime().toString(m_settings->dbFileNameTemplate()));
    setTargetFileExtension(QStringLiteral("mtdb"));
    setTargetFileDescription(tr("Muckturnier-Datenbanken"));
    setFileDialogTitle(tr("Bitte einen Dateinamen für die Turnierdatenbank auswählen"));

    const Settings::TournamentTemplate tournamentTemplate = m_settings->tournamentTemplate();

    // Tournament type

    QGroupBox *tournamentTypeBox = new QGroupBox(tr("Turniertyp"));
    QGridLayout *tournamentTypeBoxLayout = new QGridLayout(tournamentTypeBox);
    layout()->addWidget(tournamentTypeBox);

    m_fixedPairs = new QRadioButton(tr("Feste Paare"));
    m_fixedPairs->setChecked(tournamentTemplate.tournamentMode
                                 == Muckturnier::TournamentMode::FixedPairs);
    tournamentTypeBoxLayout->addWidget(m_fixedPairs, 0, 0, Qt::AlignTop);

    QLabel *fixedPairsDescription = new QLabel(tr("Es werden immer zwei Spieler als festes Paar "
                                                  "angemeldet, das das ganze Turnier zusammen "
                                                  "spielt."));
    fixedPairsDescription->setWordWrap(true);
    fixedPairsDescription->setAlignment(Qt::AlignTop);
    fixedPairsDescription->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Minimum);
    tournamentTypeBoxLayout->addWidget(fixedPairsDescription, 0, 1);

    m_singlePlayers = new QRadioButton(tr("Einzelne Spieler"));
    m_singlePlayers->setChecked(! m_fixedPairs->isChecked());
    tournamentTypeBoxLayout->addWidget(m_singlePlayers, 1, 0, Qt::AlignTop);

    QLabel *singlePlayersDescription = new QLabel(tr("Es werden einzelne Spieler angemeldet. Die "
                                                     "Paare werden pro Runde ausgelost."));
    singlePlayersDescription->setWordWrap(true);
    singlePlayersDescription->setAlignment(Qt::AlignTop);
    singlePlayersDescription->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Minimum);
    tournamentTypeBoxLayout->addWidget(singlePlayersDescription, 1, 1);

    // Tournament settings

    QGroupBox *tournamentSettingsBox = new QGroupBox(tr("Turniereinstellungen"));
    QHBoxLayout *tournamentSettingsBoxWrapperLayout = new QHBoxLayout(tournamentSettingsBox);
    QGridLayout *tournamentSettingsBoxLayout = new QGridLayout;
    tournamentSettingsBoxLayout->setVerticalSpacing(0);
    tournamentSettingsBoxWrapperLayout->addLayout(tournamentSettingsBoxLayout);
    tournamentSettingsBoxWrapperLayout->addStretch();
    layout()->addWidget(tournamentSettingsBox);

    tournamentSettingsBoxLayout->addWidget(new QLabel(tr("Punkte pro Bobbl:")), 0, 0);

    m_boogerScore = new QSpinBox;
    m_boogerScore->setMinimum(1);
    m_boogerScore->setMaximum(MAXIMUM_BOOGER_SCORE);
    m_boogerScore->setValue(tournamentTemplate.boogerScore);
    tournamentSettingsBoxLayout->addWidget(m_boogerScore, 0, 1);

    tournamentSettingsBoxLayout->addWidget(new QLabel(tr("Bobbl pro Runde:")), 1, 0);

    m_boogersPerRound = new QSpinBox;
    m_boogersPerRound->setMinimum(1);
    m_boogersPerRound->setMaximum(MAXIMUM_BOOGERS_PER_ROUND);
    m_boogersPerRound->setValue(tournamentTemplate.boogersPerRound);
    tournamentSettingsBoxLayout->addWidget(m_boogersPerRound, 1, 1);

    // Initial input settings

    QGroupBox *inputSettingsBox = new QGroupBox(tr("Eingabevoreinstellungen"));
    QVBoxLayout *inputSettingsBoxLayout = new QVBoxLayout(inputSettingsBox);
    layout()->addWidget(inputSettingsBox);

    QLabel *inputSettingsBoxInfo = new QLabel(tr(
        "<i>Diese Einstellungen können auch noch nach dem Starten des Turniers auf der "
        "Ergebnisse-Seite angepasst (aber per Vorlage hier voreingestellt) werden:</i>"));
    inputSettingsBoxInfo->setWordWrap(true);
    inputSettingsBoxInfo->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Minimum);
    inputSettingsBoxLayout->addWidget(inputSettingsBoxInfo);


    m_autoSelectPairs = new QCheckBox(tr("Tischnummer bzw. Paare automatisch auswählen"));
    m_lastAutoSelectPairsState = tournamentTemplate.autoSelectPairs;
    inputSettingsBoxLayout->addWidget(m_autoSelectPairs);

    m_noTableNumbers = new QCheckBox(tr("Eingabe ohne Tischnummern"));
    m_noTableNumbers->setChecked(tournamentTemplate.noTableNumbers);
    inputSettingsBoxLayout->addWidget(m_noTableNumbers);
    connect(m_noTableNumbers, &QCheckBox::stateChanged,
            this, &NewTournamentDialog::noTableNumbersChanged);
    connect(m_singlePlayers, &QRadioButton::toggled,
            this, &NewTournamentDialog::singlePlayerModeChanged);

    singlePlayerModeChanged(m_singlePlayers->isChecked());
}

void NewTournamentDialog::singlePlayerModeChanged(bool state)
{
    if (! state && ! m_noTableNumbers->isChecked()) {
        m_autoSelectPairs->setEnabled(true);
        m_autoSelectPairs->setChecked(m_lastAutoSelectPairsState);
    } else {
        if (m_autoSelectPairs->isEnabled()) {
            m_lastAutoSelectPairsState = m_autoSelectPairs->isChecked();
        }
        m_autoSelectPairs->setChecked(false);
        m_autoSelectPairs->setEnabled(false);
    }
}

void NewTournamentDialog::noTableNumbersChanged(int state)
{
    if (! state && ! m_singlePlayers->isChecked()) {
        m_autoSelectPairs->setEnabled(true);
        m_autoSelectPairs->setChecked(m_lastAutoSelectPairsState);
    } else {
        if (m_autoSelectPairs->isEnabled()) {
            m_lastAutoSelectPairsState = m_autoSelectPairs->isChecked();
        }
        m_autoSelectPairs->setChecked(false);
        m_autoSelectPairs->setEnabled(false);
    }
}

void NewTournamentDialog::accept()
{
    const QString lastDbPath = db()->dbFile();

    QString fileName = getTargetFileName(m_settings->lastDbPath());
    if (fileName.isEmpty()) {
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    setEnabled(false);
    QApplication::processEvents();

    emit updateStatus(tr("Lege neue Turnierdatenbank an …"));

    if (! db()->create(fileName)) {
        emit updateStatus(tr("Anlegen der neuen Turnierdatenbank fehlgeschlagen"));
        QApplication::restoreOverrideCursor();
        setEnabled(true);
        QMessageBox::critical(this,
            tr("Neue Turnierdatenbank anlegen"),
            tr("<p><b>Anlegen der Datenbank ist fehlgeschlagen</b></p>"
               "<p>Das Anlegen der Turnierdatenbank „%1“ ist nicht möglich:</p><p>%2</p>").arg(
               QDir::toNativeSeparators(fileName), db()->errorMessage()),
            QMessageBox::Ok);
        if (! lastDbPath.isEmpty()) {
            emit reopenLastOpenedDb(lastDbPath);
        }
        return;
    }

    // Fill in the markers defintion from the template
    if (! db()->addMarkers(m_settings->markersTemplate())) {
        QDialog::reject();
        return;
    }

    // Setup the tournament settings

    // Fill in the values not set here
    m_tournamentSettings->initialize();

    // Set the chosen parameters

    m_tournamentSettings->setBoogerScore(m_boogerScore->value());
    m_tournamentSettings->setBoogersPerRound(m_boogersPerRound->value());

    if (m_fixedPairs->isChecked()) {
        m_tournamentSettings->setTournamentMode(Muckturnier::TournamentMode::FixedPairs);
    } else {
        m_tournamentSettings->setTournamentMode(Muckturnier::TournamentMode::SinglePlayers);
        m_tournamentSettings->setAutoSelectPairs(false);
    }

    m_tournamentSettings->setAutoSelectPairs(m_autoSelectPairs->isChecked());
    m_tournamentSettings->setNoTableNumbers(m_noTableNumbers->isChecked());

    // Save the settings to the database
    if (! db()->saveSettings()) {
        QDialog::reject();
        return;
    }

    emit updateStatus(tr("Turnier „%1“ gestartet.").arg(QDir::toNativeSeparators(fileName)));
    QApplication::restoreOverrideCursor();

    if (m_settings->saveLastDbPathEnabled()) {
        m_settings->saveLastDbPath(db()->dbPath());
    }

    QDialog::accept();
}
