/* Copyright (c) 2019 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes

#include "AssignmentPopup.h"

#include "shared/defaultValues.h"
#include "shared/SharedStyles.h"

// Qt includes
#include <QDebug>
#include <QVBoxLayout>
#include <QGridLayout>
#include <QLabel>
#include <QSpinBox>
#include <QRadioButton>
#include <QPushButton>
#include <QCheckBox>

AssignmentPopup::AssignmentPopup(QWidget *parent) : SelectionPopup(parent)
{
    m_main = new QWidget;
    QVBoxLayout *layout = new QVBoxLayout(m_main);

    m_selection = new QWidget;
    QGridLayout *selectionLayout = new QGridLayout(m_selection);

    m_isSinglePlayer = new QLabel(tr("<span style=\"%1\">Allein gekommener Spieler!</span>").arg(
                                     SharedStyles::STYLE_RED_TEXT));
    m_isSinglePlayer->setAlignment(Qt::AlignCenter);
    layout->addWidget(m_isSinglePlayer);

    selectionLayout->addWidget(new QLabel(tr("Tisch")), 0, 0, 2, 1);

    m_table = new QSpinBox;
    m_table->setMinimum(1);
    m_table->setMaximum(MAXIMUM_PREASSIGNMENT_TABLES);
    connect(m_table, QOverload<int>::of(&QSpinBox::valueChanged),
            this, &AssignmentPopup::checkSelection);
    selectionLayout->addWidget(m_table, 0, 1, 2, 1);

    m_pair1 = new QRadioButton(tr("Paar 1"));
    connect(m_pair1, &QRadioButton::toggled, this, &AssignmentPopup::checkSelection);
    selectionLayout->addWidget(m_pair1, 0, 2);

    m_pair2 = new QRadioButton(tr("Paar 2"));
    selectionLayout->addWidget(m_pair2, 1, 2);

    layout->addWidget(m_selection);

    m_taken = new QLabel(tr("<span style=\"%1\">Bereits vergeben!</span>").arg(
                            SharedStyles::STYLE_RED_TEXT));
    m_taken->setAlignment(Qt::AlignCenter);
    layout->addWidget(m_taken);

    m_save = new QPushButton(tr("Auslosung speichern"));
    connect(m_save, &QPushButton::clicked, this, &SelectionPopup::accept);
    layout->addWidget(m_save);

    m_removeMarker = new QCheckBox(tr("und Markierung entfernen"));
    m_removeMarker->setChecked(true);
    layout->addWidget(m_removeMarker);

    m_delete = new QPushButton(tr("Auslosung löschen"));
    connect(m_delete, &QPushButton::clicked, this, &SelectionPopup::accept);
    layout->addWidget(m_delete);

    setScrollWidget(m_main);
}

void AssignmentPopup::show(bool existing, bool markerSet, bool isSinglePlayer, int table, int pair,
                           bool restore, const QPoint &position)
{
    m_isSinglePlayer->setVisible(isSinglePlayer);
    m_markerSet = markerSet;

    if (! existing && ! restore) {
        const int firstFreePair1Table = findFirstFree(m_assignedPair1Tables);
        const int firstFreePair2Table = findFirstFree(m_assignedPair2Tables);
        if (firstFreePair1Table <= firstFreePair2Table) {
            table = firstFreePair1Table;
            pair = 1;
        } else {
            table = firstFreePair2Table;
            pair = 2;
        }
    }

    m_table->blockSignals(true);
    m_pair1->blockSignals(true);
    m_pair2->blockSignals(true);
    m_table->setValue(table);
    m_pair1->setChecked(pair == 1);
    m_pair2->setChecked(pair == 2);
    m_table->blockSignals(false);
    m_pair1->blockSignals(false);
    m_pair2->blockSignals(false);

    m_selection->setEnabled(! existing);
    m_taken->hide();
    m_save->setVisible(! existing);
    m_removeMarker->setVisible(! existing && m_markerSet);
    m_delete->setVisible(existing);

    if (restore && ! existing) {
        checkSelection();
    }

    m_main->adjustSize();

    if (position == QPoint(-1, -1)) {
        SelectionPopup::show();
    } else {
        SelectionPopup::show(position);
    }
}

void AssignmentPopup::setAssignedTables(QVector<int> &assignedPair1Tables,
                                        QVector<int> &assignedPair2Tables)
{
    m_assignedPair1Tables = assignedPair1Tables;
    m_assignedPair2Tables = assignedPair2Tables;
}

int AssignmentPopup::findFirstFree(const QVector<int> &list) const
{
    int firstFree = 1;

    if (m_assignmentsPerPair == 1) {
        // It's a bit easier for fixed pairs tournaments ...
        for (const int i : list) {
            if (i != firstFree) {
                return firstFree;
            }
            firstFree++;
        }
    } else {
        // ... and a bit more complicated for single players ;-)
        QVectorIterator<int> table(list);
        while (table.hasNext()) {
            table.next();
            if (table.hasNext() && table.next() == firstFree) {
                firstFree++;
            } else {
                return firstFree;
            }
        }
    }

    return firstFree;
}

void AssignmentPopup::checkSelection()
{
    bool taken;
    if (m_pair1->isChecked()) {
        taken = m_assignedPair1Tables.count(m_table->value()) == m_assignmentsPerPair;
    } else {
        taken = m_assignedPair2Tables.count(m_table->value()) == m_assignmentsPerPair;
    }

    m_taken->setVisible(taken);
    m_save->setVisible(! taken);
    m_removeMarker->setVisible(m_markerSet && ! taken);
}

int AssignmentPopup::table() const
{
    return m_selection->isEnabled() ? m_table->value() : 0;
}

int AssignmentPopup::pair() const
{
    return m_pair1->isChecked() ? 1 : 2;
}

bool AssignmentPopup::removeMarker() const
{
    return m_markerSet && m_removeMarker->isChecked();
}

void AssignmentPopup::setPairMode(const bool state)
{
    m_assignmentsPerPair = state ? 1 : 2;
}
