/* Copyright (c) 2018-2019 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes

#include "MarkersWidget.h"
#include "MarkerDialog.h"

#include "ServerPage/Server.h"
#include "ClientPage/Client.h"

#include "shared/SharedObjects.h"
#include "shared/Database.h"
#include "shared/Settings.h"
#include "shared/TournamentSettings.h"
#include "shared/ResourceFinder.h"
#include "shared/TableDelegate.h"
#include "shared/TitleMenu.h"

// Qt includes
#include <QVBoxLayout>
#include <QTableWidget>
#include <QDebug>
#include <QHeaderView>
#include <QGroupBox>
#include <QApplication>
#include <QMessageBox>
#include <QFontMetrics>
#include <QCheckBox>
#include <QComboBox>
#include <QTimer>
#include <QLabel>
#include <QLineEdit>

// C++ includes
#include <functional>

MarkersWidget::MarkersWidget(QWidget *parent, SharedObjects *sharedObjects)
    : QWidget(parent),
      m_db(sharedObjects->database()),
      m_settings(sharedObjects->settings()),
      m_tournamentSettings(sharedObjects->tournamentSettings())
{
    QVBoxLayout *mainLayout = new QVBoxLayout(this);
    // This eliminates additional space around this widget when displayed inside the players page
    mainLayout->setContentsMargins(0, 0, 0, 0);

    // Markers list

    QGroupBox *markersBox = new QGroupBox(tr("Markierungen"));
    QVBoxLayout *markersBoxLayout = new QVBoxLayout(markersBox);
    mainLayout->addWidget(markersBox);

    m_markers = new QTableWidget;
    m_markers->setEditTriggers(QAbstractItemView::NoEditTriggers);
    m_markers->setSelectionMode(QAbstractItemView::NoSelection);
    m_markers->setFocusPolicy(Qt::NoFocus);
    m_markers->setColumnCount(4);
    m_markers->setHorizontalHeaderLabels( { tr("n", "Anzahl"), tr("Name"), tr("Sortierung"),
                                            QString() } );
    m_markers->verticalHeader()->hide();
    const int iconSize = QFontMetrics(QFont(font())).boundingRect(QStringLiteral("M")).height()
#ifdef Q_OS_WIN
                         * 1.6; // On Windows, the icon looks better if it's a bit bigger
#else
                         * 1.3; // On Linux, a bit smaller looks better
#endif
    m_markers->setIconSize(QSize(iconSize, iconSize));
    m_markers->setShowGrid(false);
    m_markers->setItemDelegate(new TableDelegate(this));
    m_markers->setContextMenuPolicy(Qt::CustomContextMenu);
    connect(m_markers, &QWidget::customContextMenuRequested, this, &MarkersWidget::showContextMenu);
    connect(m_markers, &QTableWidget::cellDoubleClicked,
            this, &MarkersWidget::checkShowContextMenu);
    connect(m_markers, &QTableWidget::cellClicked, this, &MarkersWidget::changeVisibility);
    markersBoxLayout->addWidget(m_markers);

    m_defaultMarkerWidget = new QWidget;
    QHBoxLayout *defaultMarkerLayout = new QHBoxLayout(m_defaultMarkerWidget);
    markersBoxLayout->addWidget(m_defaultMarkerWidget);
    QLabel *defaultMarkerLabel = new QLabel(tr("Neue Anmeldungen\nmarkieren als:"));
    defaultMarkerLabel->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);
    defaultMarkerLayout->addWidget(defaultMarkerLabel);
    m_defaultMarker = new QComboBox;
    m_defaultMarker->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Maximum);
    connect(m_defaultMarker, QOverload<int>::of(&QComboBox::currentIndexChanged),
            [this] {
                m_tournamentSettings->setDefaultMarker(m_defaultMarker->currentData().toInt());
            });
    defaultMarkerLayout->addWidget(m_defaultMarker);

    // Ignored registrations marker

    m_absentMarkerBox = new QGroupBox(tr("Markierung für abwesende Anmeldungen"));
    m_absentMarkerBox->setCheckable(true);
    m_absentMarkerBox->setChecked(false);
    QHBoxLayout *absentMarkerBoxLayout = new QHBoxLayout(m_absentMarkerBox);
    connect(m_absentMarkerBox, &QGroupBox::toggled, this, &MarkersWidget::absentMarkerChanged);
    mainLayout->addWidget(m_absentMarkerBox);

    QLabel *absentMarkerLabel = new QLabel(tr("Anmeldungen mit folgender\n"
                                              "Markierung ignorieren:"));
    absentMarkerBoxLayout->addWidget(absentMarkerLabel);

    m_absentMarker = new QComboBox;
    m_absentMarker->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Maximum);
    connect(m_absentMarker, QOverload<int>::of(&QComboBox::currentIndexChanged),
            this, &MarkersWidget::absentMarkerChanged);
    absentMarkerBoxLayout->addWidget(m_absentMarker);

    // Singles management

    m_singlesManagementBox = new QGroupBox(tr("Allein gekommene Spieler berücksichtigen"));
    m_singlesManagementBox->setCheckable(true);
    m_singlesManagementBox->setChecked(false);
    QGridLayout *singlesManagementLayout = new QGridLayout(m_singlesManagementBox);
    mainLayout->addWidget(m_singlesManagementBox);
    connect(m_singlesManagementBox, &QGroupBox::toggled,
            this, &MarkersWidget::singlesManagementChanged);

    QLabel *singlesMarkerLabel = new QLabel(tr("Markierung für „Allein da“:"));
    singlesManagementLayout->addWidget(singlesMarkerLabel, 0, 0);
    m_singleMarker = new QComboBox;
    connect(m_singleMarker, QOverload<int>::of(&QComboBox::currentIndexChanged),
            this, &MarkersWidget::singlesManagementChanged);
    m_singleMarker->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Maximum);
    singlesManagementLayout->addWidget(m_singleMarker, 0, 1, 1, 2);

    m_markSingles = new QCheckBox(tr("automatisch setzen\nwenn der Name"));
    connect(m_markSingles, &QCheckBox::toggled,
            this, &MarkersWidget::singlesManagementChanged);
    singlesManagementLayout->addWidget(m_markSingles, 1, 0);

    m_singleConditionString = new QLineEdit;
    m_singleConditionString->setSizePolicy(QSizePolicy::Ignored, QSizePolicy::Maximum);
    m_singleConditionString->setMinimumWidth(50);
    connect(m_singleConditionString, &QLineEdit::editingFinished,
            this, &MarkersWidget::singlesManagementChanged);
    singlesManagementLayout->addWidget(m_singleConditionString, 1, 1);

    m_singleCondition = new QComboBox;
    m_singleCondition->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);
    m_singleCondition->addItem(tr("nicht enthält"));
    m_singleCondition->addItem(tr("enthält"));
    connect(m_singleCondition, QOverload<int>::of(&QComboBox::currentIndexChanged),
            this, &MarkersWidget::singlesManagementChanged);
    singlesManagementLayout->addWidget(m_singleCondition, 1, 2);

    QLabel *assignedMarkerlabel = new QLabel(tr("Markierung nach Zuordnung:"));
    singlesManagementLayout->addWidget(assignedMarkerlabel, 2, 0);
    m_assignedMarker = new QComboBox;
    m_assignedMarker->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Maximum);
    connect(m_assignedMarker, QOverload<int>::of(&QComboBox::currentIndexChanged),
            this, &MarkersWidget::singlesManagementChanged);
    singlesManagementLayout->addWidget(m_assignedMarker, 2, 1, 1, 2);

    // Context menu

    m_markerMenu = new TitleMenu(this);

    m_moveUpAction = m_markerMenu->addAction(tr("Nach oben bewegen"));
    connect(m_moveUpAction, &QAction::triggered,
            this, std::bind(&MarkersWidget::moveMarker, this, -1));

    m_moveDownAction = m_markerMenu->addAction(tr("Nach unten bewegen"));
    connect(m_moveDownAction, &QAction::triggered,
            this, std::bind(&MarkersWidget::moveMarker, this, 1));

    m_markerMenu->addSeparator();

    m_editMarkerAction = m_markerMenu->addAction(tr("Bearbeiten"));
    connect(m_editMarkerAction, &QAction::triggered, this, &MarkersWidget::editMarker);

    m_deleteMarkerAction = m_markerMenu->addAction(tr("Löschen"));
    connect(m_deleteMarkerAction, &QAction::triggered, this, &MarkersWidget::deleteMarker);

    m_markerMenu->addSeparator();

    m_addMarkerAction = m_markerMenu->addAction(tr("Neue Markierung"));
    connect(m_addMarkerAction, &QAction::triggered, this, &MarkersWidget::addMarker);

    m_addMarkerMenu = new QMenu(this);
    m_addMarkerMenu->addAction(m_addMarkerAction);

    // Hide/Show context menu

    m_hideShowMarkedMenu = new QMenu(this);

    m_showOnlyAction = m_hideShowMarkedMenu->addAction(QString());

    m_showAllAction = m_hideShowMarkedMenu->addAction(tr("Alle anzeigen"));
    connect(m_showAllAction, &QAction::triggered, this, &MarkersWidget::showAll);

    // Icons
    m_unused = QIcon(sharedObjects->resourceFinder()->find(QStringLiteral("unused.svg")));
    m_visible = QIcon(sharedObjects->resourceFinder()->find(QStringLiteral("visible.svg")));
    m_hidden = QIcon(sharedObjects->resourceFinder()->find(QStringLiteral("hidden.svg")));
}

int MarkersWidget::markersComboBoxSelection(QComboBox *comboBox) const
{
    const QVariant data = comboBox->currentData();
    return data.isValid() ? data.toInt() : -1;
}

void MarkersWidget::setMarkersComboBoxSelection(QComboBox *comboBox, int markerId)
{
    const int index = comboBox->findData(markerId);
    comboBox->setCurrentIndex(index != -1 ? index : comboBox->findData(0));
}

void MarkersWidget::updateTournamentMode()
{
    if (m_tournamentSettings->isSinglePlayerMode()) {
        m_markSingles->setChecked(false);
        m_singlesManagementBox->setVisible(false);
    } else {
        m_singlesManagementBox->setVisible(true);
    }
}

void MarkersWidget::reload()
{
    const int lastDefaultMarker = markersComboBoxSelection(m_defaultMarker);
    const int lastAbsentMarker = markersComboBoxSelection(m_absentMarker);
    const int lastSingleMarker = markersComboBoxSelection(m_singleMarker);
    const int lastAssignedMarker = markersComboBoxSelection(m_assignedMarker);

    bool success;
    const QVector<Muckturnier::Marker> markers(m_db->markers(&success));
    if (! success) {
        return;
    }

    m_markers->setRowCount(markers.count());

    m_defaultMarker->blockSignals(true);
    m_absentMarker->blockSignals(true);
    m_singleMarker->blockSignals(true);
    m_assignedMarker->blockSignals(true);

    m_defaultMarker->clear();
    m_absentMarker->clear();
    m_singleMarker->clear();
    m_assignedMarker->clear();

    QVector<int> ids;
    ids.reserve(markers.count());
    int row = 0;
    for (const Muckturnier::Marker &marker : markers) {
        ids.append(marker.id);
        m_markers->setItem(row, 0, new QTableWidgetItem);

        QTableWidgetItem *item = new QTableWidgetItem(marker.name);
        if (marker.name.isEmpty()) {
            item->setText(tr("(unmarkiert)"));
        }
        item->setForeground(marker.color);
        item->setData(MarkerData::Id, QVariant(marker.id));
        item->setData(MarkerData::Sorting, QVariant(marker.sorting));
        m_markers->setItem(row, 1, item);

        m_defaultMarker->addItem(item->text(), marker.id);
        m_absentMarker->addItem(item->text(), marker.id);
        m_singleMarker->addItem(item->text(), marker.id);
        m_assignedMarker->addItem(item->text(), marker.id);

        switch (marker.sorting) {
        case Muckturnier::MarkerSorting::NewBlock:
            item = new QTableWidgetItem(tr("Neuer Block"));
            break;
        case Muckturnier::MarkerSorting::Merged:
            item = new QTableWidgetItem(tr("Fortlaufend"));
            break;
        }
        m_markers->setItem(row, 2, item);

        item = new QTableWidgetItem;
        m_markers->setItem(row, 3, item);

        row++;
    }

    // Set new markers visible
    for (const int id : ids) {
        if (! m_isVisible.contains(id)) {
            m_isVisible[id] = true;
        }
    }

    // Remove visibility entries for deleted markers
    for (const int id : m_isVisible.keys()) {
        if (! ids.contains(id)) {
            m_isVisible.remove(id);
        }
    }

    m_markers->horizontalHeaderItem(0)->setWhatsThis(tr(
        "In dieser Spalte wird die Anzahl der mit der entsprechenden Markierung gekennzeichneten "
        "%1 angezeigt.").arg(m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler")));
    m_markers->horizontalHeaderItem(3)->setWhatsThis(tr(
        "Durch Klicken auf das Augensymbol können alle mit der entsprechenden Markierung "
        "gekennzeichneten %1 ein- bzw. ausgeblendet werden.").arg(
        m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler")));
    m_markers->resizeColumnsToContents();

    setMarkersComboBoxSelection(m_defaultMarker, lastDefaultMarker);
    setMarkersComboBoxSelection(m_absentMarker, lastAbsentMarker);
    setMarkersComboBoxSelection(m_singleMarker, lastSingleMarker);
    setMarkersComboBoxSelection(m_assignedMarker, lastAssignedMarker);

    m_defaultMarker->blockSignals(false);
    m_absentMarker->blockSignals(false);
    m_singleMarker->blockSignals(false);
    m_assignedMarker->blockSignals(false);

    emit markersChanged();
}

void MarkersWidget::showContextMenu(const QPoint &position)
{
    int column = -1;
    QString name;
    int id;
    int row;
    const QPoint target = m_markers->viewport()->mapToGlobal(position);

    // Collect data
    QTableWidgetItem *markerItem = m_markers->itemAt(position);
    if (markerItem != nullptr) {
        column = markerItem->column();
        row = markerItem->row();
        if (column != 1) {
            markerItem = m_markers->item(row, 1);
        }
        name = m_markers->item(row, 1)->text();
        id = m_markers->item(row, 1)->data(MarkerData::Id).toInt();
    }

    // Show no context menu on the counter column
    if (column == 0) {
        return;
    }

    // Show the "Hide/Show marked entries" context menu on the "Hide/Show" column
    if (column == 3) {
        m_showAllAction->setEnabled(false);
        for (bool visible : m_isVisible.values()) {
            if (! visible) {
                m_showAllAction->setEnabled(true);
                break;
            }
        }

        // Set the context menu to handle the selected marker
        m_showOnlyAction->disconnect();
        m_showOnlyAction->setText(tr("Nur „%1“ anzeigen").arg(name));
        connect(m_showOnlyAction, &QAction::triggered,
                this, std::bind(&MarkersWidget::showOnly, this, id));

        m_hideShowMarkedMenu->exec(target);
        return;
    }

    // All other actions require write access, so return here if the database if ro
    if (! m_db->isWritable()) {
        return;
    }

    // Show the "Add marker" context menu if no item was clicked
    if (column == -1) {
        m_addMarkerMenu->exec(target);
        return;
    }

    // Otherwise, show the marker context menu

    m_markerMenu->setTitle(tr("Markierung „%1“").arg(name));
    m_selectedMarker.id = id;
    m_selectedMarker.name = name;
    m_selectedMarker.color = markerItem->foreground().color();
    m_selectedMarker.sorting = static_cast<Muckturnier::MarkerSorting>(
                                   markerItem->data(MarkerData::Sorting).toInt());
    m_selectedMarkerSequence = markerItem->row();
    m_moveUpAction->setEnabled(m_selectedMarkerSequence > 0);
    m_moveDownAction->setEnabled(m_selectedMarkerSequence < m_markers->rowCount() - 1);
    m_editMarkerAction->setEnabled(m_selectedMarker.id != 0);
    m_deleteMarkerAction->setEnabled(m_selectedMarker.id != 0);

    m_markerMenu->exec(target);
}

void MarkersWidget::addMarker()
{
    MarkerDialog *dialog = new MarkerDialog(this, m_markers->palette().color(QPalette::Base),
                                            m_markers);
    connect(dialog, &MarkerDialog::markerSaved, this, &MarkersWidget::saveNewMarker);

    QString name = tr("Neue Markierung");
    int counter = 1;
    QString newName = name;
    while (m_markers->findItems(newName, Qt::MatchFixedString).size() > 0) {
        counter++;
        newName = QStringLiteral("%1 %2").arg(name, QString::number(counter));
    }

    dialog->setName(newName);
    dialog->setTitle(tr("Neue Markierung"));
    dialog->setButtonText(QDialogButtonBox::Ok, tr("Hinzufügen"));

    dialog->exec();
}

void MarkersWidget::saveNewMarker(const QString &name, const QColor &color,
                                  Muckturnier::MarkerSorting sorting)
{
    if (m_isClient) {
        m_client->requestAddMarker(name, color, sorting);
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);

    emit statusUpdate(tr("Lege neue Markierung an …"));

    if (! m_db->addMarker(name, color, sorting)) {
        return;
    }

    reload();

    if (m_isServer) {
        m_server->broadcastAddMarker(name, color, sorting);
    }

    QApplication::restoreOverrideCursor();
    emit statusUpdate(tr("Markierung „%1“ hinzugefügt").arg(name));
}

void MarkersWidget::deleteMarker()
{
    if (QMessageBox::question(this,
            tr("Markierung löschen"),
            tr("Soll die Markierung „%1“ wirklich gelöscht werden?").arg(m_selectedMarker.name),
            QMessageBox::Yes | QMessageBox::Cancel, QMessageBox::Cancel) == QMessageBox::Cancel) {

        return;
    }

    if (m_isClient) {
        m_client->requestDeleteMarker(m_selectedMarker.name, m_selectedMarker.id,
                                      m_selectedMarkerSequence);
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);

    emit statusUpdate(tr("Lösche Markierung „%1“ …").arg(m_selectedMarker.name));

    if (! m_db->deleteMarker(m_selectedMarker.id, m_selectedMarkerSequence)) {
        return;
    }

    reload();

    if (m_isServer) {
        m_server->broadcastDeleteMarker(m_selectedMarker.name, m_selectedMarker.id,
                                        m_selectedMarkerSequence);
    }

    QApplication::restoreOverrideCursor();
    emit statusUpdate(tr("Markierung „%1“ gelöscht").arg(m_selectedMarker.name));
}

void MarkersWidget::editMarker()
{
    MarkerDialog *dialog = new MarkerDialog(this, m_markers->palette().color(QPalette::Base),
                                            m_markers);
    connect(dialog, &MarkerDialog::markerSaved, this, &MarkersWidget::saveEdit);

    dialog->setTitle(tr("Markierung bearbeiten"));
    dialog->setButtonText(QDialogButtonBox::Ok, tr("Speichern"));
    dialog->setName(m_selectedMarker.name);
    dialog->setEditedName(m_selectedMarker.name);
    dialog->setColor(m_selectedMarker.color);
    dialog->setSorting(m_selectedMarker.sorting);
    if (m_selectedMarkerSequence == 0) {
        dialog->disableSortingSelection();
    }
    dialog->setIsEditing();

    dialog->exec();
}

void MarkersWidget::saveEdit(const QString &name, const QColor &color,
                             Muckturnier::MarkerSorting sorting)
{
    if (name == m_selectedMarker.name
        && color == m_selectedMarker.color
        && sorting == m_selectedMarker.sorting) {

        emit statusUpdate(tr("Markierung „%1“ wurde nicht verändert").arg(m_selectedMarker.name));
        return;
    }

    if (m_isClient) {
        m_client->requestEditMarker(m_selectedMarker.name, m_selectedMarker.id,
                                    name, color, sorting);
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    emit statusUpdate(tr("Speichere Änderungen an Markierung „%1“ …").arg(m_selectedMarker.name));

    if (! m_db->editMarker(m_selectedMarker.id, name, color, sorting)) {
        return;
    }

    reload();

    if (m_isServer) {
        m_server->broadcastEditMarker(m_selectedMarker.name, m_selectedMarker.id,
                                      name, color, sorting);
    }

    QApplication::restoreOverrideCursor();
    emit statusUpdate(tr("Änderungen an Markierung „%1“ gespeichert").arg(m_selectedMarker.name));
}

void MarkersWidget::moveMarker(int direction)
{
    if (m_isClient) {
        m_client->requestMoveMarker(m_selectedMarker.name, m_selectedMarker.id,
                                    m_selectedMarkerSequence, direction);
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    emit statusUpdate(tr("Verschiebe Markierung „%1“ %2 …").arg(
                         m_selectedMarker.name, direction == 1 ? tr("nach unten")
                                                               : tr("nach oben")));

    if (! m_db->moveMarker(m_selectedMarker.id, m_selectedMarkerSequence, direction)) {
        return;
    }

    reload();

    if (m_isServer) {
        m_server->broadcastMoveMarker(m_selectedMarker.name, m_selectedMarker.id,
                                      m_selectedMarkerSequence, direction);
    }

    QApplication::restoreOverrideCursor();
    emit statusUpdate(tr("Markierung „%1“ %2 verschoben").arg(
                         m_selectedMarker.name, direction == 1 ? tr("nach unten")
                                                               : tr("nach oben")));
}

QVector<Muckturnier::Marker> MarkersWidget::getMarkers() const
{
    QVector<Muckturnier::Marker> markers;
    markers.reserve(m_markers->rowCount());
    for (int row = 0; row < m_markers->rowCount(); row++) {
        Muckturnier::Marker marker;
        QTableWidgetItem *item = m_markers->item(row, 1);
        marker.id = item->data(MarkerData::Id).toInt();
        marker.name = item->text();
        marker.color = item->foreground().color();
        marker.sorting = static_cast<Muckturnier::MarkerSorting>(
                             item->data(MarkerData::Sorting).toInt());
        markers.append(marker);
    }
    return markers;
}

void MarkersWidget::checkShowContextMenu(int, int column)
{
    if (column == 1 || column == 2) {
        showContextMenu(m_markers->viewport()->mapFromGlobal(QCursor::pos()));
    }
}

void MarkersWidget::updateVisibilityIcon(int id, int row)
{
    if (! m_isVisible[id]) {
        m_markers->item(row, 3)->setIcon(m_hidden);
    } else {
        if (m_markers->item(row, 0)->text().toInt() == 0) {
            m_markers->item(row, 3)->setIcon(m_unused);
        } else {
            m_markers->item(row, 3)->setIcon(m_visible);
        }
    }
}

void MarkersWidget::changeVisibility(int row, int column)
{
    if (column != 3) {
        // Not clicked on eye symbol
        return;
    }

    const int id = m_markers->item(row, 1)->data(MarkerData::Id).toInt();
    m_isVisible[id] = ! m_isVisible[id];
    updateVisibilityIcon(id, row);

    emit visibilityChanged();
}

void MarkersWidget::updateVisibilityIcons()
{
    for (int row = 0; row < m_markers->rowCount(); row++) {
        updateVisibilityIcon(m_markers->item(row, 1)->data(MarkerData::Id).toInt(), row);
    }
    emit visibilityChanged();
}

void MarkersWidget::reset()
{
    for (const int id : m_isVisible.keys()) {
        m_isVisible[id] = true;
    }

    setAbsentMarker(m_tournamentSettings->absentMarker());
    setSinglesManagement(m_tournamentSettings->singlesManagement());
    setMarkersComboBoxSelection(m_defaultMarker, m_tournamentSettings->defaultMarker());

    m_absentMarkerBox->setEnabled(m_db->isWritable());
    m_singlesManagementBox->setEnabled(m_db->isWritable());
    m_defaultMarkerWidget->setEnabled(m_db->isWritable());
}

bool MarkersWidget::isMarkerVisible(int id) const
{
    return m_isVisible.value(id);
}

void MarkersWidget::updateCounters(const QHash<int, int> &counter)
{
    m_allMarkersCount = 0;
    int markerCount;
    int id;

    for (int row = 0; row < m_markers->rowCount(); row++) {
        id = m_markers->item(row, 1)->data(MarkerData::Id).toInt();
        markerCount = counter[id];
        m_markers->item(row, 0)->setText(QString::number(markerCount));
        m_allMarkersCount += markerCount;
    }

    updateVisibilityIcons();
    m_markers->resizeColumnsToContents();
}

void MarkersWidget::showAll()
{
    for (int id : m_isVisible.keys()) {
        m_isVisible[id] = true;
    }
    updateVisibilityIcons();
}

void MarkersWidget::showOnly(int idToShow)
{
    for (int id : m_isVisible.keys()) {
        m_isVisible[id] = (id == idToShow);
    }
    updateVisibilityIcons();
}

int MarkersWidget::newRegistrationMarker(const QString &name) const
{
    int marker = m_defaultMarker->currentData().toInt();
    if (m_markSingles->isEnabled() && m_markSingles->isChecked()) {
        bool contains = name.contains(m_singleConditionString->text().simplified());
        // The first m_singleCondition entry is "doesn't contain" (index 0), the second is
        // "contains" (index 1), so we can simply use the selected index here, because it will be
        // (per official C++ behavior) silently casted to a bool value in the following comparison
        if (contains == m_singleCondition->currentIndex()) {
            marker = m_singleMarker->currentData().toInt();
        }
    }
    return marker;
}

void MarkersWidget::addChecksumData(QCryptographicHash &hash) const
{
    // Add the markers data
    for (int i = 0; i < m_markers->rowCount(); i++) {
        QTableWidgetItem *item = m_markers->item(i, 1);

        const int id = item->data(MarkerData::Id).toInt();
        const QString name = id != 0 ? item->text() : QString();

        hash.addData(QByteArray::number(id));
        hash.addData(name.toUtf8());
        hash.addData(QByteArray::number(item->foreground().color().rgb()));
        hash.addData(QByteArray::number(item->data(MarkerData::Sorting).toInt()));
    }

    // Add the absent marker
    hash.addData(QByteArray::number(m_absentMarker->currentData().toInt()));

    // Add the singles management data
    hash.addData(QByteArray::number(m_singlesManagementBox->isChecked()));
    hash.addData(QByteArray::number(m_singleMarker->currentData().toInt()));
    hash.addData(QByteArray::number(m_markSingles->isChecked()));
    hash.addData(m_singleConditionString->text().toUtf8());
    hash.addData(QByteArray::number(m_singleCondition->currentData().toInt()));
    hash.addData(QByteArray::number(m_assignedMarker->currentData().toInt()));
}

void MarkersWidget::setServer(Server *server)
{
    m_server = server;
    m_isServer = true;
}

void MarkersWidget::setClient(Client *client)
{
    m_client = client;
    m_isClient = true;
}

void MarkersWidget::serverStopped()
{
    m_isServer = false;
}

void MarkersWidget::clientDisconnected()
{
    m_isClient = false;
}

void MarkersWidget::processNetworkChange()
{
    if (m_markerMenu->isVisible()) {
        m_markerMenu->hide();
    }

    reload();

    MarkerDialog *dialog = findChild<MarkerDialog *>();
    if (dialog != nullptr && dialog->isEditing()) {
        QTableWidgetItem *item = m_markers->item(m_selectedMarkerSequence, 1);
        if (item == nullptr ||
            (item->text() != m_selectedMarker.name
             || item->data(MarkerData::Id).toInt() != m_selectedMarker.id
             || item->data(MarkerData::Sorting) != m_selectedMarker.sorting
             || item->foreground().color() != m_selectedMarker.color)) {

            dialog->reject();
            dialog->deleteLater();

            // This is invoked via a QTimer::singleShot call so that the function returns
            // immediately and the network processing is not blocked until the dialog is closed
            QTimer::singleShot(0, this, [this] {
                QMessageBox::warning(this,
                    tr("Kollidierende Netzwerkänderung"),
                    tr("Die Markierung „%1“ wurde zwischenzeitlich von einem anderen "
                       "Netzwerkteilnehmer entweder gelöscht, umbenannt oder geändert. Das Ändern "
                       "ist nicht mehr möglich und wird abgebrochen.").arg(
                       m_selectedMarker.name)); } );
            return;
        }
    }
}

void MarkersWidget::setProcessingRequest(bool processing)
{
    setEnabled(! processing);
    if (processing) {
        QApplication::setOverrideCursor(Qt::WaitCursor);
    }
}

void MarkersWidget::updateTournamentState()
{
    m_defaultMarkerWidget->setVisible(! m_db->tournamentStarted());
    m_absentMarkerBox->setVisible(! m_db->tournamentStarted());
    m_singlesManagementBox->setVisible(m_tournamentSettings->isPairMode()
                                       && ! m_db->tournamentStarted());
}

void MarkersWidget::singlesManagementChanged()
{
    if (! m_tournamentSettings->singlesManagement().enabled) {
        warnAboutSpecialMarkers();
    }

    // First, we cache the currently entered data
    const Muckturnier::SinglesManagement singlesManagement {
        m_singlesManagementBox->isChecked(),    // enabled
        m_singleMarker->currentData().toInt(),  // singleMarker
        m_markSingles->isChecked(),             // markSingles
        m_singleConditionString->text(),        // conditionString
        m_singleCondition->currentIndex(),      // condition
        m_assignedMarker->currentData().toInt() // assignedMarker
    };

    if (m_isClient) {
        // Reset the GUI to the last state in case the request fails
        setSinglesManagement(m_tournamentSettings->singlesManagement());
        // Request the change
        m_client->requestSinglesManagementChange(singlesManagement);
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    m_tournamentSettings->setSinglesManagement(singlesManagement);
    emit updatePlayersList();

    if (m_isServer) {
        m_server->broadcastSinglesManagementChange(singlesManagement);
    }

    QApplication::restoreOverrideCursor();
}

int MarkersWidget::singleMarker() const
{
    return m_singlesManagementBox->isChecked() ? m_singleMarker->currentData().toInt() : -1;
}

int MarkersWidget::assignedMarker() const
{
    return m_assignedMarker->currentData().toInt();
}

void MarkersWidget::setSinglesManagement(const Muckturnier::SinglesManagement &data)
{
    m_singlesManagementBox->blockSignals(true);
    m_singleMarker->blockSignals(true);
    m_markSingles->blockSignals(true);
    m_singleConditionString->blockSignals(true);
    m_singleCondition->blockSignals(true);
    m_assignedMarker->blockSignals(true);

    m_singlesManagementBox->setChecked(data.enabled);
    setMarkersComboBoxSelection(m_singleMarker, data.singleMarker);
    m_markSingles->setChecked(data.markSingles);
    m_singleConditionString->setText(data.conditionString);
    m_singleCondition->setCurrentIndex(data.condition == 0 ? 0 : 1);
    setMarkersComboBoxSelection(m_assignedMarker, data.assignedMarker);

    m_singlesManagementBox->blockSignals(false);
    m_singleMarker->blockSignals(false);
    m_markSingles->blockSignals(false);
    m_singleConditionString->blockSignals(false);
    m_singleCondition->blockSignals(false);
    m_assignedMarker->blockSignals(false);
}

void MarkersWidget::resetSingleConditionString()
{
    m_singleConditionString->setText(m_settings->namesSeparator().simplified());
    if (m_isClient || m_isServer) {
        singlesManagementChanged();
    }
}

QString MarkersWidget::singleConditionString() const
{
    return m_singleConditionString->text();
}

void MarkersWidget::setAbsentMarker(int markerId)
{
    m_absentMarkerBox->blockSignals(true);
    m_absentMarker->blockSignals(true);
    m_absentMarkerBox->setChecked(markerId != -1);
    if (markerId != -1) {
        setMarkersComboBoxSelection(m_absentMarker, markerId);
    }
    m_absentMarkerBox->blockSignals(false);
    m_absentMarker->blockSignals(false);
}

void MarkersWidget::absentMarkerChanged()
{
    if (m_tournamentSettings->absentMarker() == -1) {
        warnAboutSpecialMarkers();
    }

    QString absentMarkerName;
    int absentMarkerId = -1;
    if (m_absentMarkerBox->isChecked()) {
        absentMarkerName = m_absentMarker->currentText();
        absentMarkerId = m_absentMarker->currentData().toInt();
    }

    if (m_isClient) {
        // First reset the GUI to the last setting in case the request fails
        setAbsentMarker(m_tournamentSettings->absentMarker());
        // Then request the change
        m_client->requestSetAbsentMarker(absentMarkerName, absentMarkerId);
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    m_tournamentSettings->setAbsentMarker(absentMarkerId);

    if (m_isServer) {
        m_server->broadcastSetAbsentMarker(absentMarkerName, absentMarkerId);
    }

    emit updatePlayersList();

    QApplication::restoreOverrideCursor();
}

void MarkersWidget::warnAboutSpecialMarkers()
{
    if (m_specialMarkersWarnigAlreadyShown) {
        return;
    }

    QTimer::singleShot(0, this, [this] {
        QMessageBox::information(this, tr("Beenden der Anmeldung"),
            tr("<p><b>Bitte beachten:</b></p>"
               "<p>So lange es ignorierte und/oder als allein gekommene Anmeldungen gibt, kann "
               "das Turnier nicht gestartet werden.<p>"
               "<p>Bitte zum Starten des Turniers alle entsprechenden Markierungen entfernen, "
               "die so markierten Anmeldungen löschen oder die entsprechende Option deaktivieren."
               "</p>"));
    });

    m_specialMarkersWarnigAlreadyShown = true;
}
