/* Copyright (c) 2018-2019 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef MARKERSWIDGET_H
#define MARKERSWIDGET_H

// Local includes
#include "shared/Muckturnier.h"

// Qt includes
#include <QWidget>
#include <QIcon>
#include <QCryptographicHash>

// Local classes
class SharedObjects;
class Database;
class Settings;
class TournamentSettings;
class TitleMenu;
class Server;
class Client;

// Qt classes
class QTableWidget;
class QAction;
class QMenu;
class QComboBox;
class QLineEdit;
class QGroupBox;
class QCheckBox;

class MarkersWidget : public QWidget
{
    Q_OBJECT

public:
    enum MarkerData {
        Id = Qt::UserRole,
        Sorting
    };

    explicit MarkersWidget(QWidget *parent, SharedObjects *sharedObjects);
    void updateTournamentMode();
    void reload();
    QVector<Muckturnier::Marker> getMarkers() const;
    void reset();
    bool isMarkerVisible(int id) const;
    void updateCounters(const QHash<int, int> &counter);
    int newRegistrationMarker(const QString &name) const;
    void addChecksumData(QCryptographicHash &hash) const;
    void setServer(Server *server);
    void setClient(Client *client);
    void serverStopped();
    void clientDisconnected();
    void processNetworkChange();
    void setProcessingRequest(bool processing);
    void updateTournamentState();
    int singleMarker() const;
    int assignedMarker() const;
    void resetSingleConditionString();
    QString singleConditionString() const;
    void setAbsentMarker(int markerId);
    void setSinglesManagement(const Muckturnier::SinglesManagement &data);
    void warnAboutSpecialMarkers();

signals:
    void markersChanged();
    void statusUpdate(const QString &text);
    void visibilityChanged();
    void updatePlayersList();

private slots:
    void showContextMenu(const QPoint &position);
    void addMarker();
    void saveNewMarker(const QString &name, const QColor &color,
                       Muckturnier::MarkerSorting sorting);
    void moveMarker(int direction);
    void editMarker();
    void saveEdit(const QString &name, const QColor &color, Muckturnier::MarkerSorting sorting);
    void deleteMarker();
    void checkShowContextMenu(int, int column);
    void changeVisibility(int row, int column);
    void showAll();
    void showOnly(int idToShow);
    void absentMarkerChanged();
    void singlesManagementChanged();

private: // Functions
    void updateVisibilityIcons();
    void updateVisibilityIcon(int id, int row);
    bool checkSelectedMarkerChanged(const QString &name, const QString &action);
    int markersComboBoxSelection(QComboBox *comboBox) const;
    void setMarkersComboBoxSelection(QComboBox *comboBox, const int markerId);

private: // Variables
    Database *m_db;
    Settings *m_settings;
    TournamentSettings *m_tournamentSettings;

    QTableWidget *m_markers;
    TitleMenu *m_markerMenu;
    QAction *m_editMarkerAction;
    QAction *m_moveUpAction;
    QAction *m_moveDownAction;
    QAction *m_deleteMarkerAction;
    QAction *m_addMarkerAction;
    QMenu *m_addMarkerMenu;
    QMenu *m_hideShowMarkedMenu;
    QAction *m_showOnlyAction;
    QAction *m_showAllAction;
    Muckturnier::Marker m_selectedMarker;
    int m_selectedMarkerSequence;
    QIcon m_unused;
    QIcon m_visible;
    QIcon m_hidden;
    QHash<int, bool> m_isVisible;
    int m_allMarkersCount;

    QWidget *m_defaultMarkerWidget;
    QComboBox *m_defaultMarker;

    QGroupBox *m_absentMarkerBox;
    QComboBox *m_absentMarker;

    QGroupBox *m_singlesManagementBox;
    QComboBox *m_singleMarker;
    QCheckBox *m_markSingles;
    QComboBox *m_singleCondition;
    QLineEdit *m_singleConditionString;
    QComboBox *m_assignedMarker;

    bool m_specialMarkersWarnigAlreadyShown = false;

    Server *m_server;
    Client *m_client;
    bool m_isServer = false;
    bool m_isClient = false;
};

#endif // MARKERSWIDGET_H
