/* Copyright (c) 2011-2020 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes

#include "PlayersPage.h"
#include "PlayersPageDelegate.h"
#include "MarkersWidget.h"
#include "MarkersPopup.h"
#include "AssignmentPopup.h"
#include "AssignmentPage.h"

#include "ServerPage/Server.h"
#include "ClientPage/Client.h"

#include "shared/SharedObjects.h"
#include "shared/StringTools.h"
#include "shared/Database.h"
#include "shared/SearchWidget.h"
#include "shared/SearchEngine.h"
#include "shared/TitleMenu.h"
#include "shared/BackupEngine.h"
#include "shared/ResourceFinder.h"
#include "shared/SharedStyles.h"
#include "shared/Settings.h"
#include "shared/defaultValues.h"
#include "shared/TournamentSettings.h"
#include "shared/Muckturnier.h"

// Qt includes
#include <QDebug>
#include <QVBoxLayout>
#include <QGroupBox>
#include <QHBoxLayout>
#include <QLabel>
#include <QLineEdit>
#include <QPushButton>
#include <QTableWidget>
#include <QHeaderView>
#include <QScrollBar>
#include <QAction>
#include <QMessageBox>
#include <QApplication>
#include <QSplitter>
#include <QCryptographicHash>
#include <QCheckBox>
#include <QDir>
#include <QTimer>

// C++ includes
#include <functional>

PlayersPage::PlayersPage(QWidget *parent, SharedObjects *sharedObjects)
    : QWidget(parent),
      m_db(sharedObjects->database()),
      m_settings(sharedObjects->settings()),
      m_tournamentSettings(sharedObjects->tournamentSettings()),
      m_stringTools(sharedObjects->stringTools()),
      m_searchEngine(sharedObjects->searchEngine()),
      m_columnIndex { { Column::Name,       0 },
                      { Column::Mark,       1 },
                      { Column::Assignment, 2 },
                      { Column::Marker,     3 } },
      m_triggerHeaderColumns { Column::Assignment,
                               Column::Marker }
{
    QVBoxLayout *layout = new QVBoxLayout(this);

    // Register new players/pairs

    QWidget *registerWidget = new QWidget;
    QVBoxLayout *registerLayout = new QVBoxLayout(registerWidget);

    m_newPlayersBox = new QGroupBox;
    m_newPlayersBox->setSizePolicy(QSizePolicy::Minimum, QSizePolicy::Maximum);
    QHBoxLayout *newPlayersBoxLayout = new QHBoxLayout(m_newPlayersBox);
    registerLayout->addWidget(m_newPlayersBox);

    m_nameLabel = new QLabel;
    newPlayersBoxLayout->addWidget(m_nameLabel);

    m_playersName = new QLineEdit;
    newPlayersBoxLayout->addWidget(m_playersName);
    connect(m_playersName, &QLineEdit::textChanged, [this] (const QString &text) {
            m_registerButton->setEnabled(! text.simplified().isEmpty()); } );
    connect(m_playersName, &QLineEdit::returnPressed, this, &PlayersPage::registerPlayers);

    m_registerButton = new QPushButton(tr("Anmelden"));
    m_registerButton->setEnabled(false);
    newPlayersBoxLayout->addWidget(m_registerButton);
    connect(m_registerButton, &QPushButton::clicked, this, &PlayersPage::registerPlayers);

    // Registered players/pairs

    m_playersBox = new QGroupBox;
    QVBoxLayout *playersBoxLayout = new QVBoxLayout(m_playersBox);
    registerLayout->addWidget(m_playersBox);

    m_isSearchLabel = new QLabel;
    m_isSearchLabel->setAlignment(Qt::AlignCenter);
    m_isSearchLabel->hide();
    playersBoxLayout->addWidget(m_isSearchLabel);

    m_playersList = new QTableWidget(this);
    m_playersList->setColumnCount(4);
    m_playersList->setColumnHidden(m_columnIndex[Column::Mark], true);
    m_playersList->setColumnHidden(m_columnIndex[Column::Assignment], true);
    m_playersList->setColumnHidden(m_columnIndex[Column::Marker], true);
    m_delegate = new PlayersPageDelegate(m_playersList);
    m_playersList->setItemDelegate(m_delegate);
    connect(m_delegate, &PlayersPageDelegate::startedEdit,
            this, [this](int row) { m_editedRow = row; } );
    connect(m_delegate, &PlayersPageDelegate::editEntered, this, &PlayersPage::editEntered);
    m_playersList->setHorizontalHeaderLabels( { tr("Name"),
                                                QString(),
                                                tr("Auslosung"),
                                                tr("Markierung") } );
    m_playersList->horizontalHeader()->hide();
    m_playersList->verticalHeader()->hide();
    const int iconSize = QFontMetrics(QFont(font())).boundingRect(QStringLiteral("M")).height()
#ifdef Q_OS_WIN
                         * 1.9; // On Windows, the icon looks better if it's a bit bigger
#else
                         * 1.5; // On Linux, a bit smaller looks better
#endif
    m_playersList->setIconSize(QSize(iconSize, iconSize));
    m_playersList->setShowGrid(false);
    m_playersList->setWordWrap(false);
    m_playersList->setSelectionMode(QTableWidget::NoSelection);
    m_playersList->setFocusPolicy(Qt::NoFocus);
    m_playersList->setHorizontalScrollMode(QAbstractItemView::ScrollPerPixel);
    connect(m_playersList->verticalScrollBar(), &QScrollBar::valueChanged,
            m_playersList, &QTableWidget::resizeColumnsToContents);
    m_playersList->setContextMenuPolicy(Qt::CustomContextMenu);
    connect(m_playersList, &QTableWidget::customContextMenuRequested,
            this, &PlayersPage::showContextMenu);
    connect(m_playersList, &QTableWidget::cellClicked, this, &PlayersPage::checkShowPopup);

    m_markerIcon = QIcon(sharedObjects->resourceFinder()->find(QStringLiteral("marker.svg")));

    playersBoxLayout->addWidget(m_playersList);

    // Search widget
    m_searchWidget = new SearchWidget;
    connect(m_searchWidget, &SearchWidget::searchChanged, this, &PlayersPage::updateSearch);
    playersBoxLayout->addWidget(m_searchWidget);

    // Markers

    m_markersWidget = new MarkersWidget(this, sharedObjects);
    connect(m_markersWidget, &MarkersWidget::markersChanged, this, &PlayersPage::markersChanged);
    connect(m_markersWidget, &MarkersWidget::statusUpdate, this, &PlayersPage::statusUpdate);
    connect(m_markersWidget, &MarkersWidget::visibilityChanged,
            this, &PlayersPage::markersVisibilityChanged);
    connect(m_markersWidget, &MarkersWidget::updatePlayersList,
            this, &PlayersPage::updatePlayersList);

    QWidget *markersPartWidget = new QWidget;
    QVBoxLayout *markersPartWidgetLayout = new QVBoxLayout(markersPartWidget);
    markersPartWidgetLayout->addWidget(m_markersWidget);

    // Column options

    markersPartWidgetLayout->addWidget(new QLabel(tr("Spalten anzeigen:")));
    QHBoxLayout *showColumnsLayout = new QHBoxLayout;
    markersPartWidgetLayout->addLayout(showColumnsLayout);

    for (Column column : ALL_COLUMNS) {
        m_showColumn[column] = new QCheckBox(tr(CHECKBOX_TEXT.value(column)));
        connect(m_showColumn.value(column), &QCheckBox::stateChanged,
                this, std::bind(&PlayersPage::setColumnVisible, this,
                                column, std::placeholders::_1));
        showColumnsLayout->addWidget(m_showColumn.value(column));
    }

    // Arrange the registering part and the markers part in a splitter
    QSplitter *splitter = new QSplitter(this);
    splitter->addWidget(registerWidget);
    splitter->setCollapsible(0, false);
    splitter->setStretchFactor(0, 5);
    splitter->addWidget(markersPartWidget);
    splitter->setCollapsible(1, false);
    splitter->setStretchFactor(1, 3);
    splitter->setSizePolicy(QSizePolicy::Ignored, QSizePolicy::Expanding);
    layout->addWidget(splitter);

    // Status label
    m_statusLabel = new QLabel;
    m_statusLabel->setWordWrap(true);
    m_statusLabel->setTextInteractionFlags(Qt::TextBrowserInteraction);
    m_statusLabel->setSizePolicy(QSizePolicy::Ignored, QSizePolicy::Maximum);
    connect(m_statusLabel, &QLabel::linkActivated, this, &PlayersPage::showInfoPopup);
    layout->addWidget(m_statusLabel);

    // Markers popup
    m_markersPopup = new MarkersPopup(this);
    connect(m_markersPopup, &QDialog::rejected, this, &PlayersPage::resetHighlight);
    connect(m_markersPopup, &QDialog::accepted, this, [this] {
        updateMarker(m_markersPopup->markerName(), m_markersPopup->markerId());
    });

    // Assignment popup
    m_assignmentPopup = new AssignmentPopup(this);
    connect(m_assignmentPopup, &QDialog::rejected, this, &PlayersPage::resetHighlight);
    connect(m_assignmentPopup, &QDialog::accepted, this, [this] {
        updateAssignment(m_assignmentPopup->table(),
                         m_assignmentPopup->pair(),
                         m_assignmentPopup->removeMarker());
    });

    // Flash scrollto target reset timer
    m_resetHighlightTimer = new QTimer(this);
    m_resetHighlightTimer->setInterval(320);
    m_resetHighlightTimer->setSingleShot(true);
    connect(m_resetHighlightTimer, &QTimer::timeout, this, &PlayersPage::resetHighlight);

    // Context menu

    m_contextMenu = new TitleMenu(this);
    connect(m_contextMenu, &QMenu::aboutToHide, this, &PlayersPage::resetHighlight);

    m_assignMenu = m_contextMenu->addMenu(tr("Einzelspieler zuordnen"));
    m_assignMenu->menuAction()->setVisible(false);

    m_editAction = m_contextMenu->addAction(QString());
    connect(m_editAction, &QAction::triggered,
            [this] { m_playersList->editItem(m_playersList->currentItem()); });

    m_deleteAction = m_contextMenu->addAction(QString());
    connect(m_deleteAction, &QAction::triggered, this, &PlayersPage::deletePlayers);

    m_contextMenu->addSeparator();

    m_setOnlyMarkerAction = m_contextMenu->addAction(QString());
    m_setOnlyMarkerAction->setVisible(false);

    m_setMarkerMenu = m_contextMenu->addMenu(tr("Markieren als"));

    m_removeMarkerAction = m_contextMenu->addAction(tr("Markierung entfernen"));
    connect(m_removeMarkerAction, &QAction::triggered,
            this, std::bind(&PlayersPage::updateMarker, this, QString(), 0));

    m_contextMenu->addSeparator();

    m_setAssignmentAction = m_contextMenu->addAction(tr("Auslosung eingeben"));
    connect(m_setAssignmentAction, &QAction::triggered, this, [this] {
        checkShowPopup(m_playersList->currentItem()->row(), m_columnIndex[Column::Assignment]);
    });

    m_deleteAssignmentAction = m_contextMenu->addAction(tr("Auslosung löschen"));
    connect(m_deleteAssignmentAction, &QAction::triggered,
            this, std::bind(&PlayersPage::updateAssignment, this, 0, 0, false));

    // Whole list menu

    m_listMenu = new TitleMenu(tr("Ganze Liste"), this);
    m_listMenu->setTitle(tr("Ganze Liste"));

    m_contextMenu->addSeparator();
    m_contextMenu->addMenu(m_listMenu);

    m_listSetOnlyMarkerAction = m_listMenu->addAction(QString());

    m_listSetMarkerMenu = m_listMenu->addMenu(tr("Markieren als"));

    m_listRemoveOnlyMarkerAction = m_listMenu->addAction(QString());

    m_listRemoveMarkerMenu = m_listMenu->addMenu(tr("Markierungen entfernen"));

    m_listRemoveAllMarkersAction = m_listMenu->addAction(tr("Alle Markierungen entfernen"));
    connect(m_listRemoveAllMarkersAction, &QAction::triggered,
            this, std::bind(&PlayersPage::listUpdateMarker, this, QString(), 0));

    m_listMenu->addSeparator();

    m_deleteOnlyMarkedAction = m_listMenu->addAction(QString());

    m_deleteMarkedMenu = m_listMenu->addMenu(QString());

    m_deleteUnmarkedAction = m_listMenu->addAction(QString());
    connect(m_deleteUnmarkedAction, &QAction::triggered, this, &PlayersPage::deleteUnmarked);

    m_deleteAllAction = m_listMenu->addAction(QString());
    connect(m_deleteAllAction, &QAction::triggered, this, &PlayersPage::deleteAll);

    m_listMenu->addSeparator();

    m_deleteAllAssignmentsAction = m_listMenu->addAction(tr("Alle Auslosungen löschen"));
    connect(m_deleteAllAssignmentsAction, &QAction::triggered,
            this, &PlayersPage::deleteAllAssignments);
}

void PlayersPage::reload(bool updateMarkersWidget)
{
    m_newPlayersBox->show();

    switch (m_tournamentSettings->tournamentMode()) {
    case Muckturnier::TournamentMode::FixedPairs:
        m_newPlayersBox->setTitle(tr("Neue Paare anmelden"));
        m_playersBox->setTitle(tr("Angemeldete Paare"));
        m_nameLabel->setText(tr("Name(n):"));
        m_editAction->setText(tr("Name(n) ändern"));
        m_deleteAction->setText(tr("Paar löschen"));
        m_deleteMarkedMenu->menuAction()->setText(tr("Alle Paare mit Markierung löschen"));
        m_deleteUnmarkedAction->setText(tr("Alle unmarkierten Paare löschen"));
        m_deleteAllAction->setText(tr("Alle Paare löschen"));
        break;
    case Muckturnier::TournamentMode::SinglePlayers:
        m_newPlayersBox->setTitle(tr("Neue Spieler anmelden"));
        m_playersBox->setTitle(tr("Angemeldete Spieler"));
        m_nameLabel->setText(tr("Name:"));
        m_editAction->setText(tr("Name ändern"));
        m_deleteAction->setText(tr("Spieler löschen"));
        m_deleteMarkedMenu->menuAction()->setText(tr("Alle Spieler mit Markierung löschen"));
        m_deleteUnmarkedAction->setText(tr("Alle unmarkierten Spieler löschen"));
        m_deleteAllAction->setText(tr("Alle Spieler löschen"));
        break;
    }

    m_newPlayersBox->setEnabled(m_db->isWritable());
    m_playersList->setEditTriggers(m_db->isWritable() ? QAbstractItemView::DoubleClicked
                                                      : QAbstractItemView::NoEditTriggers);
    m_assignmentPopup->setPairMode(m_tournamentSettings->isPairMode());
    m_markersWidget->updateTournamentMode();

    // This will also trigger updatePlayersList() each
    if (updateMarkersWidget) {
        m_markersWidget->reload();
    } else {
        markersChanged();
    }

    if (m_assignmentPage != nullptr) {
        checkUpdateAssignmentPage();
    }
}

void PlayersPage::markersChanged()
{
    const QVector<Muckturnier::Marker> markers(m_markersWidget->getMarkers());
    const int count = markers.count();

    // No need to process count < 2, because the actions will be hidden in this case

    if (count == 2) {
        // We have exactly one marker

        m_setOnlyMarkerAction->disconnect();
        m_listSetOnlyMarkerAction->disconnect();
        m_listRemoveOnlyMarkerAction->disconnect();
        m_deleteOnlyMarkedAction->disconnect();

        for (const Muckturnier::Marker &marker : markers) {
            if (marker.id != 0) {
                m_setOnlyMarkerAction->setText(tr("Markieren als „%1“").arg(marker.name));
                connect(m_setOnlyMarkerAction, &QAction::triggered,
                        this, std::bind(&PlayersPage::updateMarker, this, marker.name, marker.id));

                m_listSetOnlyMarkerAction->setText(tr("Markieren als „%1“").arg(marker.name));
                connect(m_listSetOnlyMarkerAction, &QAction::triggered,
                        this, std::bind(&PlayersPage::listUpdateMarker, this,
                                        marker.name, marker.id));

                m_listRemoveOnlyMarkerAction->setText(tr("Markierung „%1“ entfernen").arg(
                                                          marker.name));
                connect(m_listRemoveOnlyMarkerAction, &QAction::triggered,
                        this, std::bind(&PlayersPage::listRemoveMarker,
                                        this, marker.name, marker.id));

                m_deleteOnlyMarkedAction->setText((m_tournamentSettings->isPairMode()
                    ? tr("Alle als „%1“ markierten Paare löschen")
                    : tr("Alle als „%1“ markierten Spieler löschen")).arg(marker.name));
                m_deleteOnlyMarkedAction->setData(marker.id);
                connect(m_deleteOnlyMarkedAction, &QAction::triggered,
                        this, std::bind(&PlayersPage::deleteMarked, this, marker.name, marker.id));

                break;
            }
        }
    }

    if (count > 2) {
        // We have multiple markers
        m_setMarkerMenu->clear();
        m_listSetMarkerMenu->clear();
        m_listRemoveMarkerMenu->clear();
        m_deleteMarkedMenu->clear();

        for (const Muckturnier::Marker &marker : markers) {
            if (marker.id == 0) {
                continue;
            }

            QAction *setMarkerAction = m_setMarkerMenu->addAction(marker.name);
            setMarkerAction->setData(marker.id);
            connect(setMarkerAction, &QAction::triggered,
                    this, std::bind(&PlayersPage::updateMarker, this, marker.name, marker.id));

            QAction *listSetMarkerAction = m_listSetMarkerMenu->addAction(marker.name);
            listSetMarkerAction->setData(marker.id);
            connect(listSetMarkerAction, &QAction::triggered,
                    this, std::bind(&PlayersPage::listUpdateMarker, this, marker.name, marker.id));

            QAction *listRemoveMarkerAction = m_listRemoveMarkerMenu->addAction(marker.name);
            listRemoveMarkerAction->setData(marker.id);
            connect(listRemoveMarkerAction, &QAction::triggered,
                    this, std::bind(&PlayersPage::listRemoveMarker, this, marker.name, marker.id));

            QAction *deleteMarkedAction = m_deleteMarkedMenu->addAction(marker.name);
            deleteMarkedAction->setData(marker.id);
            connect(deleteMarkedAction, &QAction::triggered,
                    this, std::bind(&PlayersPage::deleteMarked, this, marker.name, marker.id));
        }
    }

    m_setOnlyMarkerAction->setVisible(count == 2);
    m_setMarkerMenu->menuAction()->setVisible(count > 2);
    m_removeMarkerAction->setVisible(count > 1);

    m_listSetOnlyMarkerAction->setVisible(count == 2);
    m_listSetMarkerMenu->menuAction()->setVisible(count > 2);
    m_listRemoveOnlyMarkerAction->setVisible(count == 2);
    m_listRemoveMarkerMenu->menuAction()->setVisible(count > 2);
    m_listRemoveAllMarkersAction->setVisible(count > 2);

    m_deleteOnlyMarkedAction->setVisible(count == 2);
    m_deleteMarkedMenu->menuAction()->setVisible(count > 2);
    m_deleteUnmarkedAction->setVisible(count > 2);

    listModified();

    // Update the markers popup
    m_markersPopup->updateMarkers(markers);
}

void PlayersPage::updatePlayersList()
{
    // Be sure not to produce a segfault via the resetHighlight timer call
    m_highlightedItem = nullptr;
    if (m_resetHighlightTimer->isActive()) {
        m_resetHighlightTimer->stop();
    }

    // Create a list of all blocks we need
    QHash<int, int> targetBlock;
    QVector<QVector<QString>> block;
    QHash <int, QColor> markerColor;
    QHash <int, QString> markerName;
    for (const Muckturnier::Marker &marker : m_markersWidget->getMarkers()) {
        switch (marker.sorting) {
        case Muckturnier::MarkerSorting::NewBlock:
            block.append(QVector<QString>());
            targetBlock[marker.id] = block.count() - 1;
            break;
        case Muckturnier::MarkerSorting::Merged:
            targetBlock[marker.id] = block.count() - 1;
            break;
        }
        if (marker.id != 0) {
            markerColor[marker.id] = marker.color;
            markerName[marker.id] = marker.name;
        }
    }

    // In case we have invalid markers set in the players list, the program would crash. So we
    // have to catch this case. This shouldn't happen however unless the database is manipulated
    // by hand or reading the markers failed.

    // If targetBlock doesn't contain a value for 0 ("unmarked"), we have to add it
    if (! targetBlock.contains(0)) {
        block.append(QVector<QString>());
        targetBlock.insert(0, block.count() - 1);
    }

    // Fetch the players data
    bool success;
    QHash<QString, Database::PlayersData> playersData(m_db->playersData(&success));
    if (! success) {
        return;
    }
    QStringList allRegistrations(playersData.keys());
    m_stringTools->sortList(allRegistrations);

    // Search for invalid markers and map them to "unmarked"
    for (const QString &name : allRegistrations) {
        Database::PlayersData &data = playersData[name];
        if (! targetBlock.contains(data.marker)) {
            data.marker = 0;
        }
    }

    const int oldRegistrationsCount = m_playersList->rowCount();
    const int registrationsCount = allRegistrations.count();
    m_playersList->blockSignals(true);
    m_playersList->setRowCount(registrationsCount);
    m_searchEngine->clearSearchCache();

    // Iterate through all players, add them to their respective blocks and count the markers

    QHash<int, int> markersCount;

    for (const QString &name : allRegistrations) {
        m_searchEngine->updateSearchCache(name);
        block[targetBlock[playersData.value(name).marker]].append(name);
        markersCount[playersData.value(name).marker]++;
    }

    // Display all blocks

    int blockRow = 0;
    QVector<int> assignedPair1Tables;
    QVector<int> assignedPair2Tables;

    for (const QVector<QString> &list : block) {
        for (const QString &name : list) {
            const Database::PlayersData &data = playersData.value(name);

            QTableWidgetItem *nameCell = new QTableWidgetItem(name);
            nameCell->setData(PlayersData::MarkerId, data.marker);
            nameCell->setData(PlayersData::AssignedTable, data.assignedTable);
            nameCell->setData(PlayersData::AssignedPairNumber, data.assignedPairNumber);
            if (data.assignedTable != 0) {
                if (data.assignedPairNumber == 1) {
                    assignedPair1Tables.append(data.assignedTable);
                } else {
                    assignedPair2Tables.append(data.assignedTable);
                }
            }

            QTableWidgetItem *setMarkerCell = new QTableWidgetItem;
            setMarkerCell->setFlags(setMarkerCell->flags() & (~Qt::ItemIsEditable));
            setMarkerCell->setIcon(m_markerIcon);

            QTableWidgetItem *assignmentCell = new QTableWidgetItem;
            assignmentCell->setFlags(assignmentCell->flags() & (~Qt::ItemIsEditable));
            if (data.assignedTable != 0) {
                assignmentCell->setText(tr("Tisch %1 Paar %2").arg(
                                        QString::number(data.assignedTable),
                                        QString::number(data.assignedPairNumber)));
            }

            QTableWidgetItem *markerCell = new QTableWidgetItem;
            markerCell->setFlags(markerCell->flags() & (~Qt::ItemIsEditable));
            if (playersData.value(name).marker != 0) {
                nameCell->setForeground(markerColor.value(playersData.value(name).marker));
                markerCell->setText(markerName.value(playersData.value(name).marker));
            }

            m_playersList->setItem(blockRow, m_columnIndex[Column::Name], nameCell);
            m_playersList->setItem(blockRow, m_columnIndex[Column::Mark], setMarkerCell);
            m_playersList->setItem(blockRow, m_columnIndex[Column::Assignment], assignmentCell);
            m_playersList->setItem(blockRow, m_columnIndex[Column::Marker], markerCell);

            blockRow++;
        }
    }

    std::sort(assignedPair1Tables.begin(), assignedPair1Tables.end());
    std::sort(assignedPair2Tables.begin(), assignedPair2Tables.end());
    const bool assignmentsSet = assignedPair1Tables.count() > 0 || assignedPair2Tables.count() > 0;
    m_deleteAllAssignmentsAction->setEnabled(assignmentsSet);
    m_assignmentPopup->setAssignedTables(assignedPair1Tables, assignedPair2Tables);
    updateVisibility();

    // Update the context menus

    if (m_listSetMarkerMenu->menuAction()->isVisible()) {
        for (QAction *action : m_listSetMarkerMenu->actions()) {
            action->setEnabled(markersCount[action->data().toInt()] < registrationsCount);
        }
        for (QAction *action : m_listRemoveMarkerMenu->actions()) {
            action->setEnabled(markersCount[action->data().toInt()] > 0);
        }
        for (QAction *action : m_deleteMarkedMenu->actions()) {
            action->setEnabled(markersCount[action->data().toInt()] > 0);
        }
    }

    if (markersCount.value(0) == registrationsCount) {
        m_listRemoveOnlyMarkerAction->setEnabled(false);
        m_listRemoveMarkerMenu->setEnabled(false);
        m_listRemoveAllMarkersAction->setEnabled(false);
    } else {
        m_listRemoveOnlyMarkerAction->setEnabled(true);
        m_listRemoveMarkerMenu->setEnabled(true);
        m_listRemoveAllMarkersAction->setEnabled(true);
    }

    if (m_db->tournamentStarted() || markersCount.value(0) == registrationsCount) {
        m_deleteOnlyMarkedAction->setEnabled(false);
        m_deleteMarkedMenu->setEnabled(false);
        m_deleteUnmarkedAction->setEnabled(false);
    } else {
        m_deleteOnlyMarkedAction->setEnabled(
            markersCount.value(m_deleteOnlyMarkedAction->data().toInt()) < registrationsCount);
        m_deleteMarkedMenu->setEnabled(true);
        m_deleteUnmarkedAction->setEnabled(markersCount.value(0) > 0);
    }

    // Update the GUI

    m_markersWidget->updateCounters(markersCount);

    m_playersList->resizeColumnsToContents();
    m_playersList->blockSignals(false);

    m_searchWidget->setVisible(registrationsCount > 0);
    emit registrationsAvailable(registrationsCount > 0);

    const int singleMarker = m_db->tournamentStarted() ? -1 : m_markersWidget->singleMarker();
    const bool oldSinglesPresent = m_singlesPresent;
    m_singlesPresent = markersCount.value(singleMarker) > 0;
    const bool singlesPresentChanged = oldSinglesPresent != m_singlesPresent;

    const int absentMarker = m_db->tournamentStarted() ? -1 : m_tournamentSettings->absentMarker();
    const bool oldAbsentsPresent = m_absentsPresent;
    m_absentsPresent = markersCount.value(absentMarker) > 0;
    const bool absentsPresentChanged = oldAbsentsPresent != m_absentsPresent;

    int presentPairsWithoutSingles = registrationsCount;
    if (singleMarker != -1) {
        presentPairsWithoutSingles -= markersCount[singleMarker];
    }
    if (absentMarker != -1) {
        presentPairsWithoutSingles -= markersCount[absentMarker];
    }

    QString registrationStatus;

    // Update the registration status

    if (m_absentsPresent) {
        registrationStatus.append(tr("<span style=\"%1\">%n Anmeldung(en) ignoriert</span> – ", "",
                                     markersCount.value(absentMarker)).arg(
                                     SharedStyles::STYLE_RED_TEXT));
    }

    registrationStatus.append(m_tournamentSettings->isPairMode()
                                  ? tr("%n Paar(e)", "", presentPairsWithoutSingles)
                                  : tr("%n Spieler", "", presentPairsWithoutSingles));

    if (! m_singlesPresent) {
        // All registrations are valid and can be counted as-is
        registrationStatus.append(tr(" angemeldet"));

        if (presentPairsWithoutSingles > 0) {
            if (m_db->playersDividable(presentPairsWithoutSingles)) {
                registrationStatus.append(tr(" (%n Tisch(e) besetzt)", "",
                    presentPairsWithoutSingles / m_db->necessaryDivisor()));
            } else {
                registrationStatus.append(m_tournamentSettings->isPairMode()
                    ? tr(" – <span style=\"%1\">ein Paar fehlt noch</span>").arg(
                         SharedStyles::STYLE_RED_TEXT)
                    : tr(" – <span style=\"%1\">%n Spieler fehlt/fehlen noch</span>", "",
                         m_db->necessaryDivisor()
                         - presentPairsWithoutSingles % m_db->necessaryDivisor()).arg(
                         SharedStyles::STYLE_RED_TEXT));
                registrationStatus.append(tr(" (dann wäre(n) %n Tisch(e) besetzt)", "",
                    presentPairsWithoutSingles / m_db->necessaryDivisor() + 1));
            }
        }

    } else {
        // We're in pair mode and have to consider single players
        const int remainder = (presentPairsWithoutSingles * 2
                               + markersCount.value(singleMarker)) % 4;
        registrationStatus.append(tr(" und <span style=\"%1\"> %n allein gekommene(r) Spieler"
                                     "</span> angemeldet<br/>", "",
                                     markersCount.value(singleMarker)).arg(
                                     SharedStyles::STYLE_RED_TEXT));
        if (remainder == 0) {
            registrationStatus.append(tr("<span style=\"%1\">Wenn alle zugeordnet sind, kann es "
                                         "losgehen</span>").arg(SharedStyles::STYLE_GREEN_TEXT));
        } else if (remainder == 1) {
            registrationStatus.append(tr("<span style=\"%1\">Es fehlt noch ein Paar und ein "
                                         "einzelner Spieler</span>").arg(
                                         SharedStyles::STYLE_RED_TEXT));
        } else if (remainder == 2) {
            registrationStatus.append(tr("<span style=\"%1\">Es fehlt noch ein Paar</span>").arg(
                                         SharedStyles::STYLE_RED_TEXT));
        } else if (remainder == 3) {
            registrationStatus.append(tr("<span style=\"%1\">Es fehlt noch ein einzelner Spieler"
                                         "</span>").arg(SharedStyles::STYLE_RED_TEXT));
        }
        registrationStatus.append(tr(" (dann wäre(n) %n Tisch(e) besetzt)", "",
                                     (presentPairsWithoutSingles * 2
                                      + markersCount.value(singleMarker)) / 4 + (remainder != 0)));
    }

    // Add the assignment line if we have a preassignment

    if (m_showColumn[Column::Assignment]->isChecked() && assignmentsSet) {
        // Check if all pairs/players have an assignment
        bool unfinished =    assignedPair1Tables.count() * 2 != registrationsCount
                          || assignedPair2Tables.count() * 2 != registrationsCount;

        QVector<int> allAssignedTables;
        for (const int table : assignedPair1Tables) {
            if (! allAssignedTables.contains(table)) {
                allAssignedTables.append(table);
            }
        }
        for (const int table : assignedPair2Tables) {
            if (! allAssignedTables.contains(table)) {
                allAssignedTables.append(table);
            }
        }
        std::sort(allAssignedTables.begin(), allAssignedTables.end());

        // Check for tables with incomplete (missing pair 1 or 2 or incomplete pairs) assignment
        m_incompleteTables.clear();
        if (m_tournamentSettings->isPairMode()) {
            for (const int table : allAssignedTables) {
                if (! assignedPair1Tables.contains(table)
                    || ! assignedPair2Tables.contains(table)) {

                    m_incompleteTables.append(table);
                }
            }
        } else {
            for (const int table : allAssignedTables) {
                if (assignedPair1Tables.count(table) != 2
                    || assignedPair2Tables.count(table) != 2) {

                    m_incompleteTables.append(table);
                }
            }
        }

        // Search for gaps in the assignment
        m_tableGaps.clear();
        int consecutiveNumber = 1;
        for (int i = 0; i < allAssignedTables.count(); i++) {
            if (consecutiveNumber < allAssignedTables.at(i)) {
                for (int j = consecutiveNumber; j < allAssignedTables.at(i); j++) {
                    m_tableGaps.append(j);
                }
                consecutiveNumber = allAssignedTables.at(i) + 1;
            } else {
                consecutiveNumber++;
            }
        }

        const bool incomplete = m_incompleteTables.count() > 0;
        const bool gapsPresent = m_tableGaps.count() > 0;
        if (unfinished || incomplete || gapsPresent) {
            if (unfinished && ! incomplete && ! gapsPresent) {
                registrationStatus.append(tr(
                    "<br/>Alle Tische wurden fortlaufend und vollständig vergeben, aber es haben "
                    "noch nicht alle %1 eine Auslosung.").arg(
                    m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler")));
            } else {
                registrationStatus.append(tr("<br/>"));
            }
            if (incomplete) {
                if (m_tournamentSettings->isPairMode()) {
                    registrationStatus.append(tr("Nicht alle Tische haben ein Paar 1 und 2"));
                } else {
                    registrationStatus.append(tr("Nicht alle Tische haben (ein jeweils "
                                                 "vollständiges) Paar 1 und 2"));
                }
                registrationStatus.append(tr(" (<a href=\"#incompleteTables\">welche?</a>). "));
            }
            if (gapsPresent) {
                registrationStatus.append(tr("Die Tische sind nicht fortlaufend vergeben "
                                             "(<a href=\"#tableGaps\">welche fehlen?</a>)."));
            }
        } else {
            registrationStatus.append(tr("<br/>Die Auslosung ist vollständig und alle Tische "
                                         "wurden fortlaufend und vollständig vergeben."));
        }
    }

    // Finish the update

    m_statusLabel->setText(registrationStatus);

    if (registrationsCount != oldRegistrationsCount
        || singlesPresentChanged || absentsPresentChanged) {

        m_db->setPairsOrPlayersCount(registrationsCount);
        // Emitting the registrationsCountChanged signal will make the MainWindow check for single
        // players being present and hide/show the score dock accordingly, also if the number of
        // registered players wasn't changed by the previous action (e. g. marking an entry as
        // single)
        emit playersCountChanged(registrationsCount);
    }

    if (m_db->tournamentStarted()) {
        m_newPlayersBox->hide();
    }

    scoreChanged();
}

void PlayersPage::scoreChanged()
{
    m_deleteAction->setEnabled(! m_db->tournamentStarted());
    m_deleteAllAction->setEnabled(! m_db->tournamentStarted());
    m_markersWidget->updateTournamentState();
}

void PlayersPage::reset()
{
    m_playersList->clearContents();
    m_newPlayersBox->show();
}

void PlayersPage::registerPlayers()
{
    if (! m_registerButton->isEnabled()) {
        return;
    }

    const QString name = m_playersName->text().simplified();

    // Check for an exact match. In this case, we can't register the pair/player
    const QString lowerName = m_stringTools->toLower(name);
    for (int row = 0; row < m_playersList->rowCount(); row++) {
        if (lowerName == m_stringTools->toLower(m_playersList->item(
                row, m_columnIndex[Column::Name])->text())) {

            QMessageBox::warning(this, m_newPlayersBox->title(),
                tr("<p><b>%1 „%2“ kann nicht angemeldet werden</b></p>"
                   "<p>Der Name ist bereits vergeben.</p>").arg(
                   m_tournamentSettings->isPairMode() ? tr("Das Paar") : tr("Der Spieler"), name));
            m_playersName->selectAll();
            m_playersName->setFocus();
            return;
        }
    }

    // Search for a fuzzy match
    m_searchEngine->setSearchTerm(name);
    for (int row = 0; row < m_playersList->rowCount(); row++) {
        if (m_searchEngine->checkMatch(m_playersList->item(
                row, m_columnIndex[Column::Name])->text())) {

            if (QMessageBox::warning(this,
                m_newPlayersBox->title(),
                tr("<p>Es ist bereits ein %1 mit einem ähnlichen Namen angemeldet:</p>"
                   "<p><table>"
                   "<tr><td>Eingegeben:&nbsp;</td><td><i>%2</i></td></tr>"
                   "<tr><td>Vorhanden:</td><td><i>%3</i></td></tr>"
                   "</table></p>"
                   "<p>Das könnte zu Verwechslungen führen. Soll %4 trotzdem angemeldet "
                   "werden?</p>").arg(
                   m_tournamentSettings->isPairMode() ? tr("Paar") : tr("Spieler"),
                   name,
                   m_playersList->item(row, m_columnIndex[Column::Name])->text(),
                   m_tournamentSettings->isPairMode() ? tr("das Paar") : tr("der Spieler")),
                QMessageBox::Yes | QMessageBox::No, QMessageBox::No) == QMessageBox::No) {

                m_playersName->selectAll();
                m_playersName->setFocus();
                return;
            }
            break;
        }
    }

    if (m_isClient || m_isServer) {
        if (m_playersList->findItems(name, Qt::MatchExactly).count() != 0) {
            QMessageBox::warning(this,
                tr("Kollidierende Netzwerkänderung"),
                tr("Es wurde mittlerweile von einem anderen Netzwerkteilnehmer ein %1 mit dem "
                   "Namen „%2“ angemeldet. Die Anmeldung ist nicht mehr möglich und wird "
                   "abgebrochen.").arg(
                   m_tournamentSettings->isPairMode() ? tr("Paar") : tr("Spieler"), name));
            m_playersName->clear();
            return;
        }
    }

    if (m_isClient) {
        m_playersName->clear();
        m_playersList->setCurrentItem(nullptr);
        m_client->requestRegisterPlayers(name, m_markersWidget->newRegistrationMarker(name));
        m_focusPlayersName = true;
        return;
    }

    emit statusUpdate(m_tournamentSettings->isPairMode() ? tr("Melde Paar an …")
                                                         : tr("Melde Spieler an …"));
    QApplication::setOverrideCursor(Qt::WaitCursor);

    if (! m_db->registerPlayers(name, m_markersWidget->newRegistrationMarker(name))) {
        m_playersName->clear();
        return;
    }
    m_playersName->clear();
    m_playersName->setFocus();
    updatePlayersList();
    m_searchWidget->clear();

    emit statusUpdate((m_tournamentSettings->isPairMode()
                           ? tr("Paar „%1“ angemeldet") : tr("Spieler „%1“ angemeldet")).arg(name));

    scrollTo(name);

    if (m_isServer) {
        m_server->broadcastRegisterPlayers(name, m_markersWidget->newRegistrationMarker(name));
    }

    QApplication::restoreOverrideCursor();
}

void PlayersPage::scrollTo(const QString &text)
{
    QTableWidgetItem *target = m_playersList->findItems(text, Qt::MatchExactly).at(0);
    if (target->row() == m_playersList->rowCount() - 1) {
        m_playersList->scrollToBottom();
    } else {
        m_playersList->scrollToItem(target, QAbstractItemView::PositionAtCenter);
    }

    highlightItem(target);
    m_resetHighlightTimer->start();
}

bool PlayersPage::checkForExactMatch(const QString &name)
{
    const QString lowerName = m_stringTools->toLower(name);
    for (int row = 0; row < m_playersList->rowCount(); row++) {
        if (row == m_editedRow) {
            continue;
        }
        if (lowerName == m_stringTools->toLower(m_playersList->item(
                row, m_columnIndex[Column::Name])->text())) {

            return true;
        }
    }
    return false;
}

void PlayersPage::editEntered(const QString &originalName, const QString &name)
{
    // Check for an exact match. In this case, we can't rename the pair/player
    if (checkForExactMatch(name)) {
        QMessageBox::critical(this,
            m_newPlayersBox->title(),
            tr("<p><b>%1 „%2“ kann nicht umbenannt werden:</b></p>"
               "<p>Der Name „%3“ ist bereits vergeben.</p>").arg(
               m_tournamentSettings->isPairMode() ? tr("Das Paar") : tr("Der Spieler"),
               originalName, name));
        return;
    }

    // Seach for a fuzzy match
    m_searchEngine->setSearchTerm(name);
    for (int row = 0; row < m_playersList->rowCount(); row++) {
        if (row == m_editedRow) {
            continue;
        }
        if (m_searchEngine->checkMatch(m_playersList->item(
                row, m_columnIndex[Column::Name])->text())) {

            if (QMessageBox::warning(this,
                m_newPlayersBox->title(),
                tr("<p>Es ist bereits ein %1 mit einem ähnlichen Namen angemeldet:</p>"
                   "<p><table>"
                   "<tr><td>Eingegeben:&nbsp;</td><td><i>%2</i></td></tr>"
                   "<tr><td>Vorhanden:</td><td><i>%3</i></td></tr>"
                   "</table></p>"
                   "<p>Das könnte zu Verwechslungen führen. Soll %4 trotzdem umbenannt "
                   "werden?</p>").arg(
                   m_tournamentSettings->isPairMode() ? tr("Paar") : tr("Spieler"),
                   name,
                   m_playersList->item(row, m_columnIndex[Column::Name])->text(),
                   m_tournamentSettings->isPairMode() ? tr("das Paar") : tr("Der Spieler")),
                QMessageBox::Yes | QMessageBox::No, QMessageBox::No) == QMessageBox::No) {

                return;
            }
            break;
        }
    }

    if (m_isClient || m_isServer) {
        if (m_playersList->findItems(originalName, Qt::MatchExactly).count() != 1) {
            QMessageBox::warning(this,
                tr("Kollidierende Netzwerkänderung"),
                tr("%1 „%2“ wurde mittlerweile von einem anderen Netzwerkteilnehmer entweder "
                   "umbenannt oder gelöscht. Das Umbenennen ist nicht mehr möglich und wird "
                   "abgebrochen.").arg(
                   m_tournamentSettings->isPairMode() ? tr("Das Paar") : tr("Der Spieler"),
                   originalName));
            return;
        }
        if (m_playersList->findItems(name, Qt::MatchExactly).count() != 0) {
            QMessageBox::warning(this,
                tr("Kollidierende Netzwerkänderung"),
                tr("Es wurde mittlerweile von einem anderen Netzwerkteilnehmer %1 mit dem "
                   "Namen „%2“ angemeldet oder %1 entsprechend umbenannt. Das Umbenennen ist "
                   "nicht mehr möglich und wird abgebrochen.").arg(
                   m_tournamentSettings->isPairMode() ? tr("ein Paar") : tr("ein Spieler"), name));
            return;
        }
    }

    if (m_isClient) {
        m_client->requestRenamePlayers(originalName, name);
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    emit statusUpdate(m_tournamentSettings->isPairMode() ? tr("Benenne Paar um …")
                                                         : tr("Benenne Spieler um …"));

    if (! m_db->renamePlayers(originalName, name)) {
        return;
    }

    emit statusUpdate((m_tournamentSettings->isPairMode()
                          ? tr("Paar „%1“ umbenannt zu „%2“")
                          : tr("Spieler „%1“ umbenannt zu „%2“")).arg(originalName, name));
    listModified();
    checkUpdateAssignmentPage();

    if (m_playersList->findItems(name, Qt::MatchExactly).at(0)->row() != m_editedRow) {
        scrollTo(name);
    }

    if (m_isServer) {
        m_server->broadcastRenamePlayers(originalName, name);
    }

    emit nameEdited();
    QApplication::restoreOverrideCursor();
}

void PlayersPage::showContextMenu(QPoint point)
{
    if (m_playersList->rowCount() == 0 || ! m_db->isWritable()) {
        return;
    }

    QTableWidgetItem *selectedItem = m_playersList->itemAt(point);

    if (m_playersList->horizontalHeader()->isVisible()) {
        point.setY(point.y() + m_playersList->horizontalHeader()->height());
    }
    if (selectedItem == nullptr) {
        m_listMenu->exec(m_playersList->mapToGlobal(point));
        return;
    }

    if (selectedItem->column() != m_columnIndex[Column::Name]) {
        selectedItem = m_playersList->item(selectedItem->row(), m_columnIndex[Column::Name]);
    }

    highlightItem(selectedItem);

    m_selectedName = selectedItem->text();
    const int selectedMark = selectedItem->data(PlayersData::MarkerId).toInt();

    const int singleMarker = m_markersWidget->singleMarker();
    const bool isSinglePlayer = selectedMark == singleMarker;
    m_assignMenu->menuAction()->setVisible(isSinglePlayer);
    m_contextMenu->setTitle((m_tournamentSettings->isPairMode() && ! isSinglePlayer
                                ? tr("Paar „%1“") : tr("Spieler „%1“")).arg(m_selectedName));
    if (isSinglePlayer) {
        m_assignMenu->clear();
        for (int row = 0; row < m_playersList->rowCount(); row++) {
            const QString name = m_playersList->item(row, m_columnIndex[Column::Name])->text();
            if (m_playersList->item(row,
                    m_columnIndex[Column::Name])->data(PlayersData::MarkerId).toInt()
                != singleMarker || name == m_selectedName) {

                continue;
            }
            QAction *action = m_assignMenu->addAction(name);
            connect(action, &QAction::triggered,
                    this, std::bind(&PlayersPage::assignPlayer, this, name));
        }
        if (m_assignMenu->actions().count() == 0) {
            QAction *action = m_assignMenu->addAction(tr("keiner sonst angemeldet"));
            action->setEnabled(false);
        }
    }

    m_removeMarkerAction->setEnabled(selectedMark != 0);
    for (QAction *action : m_setMarkerMenu->actions()) {
        action->setEnabled(action->data().toInt() != selectedMark);
    }

    const int assignedTable = selectedItem->data(PlayersData::AssignedTable).toInt();
    m_setAssignmentAction->setVisible(assignedTable == 0);
    m_deleteAssignmentAction->setVisible(assignedTable != 0);

    m_contextMenu->exec(m_playersList->mapToGlobal(point));
}

void PlayersPage::updateSearch(const QString &searchText)
{
    if (searchText.isEmpty()) {
        for (int row = 0; row < m_playersList->rowCount(); row++) {
            m_playersList->setRowHidden(row, ! isRowVisible(row));
        }
        m_playersList->resizeColumnsToContents();
        m_isSearchLabel->hide();
        return;
    }

    int matches = 0;
    m_searchEngine->setSearchTerm(searchText);
    for (int row = 0; row < m_playersList->rowCount(); row++) {
        const bool rowMatches = m_searchEngine->checkMatch(
                                    m_playersList->item(row, m_columnIndex[Column::Name])->text());
        m_playersList->setRowHidden(row, ! rowMatches);
        matches += rowMatches;
    }

    m_searchWidget->updateMatches(matches);
    m_playersList->resizeColumnsToContents();
    m_isSearchLabel->setText(tr("(Suchergebnis für „%1“)").arg(searchText));
    m_isSearchLabel->show();
}

void PlayersPage::listModified()
{
    m_searchWidget->saveSearch();
    updatePlayersList();
    m_searchWidget->restoreSearch();
}

void PlayersPage::resetSearch()
{
    m_searchWidget->clear();
}

void PlayersPage::displayMarkedEntry()
{
    const int lastRow = m_playersList->currentItem()->row();
    listModified();
    if (m_playersList->findItems(m_selectedName, Qt::MatchExactly).at(0)->row() != lastRow) {
        scrollTo(m_selectedName);
    }
}

void PlayersPage::updateMarker(const QString &markerName, int markerId)
{
    if (m_isClient) {
        m_client->requestUpdateMarker(m_selectedName, markerName, markerId);
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);

    if (markerId != 0) {
        emit statusUpdate((m_tournamentSettings->isPairMode()
                             ? tr("Markiere Paar „%1“ als „%2“ …")
                             : tr("Markiere Spieler „%1“ als „%2“ …")).arg(
                          m_selectedName, markerName));
    } else {
        emit statusUpdate((m_tournamentSettings->isPairMode()
                             ? tr("Entferne Markierung von Paar „%2“ …")
                             : tr("Entferne Markierung von Spieler „%2“ …")).arg(
                          m_selectedName));
    }

    if (! m_db->updateMarker(m_selectedName, markerId)) {
        return;
    }

    if (markerId != 0) {
        emit statusUpdate((m_tournamentSettings->isPairMode()
                             ? tr("Paar „%1“ als „%2“ markiert")
                             : tr("Spieler „%1“ als „%2“ markiert")).arg(
                          m_selectedName, markerName));
    } else {
        emit statusUpdate((m_tournamentSettings->isPairMode()
                             ? tr("Markierung von Paar „%2“ entfernt")
                             : tr("Markierung von Spieler „%2“ entfernt")).arg(
                          m_selectedName));
    }
    displayMarkedEntry();

    if (m_isServer) {
        m_server->broadcastUpdateMarker(m_selectedName, markerName, markerId);
    }

    QApplication::restoreOverrideCursor();
}

void PlayersPage::listUpdateMarker(const QString &markerName, int markerId)
{
    if (m_isClient) {
        m_client->requestListUpdateMarker(markerName, markerId);
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    if (markerId != 0) {
        emit statusUpdate(tr("Markiere alle %1 als „%2“").arg(
                             m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler"),
                             markerName));
    } else {
        emit statusUpdate(tr("Entferne alle Markierungen …"));
    }

    if (! m_db->listUpdateMarker(markerId)) {
        return;
    }

    if (markerId != 0) {
        emit statusUpdate(tr("Alle %1 als „%2“ markiert").arg(
                             m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler"),
                             markerName));
    } else {
        emit statusUpdate(tr("Alle Markierungen entfernt"));
    }
    listModified();

    if (m_isServer) {
        m_server->broadcastListUpdateMarker(markerName, markerId);
    }

    QApplication::restoreOverrideCursor();
}

void PlayersPage::listRemoveMarker(const QString &markerName, int markerId)
{
    if (m_isClient) {
        m_client->requestListRemoveMarker(markerName, markerId);
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    emit statusUpdate(tr("Entferne alle „%1“-Markierungen …").arg(markerName));

    if (! m_db->listRemoveMarker(markerId)) {
        return;
    }

    emit statusUpdate(tr("Alle „%1“-Markierungen entfernt").arg(markerName));
    listModified();

    if (m_isServer) {
        m_server->broadcastListRemoveMarker(markerName, markerId);
    }

    QApplication::restoreOverrideCursor();
}

void PlayersPage::deletePlayers()
{
    if (QMessageBox::question(this,
            m_tournamentSettings->isPairMode() ? tr("Paar löschen") : tr("Spieler löschen"),
            (m_tournamentSettings->isPairMode()
                ? tr("Soll Paar „%1“ wirklich gelöscht werden?")
                : tr("Soll Spieler „%1“ wirklich gelöscht werden?")).arg(
            m_selectedName),
            QMessageBox::Yes | QMessageBox::Cancel,
            QMessageBox::Cancel) == QMessageBox::Cancel) {
        return;
    }

    if (m_isClient || m_isServer) {
        if (m_playersList->findItems(m_selectedName, Qt::MatchExactly).count() != 1) {
            QMessageBox::warning(this,
                tr("Kollidierende Netzwerkänderung"),
                tr("%1 „%2“ wurde mittlerweile von einem anderen Netzwerkteilnehmer entweder "
                   "gelöscht oder umbenannt. Das Löschen ist nicht mehr möglich und wird "
                   "abgebrochen.").arg(
                   m_tournamentSettings->isPairMode() ? tr("Das Paar") : tr("Der Spieler"),
                   m_selectedName));
            return;
        }
    }

    if (m_isClient) {
        m_client->requestDeletePlayers(m_selectedName);
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    emit statusUpdate((m_tournamentSettings->isPairMode()
                          ? tr("Lösche Paar „%1“ …") : tr("Lösche Spieler „%1“ …")).arg(
                       m_selectedName));

    if (! m_db->deletePlayers(m_selectedName)) {
        return;
    }

    emit statusUpdate((m_tournamentSettings->isPairMode()
                          ? tr("Paar „%1“ gelöscht") : tr("Spieler „%1“ gelöscht")).arg(
                      m_selectedName));

    listModified();
    checkUpdateAssignmentPage();

    if (m_isServer) {
        m_server->broadcastDeletePlayers(m_selectedName);
    }

    QApplication::restoreOverrideCursor();
}

void PlayersPage::cacheChecksum()
{
    if (m_isServer || m_isClient) {
        m_cachedChecksum = checksum();
    }
}

bool PlayersPage::checkCachedChecksum(const QString &title)
{
    if (! m_isClient && ! m_isServer) {
        return true;
    }

    if (checksum() != m_cachedChecksum) {
        QMessageBox::warning(this, title,
            tr("<p><b>Kollidierende Netzwerkänderung</b></p>"
               "<p>Die Anmeldungsliste wurde zwischenzeitlich von einem anderen Netzwerkteilnehmer "
               "verändert. Die Aktion kann nicht mehr durchgeführt werden und wird abgebrochen."
               "</p>"));
        return false;
    }

    return true;
}

bool PlayersPage::confirmListDeleteAction(const QString &title, const QString &text)
{
    m_backupFile.clear();

    QMessageBox messageBox(QMessageBox::Warning, title, text,
                           QMessageBox::Yes | QMessageBox::Cancel, this);
    messageBox.setDefaultButton(QMessageBox::Cancel);
    QCheckBox *createBackup = new QCheckBox(tr("Vorher ein Backup der Datenbank anlegen"));
    createBackup->setChecked(true);
    messageBox.setCheckBox(createBackup);

    if (messageBox.exec() == QMessageBox::Cancel) {
        return false;
    }

    if (createBackup->isChecked()) {
        const QString openedDb = m_db->dbFile();
        BackupEngine backupEngine(openedDb);
        if (! m_db->saveSettings()) {
            return false;
        }
        m_db->closeDatabaseConnection();
        bool success = backupEngine.createBackup();
        m_db->openDatabase(openedDb);

        if (! success) {
            QMessageBox::critical(this, title,
                tr("<p><b>Das Anlegen eines Backups ist fehlgeschlagen!</b></p>"
                   "<p>Die Datei konnte nicht angelegt werden. Bitte die Zugriffsrechte von „%1“ "
                   "bzw. dem Verzeichnis „%2“) überprüfen!</p>").arg(
                   backupEngine.backupFile(), QDir::toNativeSeparators(backupEngine.path())));
            return false;
        }

        m_backupFile = backupEngine.backupFile();
    }

    return true;
}

void PlayersPage::checkDisplayBackup(const QString &title)
{
    if (! m_backupFile.isEmpty()) {
        QMessageBox::information(this, title,
            tr("Es wurde erfolgreich ein Backup der Datenbank mit dem Dateinamen „%1“ "
               "erstellt.").arg(m_backupFile));
    }
}

void PlayersPage::deleteMarked(const QString &markerName, int markerId)
{
    cacheChecksum();
    const QString title = tr("Markierte %1 löschen").arg(
                             m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler"));

    if (! confirmListDeleteAction(title,
            tr("Sollen wirklich alle als „%1“ markierten %2 gelöscht werden?").arg(markerName,
               m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler")))
        || ! checkCachedChecksum(title)) {

        return;
    }

    if (m_isClient) {
        m_client->requestDeleteMarkedPlayers(markerName, markerId);
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    emit statusUpdate(tr("Lösche alle als „%1“ markierten %2 …").arg(
                         markerName, m_tournamentSettings->isPairMode() ? tr("Paare")
                                                                        : tr("Spieler")));
    if (! m_db->deleteMarkedPlayers(markerId)) {
        return;
    }

    emit statusUpdate(tr("Alle als „%1“ markierten %2 gelöscht").arg(
                         markerName, m_tournamentSettings->isPairMode() ? tr("Paare")
                                                                        : tr("Spieler")));

    listModified();
    checkUpdateAssignmentPage();

    if (m_isServer) {
        m_server->broadcastDeleteMarkedPlayers(markerName, markerId);
    }

    QApplication::restoreOverrideCursor();
    checkDisplayBackup(title);
}

void PlayersPage::deleteUnmarked()
{
    cacheChecksum();
    const QString title = tr("Unmarkierte %1 löschen").arg(
                             m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler"));

    if (! confirmListDeleteAction(title,
            tr("Sollen wirklich alle unmarkierten %1 gelöscht werden?").arg(
               m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler")))
        || ! checkCachedChecksum(title)) {

        return;
    }

    if (m_isClient) {
        m_client->requestDeleteUnmarkedPlayers();
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    emit statusUpdate(tr("Lösche alle unmarkierten %1 …").arg(
                         m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler")));

    if (! m_db->deleteUnmarkedPlayers()) {
        return;
    }

    emit statusUpdate(tr("Alle unmarkierten %1 gelöscht").arg(
                         m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler")));

    listModified();
    checkUpdateAssignmentPage();

    if (m_isServer) {
        m_server->broadcastDeleteUnmarkedPlayers();
    }

    QApplication::restoreOverrideCursor();
    checkDisplayBackup(title);
}

void PlayersPage::deleteAll()
{
    const QString title = tr("Alle %1 löschen").arg(
                             m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler"));

    if (! confirmListDeleteAction(title,
              tr("Sollen wirklich alle %1 gelöscht werden?").arg(
                 m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler")))) {

        return;
    }

    if (m_isClient) {
        m_client->requestDeleteAllPlayers();
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    emit statusUpdate(tr("Lösche alle %1 …").arg(
                         m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler")));

    if (! m_db->deleteAllPlayers()) {
        return;
    }

    emit statusUpdate(tr("Alle %1 gelöscht").arg(
                         m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler")));

    listModified();
    checkUpdateAssignmentPage();

    if (m_isServer) {
        m_server->broadcastDeleteAllPlayers();
    }

    QApplication::restoreOverrideCursor();
    checkDisplayBackup(title);
}

void PlayersPage::markersVisibilityChanged()
{
    m_searchWidget->saveSearch();
    updateVisibility();
    m_searchWidget->restoreSearch();
}

void PlayersPage::updateVisibility()
{
    int hidden = 0;
    for (int row = 0; row < m_playersList->rowCount(); row++) {
        bool visible = isRowVisible(row);
        m_playersList->setRowHidden(row, ! visible);
        if (! visible) {
            hidden++;
        }
    }
    if (hidden > 0) {
        m_playersBox->setTitle(m_tournamentSettings->isPairMode()
                                   ? tr("Angemeldete Paare (%n ausgeblendet)", "", hidden)
                                   : tr("Angemeldete Spieler (%n ausgeblendet)", "", hidden));
    } else {
        m_playersBox->setTitle(m_tournamentSettings->isPairMode() ? tr("Angemeldete Paare")
                                                                  : tr("Angemeldete Spieler"));
    }
    m_playersList->resizeColumnsToContents();
}

bool PlayersPage::isRowVisible(int row) const
{
    return m_markersWidget->isMarkerVisible(m_playersList->item(row,
               m_columnIndex[Column::Name])->data(PlayersData::MarkerId).toInt());
}

void PlayersPage::setColumnVisible(const Column column, const int state)
{
    m_playersList->setColumnHidden(m_columnIndex[column], state == Qt::Unchecked);

    bool headerNeeded = false;
    for (int i = 0; i < m_playersList->columnCount(); i++) {
        if (! m_playersList->isColumnHidden(i)
            && m_triggerHeaderColumns.contains(m_columnIndex.key(i))) {

            headerNeeded = true;
            break;
        }
    }

    m_playersList->horizontalHeader()->setVisible(headerNeeded);
    m_playersList->resizeColumnsToContents();

    if (column == Column::Assignment && m_db->isOpen()) {
        // We need to call this to update the status
        updatePlayersList();
    }

    // Inform the database
    QHash<Column, bool> visible;
    for (Column column : ALL_COLUMNS) {
        visible[column] = m_showColumn[column]->isChecked();
    }
    m_tournamentSettings->setPlayersPageColumns(visible);
}

QString PlayersPage::checksum() const
{
    QCryptographicHash hash(QCryptographicHash::Md5);

    // Add the players data

    for (int row = 0; row < m_playersList->rowCount(); row++) {
        QTableWidgetItem *item = m_playersList->item(row, m_columnIndex[Column::Name]);

        // Name
        hash.addData(item->text().toUtf8());

        // Marker
        hash.addData(QByteArray::number(item->data(PlayersData::MarkerId).toInt()));

        // Assignment
        hash.addData(QByteArray::number(item->data(PlayersData::AssignedTable).toInt()));
        hash.addData(QByteArray::number(item->data(PlayersData::AssignedPairNumber).toInt()));
    }

    // Add the marker data
    m_markersWidget->addChecksumData(hash);

    // Add the tournament state
    hash.addData(QByteArray::number(m_db->networkTournamentStarted()));

    return QLatin1String(hash.result().toHex());
}

void PlayersPage::setServer(Server *server)
{
    m_server = server;
    m_isServer = true;
    m_markersWidget->setServer(server);
    connect(server, &Server::nameEdited, this, &PlayersPage::nameEdited);
}

void PlayersPage::serverStopped()
{
    m_isServer = false;
    m_markersWidget->serverStopped();
    markersChanged();
}

void PlayersPage::hideContextMenus()
{
    for (QMenu *menu : findChildren<QMenu *>()) {
        menu->hide();
    }
}

void PlayersPage::processNetworkChange(int clientId)
{
    const bool contextMenuVisible = m_contextMenu->isVisible();
    const bool listMenuVisible = m_listMenu->isVisible();
    const bool editorOpen = m_delegate->isEditing();
    const bool markersPopupVisible = m_markersPopup->isVisible();
    const bool assignmentPopupVisible = m_assignmentPopup->isVisible();

    int processedRow = -1;
    int verticalOffset = 0;
    int originalY = 0;

    QTableWidgetItem *currentItem = m_playersList->currentItem();
    if (currentItem != nullptr) {
        processedRow = currentItem->row();
        originalY = m_playersList->mapToGlobal(
                        m_playersList->visualItemRect(currentItem).topLeft()).y();
    }

    // Check if we're doing anything. If not, we can simply display the change (possibly our own)

    if (! contextMenuVisible && ! listMenuVisible && ! editorOpen && ! markersPopupVisible
        && ! assignmentPopupVisible) {

        m_searchWidget->saveSearch();
        updatePlayersList();

        if (m_isClient && clientId == m_client->id() && ! m_scrollToTarget.isEmpty()) {
            if (m_playersList->findItems(m_scrollToTarget, Qt::MatchExactly).at(0)->row()
                != processedRow) {

                // The last selected row doesn't match the row of the scrollTarget, so we want to
                // be sure that the item is shown and scroll to it
                scrollTo(m_scrollToTarget);
            }

            // Otherwise, we can assume that the item is inside the viewport anyway.
        }

        m_searchWidget->restoreSearch();

        if (m_focusPlayersName) {
            m_focusPlayersName = false;
            m_playersName->setFocus();
        }

        return;
    }

    // Some action is currently executed, so we want to restore it

    // Cache what we have

    QString processedText;
    if (contextMenuVisible || markersPopupVisible || assignmentPopupVisible) {
        processedText = m_selectedName;
    } else if (editorOpen) {
        processedText = m_delegate->editedText();
    }

    QPoint listMenuPos;
    if (listMenuVisible) {
        // Process this first so that the list menu will be restored if it's visible,
        // even if it was shown by using the context menu's list menu action
        listMenuPos = m_listMenu->pos();
    }

    QPoint contextMenuPos;
    if (contextMenuVisible) {
        contextMenuPos = m_playersList->mapFromGlobal(m_contextMenu->pos());
    }

    QPoint popupPos;
    if (markersPopupVisible) {
        popupPos = m_markersPopup->pos();
    }

    // The default values are only set to make the Windows compiler happy ;-)
    // assignedTable and assignedPairNumber are only used if assignmentPopupVisible is true, and if
    // so, they are both initialized in the following if statement.
    int assignedTable = -1;
    int assignedPairNumber = -1;
    if (assignmentPopupVisible) {
        popupPos = m_assignmentPopup->pos();
        assignedTable = currentItem->data(PlayersData::AssignedTable).toInt();
        assignedPairNumber = currentItem->data(PlayersData::AssignedPairNumber).toInt();
    }

    // Abort all actions
    if (contextMenuVisible || listMenuVisible) {
        hideContextMenus();
    }
    if (editorOpen) {
        m_delegate->saveEditorState();
    }
    if (markersPopupVisible) {
        m_markersPopup->reject();
    }
    if (assignmentPopupVisible) {
        m_assignmentPopup->reject();
    }

    // Display the change
    listModified();

    // Restore the "whole list" menu (we don't need a specific item to still exist for this one)
    if (listMenuVisible) {
        if (m_playersList->rowCount() > 0) {
            QTimer::singleShot(0, this, [this, listMenuPos] { m_listMenu->exec(listMenuPos); } );
        } else {
            QTimer::singleShot(0, this, [this] {
                QMessageBox::warning(this,
                    tr("Kollidierende Netzwerkänderung"),
                    tr("Es wurden von einem anderen Netzwerkteilnehmer alle Anmeldungen gelöscht. "
                       "Ein Bearbeiten der ganzen Liste ist nicht mehr möglich.")); } );
        }
        return;
    }

    // From here on, we need the processed item, so check if we still have it

    QTableWidgetItem *processedItem;
    const QList<QTableWidgetItem *> matches = m_playersList->findItems(processedText,
                                                                       Qt::MatchExactly);
    if (matches.isEmpty()) {
        QTimer::singleShot(0, this, [this, processedText] {
            QMessageBox::warning(this,
                tr("Kollidierende Netzwerkänderung"),
                tr("<p>%1 „%2“ wurde durch eine Änderung eines anderen Netzwerkteilnehmers "
                   "entweder umbenannt oder gelöscht.</p>"
                   "<p>Das Bearbeiten wird abgebrochen.</p>").arg(
                   m_tournamentSettings->isPairMode() ? tr("Der gerade bearbeitete Spieler")
                                                      : tr("Das gerade bearbeitete Paar"),
                   processedText)); } );
        return;
    } else {
        processedItem = matches.at(0);
        m_playersList->setCurrentItem(processedItem);
    }

    // Try to scroll so that it's at the same place on screen
    if (processedItem->row() != processedRow) {
        m_playersList->verticalScrollBar()->setValue(
            m_playersList->verticalScrollBar()->value() - processedRow + processedItem->row());
        verticalOffset = m_playersList->mapToGlobal(
                             m_playersList->visualItemRect(processedItem).topLeft()).y()
                         - originalY;
    }

    // Restore the context menu

    if (contextMenuVisible) {
        contextMenuPos.setY(contextMenuPos.y() + verticalOffset);

        QTableWidgetItem *iteamAtContextMenuPos = m_playersList->itemAt(contextMenuPos);
        if (iteamAtContextMenuPos == nullptr
            || m_playersList->item(iteamAtContextMenuPos->row(), m_columnIndex[Column::Name])
               != processedItem) {

            // This can happen if the context menu has been displayed from bottom to top and now,
            // it can be displayed from top to bottom or vice versa. In this case, display the
            // context menu at the vertical middle of the processed item
            contextMenuPos.setY(m_playersList->visualItemRect(processedItem).topLeft().y()
                                + m_playersList->rowHeight(processedItem->row()) / 2);
        }

        QTimer::singleShot(0, this, [this, contextMenuPos] { showContextMenu(contextMenuPos); } );
        return;
    }

    // Restore the editor
    if (editorOpen) {
        QTimer::singleShot(0, this, [this, processedItem] {
            m_playersList->editItem(processedItem);
            m_delegate->restoreEditorState();
        } );
        return;
    }

    // Restore the markers popup
    if (markersPopupVisible) {
        popupPos.setY(popupPos.y() + verticalOffset);
        QTimer::singleShot(0, this, [this, processedItem, popupPos] {
            highlightItem(processedItem);
            m_markersPopup->show(processedItem->data(PlayersData::MarkerId).toInt(), popupPos);
        } );
        return;
    }

    // Restore the assignment popup
    if (assignmentPopupVisible) {
        const int newAssignedTable = processedItem->data(PlayersData::AssignedTable).toInt();
        const int newAssignedPairNumber = processedItem->data(
                                              PlayersData::AssignedPairNumber).toInt();

        int table;
        int pair;
        if (assignedTable == newAssignedTable && assignedPairNumber == newAssignedPairNumber) {
            table = m_assignmentPopup->table();
            pair = m_assignmentPopup->pair();
        } else {
            table = newAssignedTable;
            pair = newAssignedPairNumber;
        }

        popupPos.setY(popupPos.y() + verticalOffset);
        QTimer::singleShot(0, this, [this, processedItem, newAssignedTable, table, pair, popupPos] {
            highlightItem(processedItem);
            const int marker = processedItem->data(PlayersData::MarkerId).toInt();
            m_assignmentPopup->show(newAssignedTable != 0, marker != 0,
                                    marker == m_markersWidget->singleMarker(), table, pair, true,
                                    popupPos);
        } );
        return;
    }
}

void PlayersPage::setClient(Client *client)
{
    m_isClient = true;
    m_client = client;
    m_markersWidget->setClient(client);
    connect(client, &Client::nameEdited, this, &PlayersPage::nameEdited);
}

void PlayersPage::clientDisconnected()
{
    m_isClient = false;
    m_markersWidget->clientDisconnected();
    markersChanged();
}

void PlayersPage::setProcessingRequest(bool processing, const QString &scrollToTarget)
{
    m_newPlayersBox->setEnabled(! processing);
    m_playersBox->setEnabled(! processing);
    if (processing) {
        m_scrollToTarget = scrollToTarget;
        QApplication::setOverrideCursor(Qt::WaitCursor);
    }
}

MarkersWidget *PlayersPage::markersWidget() const
{
    return m_markersWidget;
}

void PlayersPage::finishRegistration()
{
    if (QMessageBox::warning(this,
        tr("Anmeldung beenden"),
        tr("<p>Wenn die Anmeldung beendet ist, kann (mindestens während das Netzwerk läuft) kein "
           "Netzwerkteilnehmer mehr neue %1 anmelden oder löschen!</p>"
           "<p>Soll die Anmeldung wirklich beendet werden?</p>").arg(
           m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler")),
        QMessageBox::Yes | QMessageBox::Cancel, QMessageBox::Cancel) == QMessageBox::Cancel) {

        return;
    }

    if (m_isClient) {
        m_client->requestFinishRegistration();
        return;
    }

    // Otherwise, we're the server
    networkFinishRegistration();
    m_server->broadcastFinishRegistration();
}

void PlayersPage::networkFinishRegistration(int clientId)
{
    QString networkChange;
    if ((m_isClient && clientId != m_client->id()) || (m_isServer && clientId != 0)) {
        networkChange = tr(" von einem anderen Netzwerkteilnehmer");
        if (m_contextMenu->isVisible()) {
            hideContextMenus();
        }
        if (m_delegate->isEditing()) {
            m_delegate->abortEdit();
        }
    }

    m_db->setNetworkTournamentStarted(true);
    reload();
    emit networkTournamentStarted(true);

    QTimer::singleShot(0, this, [this, networkChange] {
        QMessageBox::information(this,
            tr("Anmeldung beendet"),
            tr("<p>Die Anmeldung wurde%1 beendet. Es können ab jetzt keine neuen %2 mehr "
               "angemeldet oder bestehende Anmeldungen gelöscht werden.</p>"
               "<p>Sollte das trotzdem noch nötig sein, dann bitte den Server stoppen, neu "
               "starten und alle Clients neu verbinden. Sofern noch keine Ergebnisse eingegeben "
               "wurden, ist danach das komplette Bearbeiten der Anmeldungsliste wieder möglich."
               "</p>").arg(networkChange, m_tournamentSettings->isPairMode() ? tr("Paare")
                                                                             : tr("Spieler")));
    } );
}

void PlayersPage::checkShowPopup(int row, int column)
{
    if ((column != m_columnIndex[Column::Mark] && column != m_columnIndex[Column::Assignment])
        || ! m_db->isWritable()) {

        return;
    }

    QTableWidgetItem *item = m_playersList->item(row, m_columnIndex[Column::Name]);
    m_selectedName = item->text();
    highlightItem(item);

    if (column == m_columnIndex[Column::Mark]) {
        m_markersPopup->show(item->data(PlayersData::MarkerId).toInt());
    } else if (column == m_columnIndex[Column::Assignment]) {
        const int table = item->data(PlayersData::AssignedTable).toInt();
        const int marker = item->data(PlayersData::MarkerId).toInt();
        m_assignmentPopup->show(table != 0, marker != 0, marker == m_markersWidget->singleMarker(),
                                table, item->data(PlayersData::AssignedPairNumber).toInt());
    }
}

void PlayersPage::highlightItem(QTableWidgetItem *item)
{
    m_lastForeground = item->foreground();
    m_lastBackground = item->background();
    item->setForeground(m_playersList->palette().highlightedText());
    item->setBackground(m_playersList->palette().highlight());
    m_highlightedItem = item;
}

void PlayersPage::resetHighlight()
{
    if (m_highlightedItem == nullptr) {
        return;
    }

    m_highlightedItem->setForeground(m_lastForeground);
    m_highlightedItem->setBackground(m_lastBackground);
    m_highlightedItem = nullptr;
}

void PlayersPage::updateAssignment(int table, int pair, bool alsoRemoveMarker)
{
    if (table >= PREASSIGNMENT_TABLE_WARNING) {
        if (QMessageBox::question(this,
                tr("Auslosung speichern"),
                tr("Soll wirklich eine Auslosung für Tisch Nr. %1 eingegeben werden (vielleicht "
                   "war es ein Tippfehler)?").arg(QString::number(table)),
                QMessageBox::Yes | QMessageBox::Cancel, QMessageBox::Cancel)
            == QMessageBox::Cancel) {

            return;
        }
    }

    if (m_isClient) {
        m_client->requestUpdatePreAssignment(m_selectedName, table, pair, alsoRemoveMarker);
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    emit statusUpdate(table == 0 ? tr("Lösche Auslosung …") : tr("Speichere Auslosung …"));

    if (! m_db->updatePreAssignment(m_selectedName, table, pair, alsoRemoveMarker)) {
        return;
    }

    emit assignmentChanged();
    if (alsoRemoveMarker) {
        displayMarkedEntry();
    } else {
        listModified();
    }
    checkUpdateAssignmentPage();

    if (m_isServer) {
        m_server->broadcastUpdatePreAssignment(m_selectedName, table, pair, alsoRemoveMarker);
    }

    if (table == 0) {
        emit statusUpdate(tr("Auslosung für %1 „%2“ gelöscht").arg(
                             m_tournamentSettings->isPairMode() ? tr("Paar") : tr("Spieler"),
                             m_selectedName));
    } else {
        emit statusUpdate(tr("Auslosung für %1 „%2“ gespeichert (Tisch %3 Paar %4)").arg(
                             m_tournamentSettings->isPairMode() ? tr("Paar") : tr("Spieler"),
                             m_selectedName, QString::number(table), QString::number(pair)));
    }
    QApplication::restoreOverrideCursor();
}

void PlayersPage::assignPlayer(const QString &name)
{
    const QString pairName = m_selectedName + m_settings->namesSeparator() + name;
    if (checkForExactMatch(pairName)) {
        QMessageBox::warning(this, tr("Allein gekommene Spieler zuordnen"),
            tr("<p><b>Zuordnung nicht möglich</b></p>"
               "<p>Der Spieler „%1“ kann dem Spieler „%2“ nicht zugeordnet werden, da der Paarname "
               "„%3“ bereits vergeben ist.</p>"
               "<p>Bitte mindestestens einen Namen ändern, damit ein eindeutiger Paarname erstellt "
               "werden kann!</p>").arg(
               name, m_selectedName, pairName));
        return;
    }

    if (m_isClient) {
        m_client->requestAssemblePair(m_selectedName, name, m_settings->namesSeparator(),
                                      m_markersWidget->assignedMarker());
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    emit statusUpdate(tr("Erstelle Paar aus Einzelspielern …"));

    if (! m_db->assemblePair(m_selectedName, name, m_settings->namesSeparator(),
                             m_markersWidget->assignedMarker())) {
        return;
    }

    listModified();
    checkUpdateAssignmentPage();

    if (m_playersList->findItems(pairName, Qt::MatchExactly).at(0)->row() != m_editedRow) {
        scrollTo(pairName);
    }

    if (m_isServer) {
        m_server->broadcastAssemblePair(m_selectedName, name, m_settings->namesSeparator(),
                                        m_markersWidget->assignedMarker());
    }

    emit statusUpdate(tr("Paar „%1“ erstellt").arg(pairName));
    QApplication::restoreOverrideCursor();
}

bool PlayersPage::absentsPresent() const
{
    return m_absentsPresent;
}

bool PlayersPage::singlesPresent() const
{
    return m_singlesPresent;
}

void PlayersPage::deleteAllAssignments()
{
    const QString title = tr("Alle Auslosungen löschen");
    if (! confirmListDeleteAction(title, tr("Sollen wirklich alle bisher eingegebenen Auslosungen "
                                            "gelöscht werden?"))) {
        return;
    }

    if (m_isClient) {
        m_client->requestDeleteAllPreAssignments();
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    emit statusUpdate(tr("Lösche alle Auslosungen …"));

    if (! m_db->deleteAllPreAssignments()) {
        return;
    }

    emit assignmentChanged();
    emit statusUpdate(tr("Alle Auslosungen gelöscht"));
    listModified();
    checkUpdateAssignmentPage();

    if (m_isServer) {
        m_server->broadcastDeleteAllPreAssignments();
    }

    QApplication::restoreOverrideCursor();
    checkDisplayBackup(title);
}

void PlayersPage::showInfoPopup(const QString &anchor)
{
    QString text;

    if (anchor == QLatin1String("#incompleteTables")) {
        text = tr("<p><b>Tische mit unvollständiger Auslosung</b></p>");
        const int incompleteCount = m_incompleteTables.count();
        if (incompleteCount == 1) {
            text.append(tr("<p>Zu Tisch <b>%1</b> wurde bisher nur ein Paar 1 oder ein Paar 2 "
                           "zugelost%2.</p>").arg(
                           QString::number(m_incompleteTables.at(0)),
                           m_tournamentSettings->isPairMode()
                               ? QString() : tr(" (bzw. die Paare sind unvollständig)")));
        } else if (incompleteCount == 2) {
            text.append(tr("<p>Zu den Tischen <b>%1</b> und <b>%2</b> wurden bisher nur ein Paar 1 "
                           "oder ein Paar 2 zugelost%3.</p>").arg(
                           QString::number(m_incompleteTables.at(0)),
                           QString::number(m_incompleteTables.at(1)),
                           m_tournamentSettings->isPairMode()
                               ? QString() : tr(" (bzw. die Paare sind unvollständig)")));
        } else {
            text.append(tr("<p>Zu folgenden Tischen wurden bisher nur ein Paar 1 oder ein Paar 2 "
                           "zugelost%2:</p><p>").arg(
                           m_tournamentSettings->isPairMode()
                               ? QString() : tr(" (bzw. die Paare sind unvollständig)")));
            for (int i = 0; i < incompleteCount - 1; i++) {
                text.append(tr("<b>%1</b>, ").arg(QString::number(m_incompleteTables.at(i))));
            }
            text = text.left(text.size() - 2);
            text.append(tr(" und <b>%1</b>.</p>").arg(m_incompleteTables.at(incompleteCount - 1)));
        }

    } else if (anchor == QLatin1String("#tableGaps")) {
        text = tr("<p><b>Tische, die in der Auslosung fehlen</b></p>");
        const int gapsCount = m_tableGaps.count();
        if (gapsCount == 1) {
            text.append(tr("<p>Tisch <b>%1</b> ist bisher nicht ausgelost, müsste es aber sein (da "
                           "es sonst Lücken gibt)</p>").arg(
                           QString::number(m_tableGaps.at(0))));
        } else if (gapsCount == 2) {
            text.append(tr("<p>Die Tische <b>%1</b> und <b>%2</b> sind bisher nicht zugewiesen, "
                           "müssten es aber sein (da es sonst Lücken gibt).</p>").arg(
                           QString::number(m_tableGaps.at(0)),
                           QString::number(m_tableGaps.at(1))));
        } else {
            text.append(tr("<p>Folgende Tische sind bisher nicht ausgelost, müssten es aber "
                           "sein (da es sonst Lücken gibt):</p><p>"));
            for (int i = 0; i < gapsCount - 1; i++) {
                text.append(tr("<b>%1</b>, ").arg(QString::number(m_tableGaps.at(i))));
            }
            text = text.left(text.size() - 2);
            text.append(tr(" und <b>%1</b>.</p>").arg(m_tableGaps.at(gapsCount - 1)));
        }
    }

    QMessageBox::information(this, tr("Auslosung"), text);
}

void PlayersPage::setAssignmentPage(AssignmentPage *assignmentPage)
{
    m_assignmentPage = assignmentPage;
}

void PlayersPage::checkUpdateAssignmentPage()
{
    if (m_assignmentPage == nullptr) {
        return;
    }

    int lastTable = 0;
    QHash<int, QHash<int, QString>> preAssignment;
    QHash<int, QHash<int, bool>> pairComplete;

    for (int row = 0; row < m_playersList->rowCount(); row++) {
        QTableWidgetItem *item = m_playersList->item(row, m_columnIndex[Column::Name]);
        const int table = item->data(PlayersData::AssignedTable).toInt();
        if (table == 0) {
            continue;
        }
        if (table > lastTable) {
            lastTable = table;
        }

        const int pair = item->data(PlayersData::AssignedPairNumber).toInt();
        const QString name = item->text();
        switch (m_tournamentSettings->tournamentMode()) {
        case Muckturnier::TournamentMode::FixedPairs:
            preAssignment[table][pair] = name;
            break;
        case Muckturnier::TournamentMode::SinglePlayers:
            if (preAssignment.value(table).contains(pair)) {
                preAssignment[table][pair] = preAssignment[table][pair]
                                             + m_settings->namesSeparator()
                                             + name;
                pairComplete[table][pair] = true;
            } else {
                preAssignment[table][pair] = name;
                pairComplete[table][pair] = false;
            }
            break;
        }
    }

    m_assignmentPage->updateAssignment(preAssignment, pairComplete, m_settings->namesSeparator(),
                                       lastTable);
}

void PlayersPage::setColumnsVisible(const QHash<Column, bool> &visible)
{
    // We have to call this via a QTimer call if the database is opened via command line argument
    // to be sure that everything is constructed and set up before we do this
    QTimer::singleShot(0, this, [this, visible] {
        for (Column column : ALL_COLUMNS) {
            m_showColumn[column]->setChecked(visible.value(column));
        }
    });
}
