/* Copyright (c) 2011-2020 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef PLAYERSPAGE_H
#define PLAYERSPAGE_H

// Local includes
#include "shared/PlayersPageColumns.h"

// Qt includes
#include <QWidget>
#include <QIcon>

// Local classes

class SharedObjects;
class StringTools;
class Database;
class TournamentSettings;
class SearchEngine;
class PlayersPageDelegate;
class TitleMenu;
class SearchWidget;
class MarkersWidget;
class MarkersPopup;
class Settings;
class Server;
class Client;
class AssignmentPopup;
class AssignmentPage;

// Qt classes
class QGroupBox;
class QLabel;
class QLineEdit;
class QPushButton;
class QTableWidget;
class QAction;
class QMenu;
class QTableWidgetItem;
class QCheckBox;

using namespace PlayersPageColumns;

class PlayersPage : public QWidget
{
    Q_OBJECT

public:
    explicit PlayersPage(QWidget *parent, SharedObjects *sharedObjects);
    void reload(bool updateMarkersWidget = true);
    void reset();
    bool absentsPresent() const;
    bool singlesPresent() const;
    void setAssignmentPage(AssignmentPage *assignmentPage);
    void checkUpdateAssignmentPage();
    void setColumnsVisible(const QHash<Column, bool> &visible);

    void setServer(Server *server);
    void serverStopped();
    void setClient(Client *client);
    void clientDisconnected();
    QString checksum() const;
    MarkersWidget *markersWidget() const;
    void processNetworkChange(int clientId = 0);
    void setProcessingRequest(bool processing, const QString &scrollToTarget = QString());
    void networkFinishRegistration(int clientId = 0);

public slots:
    void updatePlayersList();
    void scoreChanged();
    void finishRegistration();
    void resetSearch();

signals:
    void statusUpdate(const QString &message);
    void playersCountChanged(int count);
    void registrationsAvailable(bool state);
    void nameEdited();
    void networkTournamentStarted(bool state);
    void assignmentChanged();

private: // Enums
    enum PlayersData {
        MarkerId = Qt::UserRole,
        AssignedTable,
        AssignedPairNumber
    };

private slots:
    void registerPlayers();
    void editEntered(const QString &originalName, const QString &name);
    void showContextMenu(QPoint point);
    void scrollTo(const QString &text);
    void deletePlayers();
    void updateSearch(const QString &searchText);
    void markersChanged();
    void updateMarker(const QString &markerName, int markerId);
    void listUpdateMarker(const QString &markerName, int markerId);
    void listRemoveMarker(const QString &markerName, int markerId);
    void deleteMarked(const QString &markerName, int markerId);
    void deleteUnmarked();
    void deleteAll();
    void updateVisibility();
    void markersVisibilityChanged();
    void setColumnVisible(Column column, int state);
    void checkShowPopup(int row, int column);
    void highlightItem(QTableWidgetItem *item);
    void resetHighlight();
    void assignPlayer(const QString &name);
    void updateAssignment(int table, int pair, bool alsoRemoveMarker);
    void deleteAllAssignments();
    void showInfoPopup(const QString &anchor);

private: // Functions
    void listModified();
    bool isRowVisible(int row) const;
    void updateStatusLabel();
    bool confirmListDeleteAction(const QString &title, const QString &text);
    void checkDisplayBackup(const QString &title);
    void cacheChecksum();
    bool checkCachedChecksum(const QString &title);
    void hideContextMenus();
    void displayMarkedEntry();
    bool checkForExactMatch(const QString &name);

private: // Variables
    Database *m_db;
    Settings *m_settings;
    TournamentSettings *m_tournamentSettings;
    StringTools *m_stringTools;
    SearchEngine *m_searchEngine;
    const QHash<Column, int> m_columnIndex;
    const QVector<Column> m_triggerHeaderColumns;

    QGroupBox *m_newPlayersBox;
    QLabel *m_nameLabel;
    QLineEdit *m_playersName;
    QPushButton *m_registerButton;
    QTimer *m_resetHighlightTimer;
    QGroupBox *m_playersBox;
    QTableWidget *m_playersList;
    QIcon m_markerIcon;
    QLabel *m_statusLabel;
    QLabel *m_isSearchLabel;
    QHash<Column, QCheckBox *> m_showColumn;
    PlayersPageDelegate *m_delegate;
    SearchWidget *m_searchWidget;
    MarkersWidget *m_markersWidget;
    MarkersPopup *m_markersPopup;
    AssignmentPopup *m_assignmentPopup;
    AssignmentPage *m_assignmentPage = nullptr;

    QTableWidgetItem *m_highlightedItem;
    int m_editedRow;
    QString m_selectedName;
    QVector<int> m_incompleteTables;
    QVector<int> m_tableGaps;
    QBrush m_lastForeground;
    QBrush m_lastBackground;

    // Context menus

    TitleMenu *m_contextMenu;

    QMenu *m_assignMenu;

    QAction *m_editAction;
    QAction *m_deleteAction;

    QAction *m_setOnlyMarkerAction;
    QMenu *m_setMarkerMenu;
    QAction *m_removeMarkerAction;

    QAction *m_setAssignmentAction;
    QAction *m_deleteAssignmentAction;

    TitleMenu *m_listMenu;

    QAction *m_listSetOnlyMarkerAction;
    QMenu *m_listSetMarkerMenu;
    QAction *m_listRemoveOnlyMarkerAction;
    QMenu *m_listRemoveMarkerMenu;
    QAction *m_listRemoveAllMarkersAction;

    QAction *m_deleteOnlyMarkedAction;
    QMenu *m_deleteMarkedMenu;
    QAction *m_deleteUnmarkedAction;
    QAction *m_deleteAllAction;

    QAction *m_deleteAllAssignmentsAction;

    // Network

    Server *m_server;
    Client *m_client;
    bool m_isServer = false;
    bool m_isClient = false;
    QString m_scrollToTarget;
    QString m_cachedChecksum;
    QString m_backupFile;
    bool m_absentsPresent = false;
    bool m_singlesPresent = false;
    bool m_focusPlayersName = false;

};

#endif // PLAYERSPAGE_H
