/* Copyright (c) 2018-2019 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes

#include "RankingDialog.h"

#include "ClientPage/Client.h"

#include "shared/ServerProtocol.h"
#include "shared/SharedStyles.h"
#include "shared/TableDelegate.h"
#include "shared/SharedObjects.h"
#include "shared/StringTools.h"
#include "shared/TournamentSettings.h"

// Qt includes
#include <QDebug>
#include <QApplication>
#include <QMessageBox>
#include <QGroupBox>
#include <QTableWidget>
#include <QHeaderView>
#include <QVBoxLayout>
#include <QLabel>
#include <QPushButton>
#include <QTimer>
#include <QScrollBar>
#include <QJsonObject>
#include <QJsonArray>

using namespace ServerProtocol;

RankingDialog::RankingDialog(QWidget *parent, SharedObjects *sharedObjects, Client *client)
    : TitleDialog(parent),
      m_db(sharedObjects->database()),
      m_tournamentSettings(sharedObjects->tournamentSettings()),
      m_client(client),
      m_stringTools(sharedObjects->stringTools())
{
    setTitle(tr("Ranglistenvergleich"));

    QGroupBox *resultBox = new QGroupBox;
    resultBox->setProperty("notitle", true);
    QVBoxLayout *resultBoxLayout = new QVBoxLayout(resultBox);
    layout()->addWidget(resultBox);

    m_resultLabel = new QLabel;
    m_resultLabel->setWordWrap(true);
    resultBoxLayout->addWidget(m_resultLabel);

    m_differences = new QTableWidget;
    m_differences->setEditTriggers(QAbstractItemView::NoEditTriggers);
    m_differences->setItemDelegate(new TableDelegate(this));
    m_differences->verticalHeader()->hide();
    m_differences->setShowGrid(false);
    m_differences->setWordWrap(false);
    m_differences->setSelectionMode(QTableWidget::NoSelection);
    m_differences->setFocusPolicy(Qt::NoFocus);
    m_differences->setContextMenuPolicy(Qt::CustomContextMenu);
    resultBoxLayout->addWidget(m_differences);
    m_differences->hide();

    m_dataChangedLabel = new QLabel(tr("<span style=\"%1\">Der lokale Datensatz wurde seit dem "
                                       "letzten Abruf verändert. Bitte die Rangliste erneut "
                                       "abgleichen!</span>").arg(SharedStyles::STYLE_RED_TEXT));
    m_dataChangedLabel->setWordWrap(true);
    resultBoxLayout->addWidget(m_dataChangedLabel);
    m_dataChangedLabel->hide();

    // Button box

    addButtonBox();

    m_requestRanking = buttonBox()->addButton(QDialogButtonBox::Yes);
    m_requestRanking->setText(tr("Rangliste abfragen"));
    m_requestRanking->setFocus();
    connect(m_requestRanking, &QPushButton::clicked, this, &RankingDialog::requestRanking);

    m_close = buttonBox()->addButton(QDialogButtonBox::Close);
    connect(m_close, &QPushButton::clicked, this, &QDialog::reject);

    if (checkTournamentStarted()) {
        requestRanking();
    }
}

bool RankingDialog::checkTournamentStarted()
{
    if (m_db->localTournamentStarted()) {
        m_requestRanking->setEnabled(true);
        return true;
    } else {
        m_requestRanking->setEnabled(false);
        m_differences->hide();
        m_resultLabel->setText(tr("Es wurden noch keine Ergebnisse eingegeben."));
        m_dataChangedLabel->hide();
        m_close->setFocus();
        resizeToContents();
        return false;
    }
}

void RankingDialog::clientDisconnected()
{
    QApplication::restoreOverrideCursor();
    QMessageBox::critical(this, tr("Ranglistenvergleich"),
                                tr("Die Netzwerkverbindung wurde beendet. Der Dialog wird "
                                   "geschlossen."));
    QDialog::reject();
}

void RankingDialog::requestRanking()
{
    QApplication::setOverrideCursor(Qt::WaitCursor);
    m_requestRanking->setEnabled(false);
    m_close->setEnabled(false);
    m_client->requestRanking(this);
}

void RankingDialog::dataReceived()
{
    m_requestRanking->setEnabled(true);
    m_close->setEnabled(true);
    m_dataChangedLabel->hide();
}

void RankingDialog::noRanking()
{
    dataReceived();
    m_differences->hide();
    m_resultLabel->setText(tr("Auf dem Server wurden noch keine Ergebnisse eingegeben."));
    m_close->setFocus();
    resizeToContents();
}

void RankingDialog::compareRanking(const QVector<Database::RankData> &remoteRanking)
{
    dataReceived();

    bool success;
    const QVector<Database::RankData> localRanking = m_db->ranking(-1, &success);
    if (! success) {
        return;
    }

    QHash<QString, int> localIndex;
    for (int i = 0; i < localRanking.count(); i++) {
        localIndex[localRanking.at(i).name] = i;
    }

    // Check if the same number of results has been stored

    QHash<QString, QPair<int, int>> roundDifferences;
    for (const Database::RankData &rank : remoteRanking) {
        const int localPlayedBoogers = localRanking.at(localIndex.value(rank.name)).playedBoogers;
        if (rank.playedBoogers != localPlayedBoogers) {
            roundDifferences[rank.name] = QPair<int, int>(rank.playedBoogers, localPlayedBoogers);
        }
    }

    if (! roundDifferences.isEmpty()) {
        m_resultLabel->setText(tr(
            "<p style=\"%1\">Es wurden hier und auf dem Server nicht für alle %2 gleich viele "
            "Ergebnisse eingegeben!</p>"
            "<p>Bei folgenden %3 fehlen noch Spielstände (links stehen die eingegebenen Bobbl auf "
            "dem Server, rechts die hier eingegebenen:</p>").arg(
            SharedStyles::STYLE_RED_TEXT,
            m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler"),
            m_tournamentSettings->isPairMode() ? tr("Paaren") : tr("Spielern")));

        m_differences->show();
        m_differences->clear();
        m_differences->setColumnCount(2);
        m_differences->setHorizontalHeaderLabels( {
            m_tournamentSettings->isPairMode() ? tr("Paar") : tr("Spieler"),
            tr("Eingetragene Bobbl") });
        m_differences->setRowCount(roundDifferences.count());

        QStringList sortedNames(roundDifferences.keys());
        m_stringTools->sortList(sortedNames);
        int row = 0;
        for (const QString &name : sortedNames) {
            m_differences->setItem(row, 0, new QTableWidgetItem(name));
            m_differences->setItem(row, 1, new QTableWidgetItem(differenceString(
                roundDifferences.value(name).first, roundDifferences.value(name).second)));
            row++;
        }

        finishDisplayDifferences();
        return;
    }

    // Check if the ranking is identical

    QHash<QString, QStringList> rankingDifferences;
    for (const Database::RankData &rank : remoteRanking) {
        const int localBoogers = localRanking.at(localIndex.value(rank.name)).boogers;
        const int localGoals = localRanking.at(localIndex.value(rank.name)).goals;
        const int localOpponentGoals = localRanking.at(localIndex.value(rank.name)).opponentGoals;

        if (rank.boogers != localBoogers
            || rank.goals != localGoals
            || rank.opponentGoals != localOpponentGoals) {

            rankingDifferences[rank.name] = QStringList {
                differenceString(rank.boogers, localBoogers),
                differenceString(rank.goals, localGoals),
                differenceString(rank.opponentGoals, localOpponentGoals) };
        }
    }

    if (! rankingDifferences.isEmpty()) {
        m_resultLabel->setText(tr(
            "<p style=\"%1\">Die lokale Rangliste ist nicht identisch mit der des Servers!</p>"
            "<p>Folgende %2 haben auf dem Server andere Werte (links stehen die Werte des Servers, "
            "rechts die hier eingegebenen):</p>").arg(
            SharedStyles::STYLE_RED_TEXT,
            m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler")));

        m_differences->show();
        m_differences->clear();
        m_differences->setColumnCount(4);
        m_differences->setHorizontalHeaderLabels( {
            m_tournamentSettings->isPairMode() ? tr("Paar") : tr("Spieler"),
            tr("Bobbl"), tr("Tore"), tr("Gegn. Tore") });
        m_differences->setRowCount(rankingDifferences.count());

        QStringList sortedNames(rankingDifferences.keys());
        m_stringTools->sortList(sortedNames);
        int row = 0;
        for (const QString &name : sortedNames) {
            m_differences->setItem(row, 0,
                                   new QTableWidgetItem(name));
            m_differences->setItem(row, 1,
                                   new QTableWidgetItem(rankingDifferences.value(name).at(0)));
            m_differences->setItem(row, 2,
                                   new QTableWidgetItem(rankingDifferences.value(name).at(1)));
            m_differences->setItem(row, 3,
                                   new QTableWidgetItem(rankingDifferences.value(name).at(2)));
            row++;
        }

        finishDisplayDifferences();
        return;
    }

    // If we reach here, the rankings are identical

    m_resultLabel->setText(tr("<span style=\"%1\">Die lokale Rangliste ist identisch mit der des "
                              "Servers.</span>").arg(SharedStyles::STYLE_GREEN_TEXT));
    m_differences->hide();
    resizeToContents();
    m_close->setFocus();
}

void RankingDialog::finishDisplayDifferences()
{
    m_differences->resizeColumnsToContents();
    resizeToContents();
    resizeToTableWidth();
    m_requestRanking->setFocus();
}

const QString RankingDialog::differenceString(int local, int remote) const
{
    if (local == remote) {
        return QString::number(local);
    } else {
        return QStringLiteral("%1 %2 %3").arg(QString::number(local),
                                              QStringLiteral("\u2260"),
                                              QString::number(remote));
    }
}

void RankingDialog::resizeToTableWidth()
{
    QApplication::processEvents();
    QTimer::singleShot(0, this, [this] {
        resize(width() + m_differences->horizontalHeader()->length() + 4
                       + (m_differences->verticalScrollBar()->isVisible()
                              ? m_differences->verticalScrollBar()->width() : 0)
                       - m_differences->width(),
               height());
    });
}

void RankingDialog::dataChanged()
{
    checkTournamentStarted();
    m_dataChangedLabel->setVisible(m_requestRanking->isEnabled());
    resizeToContents();
}
