/* Copyright (c) 2011-2020 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes

#include "ScorePage.h"
#include "BoogerWidget.h"
#include "NumbersComboBox.h"
#include "PlayersComboBox.h"
#include "BoogerSpinBox.h"
#include "ScorePopup.h"
#include "TableComboBox.h"

#include "shared/SharedStyles.h"
#include "shared/SharedObjects.h"
#include "shared/StringTools.h"
#include "shared/SearchWidget.h"
#include "shared/TableDelegate.h"
#include "shared/SearchEngine.h"
#include "shared/TitleMenu.h"
#include "shared/Settings.h"
#include "shared/TournamentSettings.h"
#include "shared/Muckturnier.h"

// Qt includes
#include <QDebug>
#include <QVBoxLayout>
#include <QGroupBox>
#include <QHBoxLayout>
#include <QLabel>
#include <QGridLayout>
#include <QCheckBox>
#include <QScrollArea>
#include <QPushButton>
#include <QTableWidget>
#include <QHeaderView>
#include <QAction>
#include <QMessageBox>
#include <QApplication>
#include <QScrollBar>
#include <QLineEdit>

ScorePage::ScorePage(QWidget *parent, SharedObjects *sharedObjects)
    : QWidget(parent),
      m_db(sharedObjects->database()),
      m_settings(sharedObjects->settings()),
      m_tournamentSettings(sharedObjects->tournamentSettings()),
      m_stringTools(sharedObjects->stringTools()),
      m_searchEngine(sharedObjects->searchEngine())
{
    QVBoxLayout *layout = new QVBoxLayout(this);

    // Round

    QGroupBox *roundBox = new QGroupBox;
    roundBox->setProperty("notitle", true);
    QHBoxLayout *roundBoxLayout = new QHBoxLayout(roundBox);
    layout->addWidget(roundBox);

    QLabel *roundLabel = new QLabel(tr("Runde:"));
    roundLabel->setStyleSheet(SharedStyles::STYLE_BOLD_TEXT);
    roundBoxLayout->addWidget(roundLabel);

    m_roundSelect = new NumbersComboBox;
    roundBoxLayout->addWidget(m_roundSelect);
    connect(m_roundSelect, &NumbersComboBox::currentNumberChanged, this, &ScorePage::openRound);

    roundBoxLayout->addStretch();

    m_noTableNumbers = new QCheckBox(tr("Eingabe ohne Tischnummern"));
    m_noTableNumbers->setWhatsThis(tr("Wenn ohne Tischnummern gespielt wird, dann wird die "
                                      "Tischauswahl und -anzeige ausgeblendet, wenn diese Option "
                                      "gesetzt ist."));
    roundBoxLayout->addWidget(m_noTableNumbers);
    connect(m_noTableNumbers, &QCheckBox::stateChanged, this, &ScorePage::noTableNumbersChanged);

    m_resultBox = new QGroupBox(tr("Ergebnisse eintragen"));
    QVBoxLayout *resultBoxLayout = new QVBoxLayout(m_resultBox);
    layout->addWidget(m_resultBox);

    // Table and pairs/players

    QGroupBox *playersBox = new QGroupBox;
    playersBox->setProperty("notitle", true);
    QHBoxLayout *playersBoxLayout = new QHBoxLayout(playersBox);
    resultBoxLayout->addWidget(playersBox);

    m_tableLabel = new QLabel(tr("Tisch:"));
    m_tableLabel->setStyleSheet(SharedStyles::STYLE_BOLD_TEXT);
    playersBoxLayout->addWidget(m_tableLabel);

    m_tableSelect = new TableComboBox;
    playersBoxLayout->addWidget(m_tableSelect);
    connect(m_tableSelect, &TableComboBox::currentNumberChanged, this, &ScorePage::tableChanged);
    connect(m_tableSelect, &TableComboBox::checkSavePossible, this, &ScorePage::checkSavePossible);

    QGridLayout *playersLayout = new QGridLayout;
    playersBoxLayout->addLayout(playersLayout);

    for (int pair = 1; pair <= 2; pair++) {
        QLabel *playersLabel = new QLabel(tr("Paar %1:").arg(QString::number(pair)));
        playersLabel->setStyleSheet(SharedStyles::STYLE_BOLD_TEXT);
        playersLayout->addWidget(playersLabel, pair - 1, 0);

        QHash<int, PlayersComboBox *> entry;
        for (int pairPart = 1; pairPart <= 2; pairPart++) {
            PlayersComboBox *pairPartComboBox = new PlayersComboBox(this, m_searchEngine, pair);
            connect(pairPartComboBox, &PlayersComboBox::selectionChanged,
                    this, &ScorePage::playersChanged);
            connect(pairPartComboBox, &PlayersComboBox::disableSaveScore,
                    this, [this] { m_saveScoreButton->setEnabled(false); });
            connect(pairPartComboBox, &PlayersComboBox::checkSavePossible,
                    this, &ScorePage::checkSavePossible);
            playersLayout->addWidget(pairPartComboBox, pair - 1, pairPart);
            entry[pairPart] = pairPartComboBox;
        }
        m_playersSelect[pair] = entry;
    }

    m_autoSelectPairs = new QCheckBox(tr("Tischnummer bzw. Paare\nautomatisch auswählen"));
    m_autoSelectPairs->setWhatsThis(tr(
        "<p>Wird nach dem System „Paar 1 bleibt sitzen, Paar 2 rückt jede Runde einen Tisch "
        "weiter“ gespielt, dann werden mittels dieser Option die Paare für einen Tisch oder der "
        "Tisch und das gegnerische Paar für ein anderes Paar automatisch ausgewählt.</p>"
        "<p>Hierzu muss natürlich die Auslosung bekannt sein, so dass die automatische Auswahl "
        "erst ab Runde 2 funktionieren kann. Am einfachsten ist die Auswahl über die Tischnummer!"
        "</p>"));
    playersBoxLayout->addWidget(m_autoSelectPairs);
    connect(m_autoSelectPairs, &QCheckBox::stateChanged, this, &ScorePage::autoSelectPairsChanged);

    playersBoxLayout->addStretch();

    // Enter Score
    QGroupBox *scoreBox = new QGroupBox;
    scoreBox->setProperty("notitle", true);
    QVBoxLayout *scoreBoxLayout = new QVBoxLayout(scoreBox);
    scoreBox->setSizePolicy(QSizePolicy::Ignored, QSizePolicy::Maximum);
    resultBoxLayout->addWidget(scoreBox);

    m_scoreBoxScroll = new QScrollArea(scoreBox);
    m_scoreBoxScroll->setFrameShape(QFrame::NoFrame);
    scoreBoxLayout->addWidget(m_scoreBoxScroll);

    m_saveScoreButton = new QPushButton(tr("Ergebnis\neintragen"));
    connect(m_saveScoreButton, &QPushButton::clicked, this, &ScorePage::saveScore);

    // The round's results

    m_roundSummaryBox = new QGroupBox;
    QVBoxLayout *roundSummaryBoxLayout = new QVBoxLayout(m_roundSummaryBox);
    layout->addWidget(m_roundSummaryBox);

    m_isSearchLabel = new QLabel;
    m_isSearchLabel->setAlignment(Qt::AlignCenter);
    m_isSearchLabel->hide();
    roundSummaryBoxLayout->addWidget(m_isSearchLabel);

    m_scoreList = new QTableWidget;
    m_scoreList->setEditTriggers(QAbstractItemView::NoEditTriggers);
    m_scoreList->setItemDelegate(new TableDelegate(this, TableDelegate::HighlightMode::EvenRows));
    m_scoreList->verticalHeader()->hide();
    m_scoreList->setShowGrid(false);
    m_scoreList->setWordWrap(false);
    m_scoreList->setSelectionMode(QTableWidget::NoSelection);
    m_scoreList->setFocusPolicy(Qt::NoFocus);
    m_scoreList->setContextMenuPolicy(Qt::CustomContextMenu);
    connect(m_scoreList, &QTableWidget::cellDoubleClicked,
            [this] { showContextMenu(m_scoreList->viewport()->mapFromGlobal(QCursor::pos())); } );
    connect(m_scoreList, &QTableWidget::customContextMenuRequested,
            this, &ScorePage::showContextMenu);
    roundSummaryBoxLayout->addWidget(m_scoreList);

    // Search widget

    m_searchWidget = new SearchWidget;
    connect(m_searchWidget, &SearchWidget::searchChanged, this, &ScorePage::updateSearch);
    roundSummaryBoxLayout->addWidget(m_searchWidget);

    // Context menu

    m_contextMenu = new TitleMenu(this);

    m_editAction = new QAction(tr("Ändern (Löschen und neu eingeben)"), this);
    m_contextMenu->addAction(m_editAction);
    connect(m_editAction, &QAction::triggered, this, &ScorePage::editScore);

    m_swapPairsAction = new QAction(tr("Paar 1 und Paar 2 vertauschen"), this);
    m_contextMenu->addAction(m_swapPairsAction);
    connect(m_swapPairsAction, &QAction::triggered, this, &ScorePage::swapPairs);

    m_contextMenu->addSeparator();

    m_deleteAction = new QAction(tr("Löschen"), this);
    m_contextMenu->addAction(m_deleteAction);
    connect(m_deleteAction, &QAction::triggered, this, &ScorePage::deleteScore);

    // Score dialog
    m_scorePopup = new ScorePopup(this);
    connect(m_scorePopup, &QDialog::accepted, this, &ScorePage::setSelectedScore);
}

void ScorePage::setupGui()
{
    // Show the needed input fields for pairs or players
    switch (m_tournamentSettings->tournamentMode()) {
    case Muckturnier::TournamentMode::FixedPairs:
        m_playersSelect[1][2]->hide();
        m_playersSelect[2][2]->hide();
        m_autoSelectPairs->show();
        break;
    case Muckturnier::TournamentMode::SinglePlayers:
        m_playersSelect[1][2]->show();
        m_playersSelect[2][2]->show();
        m_autoSelectPairs->hide();
        break;
    }

    // Display the correct number of booger widgets

    m_boogerWidgets.clear();
    QWidget *scoreBoxWidget = new QWidget;
    QHBoxLayout *scoreBoxWidgetLayout = new QHBoxLayout(scoreBoxWidget);
    scoreBoxWidgetLayout->setContentsMargins(0, 0, 0, 0);
    scoreBoxWidgetLayout->setSizeConstraint(QLayout::SetFixedSize);

    for (int i = 1; i <= m_tournamentSettings->boogersPerRound(); i++) {
        BoogerWidget *boogerWidget = new BoogerWidget(scoreBoxWidget, i,
                                                      m_tournamentSettings->boogersPerRound(),
                                                      m_tournamentSettings->boogerScore());
        connect(boogerWidget, &BoogerWidget::abortedStateChanged,
                this, &ScorePage::boogerAbortedStateChanged);
        connect(boogerWidget, &BoogerWidget::scorePopupRequested,
                this, &ScorePage::showScorePopup);

        m_boogerWidgets.append(boogerWidget);
        scoreBoxWidgetLayout->addWidget(boogerWidget);
    }
    connect(m_boogerWidgets.last(), &BoogerWidget::returnPressed,
            this, &ScorePage::checkEnterScore);

    m_scoreSize = m_boogerWidgets.count() * m_db->necessaryDivisor();

    // Add the "Save score" button
    scoreBoxWidgetLayout->addWidget(m_saveScoreButton);
    m_saveScoreButton->setEnabled(false);
    m_scoreBoxScroll->setWidget(scoreBoxWidget);
    scoreBoxWidget->setAutoFillBackground(false);
    setScrollAreaHeight();

    // Disable data input if we have a read-only database
    m_resultBox->setEnabled(m_db->isWritable());
    m_noTableNumbers->setEnabled(m_db->isWritable());

    // Set the correct score widget's column count
    m_scoreList->setColumnCount(m_tournamentSettings->boogersPerRound() + 2);
    QStringList headerLabels { tr("Tisch"), tr("Paare") };
    for (int i = 1; i <= m_tournamentSettings->boogersPerRound(); i++) {
        headerLabels.append(tr("%1. Bobbl").arg(QString::number(i)));
    }
    m_scoreList->setHorizontalHeaderLabels(headerLabels);

    // Update the score dialog
    m_scorePopup->setBoogerScore(m_tournamentSettings->boogerScore());
}

void ScorePage::playersNameEdited()
{
    const QVector<Database::BoogerResult> score = currentScore();
    const int lastOpenedRound = m_roundSelect->currentNumber();
    const int selectedTable = m_tableSelect->currentNumber();

    // Check if we currently search for a name
    int editedPair = -1;
    int editedPlayer = -1;
    QString editedText;
    for (int pair = 1; pair <= 2; pair++) {
        for (int player = 1; player <= (m_tournamentSettings->isSinglePlayerMode() ? 2 : 1);
             player++) {

            if (m_playersSelect[pair][player]->itemText(
                    m_playersSelect[pair][player]->currentIndex())
                != m_playersSelect[pair][player]->currentText()) {

                editedPair = pair;
                editedPlayer = player;
                editedText = m_playersSelect[pair][player]->currentText();
            }
        }
    }

    reload(m_roundSelect->currentNumber());

    fillInScore(lastOpenedRound, selectedTable, score);
    if (editedPair != -1) {
        m_playersSelect[editedPair][editedPlayer]->lineEdit()->setText(editedText);
        m_playersSelect[editedPair][editedPlayer]->lineEditTextChanged(editedText);
    }
}

void ScorePage::reload(int lastRound)
{
    if (! isVisible()) {
        return;
    }

    // Reset all booger widgets if the GUI has been already setup
    if (lastRound != -1) {
        for (BoogerWidget *boogerWidget : m_boogerWidgets) {
            boogerWidget->reset();
        }
        m_boogerWidgets.at(0)->setFocus();
    }

    // Fill the round select

    blockSelectSignals(true);

    int allRounds = 0;
    int firstOpenRound = -1;

    m_roundSelect->clear();
    for (const int roundNumber : m_db->roundsState().keys()) {
        if (m_db->roundsState().value(roundNumber)) {
            m_roundSelect->addItem(tr("%1 (abgeschlossen)").arg(QString::number(roundNumber)),
                                   roundNumber);
        } else {
            m_roundSelect->addItem(tr("%1 (offen)").arg(QString::number(roundNumber)),
                                   roundNumber);
            if (firstOpenRound == -1) {
                firstOpenRound = roundNumber;
            }
        }
        allRounds++;
    }

    if (firstOpenRound == -1) {
        firstOpenRound = 1;
    }
    if (lastRound != -1) {
        firstOpenRound = lastRound;
    }

    // Add one for a new round
    if (m_db->isWritable()) {
        allRounds++;
        m_roundSelect->addItem(tr("%1 (neue Runde)").arg(QString::number(allRounds)), allRounds);
        // The "false" value for the new round's missing roundsState entry will be provided
        // by setting the default value to "false" in openRound().
    }

    // Display the data
    m_roundSelect->setCurrentNumber(firstOpenRound);
    blockSelectSignals(false);
    openRound(firstOpenRound);
}

void ScorePage::setScrollAreaHeight()
{
    if (m_scoreBoxScroll->widget() == nullptr) {
        // This happens when no database has been created/loaded yet
        return;
    }

    if (m_scoreBoxScroll->width() < m_scoreBoxScroll->widget()->width()) {
        m_scoreBoxScroll->setMinimumHeight(m_scoreBoxScroll->widget()->height()
                                           + m_scoreBoxScroll->horizontalScrollBar()->height());
    } else {
        m_scoreBoxScroll->setMinimumHeight(m_scoreBoxScroll->widget()->height());
    }
}

void ScorePage::openRound(int round)
{
    if (round == -1) {
        return;
    }

    blockSelectSignals(true);

    m_searchWidget->saveSearch();

    // We set the "false" default value here because if a new round is selected,
    // the Database's roundsState map won't contain a value for it yet

    if (! m_db->roundsState().value(round, false)) {
        m_resultBox->setVisible(true);
        bool success;

        // Insert all missing tables

        m_tableSelect->clear();

        const QVector<int> missingTables = m_db->missingTables(round, &success);
        if (! success) {
            return;
        }

        for (const int table : missingTables) {
            m_tableSelect->addItem(QString::number(table), table);
        }
        m_tableSelect->setCurrentIndex(m_noTableNumbers->isChecked());

        // Insert all missing players/pairs

        for (int i = 1; i <= 2; i++) {
            m_playersSelect[i][1]->clear();
            if (m_tournamentSettings->isSinglePlayerMode()) {
                m_playersSelect[i][2]->clear();
            }
        }

        const QHash<QString, int> missingPlayers = m_db->missingPlayers(round, &success);
        if (! success) {
            return;
        }

        QStringList availableNames = missingPlayers.keys();
        m_stringTools->sortList(availableNames);
        for (const QString &name : availableNames) {
            for (int i = 1; i <= 2; i++) {
                m_playersSelect[i][1]->addItem(name, missingPlayers.value(name));
                if (m_tournamentSettings->isSinglePlayerMode()) {
                    m_playersSelect[i][2]->addItem(name, missingPlayers.value(name));
                }
            }
        }

    } else {
        m_resultBox->setVisible(false);
    }

    updateScore();
    updateAssignment(round);

    blockSelectSignals(false);
    setScrollAreaHeight();
    checkSavePossible();
    m_searchWidget->restoreSearch();
}

void ScorePage::updateAssignment(int round)
{
    if (round == -1) {
        // If round is -1, this is triggered via setting a pre-assignment on the players page.
        // In this case, we only have to update the assignment if round 1 is currently open
        // (one can also select round 2 with no results being entered yet, in this case, we can
        // skip the update as the preassignment is only for round 1)
        round = m_roundSelect->currentData().toInt();
        if (round != 1) {
            return;
        }
    }

    bool success;
    m_assignment = m_db->assignment(round, &success);
    Q_UNUSED(success);
}

void ScorePage::updateScore()
{
    bool success;
    const QVector<Database::Score> score(m_db->scores(m_roundSelect->currentNumber(), &success));
    if (! success) {
        return;
    }

    m_scoreList->setRowCount(score.count() / m_tournamentSettings->boogersPerRound());

    if (score.count() == 0) {
        m_searchWidget->clear();
        m_isSearchLabel->hide();
        m_searchWidget->hide();
    } else {
        m_searchWidget->show();
    }

    if (m_tournamentSettings->isSinglePlayerMode()) {
        m_searchEngine->clearSearchCache(SearchEngine::SearchType::DrawnPairsScore);
    }

    int row = -1;
    for (int i = 0; i < score.count(); i++) {
        row++;
        if (m_tournamentSettings->isSinglePlayerMode()) {
            m_searchEngine->updateSearchCache(score.at(i).pairName,
                                                    SearchEngine::SearchType::DrawnPairsScore);
        }
        m_scoreList->setItem(row, 0, new QTableWidgetItem(QString::number(score.at(i).table)));
        m_scoreList->setSpan(row, 0, 2, 1);
        m_scoreList->setItem(row, 1, new QTableWidgetItem(score.at(i).pairName));
        m_scoreList->setItem(row, 2, new QTableWidgetItem(
            m_stringTools->boogify(score.at(i).score, m_tournamentSettings->boogerScore())));
        for (int booger = 2; booger <= m_tournamentSettings->boogersPerRound(); booger++) {
            m_scoreList->setItem(row, 1 + booger, new QTableWidgetItem(
                m_stringTools->boogify(score.at(++i).score, m_tournamentSettings->boogerScore())));
        }

        row++;
        i++;
        if (m_tournamentSettings->isSinglePlayerMode()) {
            m_searchEngine->updateSearchCache(score.at(i).pairName,
                                                    SearchEngine::SearchType::DrawnPairsScore);
        }
        m_scoreList->setItem(row, 1, new QTableWidgetItem(score.at(i).pairName));
        m_scoreList->setItem(row, 2, new QTableWidgetItem(
            m_stringTools->boogify(score.at(i).score, m_tournamentSettings->boogerScore())));
        for (int booger = 2; booger <= m_tournamentSettings->boogersPerRound(); booger++) {
            m_scoreList->setItem(row, 1 + booger, new QTableWidgetItem(
                m_stringTools->boogify(score.at(++i).score, m_tournamentSettings->boogerScore())));
        }
    }
    m_scoreList->resizeColumnsToContents();

    const int openRound = m_roundSelect->currentNumber();
    if (m_db->roundsState().value(openRound)) {
        m_roundSummaryBox->setTitle(tr("Alle Ergebnisse der %1. Runde").arg(
                                       QString::number(openRound)));
    } else {
        int enteredScores = (row + 1) / 2;
        if (enteredScores == 0) {
            m_roundSummaryBox->setTitle(
                tr("Bisher keine Ergebnisse aus der %1. Runde").arg(
                    QString::number(openRound)));
        } else {
            m_roundSummaryBox->setTitle(
                tr("Bisher %n Ergebnis(se) aus der %1. Runde", "", enteredScores).arg(openRound)
                + QStringLiteral(" ")
                + tr("(%n fehlt/fehlen noch)", "", m_tableSelect->count() - 1));
        }
    }
}

void ScorePage::tableChanged(int table)
{
    if (! m_autoSelectPairs->isChecked()) {
        checkSavePossible();
        return;
    }

    if (m_assignment.assignedPair.contains(table)) {
        if (m_assignment.assignedPair[table].contains(1)) {
            m_playersSelect[1][1]->setCurrentNumber(m_assignment.assignedPair[table][1]);
        }
        if (m_assignment.assignedPair[table].contains(2)) {
            m_playersSelect[2][1]->setCurrentNumber(m_assignment.assignedPair[table][2]);
        }
    }

    checkSavePossible();
}

void ScorePage::playersChanged(int pair, int pairId)
{
    if (! m_autoSelectPairs->isChecked()) {
        return;
    }

    if (m_assignment.assignedPairNumber.contains(pairId)) {
        if (pair != m_assignment.assignedPairNumber.value(pairId)) {
            QMessageBox::warning(this,
                tr("Die Paarauswahl passt nicht zur Auslosung"),
                tr("<p>Als Paar %1 wurde gerade das Paar „%2“ ausgewählt.</p>"
                    "<p>Laut der Auslosung (bzw. dem Ergebnis der letzten Runde) "
                    "ist dieses Paar aber ein Paar %3!</p>").arg(
                    QString::number(pair),
                    m_playersSelect[pair][1]->currentText(),
                    QString::number(m_assignment.assignedPairNumber.value(pairId))));
        } else {
            if (m_assignment.assignedTable.contains(pairId)) {
                m_tableSelect->setCurrentNumber(m_assignment.assignedTable.value(pairId));
            }
        }
    }
}

void ScorePage::checkSavePossible()
{
    switch (m_tournamentSettings->tournamentMode()) {
    case Muckturnier::TournamentMode::FixedPairs:
        m_saveScoreButton->setEnabled(
            m_tableSelect->inputOkay()
            && m_tableSelect->currentNumber() != -1
            && m_playersSelect[1][1]->currentIndex() != m_playersSelect[2][1]->currentIndex());
        break;
    case Muckturnier::TournamentMode::SinglePlayers:
        m_saveScoreButton->setEnabled(
            m_tableSelect->inputOkay()
            && m_tableSelect->currentNumber() != -1
            && m_playersSelect[1][1]->currentIndex() != m_playersSelect[1][2]->currentIndex()
            && m_playersSelect[1][1]->currentIndex() != m_playersSelect[2][1]->currentIndex()
            && m_playersSelect[1][1]->currentIndex() != m_playersSelect[2][2]->currentIndex()
            && m_playersSelect[1][2]->currentIndex() != m_playersSelect[2][1]->currentIndex()
            && m_playersSelect[1][2]->currentIndex() != m_playersSelect[2][2]->currentIndex()
            && m_playersSelect[2][1]->currentIndex() != m_playersSelect[2][2]->currentIndex()
        );
        break;
    }
}

QVector<Database::BoogerResult> ScorePage::currentScore() const
{
    QVector<Database::BoogerResult> score;
    score.reserve(m_scoreSize);
    for (int i = 1; i <= 2; i++) {
        for (BoogerWidget *boogerWidget : m_boogerWidgets) {
            score.append(Database::BoogerResult { m_playersSelect[i][1]->currentNumber(),
                                                  boogerWidget->score(i) });
        }
        if (m_tournamentSettings->isSinglePlayerMode()) {
            for (BoogerWidget *boogerWidget : m_boogerWidgets) {
                score.append(Database::BoogerResult { m_playersSelect[i][2]->currentNumber(),
                                                      boogerWidget->score(i) });
            }
        }
    }
    return score;
}

void ScorePage::saveScore()
{
    emit statusUpdate(tr("Speichere Spielergebnis …"));
    QApplication::setOverrideCursor(Qt::WaitCursor);

    const bool showRankingPage = ! m_db->localTournamentStarted();

    const int lastOpenedRound = m_roundSelect->currentNumber();
    const int selectedTable = m_tableSelect->currentNumber();
    if (! m_db->saveScore(lastOpenedRound, selectedTable, currentScore())) {
        return;
    }

    QString tableIdentifier;
    if (! m_noTableNumbers->isChecked()) {
        tableIdentifier = tr("Tisch %1").arg(selectedTable);
    } else {
        switch (m_tournamentSettings->tournamentMode()) {
        case Muckturnier::TournamentMode::FixedPairs:
            tableIdentifier = formatOpponentsString(m_playersSelect[1][1]->currentText(),
                                                    m_playersSelect[2][1]->currentText());
            break;
        case Muckturnier::TournamentMode::SinglePlayers:
            tableIdentifier = formatOpponentsString(m_playersSelect[1][1]->currentText()
                                                    + m_settings->namesSeparator()
                                                    + m_playersSelect[1][2]->currentText(),
                                                    m_playersSelect[2][1]->currentText()
                                                    + m_settings->namesSeparator()
                                                    + m_playersSelect[2][2]->currentText());
            break;
        }
    }

    if (showRankingPage) {
        emit tournamentStarted();
    }

    emit scoreChanged();
    reload(lastOpenedRound);

    // Make sure the entered score is visible
    int targetRow = 0;
    for (int i = 0; i < m_scoreList->rowCount(); i += 2) {
        if (m_scoreList->item(i, 0)->text().toInt() == selectedTable) {
            targetRow = i;
            break;
        }
    }
    if (targetRow == m_scoreList->rowCount() - 2) {
        m_scoreList->scrollToBottom();
    } else {
        m_scoreList->scrollToItem(m_scoreList->item(targetRow, 1),
                                  QAbstractItemView::PositionAtCenter);
    }

    QApplication::restoreOverrideCursor();
    emit statusUpdate(tr("Spielergebnis für %1 in Runde %2 gespeichert").arg(
                         tableIdentifier, QString::number(lastOpenedRound)));
}

void ScorePage::showContextMenu(QPoint point)
{
    if (! m_db->isWritable()) {
        return;
    }

    QTableWidgetItem *selectedScore = m_scoreList->itemAt(point);
    if (selectedScore == nullptr) {
        return;
    }

    int row = selectedScore->row();
    if (row % 2 != 0) {
        row = row - 1;
    }

    QString table = m_scoreList->item(row, 0)->text();
    if (m_noTableNumbers->isChecked()) {
        m_contextMenu->setTitle(formatOpponentsString(m_scoreList->item(row, 1)->text(),
                                                      m_scoreList->item(row + 1, 1)->text()));
    } else {
        m_contextMenu->setTitle(tr("Ergebnis für Tisch %1").arg(table));
    }

    m_selectedTable = table.toInt();

    point.setY(point.y() + m_scoreList->horizontalHeader()->height());
    m_contextMenu->exec(m_scoreList->mapToGlobal(point));
}

void ScorePage::deleteScore()
{
    QMessageBox messageBox(this);
    messageBox.setIcon(QMessageBox::Question);
    messageBox.setWindowTitle(tr("Spielergebnis löschen"));
    messageBox.setText(tr("Soll das Spielergebnis von %1 wirklich gelöscht werden?").arg(
                          getStatusUpdateIdentifier()));
    messageBox.setStandardButtons(QMessageBox::Yes | QMessageBox::No);
    messageBox.setDefaultButton(QMessageBox::No);

    if (messageBox.exec() == QMessageBox::No) {
        return;
    }

    doDeleteScore(m_roundSelect->currentNumber(), m_selectedTable);
}

void ScorePage::doDeleteScore(int round, int table)
{
    QApplication::setOverrideCursor(Qt::WaitCursor);
    emit statusUpdate(tr("Lösche Spielergebnis …"));

    if (! m_db->deleteScore(round, table)) {
        return;
    }

    emit statusUpdate(tr("Spielergebnis für %1 in Runde %2 gelöscht").arg(
                         getStatusUpdateIdentifier(), QString::number(round)));
    reload(round);
    QApplication::restoreOverrideCursor();

    if (! m_db->localTournamentStarted()) {
        // Re-emitting this will cause the players page to be updated again,
        // as it checks for db.tournamentStarted() itself
        emit tournamentStarted();
    }
    emit scoreChanged();
}

void ScorePage::editScore()
{
    const int round = m_roundSelect->currentNumber();
    bool success;

    const QVector<Database::BoogerResult> score = m_db->singleScore(round, m_selectedTable,
                                                                    &success);
    if (! success) {
        return;
    }

    doDeleteScore(round, m_selectedTable);
    fillInScore(round, m_selectedTable, score);
}

void ScorePage::fillInScore(int round, int table, const QVector<Database::BoogerResult> &score)
{
    m_roundSelect->setCurrentNumber(round);

    blockSelectSignals(true);
    m_tableSelect->setCurrentNumber(table);
    switch (m_tournamentSettings->tournamentMode()) {
    case Muckturnier::TournamentMode::FixedPairs:
        m_playersSelect[1][1]->setCurrentNumber(score.at(
            0).pairOrPlayer);
        m_playersSelect[2][1]->setCurrentNumber(score.at(
            m_tournamentSettings->boogersPerRound()).pairOrPlayer);
        break;
    case Muckturnier::TournamentMode::SinglePlayers:
        m_playersSelect[1][1]->setCurrentNumber(score.at(
            0).pairOrPlayer);
        m_playersSelect[1][2]->setCurrentNumber(score.at(
            m_tournamentSettings->boogersPerRound()).pairOrPlayer);
        m_playersSelect[2][1]->setCurrentNumber(score.at(
            m_tournamentSettings->boogersPerRound() * 2).pairOrPlayer);
        m_playersSelect[2][2]->setCurrentNumber(score.at(
            m_tournamentSettings->boogersPerRound() * 3).pairOrPlayer);
        break;
    }
    blockSelectSignals(false);

    // Get the scores to set
    QVector<int> scores;
    QVectorIterator<Database::BoogerResult> boogerResult(score);
    while (boogerResult.hasNext()) {
        scores.append(boogerResult.next().score);
        for (int i = 1; i < m_tournamentSettings->boogersPerRound(); i++) {
            scores.append(boogerResult.next().score);
        }
        if (m_tournamentSettings->isSinglePlayerMode()) {
            // Leave out the double scores
            for (int i = 0; i < m_tournamentSettings->boogersPerRound(); i++) {
                boogerResult.next();
            }
        }
    }

    // Check if we have an aborted score
    int abortedBooger = -1;
    for (int i = 0; i < m_boogerWidgets.count(); i++) {
        if (scores.at(i) < m_tournamentSettings->boogerScore()
            && scores.at(i + scores.count() / 2) < m_tournamentSettings->boogerScore()) {

            abortedBooger = i;
            break;
        }
    }

    // First reset all boogers
    for (BoogerWidget *booger : m_boogerWidgets) {
        booger->reset();
    }

    // Then mark the first aborted booger if we have one
    if (abortedBooger > -1) {
        boogerAbortedStateChanged(abortedBooger, true);
        m_boogerWidgets.at(abortedBooger)->setAborted(true);
        m_boogerWidgets.at(abortedBooger)->setAsFirstAborted();
    }

    // Pre-set all scores
    for (int i = 0; i < m_boogerWidgets.count(); i++) {
        m_boogerWidgets.at(i)->setScore(scores.at(i), scores.at(i + scores.count() / 2));
    }

    checkSavePossible();
}

void ScorePage::blockSelectSignals(bool state)
{
    m_roundSelect->blockSignals(state);
    m_tableSelect->blockSignals(state);
    for (int i = 1; i <= 2; i++) {
        for (int j = 1; j <= 2; j++) {
            m_playersSelect[i][j]->blockSignals(state);
        }
    }
}

void ScorePage::checkEnterScore()
{
    if (! m_saveScoreButton->isEnabled()) {
        return;
    }
    saveScore();
}

void ScorePage::showEvent(QShowEvent *event)
{
    QWidget::showEvent(event);
    setScrollAreaHeight();
}

void ScorePage::updateSearch(const QString &searchText)
{
    if (searchText.isEmpty()) {
        for (int row = 0; row < m_scoreList->rowCount(); row++) {
            m_scoreList->showRow(row);
        }
        m_scoreList->resizeColumnsToContents();
        m_isSearchLabel->hide();
        return;
    }

    m_searchEngine->setSearchTerm(searchText);
    const SearchEngine::SearchType searchType
        = m_tournamentSettings->isPairMode() ? SearchEngine::SearchType::Default
                                             : SearchEngine::SearchType::DrawnPairsScore;
    int matches = 0;
    for (int row = 0; row < m_scoreList->rowCount(); row += 2) {
        const bool rowMatches
            =    m_searchEngine->checkMatch(m_scoreList->item(row,     1)->text(), searchType)
              || m_searchEngine->checkMatch(m_scoreList->item(row + 1, 1)->text(), searchType);
        m_scoreList->setRowHidden(row,     ! rowMatches);
        m_scoreList->setRowHidden(row + 1, ! rowMatches);
        matches += rowMatches;
    }

    m_searchWidget->updateMatches(matches);
    m_scoreList->resizeColumnsToContents();
    m_isSearchLabel->setText(tr("(Suchergebnis für „%1“)").arg(searchText));
    m_isSearchLabel->show();
}

void ScorePage::swapPairs()
{
    const int round = m_roundSelect->currentNumber();
    bool success;

    QVector<Database::BoogerResult> score = m_db->singleScore(round, m_selectedTable, &success);
    if (! success) {
        return;
    }

    if (! m_db->deleteScore(round, m_selectedTable)) {
        return;
    }

    QVector<Database::BoogerResult> newScore = score.mid(score.count() / 2, score.count() / 2);
    newScore.append(score.mid(0, score.count() / 2));
    if (! m_db->saveScore(round, m_selectedTable, newScore)) {
        return;
    }

    emit statusUpdate(tr("Paar 1 und Paar 2 für %1 in Runde %2 vertauscht").arg(
                         getStatusUpdateIdentifier(), QString::number(round)));

    // This will reload the page and restore the currently entered score
    playersNameEdited();
}

void ScorePage::noTableNumbersChanged(bool state)
{
    m_scoreList->setColumnHidden(0, state);
    m_tableLabel->setVisible(! state);
    m_tableSelect->setVisible(! state);
    if (state) {
        m_lastAutoSelectState = m_autoSelectPairs->isChecked();
        m_autoSelectPairs->setChecked(false);
        m_autoSelectPairs->hide();
    } else {
        m_autoSelectPairs->show();
        m_scoreList->resizeColumnsToContents();
        m_autoSelectPairs->setChecked(m_lastAutoSelectState);
    }

    m_tournamentSettings->setNoTableNumbers(state);
    m_tableSelect->setCurrentIndex(state);
    checkSavePossible();
}

QString ScorePage::getStatusUpdateIdentifier() const
{
    return m_noTableNumbers->isChecked() ? m_contextMenu->title()
                                         : tr("Tisch %1").arg(QString::number(m_selectedTable));
}

void ScorePage::autoSelectPairsChanged(bool state)
{
    m_tournamentSettings->setAutoSelectPairs(state);
    if (state) {
        tableChanged(m_tableSelect->currentNumber());
    }
}

QString ScorePage::formatOpponentsString(const QString &pair1, const QString &pair2) const
{
    if (pair1.size() <= 15 && pair2.size() <= 15) {
        return tr("„%1“ gegen „%2“").arg(pair1, pair2);
    } else if (pair1.size() > 15 && pair2.size() <= 15) {
        return tr("„%1…“ gegen „%2“").arg(pair1.left(15), pair2);
    } else if (pair1.size() <= 15 && pair2.size() > 15) {
        return tr("„%1“ gegen „%2…“").arg(pair1, pair2.left(15));
    } else {
        return tr("„%1…“ gegen „%2…“").arg(pair1.left(15), pair2.left(15));
    }
}

void ScorePage::setNoTableNumbers(bool state)
{
    m_noTableNumbers->setChecked(state);
}

void ScorePage::setAutoSelectPairs(bool state)
{
    m_autoSelectPairs->setChecked(state);
}

void ScorePage::boogerAbortedStateChanged(int booger, bool aborted)
{
    if (aborted) {
        for (int i = booger + 1; i < m_boogerWidgets.count(); i++) {
            m_boogerWidgets.at(i)->setAborted(true);
        }
    } else {
        for (int i = 0; i <= booger; i++) {
            if (m_boogerWidgets.at(i)->isAborted()) {
                m_boogerWidgets.at(i)->reset();
            }
        }
        if (booger + 1 < m_boogerWidgets.count()) {
            m_boogerWidgets.at(booger + 1)->setAsFirstAborted();
        }
    }
}

void ScorePage::resetSearch()
{
    m_searchWidget->clear();
}

void ScorePage::showScorePopup(BoogerSpinBox *spinBox)
{
    m_targetSpinBox = spinBox;
    m_scorePopup->show();
}

void ScorePage::setSelectedScore()
{
    m_targetSpinBox->setValue(m_scorePopup->score());
    m_targetSpinBox->setFocus();
    m_targetSpinBox->selectAll();
}
