/* Copyright (c) 2018-2019 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes

#include "Server.h"
#include "Connection.h"

#include "PlayersPage/PlayersPage.h"
#include "PlayersPage/MarkersWidget.h"

#include "shared/SharedObjects.h"
#include "shared/Database.h"
#include "shared/TournamentSettings.h"
#include "shared/ServerProtocol.h"
#include "shared/StringTools.h"
#include "shared/JsonHelper.h"

// Qt includes
#include <QDebug>
#include <QTimer>
#include <QJsonArray>
#include <QApplication>

// C++ includes
#include <functional>

using namespace ServerProtocol;

Server::Server(QObject *parent, int connectTimeout, SharedObjects *sharedObjects,
               PlayersPage *playersPage)
    : QTcpServer(parent),
      m_connectTimeout(connectTimeout),
      m_db(sharedObjects->database()),
      m_tournamentSettings(sharedObjects->tournamentSettings()),
      m_playersPage(playersPage),
      m_markersWidget(playersPage->markersWidget()),
      m_locale(sharedObjects->stringTools()->locale()),
      m_requestHandlers {
          // Registration changes

          { ACTION_PLAYERS_REGISTER,           &Server::processRegisterPlayers },
          { ACTION_PLAYERS_RENAME,             &Server::processRenamePlayers },
          { ACTION_PLAYERS_ASSEMBLE_PAIR,      &Server::processAssemblePair },
          { ACTION_PLAYERS_DELETE,             &Server::processDeletePlayers },
          { ACTION_PLAYERS_UPDATE_MARKER,      &Server::processUpdatePlayersMarker },

          { ACTION_PLAYERS_DELETE_MARKED,      &Server::processDeleteMarkedPlayers },
          { ACTION_PLAYERS_DELETE_UNMARKED,    &Server::processDeleteUnmarkedPlayers },
          { ACTION_PLAYERS_DELETE_ALL,         &Server::processDeleteAllPlayers },
          { ACTION_PLAYERS_LIST_UPDATE_MARKER, &Server::processUpdateListMarker },
          { ACTION_PLAYERS_LIST_REMOVE_MARKER, &Server::processRemoveListMarker },

          // Preassignment changes
          { ACTION_UPDATE_PREASSIGNMENT,      &Server::processUpdatePreAssignment },
          { ACTION_DELETE_ALL_PREASSIGNMENTS, &Server::processDeleteAllPreAssignments },

          // Markers defintion changes
          { ACTION_MARKERS_ADD,    &Server::processAddMarker },
          { ACTION_MARKERS_DELETE, &Server::processDeleteMarker },
          { ACTION_MARKERS_MOVE,   &Server::processMoveMaker },
          { ACTION_MARKERS_EDIT,   &Server::processEditMarker },

          // Special markers changes
          { ACTION_SET_ABSENT_MARKER,         &Server::processSetAbsentMarker },
          { ACTION_CHANGE_SINGLES_MANAGEMENT, &Server::processChangeSinglesManagement },

          // Finish registration
          { ACTION_FINISH_REGISTRATION, &Server::processFinishRegistration }

          // Ranking comparison
          // (Can't be requested by the server)
      }
{
}

void Server::close()
{
    for (Connection *connection : m_connections) {
        connection->close();
        removeConnection(connection);
    }
    QTcpServer::close();
}

void Server::incomingConnection(qintptr descriptor)
{
    // Create a new connection for the connection request
    Connection *connection = new Connection(this);
    if (! connection->setSocketDescriptor(descriptor)) {
        connection->deleteLater();
        return;
    }
    connect(connection, &Connection::disconnected,
            this, std::bind(&Server::removeConnection, this, connection));
    connect(connection, &Connection::handshakeResult,
            this, std::bind(&Server::processHandshake, this, connection, std::placeholders::_1));
    connect(connection, &Connection::clientRequest,
            this, std::bind(&Server::processClientRequest, this, connection,
                            std::placeholders::_1, std::placeholders::_2));
    connect(connection, &AbstractNetworkEngine::protocolError, this, &Server::protocolError);

    // Initiate a handshake

    m_pendingConnections.append(connection);
    emit message(tr("Verbindungsanfrage von %1, sende Handshake").arg(connection->ip()));

    // Send greeting message
    connection->sendData(HANDSHAKE_MESSAGE_INITIALIZE);

    // Start a concurrent timeout call
    QTimer::singleShot(m_connectTimeout, this,
                       std::bind(&Server::checkHandshakeTimeout, this, connection));
}

void Server::removeConnection(Connection *connection)
{
    if (m_pendingConnections.removeOne(connection)) {
        emit message(tr("Verbindung mit %1 beendet").arg(connection->ip()));
    } else if (m_connections.removeOne(connection)) {
        emit message(tr("Verbindung mit %1 beendet").arg(clientString(connection)));
        emit statusUpdate(tr("Verbindung mit %1 beendet").arg(clientString(connection)));
        emit clientCountChanged(m_connections.count());
    }
    connection->deleteLater();
}

void Server::checkHandshakeTimeout(Connection *connection)
{
    if (m_pendingConnections.contains(connection)) {
        emit message(tr("Keine Antwort von %1").arg(connection->ip()));
        removeConnection(connection);
    }
}

void Server::broadcastMessage(QJsonObject &jsonObject, int clientId, const QString &messageText)
{
    emit message(messageText);
    jsonObject.insert(KEY_CHECKSUM, m_playersPage->checksum());
    jsonObject.insert(KEY_CLIENT_ID, clientId);
    for (Connection *connection : m_connections) {
        connection->sendData(jsonObject);
    }
}

int Server::clientCount() const
{
    return m_connections.count();
}

QString Server::clientString(Connection *connection) const
{
    return QStringLiteral("%1 (Client-ID %2)").arg(connection->ip(),
                                                   QString::number(connection->id()));
}

// =========
// Handshake
// =========

void Server::processHandshake(Connection *connection, const QString &phase)
{
    if (! m_pendingConnections.contains(connection)) {
        return;
    }

    if (phase == HANDSHAKE_PHASE_SYNCHRONIZE) {
        emit message(tr("Synchronisation von %1 angefragt, sende Daten").arg(connection->ip()));

        bool success;

        const QJsonObject settings {
            { HANDSHAKE_KEY_TOURNAMENT_MODE,    static_cast<int>(
                                                    m_tournamentSettings->tournamentMode()) },
            { HANDSHAKE_KEY_BOOGER_SCORE,       m_tournamentSettings->boogerScore() },
            { HANDSHAKE_KEY_BOOGERS_PER_ROUND,  m_tournamentSettings->boogersPerRound() },
            { HANDSHAKE_KEY_TOURNAMENT_STARTED, m_db->networkTournamentStarted() } };

        const QVector<Muckturnier::Marker> definedMarkers(m_db->markers(&success));
        if (! success) {
            return;
        }

        QJsonArray markers;
        for (const Muckturnier::Marker &marker : definedMarkers) {
            markers.append(QJsonObject {
                { HANDSHAKE_KEY_MARKER_ID,      marker.id },
                { HANDSHAKE_KEY_MARKER_NAME,    marker.name },
                { HANDSHAKE_KEY_MARKER_COLOR,   Connection::colorToString(marker.color) },
                { HANDSHAKE_KEY_MARKER_SORTING, static_cast<int>(marker.sorting) }
            });
        }

        const QHash<QString, Database::PlayersData> playersData = m_db->playersData(&success);
        if (! success) {
            return;
        }

        QJsonArray players;
        for (const QString &name : playersData.keys()) {
            const Database::PlayersData data = playersData.value(name);
            QJsonObject jsonData;
            if (data.marker != 0) {
                jsonData.insert(PLAYERS_KEY_MARKER_ID, data.marker);
            }
            if (data.assignedTable != 0) {
                jsonData.insert(PLAYERS_KEY_ASSIGNED_TABLE, data.assignedTable);
                jsonData.insert(PLAYERS_KEY_ASSIGNED_PAIR_NUMBER, data.assignedPairNumber);
            }
            players.append(QJsonObject { { name, jsonData } });
        }

        connection->sendData(QJsonObject {
            { KEY_ACTION,                       ACTION_HANDSHAKE },
            { HANDSHAKE_KEY_PHASE,              HANDSHAKE_PHASE_SYNCHRONIZE },
            { HANDSHAKE_KEY_LOCALE,             m_locale },
            { HANDSHAKE_KEY_SETTINGS,           settings },
            { HANDSHAKE_KEY_MARKERS,            markers },
            { HANDSHAKE_KEY_ABSENT_MARKER,      m_tournamentSettings->absentMarker() },
            { HANDSHAKE_KEY_SINGLES_MANAGEMENT, JsonHelper::singlesManagementToJson(
                                                    m_tournamentSettings->singlesManagement()) },
            { HANDSHAKE_KEY_PLAYERS,            players },
            { KEY_CHECKSUM,                     m_playersPage->checksum() }
        });

    } else if (phase == HANDSHAKE_PHASE_ALLOCATE_ID) {
        m_clientIdCounter++;
        connection->setId(m_clientIdCounter);
        connection->sendData(QJsonObject { { KEY_ACTION,          ACTION_HANDSHAKE },
                                           { HANDSHAKE_KEY_PHASE, HANDSHAKE_PHASE_ALLOCATE_ID },
                                           { KEY_CLIENT_ID,       m_clientIdCounter } });
        m_pendingConnections.removeOne(connection);
        m_connections.append(connection);
        emit message(tr("Synchronisation mit %1 war erfolgreich, Verbindung hergestellt "
                        "(Client-ID %2)").arg(connection->ip(), QString::number(connection->id())));
        emit statusUpdate(tr("Verbindung mit %1 hergestellt (Client-ID %2)").arg(
                             connection->ip(), QString::number(connection->id())));
        emit clientCountChanged(m_connections.count());
    }
}

// ====================
// Changes broadcasting
// ====================

// Registration changes
// ====================

void Server::broadcastRegisterPlayers(const QString &name, int marker, int clientId)
{
    QJsonObject message { { KEY_ACTION,            ACTION_PLAYERS_REGISTER },
                          { PLAYERS_KEY_NAME,      name },
                          { PLAYERS_KEY_MARKER_ID, marker } };
    broadcastMessage(message, clientId, tr("Sende Anmelden von „%1“").arg(name));
}

void Server::broadcastRenamePlayers(const QString &name, const QString &newName, int clientId)
{
    QJsonObject message { { KEY_ACTION,           ACTION_PLAYERS_RENAME },
                          { PLAYERS_KEY_NAME,     name },
                          { PLAYERS_KEY_NEW_NAME, newName } };
    broadcastMessage(message, clientId, tr("Sende Umbenennen von „%1“ zu „%2“").arg(name, newName));
}

void Server::broadcastAssemblePair(const QString &player1, const QString &player2,
                                   const QString &separator, int assignedMarker, int clientId)
{
    QJsonObject message { { KEY_ACTION,            ACTION_PLAYERS_ASSEMBLE_PAIR },
                          { PLAYERS_KEY_NAME,      QJsonArray { player1, player2, separator } },
                          { PLAYERS_KEY_MARKER_ID, assignedMarker } };
    broadcastMessage(message, clientId, tr("Sende Zusammenfügen von „%1“ und „%2“ zu einem "
                                           "Paar").arg(player1, player2));
}

void Server::broadcastDeletePlayers(const QString &name, int clientId)
{
    QJsonObject message { { KEY_ACTION,       ACTION_PLAYERS_DELETE },
                          { PLAYERS_KEY_NAME, name } };
    broadcastMessage(message, clientId, tr("Sende Löschen von „%1“").arg(name));
}

void Server::broadcastUpdateMarker(const QString &name, const QString &markerName, int markerId,
                                   int clientId)
{
    QJsonObject message { { KEY_ACTION,              ACTION_PLAYERS_UPDATE_MARKER },
                          { PLAYERS_KEY_NAME,        name },
                          { PLAYERS_KEY_MARKER_NAME, markerName },
                          { PLAYERS_KEY_MARKER_ID,   markerId } };
    if (markerId != 0) {
        broadcastMessage(message, clientId,
                        tr("Sende Markieren von „%1“ als „%2“").arg(name, markerName));
    } else {
        broadcastMessage(message, clientId,
                         tr("Sende Entfernen der Markierung von „%1“").arg(name));
    }
}

void Server::broadcastDeleteMarkedPlayers(const QString &markerName, int markerId, int clientId)
{
    QJsonObject message { { KEY_ACTION,              ACTION_PLAYERS_DELETE_MARKED },
                          { PLAYERS_KEY_MARKER_NAME, markerName },
                          { PLAYERS_KEY_MARKER_ID,   markerId } };
    broadcastMessage(message, clientId,
                     tr("Sende Löschen aller als „%1“ markierten %2").arg(
                        markerName, m_tournamentSettings->isPairMode() ? tr("Paare")
                                                                       : tr("Spieler")));
}

void Server::broadcastDeleteUnmarkedPlayers(int clientId)
{
    QJsonObject message { { KEY_ACTION, ACTION_PLAYERS_DELETE_UNMARKED } };
    broadcastMessage(message, clientId, tr("Sende Löschen aller unmarkierten %2").arg(
                                           m_tournamentSettings->isPairMode() ? tr("Paare")
                                                                              : tr("Spieler")));
}

void Server::broadcastDeleteAllPlayers(int clientId)
{
    QJsonObject message { { KEY_ACTION, ACTION_PLAYERS_DELETE_ALL } };
    broadcastMessage(message, clientId, tr("Sende Löschen aller %2").arg(
                                           m_tournamentSettings->isPairMode() ? tr("Paare")
                                                                              : tr("Spieler")));
}

void Server::broadcastListUpdateMarker(const QString &markerName, int markerId, int clientId)
{
    QJsonObject message { { KEY_ACTION,              ACTION_PLAYERS_LIST_UPDATE_MARKER },
                          { PLAYERS_KEY_MARKER_NAME, markerName },
                          { PLAYERS_KEY_MARKER_ID,   markerId } };

    if (markerId != 0) {
        broadcastMessage(message, clientId,
                         tr("Sende Markieren aller %1 als „%2“").arg(
                            m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler"),
                            markerName));
    } else {
        broadcastMessage(message, clientId, tr("Sende Entfernen aller Markierungen"));
    }
}

void Server::broadcastListRemoveMarker(const QString &markerName, int markerId, int clientId)
{
    QJsonObject message { { KEY_ACTION,              ACTION_PLAYERS_LIST_REMOVE_MARKER },
                          { PLAYERS_KEY_MARKER_NAME, markerName },
                          { PLAYERS_KEY_MARKER_ID,   markerId } };
    broadcastMessage(message, clientId,
                     tr("Sende Entfernen aller „%1“-Markierungen").arg(markerName));
}

// Preassignment changes
// =====================

void Server::broadcastUpdatePreAssignment(const QString &name, int table, int pair,
                                          bool removeMarker, int clientId)
{
    QJsonObject message { { KEY_ACTION,                       ACTION_UPDATE_PREASSIGNMENT },
                          { PLAYERS_KEY_NAME,                 name },
                          { PLAYERS_KEY_ASSIGNED_TABLE,       table },
                          { PLAYERS_KEY_ASSIGNED_PAIR_NUMBER, pair },
                          { PLAYERS_KEY_REMOVE_MARKER,        removeMarker } };
    const QString text = table == 0 ? tr("Sende Löschen der Auslosung für „%1“").arg(name)
                                    : tr("Sende Speichern der Auslosung für „%1“").arg(name);
    broadcastMessage(message, clientId, text);
}

void Server::broadcastDeleteAllPreAssignments(int clientId)
{
    QJsonObject message { { KEY_ACTION, ACTION_DELETE_ALL_PREASSIGNMENTS } };
    broadcastMessage(message, clientId, tr("Sende Löschen aller Auslosungen"));
}

// Markers definition changes
// ==========================

void Server::broadcastAddMarker(const QString &name, const QColor &color,
                                Muckturnier::MarkerSorting sorting, int clientId)
{
    QJsonObject message { { KEY_ACTION,          ACTION_MARKERS_ADD },
                          { MARKERS_KEY_NAME,    name },
                          { MARKERS_KEY_COLOR,   Connection::colorToString(color) },
                          { MARKERS_KEY_SORTING, sorting } };
    broadcastMessage(message, clientId, tr("Sende Hinzufügen der Markierung „%1“").arg(name));
}

void Server::broadcastDeleteMarker(const QString &name, int id, int sequence, int clientId)
{
    QJsonObject message { { KEY_ACTION,           ACTION_MARKERS_DELETE },
                          { MARKERS_KEY_NAME,     name },
                          { MARKERS_KEY_ID,       id },
                          { MARKERS_KEY_SEQUENCE, sequence } };
    broadcastMessage(message, clientId, tr("Sende Hinzufügen der Markierung „%1“").arg(name));
}

void Server::broadcastMoveMarker(const QString &name, int id, int sequence, int direction,
                                 int clientId)
{
    QJsonObject message { { KEY_ACTION,            ACTION_MARKERS_MOVE },
                          { MARKERS_KEY_NAME,      name },
                          { MARKERS_KEY_ID,        id },
                          { MARKERS_KEY_SEQUENCE,  sequence },
                          { MARKERS_KEY_DIRECTION, direction } };
    broadcastMessage(message, clientId,
                     tr("Sende Verschieben der Markierung „%1“ %2").arg(
                        name, direction == 1 ? tr("nach unten") : tr("nach oben")));
}

void Server::broadcastEditMarker(const QString &name, int id, const QString &newName,
                                 const QColor &color, Muckturnier::MarkerSorting sorting,
                                 int clientId)
{
    QJsonObject message { { KEY_ACTION,           ACTION_MARKERS_EDIT },
                          { MARKERS_KEY_NAME,     name },
                          { MARKERS_KEY_ID,       id },
                          { MARKERS_KEY_NEW_NAME, newName },
                          { MARKERS_KEY_COLOR,    Connection::colorToString(color) },
                          { MARKERS_KEY_SORTING,  sorting } };
    broadcastMessage(message, clientId, tr("Sende Änderung der Markierung „%1“").arg(name));
}

// Special markers changes
// =======================

void Server::broadcastSetAbsentMarker(const QString &name, int id, int clientId)
{
    QJsonObject message { { KEY_ACTION,       ACTION_SET_ABSENT_MARKER },
                          { MARKERS_KEY_NAME, name },
                          { MARKERS_KEY_ID,   id } };
    const QString text = id == -1
        ? tr("Sende Deaktivierung der Markierung für abwesende Anmeldungen")
        : tr("Sende Setzen der Markierung für abwesende Anmeldungen auf „%1“").arg(name);
    broadcastMessage(message, clientId, text);
}

void Server::broadcastSinglesManagementChange(const Muckturnier::SinglesManagement &data,
                                              int clientId)
{
    QJsonObject message { { KEY_ACTION, ACTION_CHANGE_SINGLES_MANAGEMENT },
                          { KEY_DATA,   JsonHelper::singlesManagementToJson(data) } };
    broadcastMessage(message, clientId, tr("Sende Änderung für das Behandeln allein gekommener "
                                           "Spieler"));
}

// Finishing the registration
// ==========================

void Server::broadcastFinishRegistration(int clientId)
{
    QJsonObject message { { KEY_ACTION, ACTION_FINISH_REGISTRATION } };
    broadcastMessage(message, clientId, tr("Sende Beenden der Anmeldung"));
}

// Ranking comparison
// ==================

// (Can't be requested by the server)

// =======================
// Client request handlers
// =======================

void Server::processClientRequest(Connection *connection, const QString &action,
                                  const QJsonObject &jsonObject)
{
    QApplication::setOverrideCursor(Qt::BusyCursor);

    const QString client = clientString(connection);
    const int clientId = jsonObject.value(KEY_CLIENT_ID).toInt();

    if (m_playersPage->checksum() != jsonObject.value(KEY_CHECKSUM).toString()) {
        emit message(tr("Ungültige Anfrage von %1: Die Prüfsummen stimmen nicht überein").arg(
                        client));
        connection->sendData(QJsonObject {
            { KEY_ACTION,        ACTION_REJECTED },
            { KEY_REJECT_REASON, static_cast<int>(RejectReason::OutdatedRequest) },
            { KEY_CLIENT_ID,     clientId }
        });
        QApplication::restoreOverrideCursor();
        return;
    }

    // We have to handle a ranking request separately due to the different needed function arguments
    if (action == ACTION_RANKING_REQUEST) {
        processRankingRequest(connection, client, clientId);
        QApplication::restoreOverrideCursor();
        return;
    }

    // Pass the request to the respective handler function
    if (m_requestHandlers.contains(action)) {
        (this->*m_requestHandlers.value(action))(jsonObject, client, clientId);
    } else {
        // This should not happen
        emit message(tr("Ungültige Anfrage von %1: Unbekannte Aktion angefragt").arg(client));
        connection->sendData(QJsonObject {
            { KEY_ACTION,        ACTION_REJECTED },
            { KEY_REJECT_REASON, static_cast<int>(RejectReason::UnknownAction) },
            { KEY_CLIENT_ID,     clientId }
        });
    }

    QApplication::restoreOverrideCursor();
}

// Registration changes
// ====================

void Server::processRegisterPlayers(const QJsonObject &jsonObject, const QString &client,
                                    int clientId)
{
    const QString name = jsonObject.value(PLAYERS_KEY_NAME).toString();
    const int marker = jsonObject.value(PLAYERS_KEY_MARKER_ID).toInt();
    emit message(tr("Anfrage zum Anmelden von „%1“ von %2").arg(name, client));

    if (! m_db->registerPlayers(name, marker)) {
        return;
    }
    emit statusUpdate(tr("Netzwerkänderung: %1 „%2“ angemeldet").arg(
                         m_tournamentSettings->isPairMode() ? tr("Paar") : tr("Spieler"), name));

    m_playersPage->processNetworkChange();
    broadcastRegisterPlayers(name, marker, clientId);
}

void Server::processRenamePlayers(const QJsonObject &jsonObject, const QString &client,
                                  int clientId)
{
    const QString name = jsonObject.value(PLAYERS_KEY_NAME).toString();
    const QString newName = jsonObject.value(PLAYERS_KEY_NEW_NAME).toString();
    emit message(tr("Anfrage zum Umbenennen von „%1“ zu „%2“ von %3").arg(name, newName, client));

    if (! m_db->renamePlayers(name, newName)) {
        return;
    }

    emit statusUpdate(tr("Netzwerkänderung: %1 „%2“ zu „%3“ umbenannt").arg(
                         m_tournamentSettings->isPairMode() ? tr("Paar") : tr("Spieler"),
                         name, newName));

    m_playersPage->processNetworkChange();
    m_playersPage->checkUpdateAssignmentPage();
    emit nameEdited();
    broadcastRenamePlayers(name, newName, clientId);
}

void Server::processAssemblePair(const QJsonObject &jsonObject, const QString &client,
                                 int clientId)
{
    const QJsonArray namesData = jsonObject.value(PLAYERS_KEY_NAME).toArray();
    const QString player1 = namesData.at(0).toString();
    const QString player2 = namesData.at(1).toString();
    const QString separator = namesData.at(2).toString();
    const int marker = jsonObject.value(PLAYERS_KEY_MARKER_ID).toInt();
    emit message(tr("Anfrage zum Zusammenfügen von „%1“ und „%2“ zu einem Paar").arg(
                    player1, player2, client));

    if (! m_db->assemblePair(player1, player2, separator, marker)) {
        return;
    }

    emit statusUpdate(tr("Netzwerkänderung: „%1“ und „%2“ zu einem Paar zusammengefügt").arg(
                         player1, player2));

    m_playersPage->processNetworkChange();
    m_playersPage->checkUpdateAssignmentPage();
    broadcastAssemblePair(player1, player2, separator, marker, clientId);
}

void Server::processDeletePlayers(const QJsonObject &jsonObject, const QString &client,
                                  int clientId)
{
    const QString name = jsonObject.value(PLAYERS_KEY_NAME).toString();
    emit message(tr("Anfrage zum Löschen von „%1“ von %2").arg(name, client));

    if (! m_db->deletePlayers(name)) {
        return;
    }

    emit statusUpdate(tr("Netzwerkänderung: %1 „%2“ gelöscht").arg(
                         m_tournamentSettings->isPairMode() ? tr("Paar") : tr("Spieler"), name));

    m_playersPage->processNetworkChange();
    m_playersPage->checkUpdateAssignmentPage();
    broadcastDeletePlayers(name, clientId);
}

void Server::processUpdatePlayersMarker(const QJsonObject &jsonObject, const QString &client,
                                        int clientId)
{
    const QString name = jsonObject.value(PLAYERS_KEY_NAME).toString();
    const QString markerName = jsonObject.value(PLAYERS_KEY_MARKER_NAME).toString();
    const int markerId = jsonObject.value(PLAYERS_KEY_MARKER_ID).toInt();
    if (markerId != 0) {
        emit message(tr("Anfrage zum Markieren von „%1“ als „%2“ von %3").arg(
                        name, markerName, client));
    } else {
        emit message(tr("Anfrage zum Entfernen der Markierung von „%1“ von %3").arg(
                        name, client));
    }

    if (! m_db->updateMarker(name, markerId)) {
        return;
    }

    if (markerId != 0) {
        emit statusUpdate(tr("Netzwerkänderung: %1 „%2“ als „%3“ markiert").arg(
                             m_tournamentSettings->isPairMode() ? tr("Paar") : tr("Spieler"),
                             name, markerName));
    } else {
        emit statusUpdate(tr("Netzwerkänderung: Markierung von %1 „%2“ entfernt").arg(
                             m_tournamentSettings->isPairMode() ? tr("Paar") : tr("Spieler"),
                             name));
    }

    m_playersPage->processNetworkChange();
    broadcastUpdateMarker(name, markerName, markerId, clientId);
}

void Server::processDeleteMarkedPlayers(const QJsonObject &jsonObject, const QString &client,
                                        int clientId)
{
    const QString markerName = jsonObject.value(PLAYERS_KEY_MARKER_NAME).toString();
    const int markerId = jsonObject.value(PLAYERS_KEY_MARKER_ID).toInt();
    emit message(tr("Anfrage zum Löschen aller als „%1“ markierten %2 von %3").arg(
                    markerName, m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler"),
                    client));

    if (! m_db->deleteMarkedPlayers(markerId)) {
        return;
    }

    emit statusUpdate(tr("Netzwerkänderung: Alle als %1 markierten „%2“ gelöscht").arg(
                         markerName, m_tournamentSettings->isPairMode() ? tr("Paare")
                                                                        : tr("Spieler")));

    m_playersPage->processNetworkChange();
    m_playersPage->checkUpdateAssignmentPage();
    broadcastDeleteMarkedPlayers(markerName, markerId, clientId);
}

void Server::processDeleteUnmarkedPlayers(const QJsonObject &, const QString &client,
                                          int clientId)
{
    emit message(tr("Anfrage zum Löschen aller unmarkierten %1 von %2").arg(
                    m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler"), client));

    if (! m_db->deleteUnmarkedPlayers()) {
        return;
    }

    emit statusUpdate(tr("Netzwerkänderung: Alle unmarkierten %1 gelöscht").arg(
                         m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler")));

    m_playersPage->processNetworkChange();
    m_playersPage->checkUpdateAssignmentPage();
    broadcastDeleteUnmarkedPlayers(clientId);
}

void Server::processDeleteAllPlayers(const QJsonObject &, const QString &client, int clientId)
{
    emit message(tr("Anfrage zum Löschen aller %1 von %2").arg(
                    m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler"), client));

    if (! m_db->deleteAllPlayers()) {
        return;
    }

    emit statusUpdate(tr("Netzwerkänderung: Alle %1 gelöscht").arg(
                         m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler")));

    m_playersPage->processNetworkChange();
    m_playersPage->checkUpdateAssignmentPage();
    broadcastDeleteAllPlayers(clientId);
}

void Server::processUpdateListMarker(const QJsonObject &jsonObject, const QString &client,
                                     int clientId)
{
    const QString markerName = jsonObject.value(PLAYERS_KEY_MARKER_NAME).toString();
    const int markerId = jsonObject.value(PLAYERS_KEY_MARKER_ID).toInt();
    if (markerId != 0) {
        emit message(tr("Anfrage zum Markieren aller %1 als „%2“ von %3").arg(
                        m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler"),
                        markerName, client));
    } else {
        emit message(tr("Anfrage zum Entfernen aller Markierungen von %1").arg(client));
    }

    if (! m_db->listUpdateMarker(markerId)) {
        return;
    }

    if (markerId != 0) {
        emit statusUpdate(tr("Netzwerkänderung: Alle %1 als „%2“ markiert").arg(
                             m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler"),
                             markerName));
    } else {
        emit statusUpdate(tr("Netzwerkänderung: Alle Markierungen entfernt"));
    }

    m_playersPage->processNetworkChange();
    broadcastListUpdateMarker(markerName, markerId, clientId);
}

void Server::processRemoveListMarker(const QJsonObject &jsonObject, const QString &client,
                                     int clientId)
{
    const QString markerName = jsonObject.value(PLAYERS_KEY_MARKER_NAME).toString();
    const int markerId = jsonObject.value(PLAYERS_KEY_MARKER_ID).toInt();
    emit message(tr("Anfrage zum Entfernen aller „%1“-Markierungen von %2").arg(
                    markerName, client));

    if (! m_db->listRemoveMarker(markerId)) {
        return;
    }

    emit statusUpdate(tr("Netzwerkänderung: Alle „%1“-Markierungen entfernt").arg(markerName));

    m_playersPage->processNetworkChange();
    broadcastListRemoveMarker(markerName, markerId, clientId);
}

// Preassignment changes
// =====================

void Server::processUpdatePreAssignment(const QJsonObject &jsonObject, const QString &client,
                                        int clientId)
{
    const QString name = jsonObject.value(PLAYERS_KEY_NAME).toString();
    const int table = jsonObject.value(PLAYERS_KEY_ASSIGNED_TABLE).toInt();
    const int pair = jsonObject.value(PLAYERS_KEY_ASSIGNED_PAIR_NUMBER).toInt();
    const bool removeMarker = jsonObject.value(PLAYERS_KEY_REMOVE_MARKER).toBool();
    if (table == 0) {
        emit message(tr("Anfrage zum Löschen der Auslosung für „%1“ von %2").arg(name, client));
    } else {
        emit message(tr("Anfrage zum Speichern der Auslosung für „%1“ von %2").arg(name, client));
    }

    if (! m_db->updatePreAssignment(name, table, pair, removeMarker)) {
        return;
    }

    if (table == 0) {
        emit statusUpdate(tr("Netzwerkänderung: Auslosung für „%1“ gelöscht").arg(name));
    } else {
        emit statusUpdate(tr("Netzwerkänderung: Auslosung für „%1“ gespeichert (Tisch %2 Paar "
                             "%3)").arg(name, QString::number(table), QString::number(pair)));
    }

    m_playersPage->processNetworkChange();
    m_playersPage->checkUpdateAssignmentPage();
    broadcastUpdatePreAssignment(name, table, pair, removeMarker, clientId);
}

void Server::processDeleteAllPreAssignments(const QJsonObject &, const QString &client,
                                            int clientId)
{
    emit message(tr("Anfrage zum Löschen aller Auslosungen von %1").arg(client));

    if (! m_db->deleteAllPreAssignments()) {
        return;
    }

    emit statusUpdate(tr("Netzwerkänderung: Alle Auslosungen gelöscht"));

    m_playersPage->processNetworkChange();
    m_playersPage->checkUpdateAssignmentPage();
    broadcastDeleteAllPreAssignments(clientId);
}

// Markers defintion changes
// =========================

void Server::processAddMarker(const QJsonObject &jsonObject, const QString &client, int clientId)
{
    const QString name = jsonObject.value(MARKERS_KEY_NAME).toString();
    const QColor color = Connection::colorFromValue(jsonObject.value(MARKERS_KEY_COLOR));
    const Muckturnier::MarkerSorting sorting = static_cast<Muckturnier::MarkerSorting>(
                                                   jsonObject.value(MARKERS_KEY_SORTING).toInt());
    emit message(tr("Anfrage zum Hinzufügen der Markierung „%1“ von %2").arg(name, client));

    if (! m_db->addMarker(name, color, sorting)) {
        return;
    }

    emit statusUpdate(tr("Netzwerkänderung: Markierung „%1“ hinzugefügt").arg(name));

    m_markersWidget->processNetworkChange();
    broadcastAddMarker(name, color, sorting, clientId);
}

void Server::processDeleteMarker(const QJsonObject &jsonObject, const QString &client,
                                 int clientId)
{
    const QString name = jsonObject.value(MARKERS_KEY_NAME).toString();
    const int id = jsonObject.value(MARKERS_KEY_ID).toInt();
    const int sequence = jsonObject.value(MARKERS_KEY_SEQUENCE).toInt();
    emit message(tr("Anfrage zum Löschen der Markierung „%1“ von %2").arg(name, client));

    if (! m_db->deleteMarker(id, sequence)) {
        return;
    }

    emit statusUpdate(tr("Netzwerkänderung: Markierung „%1“ gelöscht").arg(name));

    m_markersWidget->processNetworkChange();
    broadcastDeleteMarker(name, id, sequence, clientId);
}

void Server::processMoveMaker(const QJsonObject &jsonObject, const QString &client, int clientId)
{
    const QString name = jsonObject.value(MARKERS_KEY_NAME).toString();
    const int id = jsonObject.value(MARKERS_KEY_ID).toInt();
    const int sequence = jsonObject.value(MARKERS_KEY_SEQUENCE).toInt();
    const int direction = jsonObject.value(MARKERS_KEY_DIRECTION).toInt();
    emit message(tr("Anfrage zum Verschieben der Markierung „%1“ %2 von %3").arg(
                    name, direction == 1 ? tr("nach unten") : tr("nach oben"), client));

    if (! m_db->moveMarker(id, sequence, direction)) {
        return;
    }

    emit statusUpdate(tr("Netzwerkänderung: Markierung „%1“ %2 verschoben").arg(
                         name, direction == 1 ? tr("nach unten") : tr("nach oben")));

    m_markersWidget->processNetworkChange();
    broadcastMoveMarker(name, id, sequence, direction, clientId);
}

void Server::processEditMarker(const QJsonObject &jsonObject, const QString &client, int clientId)
{
    const QString name = jsonObject.value(MARKERS_KEY_NAME).toString();
    const int id = jsonObject.value(MARKERS_KEY_ID).toInt();
    const QString newName = jsonObject.value(MARKERS_KEY_NEW_NAME).toString();
    const QColor color = Connection::colorFromValue(jsonObject.value(MARKERS_KEY_COLOR));
    const Muckturnier::MarkerSorting sorting = static_cast<Muckturnier::MarkerSorting>(
                                                   jsonObject.value(MARKERS_KEY_SORTING).toInt());
    emit message(tr("Anfrage zum Ändern der Markierung „%1“ von %2").arg(name, client));

    if (! m_db->editMarker(id, newName, color, sorting)) {
        return;
    }

    emit statusUpdate(tr("Netzwerkänderung: Markierung „%1“ geändert").arg(name));

    m_markersWidget->processNetworkChange();
    broadcastEditMarker(name, id, newName, color, sorting, clientId);
}

// Special markers changes
// =======================

void Server::processSetAbsentMarker(const QJsonObject &jsonObject, const QString &client,
                                    int clientId)
{
    const QString name = jsonObject.value(MARKERS_KEY_NAME).toString();
    const int id = jsonObject.value(MARKERS_KEY_ID).toInt();
    emit message(id == -1 ? tr("Anfrage zum Deaktivieren der Markierung für abwesende Anmeldungen "
                               "von %1").arg(client)
                          : tr("Anfrage zum Setzen der Markierung für abwesende Anmeldungen auf "
                               "„%1“ von %2").arg(name, client));

    m_markersWidget->setAbsentMarker(id);
    m_tournamentSettings->setAbsentMarker(id);
    emit statusUpdate(id == -1
        ? tr("Netzwerkänderung: Markierung für abwesende Anmeldungen deaktiviert")
        : tr("Netzwerkänderung: Markierung für abwesende Anmeldungen auf „%1“ gesetzt").arg(name));

    m_markersWidget->processNetworkChange();
    broadcastSetAbsentMarker(name, id, clientId);
}

void Server::processChangeSinglesManagement(const QJsonObject &jsonObject, const QString &client,
                                            int clientId)
{
    const Muckturnier::SinglesManagement singlesManagement =
        JsonHelper::singlesManagementFromJson(jsonObject.value(KEY_DATA).toObject());
    emit message(tr("Anfrage zum Ändern der Einstellungen zum Behandeln allein gekommner Spieler "
                    "von %1").arg(client));

    m_markersWidget->setSinglesManagement(singlesManagement);
    m_tournamentSettings->setSinglesManagement(singlesManagement);
    emit statusUpdate(tr("Netzwerkänderung: Änderungen der Einstellungen zum Behandeln allein "
                         "gekommener Spieler übernommen"));

    m_markersWidget->processNetworkChange();
    broadcastSinglesManagementChange(singlesManagement, clientId);
}

// Finishing the registration
// ==========================

void Server::processFinishRegistration(const QJsonObject &, const QString &client, int clientId)
{
    emit message(tr("Anfrage zum Beenden der Anmeldung von %1").arg(client));

    m_playersPage->networkFinishRegistration(clientId);
    emit statusUpdate(tr("Netzwerkänderung: Anmeldung beendet"));

    broadcastFinishRegistration(clientId);
}

// Ranking comparison
// ==================

void Server::processRankingRequest(Connection *connection, const QString &client, int clientId)
{
    emit message(tr("Rangliste von %1 angefragt").arg(client));

    if (! m_db->localTournamentStarted()) {
        connection->sendData(QJsonObject { { KEY_ACTION,            ACTION_RANKING_REQUEST },
                                           { KEY_CLIENT_ID,         clientId },
                                           { RANKING_KEY_AVAILABLE, false } });
        emit message(tr("Es ist noch keine Rangliste verfügbar"));
        return;
    }

    QJsonArray names;
    QJsonArray boogers;
    QJsonArray goals;
    QJsonArray opponentGoals;
    QJsonArray playedBoogers;

    bool success;
    const QVector<Database::RankData> ranking = m_db->ranking(-1, &success);
    if (! success) {
        return;
    }

    for (const Database::RankData &rank : ranking) {
        names.append(QJsonValue(rank.name));
        boogers.append(QJsonValue(rank.boogers));
        goals.append(QJsonValue(rank.goals));
        opponentGoals.append(QJsonValue(rank.opponentGoals));
        playedBoogers.append(QJsonValue(rank.playedBoogers));
    }

    connection->sendData(QJsonObject { { KEY_ACTION,                 ACTION_RANKING_REQUEST },
                                       { KEY_CLIENT_ID,              clientId },
                                       { RANKING_KEY_AVAILABLE,      true },
                                       { RANKING_KEY_NAMES,          names },
                                       { RANKING_KEY_BOOGERS,        boogers },
                                       { RANKING_KEY_GOALS,          goals },
                                       { RANKING_KEY_OPPONENT_GOALS, opponentGoals },
                                       { RANKING_KEY_PLAYED_BOOGERS, playedBoogers } });

    emit message(tr("Rangliste an %1 gesendet").arg(client));
}
