/* Copyright (c) 2018-2019 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef SERVER_H
#define SERVER_H

// Local includes
#include "shared/Muckturnier.h"

// Qt includes
#include <QTcpServer>

// Local classes
class SharedObjects;
class Database;
class TournamentSettings;
class PlayersPage;
class MarkersWidget;
class Connection;

class Server : public QTcpServer
{
    Q_OBJECT

public:
    explicit Server(QObject *parent, int connectTimeout, SharedObjects *sharedObjects,
                    PlayersPage *playersPage);
    int clientCount() const;
    void close();

    // Changes broadcasting
    // ====================

    // Helper macro to keep the header a bit more tidy
    #define BROADCAST_FUNCTION(name, ...) void broadcast##name(__VA_ARGS__, int clientId = 0)
    #define BROADCAST_FUNCTION_NA(name) void broadcast##name(int clientId = 0)

    // Registration changes

    BROADCAST_FUNCTION(RegisterPlayers, const QString &name, int marker);
    BROADCAST_FUNCTION(DeletePlayers, const QString &name);
    BROADCAST_FUNCTION(RenamePlayers, const QString &name, const QString &newName);
    BROADCAST_FUNCTION(AssemblePair, const QString &player1, const QString &player2,
                                     const QString &separator, int assignedMarker);
    BROADCAST_FUNCTION(UpdateMarker, const QString &name, const QString &markerName, int markerId);

    BROADCAST_FUNCTION(DeleteMarkedPlayers, const QString &markerName, int markerId);
    BROADCAST_FUNCTION_NA(DeleteUnmarkedPlayers);
    BROADCAST_FUNCTION_NA(DeleteAllPlayers);
    BROADCAST_FUNCTION(ListUpdateMarker, const QString &markerName, int markerId);
    BROADCAST_FUNCTION(ListRemoveMarker, const QString &markerName, int markerId);

    // Preassignment changes
    BROADCAST_FUNCTION(UpdatePreAssignment, const QString &name, int table, int pair,
                                            bool removeMarker);
    BROADCAST_FUNCTION_NA(DeleteAllPreAssignments);

    // Markers definition changes
    BROADCAST_FUNCTION(AddMarker, const QString &name, const QColor &color,
                                  Muckturnier::MarkerSorting sorting);
    BROADCAST_FUNCTION(DeleteMarker, const QString &name, int id, int sequence);
    BROADCAST_FUNCTION(MoveMarker, const QString &name, int id, int sequence, int direction);
    BROADCAST_FUNCTION(EditMarker, const QString &name, int id, const QString &newName,
                                   const QColor &color, Muckturnier::MarkerSorting sorting);

    // Special markers changes
    BROADCAST_FUNCTION(SetAbsentMarker, const QString &name, int id);
    BROADCAST_FUNCTION(SinglesManagementChange, const Muckturnier::SinglesManagement &data);

    // Finishing the registration
    BROADCAST_FUNCTION_NA(FinishRegistration);

    // Ranking comparison
    // (Can't be requested by the server)

    #undef BROADCAST_FUNCTION
    #undef BROADCAST_FUNCTION_NA

signals:
    void protocolError(const QString &text);
    void message(const QString &text);
    void statusUpdate(const QString &text);
    void clientCountChanged(int count);
    void nameEdited();

protected:
    virtual void incomingConnection(qintptr descriptor) override;

private slots:
    void checkHandshakeTimeout(Connection *connection);
    void removeConnection(Connection *connection);
    void processHandshake(Connection *connection, const QString &phase);
    void processClientRequest(Connection *connection, const QString &action,
                              const QJsonObject &jsonObject);

private: // Functions
    void broadcastMessage(QJsonObject &jsonObject, int clientId, const QString &messageText);
    QString clientString(Connection *connection) const;

    // Client request handlers
    // =======================

    // Helper macro to keep the header a bit more tidy
    #define REQUEST_HANDLER(name) \
            void process##name(const QJsonObject &jsonObject, const QString &client, int clientId)

    // Registration changes

    REQUEST_HANDLER(RegisterPlayers);
    REQUEST_HANDLER(RenamePlayers);
    REQUEST_HANDLER(AssemblePair);
    REQUEST_HANDLER(DeletePlayers);
    REQUEST_HANDLER(UpdatePlayersMarker);

    REQUEST_HANDLER(DeleteMarkedPlayers);
    REQUEST_HANDLER(DeleteUnmarkedPlayers);
    REQUEST_HANDLER(DeleteAllPlayers);
    REQUEST_HANDLER(UpdateListMarker);
    REQUEST_HANDLER(RemoveListMarker);

    // Preassignment changes
    REQUEST_HANDLER(UpdatePreAssignment);
    REQUEST_HANDLER(DeleteAllPreAssignments);

    // Markers definition changes
    REQUEST_HANDLER(AddMarker);
    REQUEST_HANDLER(DeleteMarker);
    REQUEST_HANDLER(MoveMaker);
    REQUEST_HANDLER(EditMarker);

    // Special markers changes
    REQUEST_HANDLER(SetAbsentMarker);
    REQUEST_HANDLER(ChangeSinglesManagement);

    // Finishing the registration
    REQUEST_HANDLER(FinishRegistration);

    // Ranking comparison
    // (this one has another signature, thus we can't generate it via REQUEST_HANDLER)
    void processRankingRequest(Connection *connection, const QString &client, int clientId);

    #undef REQUEST_HANDLER

private: // Variables
    const int m_connectTimeout;
    Database *m_db;
    TournamentSettings *m_tournamentSettings;
    PlayersPage *m_playersPage;
    MarkersWidget *m_markersWidget;

    const QString m_locale;
    const QHash<QString, void(Server::*)(const QJsonObject &, const QString &, int)>
              m_requestHandlers;

    QVector<Connection *> m_pendingConnections;
    QVector<Connection *> m_connections;
    int m_clientIdCounter = 0;

};

#endif // SERVER_H
