/* Copyright (c) 2019 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes

#include "MarkersTemplatePage.h"

#include "shared/SharedObjects.h"
#include "shared/Database.h"
#include "shared/Settings.h"
#include "shared/TournamentSettings.h"
#include "shared/TableDelegate.h"

// Qt includes
#include <QDebug>
#include <QVBoxLayout>
#include <QGroupBox>
#include <QLabel>
#include <QTableWidget>
#include <QPushButton>
#include <QHeaderView>
#include <QMessageBox>

MarkersTemplatePage::MarkersTemplatePage(QWidget *parent, SharedObjects *sharedObjects)
    : AbstractSettingsPage(parent, tr("Vorlage Markierungen")),
      m_db(sharedObjects->database()),
      m_settings(sharedObjects->settings()),
      m_tournamentSettings(sharedObjects->tournamentSettings())
{
    QLabel *markersTemplateLabel = new QLabel(tr("Folgende Markierungen werden in neuen "
                                                 "Turnierdatenbanken automatisch angelegt:"));
    markersTemplateLabel->setWordWrap(true);
    layout()->addWidget(markersTemplateLabel);

    m_markersTemplateDisplay = new QTableWidget;
    m_markersTemplateDisplay->verticalHeader()->hide();
    m_markersTemplateDisplay->setShowGrid(false);
    m_markersTemplateDisplay->setSelectionMode(QTableWidget::NoSelection);
    m_markersTemplateDisplay->setFocusPolicy(Qt::NoFocus);
    m_markersTemplateDisplay->setEditTriggers(QAbstractItemView::NoEditTriggers);
    m_markersTemplateDisplay->setItemDelegate(new TableDelegate(m_markersTemplateDisplay));
    m_markersTemplateDisplay->setColumnCount(2);
    m_markersTemplateDisplay->setHorizontalHeaderLabels({ tr("Name/Farbe"), tr("Sortierung") });
    layout()->addWidget(m_markersTemplateDisplay);

    m_defaultMarkerLabel = new QLabel;
    layout()->addWidget(m_defaultMarkerLabel);

    m_absentMarkerLabel = new QLabel;
    layout()->addWidget(m_absentMarkerLabel);

    m_noProcessSingles = new QLabel(tr("Allein gekommene Spieler im Paarmodus werden nicht "
                                       "gesondert berücksichtigt"));
    m_noProcessSingles->setWordWrap(true);
    m_noProcessSingles->hide();
    layout()->addWidget(m_noProcessSingles);

    m_processSingles = new QGroupBox(tr("Allein gekommene Spieler"));
    m_processSingles->setStyleSheet(QStringLiteral("font-weight: normal"));
    QVBoxLayout *processSingleslayout = new QVBoxLayout(m_processSingles);
    m_singlesMarker = new QLabel;
    processSingleslayout->addWidget(m_singlesMarker);
    m_automaticMarking = new QLabel;
    processSingleslayout->addWidget(m_automaticMarking);
    m_assignedMarker = new QLabel;
    processSingleslayout->addWidget(m_assignedMarker);
    layout()->addWidget(m_processSingles);

    layout()->addStretch();

    QHBoxLayout *controlLayout = new QHBoxLayout;
    layout()->addLayout(controlLayout);
    controlLayout->addStretch();

    QPushButton *adoptCurrentSettingsButton = new QPushButton(tr("Aktuelle Einstellungen "
                                                                 "übernehmen"));
    connect(adoptCurrentSettingsButton, &QPushButton::clicked,
            this, &MarkersTemplatePage::adoptCurrentSettings);
    controlLayout->addWidget(adoptCurrentSettingsButton);
    adoptCurrentSettingsButton->setEnabled(! m_db->dbFile().isEmpty());

    QPushButton *restoreTemplateButton = new QPushButton(tr("Zurücksetzen"));
    connect(restoreTemplateButton, &QPushButton::clicked,
            this, &MarkersTemplatePage::restoreTemplate);
    controlLayout->addWidget(restoreTemplateButton);

    m_markersTemplate = m_settings->markersTemplate();
    m_singlesManagementTemplate = m_settings->singlesManagementTemplate();
    m_defaultMarker = m_settings->defaultMarker();
    m_absentMarker = m_settings->absentMarker();

    updateTemplateDisplay();
}

QString MarkersTemplatePage::markerName(int id) const
{
    for (const Muckturnier::Marker &marker : m_markersTemplate) {
        if (marker.id == id) {
            if (! marker.name.isEmpty()) {
                return QStringLiteral("<span style=\"color: %1\">%2</span>").arg(
                                      marker.color.name(), marker.name);
            } else {
                return tr("(unmarkiert)");
            }
        }
    }
    return tr("(unmarkiert)");
}

void MarkersTemplatePage::updateTemplateDisplay()
{
    m_markersTemplateDisplay->setRowCount(m_markersTemplate.count());
    for (int i = 0; i < m_markersTemplate.count(); i++) {
        QTableWidgetItem *item = new QTableWidgetItem;
        m_markersTemplateDisplay->setItem(i, 0, item);
        item->setText(m_markersTemplate.at(i).name.isEmpty() ? tr("(unmarkiert)")
                                                             : m_markersTemplate.at(i).name);
        item->setForeground(m_markersTemplate.at(i).color);
        m_markersTemplateDisplay->setItem(i, 1, new QTableWidgetItem(
            m_markersTemplate.at(i).sorting == Muckturnier::MarkerSorting::NewBlock
                                                   ? tr("Neuer Block") : tr("Fortlaufend")));
    }
    m_markersTemplateDisplay->resizeColumnsToContents();

    m_defaultMarkerLabel->setText(tr("Neue Anmeldungen markieren als: %1").arg(
                                  markerName(m_defaultMarker)));

    if (m_absentMarker != -1) {
        m_absentMarkerLabel->setText(tr("Anmeldungen mit Markierung %1 werden ignoriert").arg(
                                        markerName(m_absentMarker)));
    } else {
        m_absentMarkerLabel->setText(tr("Es werden keine Anmeldungen ignoriert"));
    }

    if (m_singlesManagementTemplate.enabled) {
        m_noProcessSingles->hide();
        m_processSingles->show();
        m_singlesMarker->setText(tr("Markierung für „Allein da“: %1").arg(
                                    markerName(m_singlesManagementTemplate.singleMarker)));
        if (m_singlesManagementTemplate.markSingles) {
            m_automaticMarking->setText(
                tr("Automatisch setzen, wenn der Name „%1“ %2").arg(
                    m_singlesManagementTemplate.conditionString,
                    m_singlesManagementTemplate.condition == 0 ? tr("nicht enthält") : tr("enthält")));
            m_automaticMarking->show();
        } else {
            m_automaticMarking->hide();
        }
        m_assignedMarker->setText(tr("Markierung nach Zuweisung: %1").arg(
                                    markerName(m_singlesManagementTemplate.assignedMarker)));
    } else {
        m_noProcessSingles->show();
        m_processSingles->hide();
    }
}

void MarkersTemplatePage::restoreTemplate()
{
    m_defaultMarker = 0;
    m_absentMarker = -1;
    m_markersTemplate = m_settings->defaultMarkersTemplate();
    m_singlesManagementTemplate = m_settings->defaultSinglesManagementTemplate();
    updateTemplateDisplay();
}

void MarkersTemplatePage::adoptCurrentSettings()
{
    bool success;
    m_markersTemplate = m_db->markers(&success);
    if (! success) {
        return;
    }

    m_singlesManagementTemplate = m_tournamentSettings->singlesManagement();
    m_defaultMarker = m_tournamentSettings->defaultMarker();
    m_absentMarker = m_tournamentSettings->absentMarker();
    updateTemplateDisplay();
}

void MarkersTemplatePage::saveSettings()
{
    m_settings->saveDefaultMarker(m_defaultMarker);
    m_settings->saveAbsentMarker(m_absentMarker);
    m_settings->saveMarkersTemplate(m_markersTemplate);
    m_settings->saveSinglesManagementTemplate(m_singlesManagementTemplate);
}
