/* Copyright (c) 2019 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes

#include "TournamentTemplatePage.h"

#include "shared/SharedObjects.h"
#include "shared/Database.h"
#include "shared/TournamentSettings.h"
#include "shared/TableDelegate.h"
#include "shared/Muckturnier.h"

// Qt includes
#include <QDebug>
#include <QVBoxLayout>
#include <QLabel>
#include <QTableWidget>
#include <QPushButton>
#include <QMessageBox>
#include <QHeaderView>

TournamentTemplatePage::TournamentTemplatePage(QWidget *parent, SharedObjects *sharedObjects)
    : AbstractSettingsPage(parent, tr("Vorlage Turniereinstellungen")),
      m_db(sharedObjects->database()),
      m_settings(sharedObjects->settings()),
      m_tournamentSettings(sharedObjects->tournamentSettings())
{
    QLabel *tournamentTemplateLabel = new QLabel(tr("Folgende Turniereinstellungen werden für "
                                                    "neue Turnierdatenbanken voreingestellt:"));
    tournamentTemplateLabel->setWordWrap(true);
    layout()->addWidget(tournamentTemplateLabel);

    m_templateDisplay = new QTableWidget;
    m_templateDisplay->horizontalHeader()->hide();
    m_templateDisplay->verticalHeader()->hide();
    m_templateDisplay->setShowGrid(false);
    m_templateDisplay->setSelectionMode(QTableWidget::NoSelection);
    m_templateDisplay->setFocusPolicy(Qt::NoFocus);
    m_templateDisplay->setEditTriggers(QAbstractItemView::NoEditTriggers);
    m_templateDisplay->setItemDelegate(new TableDelegate(m_templateDisplay));
    m_templateDisplay->setColumnCount(2);
    m_templateDisplay->setRowCount(5);
    const int height = m_templateDisplay->rowHeight(0) * 5 + 4;
    m_templateDisplay->setMinimumHeight(height);
    layout()->addWidget(m_templateDisplay);

    m_templateDisplay->setItem(0, 0, new QTableWidgetItem(tr("Turniermodus:")));
    m_tournamentMode = new QTableWidgetItem;
    m_templateDisplay->setItem(0, 1, m_tournamentMode);

    m_templateDisplay->setItem(1, 0, new QTableWidgetItem(tr("Punkte pro Bobbl:")));
    m_boogerScore = new QTableWidgetItem;
    m_templateDisplay->setItem(1, 1, m_boogerScore);

    m_templateDisplay->setItem(2, 0, new QTableWidgetItem(tr("Bobbl pro Runde:")));
    m_boogersPerRound = new QTableWidgetItem;
    m_templateDisplay->setItem(2, 1, m_boogersPerRound);

    m_autoSelectPairsLabel = new QTableWidgetItem(tr("Automatische Tisch- bzw. Paarauswahl:"));
    m_enabledFlags = m_autoSelectPairsLabel->flags() | Qt::ItemIsEnabled;
    m_disabledFlags = m_autoSelectPairsLabel->flags() & (~Qt::ItemIsEnabled);
    m_templateDisplay->setItem(3, 0, m_autoSelectPairsLabel);
    m_autoSelectPairs = new QTableWidgetItem;
    m_templateDisplay->setItem(3, 1, m_autoSelectPairs);

    m_templateDisplay->setItem(4, 0, new QTableWidgetItem(tr("Eingabe ohne Tischnummern:")));
    m_noTableNumbers = new QTableWidgetItem;
    m_templateDisplay->setItem(4, 1, m_noTableNumbers);

    layout()->addStretch();

    QHBoxLayout *controlLayout = new QHBoxLayout;
    layout()->addLayout(controlLayout);
    controlLayout->addStretch();

    QPushButton *adoptCurrentSettingsButton = new QPushButton(tr("Aktuelle Einstellungen "
                                                                 "übernehmen"));
    connect(adoptCurrentSettingsButton, &QPushButton::clicked,
            this, &TournamentTemplatePage::adoptCurrentSettings);
    controlLayout->addWidget(adoptCurrentSettingsButton);
    adoptCurrentSettingsButton->setEnabled(! m_db->dbFile().isEmpty());

    QPushButton *restoreTemplateButton = new QPushButton(tr("Zurücksetzen"));
    connect(restoreTemplateButton, &QPushButton::clicked,
            this, &TournamentTemplatePage::restoreTemplate);
    controlLayout->addWidget(restoreTemplateButton);

    m_template = m_settings->tournamentTemplate();
    updateTemplateDisplay();
}

void TournamentTemplatePage::updateTemplateDisplay()
{
    const bool isPairMode = m_template.tournamentMode == Muckturnier::TournamentMode::FixedPairs;

    m_tournamentMode->setText(isPairMode ? tr("Feste Paare") : tr("Einzelne Spieler"));
    m_boogerScore->setText(QString::number(m_template.boogerScore));
    m_boogersPerRound->setText(QString::number(m_template.boogersPerRound));
    m_autoSelectPairs->setText(m_template.autoSelectPairs ? tr("Ja") : tr("Nein"));
    m_autoSelectPairsLabel->setFlags(isPairMode ? m_enabledFlags : m_disabledFlags);
    m_autoSelectPairs->setFlags(isPairMode ? m_enabledFlags : m_disabledFlags);
    m_noTableNumbers->setText(m_template.noTableNumbers ? tr("Ja") : tr("Nein"));

    m_templateDisplay->resizeColumnsToContents();
    const int width = m_templateDisplay->columnWidth(0) + m_templateDisplay->columnWidth(1) + 4;
    m_templateDisplay->setMinimumWidth(width);
}

void TournamentTemplatePage::restoreTemplate()
{
    m_template = m_settings->defaultTournamentTemplate();
    updateTemplateDisplay();
}

void TournamentTemplatePage::adoptCurrentSettings()
{
    m_template.tournamentMode = m_tournamentSettings->tournamentMode();
    m_template.boogerScore = m_tournamentSettings->boogerScore();
    m_template.boogersPerRound = m_tournamentSettings->boogersPerRound();
    m_template.autoSelectPairs = m_tournamentSettings->autoSelectPairs();
    m_template.noTableNumbers = m_tournamentSettings->noTableNumbers();
    updateTemplateDisplay();
}

void TournamentTemplatePage::saveSettings()
{
    m_settings->saveTournamentTemplate(m_template);
}
