/* Copyright (c) 2011-2020 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes

#include "TournamentInfoDialog.h"

#include "shared/SharedObjects.h"
#include "shared/Database.h"
#include "shared/TournamentSettings.h"

// Qt includes
#include <QGridLayout>
#include <QLabel>
#include <QDir>
#include <QGroupBox>
#include <QDebug>
#include <QCryptographicHash>

TournamentInfoDialog::TournamentInfoDialog(QWidget *parent, SharedObjects *sharedObjects)
    : TitleDialog(parent)
{
    // Cache the pointers to the database and the tournament settings for convenience
    Database *db = sharedObjects->database();
    TournamentSettings *ts = sharedObjects->tournamentSettings();
    const bool isPairMode = ts->isPairMode();

    setTitle(tr("Turnierinformationen"));
    addButtonBox(QDialogButtonBox::Ok);

    addGroup(tr("Datenbank"));

    addLabel(tr("Dateiname:"));
    QString dbFile = QDir::toNativeSeparators(db->dbFile());
    QFont font = this->font();
    QFontMetrics fontMetrics(font);
    if (fontMetrics.boundingRect(dbFile).width() > fontMetrics.boundingRect(
        QStringLiteral("/A/very/long/path/which/still/does/not/mess/up/the/layout")).width()) {
        addInfo(dbFile, true);
    } else {
        addInfo(dbFile, false);
    }

    addLabel(tr("Erstellt mit:"));
    addInfo(tr("Muckturnier %1").arg(ts->muckturnierVersion()));

    addGroup(tr("Einstellungen"));

    addLabel(tr("Tuniermodus:"));
    addInfo(isPairMode ? tr("Feste Paare") : tr("Einzelne Spieler"));

    addLabel(tr("Punkte pro Bobbl:"));
    addInfo(QString::number(ts->boogerScore()));

    addLabel(tr("Bobbl pro Runde:"));
    addInfo(QString::number(ts->boogersPerRound()));

    addGroup(tr("Turnier"));

    addLabel(isPairMode ? tr("Angemeldete Paare:") : tr("Angemeldete Spieler:"));
    addInfo(QString::number(db->pairsOrPlayersCount()));

    addLabel(tr("Gespielte Runden:"));
    addInfo(QString::number(db->tournamentStarted() ? db->allRounds() : 0));

    if (db->tournamentStarted()) {
        bool allFinished = true;
        for (bool finished : db->roundsState().values()) {
            if (! finished) {
                allFinished = false;
                break;
            }
        }
        addLabel(tr("Alle Runden abgeschlossen:"));
        addInfo(allFinished ? tr("Ja") : tr("Nein"));
    }
}

void TournamentInfoDialog::addGroup(const QString &text)
{
    QGroupBox *group = new QGroupBox(text);
    layout()->addWidget(group);
    QGridLayout *layout = new QGridLayout(group);
    m_currentGroupLayout = layout;
    m_row = -1;
}

void TournamentInfoDialog::addLabel(const QString &text)
{
    m_row++;
    QLabel *label = new QLabel(text);
    label->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Minimum);
    label->setTextInteractionFlags(Qt::TextBrowserInteraction | Qt::TextSelectableByMouse);
    label->setAlignment(Qt::AlignTop);
    m_currentGroupLayout->addWidget(label, m_row, 0);
}

void TournamentInfoDialog::addInfo(const QString &text, bool wordWrap)
{
    QLabel *label = new QLabel(text);
    label->setTextInteractionFlags(Qt::TextBrowserInteraction | Qt::TextSelectableByMouse);
    if (wordWrap) {
        label->setWordWrap(true);
    }
    m_currentGroupLayout->addWidget(label, m_row, 1);
}
