/* Copyright (c) 2018-2019 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes
#include "AbstractNetworkEngine.h"
#include "ServerProtocol.h"

// Qt includes
#include <QDebug>
#include <QTcpSocket>
#include <QHostAddress>
#include <QJsonObject>
#include <QDataStream>
#include <QColor>

using namespace ServerProtocol;

AbstractNetworkEngine::AbstractNetworkEngine(QObject *parent) : AbstractJsonEngine(parent)
{
    m_socket = new QTcpSocket(this);
    connect(m_socket, &QTcpSocket::readyRead, this, &AbstractNetworkEngine::readData);
}

QTcpSocket * AbstractNetworkEngine::socket() const
{
    return m_socket;
}

QString AbstractNetworkEngine::ip() const
{
    return m_socket->peerAddress().toString();
}

void AbstractNetworkEngine::sendData(const QJsonObject &jsonObject)
{
    QDataStream stream(m_socket);
    stream.setVersion(STREAM_VERSION);
    stream << jsonObjectToByteArray(jsonObject);
}

void AbstractNetworkEngine::readData()
{
    QByteArray data;
    QDataStream stream(m_socket);
    stream.setVersion(STREAM_VERSION);

    for (;;) {
        // Probably, the connection has been closed in the meantime
        if (! m_socket->isValid()) {
            break;
        }

        // Read data from the socket
        stream.startTransaction();
        stream >> data;

        if (! stream.commitTransaction()) {
            // The data is incomplete, wait for more
            break;
        }

        // Parse the data
        bool success;
        const QJsonObject jsonObject = jsonObjectFromByteArray(data, &success);

        if (! success) {
            emit protocolError(tr("%1: Fehler beim Parsen der übertragenen Daten").arg(ip()));
            continue;
        }

        const QJsonValue actionValue = jsonObject.value(KEY_ACTION);
        if (actionValue.type() != QJsonValue::String) {
            emit protocolError(tr("%1: Ungültiges Paket").arg(ip()));
            continue;
        }

        dataReceived(actionValue.toString(), jsonObject);

        // Look for other JSON objects in another loop
    }
}

QString AbstractNetworkEngine::colorToString(const QColor &color)
{
    return QString::number(color.rgb());
}

QColor AbstractNetworkEngine::colorFromValue(const QJsonValue &value)
{
    return QColor::fromRgb(value.toString().toUInt());
}
