/* Copyright (c) 2011-2020 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef DATABASE_H
#define DATABASE_H

// Local includes
#include "TournamentSettings.h"
#include "Muckturnier.h"

// Qt includes
#include <QObject>
#include <QSqlDatabase>
#include <QSqlQuery>
#include <QMap>
#include <QColor>
#include <QHash>

// Local classes
class SharedObjects;
class StringTools;

// Qt classes
class QLockFile;

class Database : public QObject
{
    Q_OBJECT

public:
    enum Error {
        NoError,
        NoErrorReadOnly,
        CouldNotOpen,
        DbVersion3,
        DbVersion4,
        DbVersion5,
        DbVersion6,
        DbVersion7
    };

    struct Assignment {
        QHash<int, int> assignedTable;
        QHash<int, int> assignedPairNumber;
        QHash<int, QHash<int, int>> assignedPair;
    };

    struct Score {
        int table;
        QString pairName;
        int score;
    };

    struct BoogerResult {
        int pairOrPlayer;
        int score;
    };

    struct RankData {
        int rank;
        QString name;
        int boogers;
        int goals;
        int opponentGoals;
        int playedBoogers;
    };

    struct PlayersData {
        int marker;
        int assignedTable;
        int assignedPairNumber;
    };

    explicit Database(QObject *parent, SharedObjects *sharedObjects);

    // Underlying database handling
    QString checkForLockFile(const QString &dbFile) const;
    bool create(const QString &dbFile);
    bool saveSettings();
    Database::Error openDatabase(const QString &dbFile);
    Database::Error checkDatabase(const QString &dbFile);
    bool updateDatabase(const QString &dbFileName, Error dbVersion);
    void closeDatabaseConnection();
    bool reset();

    // Settings
    void setNamesSeparator(const QString &text);
    void setNetworkTournamentStarted(bool state);
    bool networkTournamentStarted() const;
    bool localTournamentStarted() const;
    void setPairsOrPlayersCount(int count);

    // Functions returning const values/references or boolean operations on them

    bool isWritable() const;
    bool tournamentStarted() const;
    int dbVersion() const;
    int necessaryDivisor() const;
    bool playersDividable() const;
    bool playersDividable(int count) const;
    int pairsOrPlayersCount() const;
    int allRounds() const;
    bool isOpen() const;

    const QString &errorMessage() const;
    const QString &dbFile() const;
    const QString &dbPath() const;
    const QMap<int, bool> &roundsState() const;

    // Functions using SQL to write to the database

    bool addMarkers(const QVector<Muckturnier::Marker> &markers);
    bool registerPlayers(const QString &name, int markerId);
    bool registerPlayers(const QStringList &names);
    bool renamePlayers(const QString &oldName, const QString &newName);
    bool renamePlayers(const QStringList &oldNames, const QStringList &newNames);
    bool updatePreAssignment(const QString &name, int table, int pair, bool alsoRemoveMarker);
    bool assemblePair(const QString &player1, const QString &player2, const QString &separator,
                      int assignedMarker);
    bool deletePlayers(const QString &name);
    bool deleteMarkedPlayers(int markerId);
    bool deleteUnmarkedPlayers();
    bool deleteAllPlayers();
    bool deleteAllPreAssignments();

    bool updateMarker(const QString &name, int markerId);
    bool listUpdateMarker(int markerId);
    bool listRemoveMarker(int markerId);

    bool addMarker(const QString &name, const QColor &color, Muckturnier::MarkerSorting sorting);
    bool editMarker(int id, const QString &name, const QColor &color,
                    Muckturnier::MarkerSorting sorting);
    bool deleteMarker(int id, int sequence);
    bool moveMarker(int id, int sequence, int direction);

    bool saveScore(int round, int table, const QVector<Database::BoogerResult> &score);
    bool deleteScore(int round, int table);

    bool adoptAllData(const QHash<QString, PlayersData> &playersData,
                      const QVector<Muckturnier::Marker> &markers);
    bool adoptMarkersAndPreAssignment(const QHash<QString, PlayersData> &playersData,
                                      const QVector<Muckturnier::Marker> &markers);

    // Functions using SQL to read from the database
    QStringList players(bool *success);
    QVector<int> missingTables(int round, bool *success);
    QVector<Score> scores(int round, bool *success);
    QVector<BoogerResult> singleScore(int round, int table, bool *success);
    QVector<RankData> ranking(int upToRound, bool *success);
    QHash<QString, int> missingPlayers(int round, bool *success);
    QVector<Muckturnier::Marker> markers(bool *success);
    QHash<QString, PlayersData> playersData(bool *success);
    Assignment assignment(int round, bool *success);

signals:
    void sqlError(const QString &text) const;

private: // Functions
    // SQL helper functions
    bool startTransaction();
    bool commitTransaction();
    bool rollbackTransaction(QSqlQuery &query);
    bool queryPrepare(QSqlQuery &query, const QString &text);
    bool queryExec(QSqlQuery &query);
    bool queryExec(QSqlQuery &query, const QString &text);
    bool queryExecBatch(QSqlQuery &query);
    QString sqlErrorText() const;
    QString sqlErrorText(const QSqlQuery &query) const;
    void emitSqlError(const QString &text);
    QVariantList stringListToVariantList(const QStringList &stringList) const;

    // Helper functions for catching SQL errors

    template<typename SqlFunction>
    bool writeHelper(SqlFunction sqlFunction)
    {
        if (! startTransaction()) {
            return false;
        }

        QSqlQuery query(m_db);
        try {
            sqlFunction(query);
            commitTransaction();
        } catch (...) {
            rollbackTransaction(query);
            return false;
        }

        return true;
    }

    template<typename SqlFunction>
    void readHelper(bool *success, SqlFunction sqlFunction)
    {
        try {
            QSqlQuery query(m_db);
            sqlFunction(query);
            (*success) = true;
        } catch (...) {
            (*success) = false;
        }
    }

    // Connecting helpers
    bool connectDatabase(const QString &dbFile);
    QString lockFileName(const QString &dbFile) const;
    bool lockDatabase(const QString &dbFileName);
    void updateDbFileInfo();

    // Database revision update functions
    void updateDatabaseFromV3(QSqlQuery &query);
    void updateDatabaseFromV4(QSqlQuery &query);
    void updateDatabaseFromV5(QSqlQuery &query);
    void updateDatabaseFromV6(QSqlQuery &query);
    void updateDatabaseFromV7(QSqlQuery &query);

    bool checkIfTournamentStarted();
    bool updateRoundsState();
    bool ranksDiffer(const RankData &currentRank, const RankData &lastRank);
    void updateAssignment(QSqlQuery &query,
                          int round, int pairOrPlayer, int pairNumber, int playerNumber, int table);
    void addMarker(QSqlQuery &query, const QString &name, const QColor &color,
                   Muckturnier::MarkerSorting sorting, int id = -1);
    void updateMarker(QSqlQuery &query, const QString &name, int markerId);
    void registerPlayers(QSqlQuery &query, const QStringList &names);
    void listUpdateMarker(QSqlQuery &query, int markerId);
    void renamePlayers(QSqlQuery &query, const QString &oldName, const QString &newName);
    void deletePlayers(QSqlQuery &query, const QString &name);
    void deleteAllPreAssignments(QSqlQuery &query);
    void adoptMarkersAndPreAssignment(QSqlQuery &query,  const QStringList &names,
                                      const QHash<QString, PlayersData> &playersData,
                                      const QVector<Muckturnier::Marker> &markers);

private: // Variables
    TournamentSettings *m_tournamentSettings;
    StringTools *m_stringTools;
    QLockFile *m_lockFile;
    QSqlDatabase m_db;
    QString m_errorMessage;
    QString m_dbFileName;
    QString m_dbFilePath;
    bool m_isWritable;
    bool m_emitSqlErrors = false;

    QString m_namesSeparator;
    bool m_tournamentStarted = false;
    bool m_networkTournamentStarted = false;
    int m_pairsOrPlayersCount = 0;
    int m_scoresPerRound = 0;
    int m_tablesCount = 0;
    QMap<int, bool> m_roundsState;
    int m_allRounds = 0;

};

#endif // DATABASE_H
