/* Copyright (c) 2019 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes
#include "JsonHelper.h"

// Qt includes
#include <QString>
#include <QJsonObject>
#include <QJsonArray>
#include <QJsonDocument>
#include <QDebug>

namespace JsonHelper
{

QString jsonObjectToString(const QJsonObject &json)
{
    return QString::fromUtf8(QJsonDocument(json).toJson(QJsonDocument::Compact));
}

QString jsonArrayToString(const QJsonArray &json)
{
    return QString::fromUtf8(QJsonDocument(json).toJson(QJsonDocument::Compact));
}

QJsonObject jsonObjectFromString(const QString &data)
{
    return QJsonDocument::fromJson(data.toUtf8()).object();
}

QJsonArray jsonArrayFromString(const QString &data)
{
    return QJsonDocument::fromJson(data.toUtf8()).array();
}

const QString SINGLES_MANAGEMENT_ENABLED          = QStringLiteral("enabled");
const QString SINGLES_MANAGEMENT_MARKER_SINGLE    = QStringLiteral("single marker");
const QString SINGLES_MANAGEMENT_MARK_SINGLES     = QStringLiteral("mark singles");
const QString SINGLES_MANAGEMENT_CONDITION_STRING = QStringLiteral("condition string");
const QString SINGLES_MANAGEMENT_CONDITION        = QStringLiteral("condition");
const QString SINGLES_MANAGEMENT_MARKER_ASSIGNED  = QStringLiteral("assigned marker");

QJsonObject singlesManagementToJson(const Muckturnier::SinglesManagement &data)
{
    return QJsonObject {
        { SINGLES_MANAGEMENT_ENABLED,          data.enabled },
        { SINGLES_MANAGEMENT_MARKER_SINGLE,    data.singleMarker },
        { SINGLES_MANAGEMENT_MARK_SINGLES,     data.markSingles },
        { SINGLES_MANAGEMENT_CONDITION_STRING, data.conditionString },
        { SINGLES_MANAGEMENT_CONDITION,        data.condition },
        { SINGLES_MANAGEMENT_MARKER_ASSIGNED,  data.assignedMarker }
    };
}

QString singlesManagementToString(const Muckturnier::SinglesManagement &data)
{
    return QString::fromUtf8(
        QJsonDocument(singlesManagementToJson(data)).toJson(QJsonDocument::Compact));
}

Muckturnier::SinglesManagement singlesManagementFromString(const QString &data)
{
    return singlesManagementFromJson(jsonObjectFromString(data));
}

Muckturnier::SinglesManagement singlesManagementFromJson(const QJsonObject &jsonObject)
{
    Muckturnier::SinglesManagement data;
    QJsonValue value;

    value = jsonObject.value(SINGLES_MANAGEMENT_ENABLED);
    data.enabled = value.type() == QJsonValue::Bool
        ? value.toBool() : Muckturnier::DEFAULT_SINGLES_MANAGEMENT.enabled;

    value = jsonObject.value(SINGLES_MANAGEMENT_MARKER_SINGLE);
    data.singleMarker = value.type() == QJsonValue::Double
        ? value.toInt() : Muckturnier::DEFAULT_SINGLES_MANAGEMENT.singleMarker;

    value = jsonObject.value(SINGLES_MANAGEMENT_MARK_SINGLES);
    data.markSingles = value.type() == QJsonValue::Bool
        ? value.toBool() : Muckturnier::DEFAULT_SINGLES_MANAGEMENT.markSingles;

    value = jsonObject.value(SINGLES_MANAGEMENT_CONDITION_STRING);
    data.conditionString = value.type() == QJsonValue::String
        ? value.toString() : Muckturnier::DEFAULT_SINGLES_MANAGEMENT.conditionString;

    value = jsonObject.value(SINGLES_MANAGEMENT_CONDITION);
    data.condition = value.type() == QJsonValue::Double
        ? value.toInt() : Muckturnier::DEFAULT_SINGLES_MANAGEMENT.condition;

    value = jsonObject.value(SINGLES_MANAGEMENT_MARKER_ASSIGNED);
    data.assignedMarker = value.type() == QJsonValue::Double
        ? value.toInt() : Muckturnier::DEFAULT_SINGLES_MANAGEMENT.assignedMarker;

    return data;
}

QJsonObject playersPageColumnsToJson(const QHash<PlayersPageColumns::Column, bool> &visible)
{
    QJsonObject data;
    for (PlayersPageColumns::Column column : PlayersPageColumns::ALL_COLUMNS) {
        data.insert(PlayersPageColumns::COLUMN_ID.value(column), visible.value(column));
    }
    return data;
}

QString playersPageColumnsToString(const QHash<PlayersPageColumns::Column, bool> &visible)
{
    return jsonObjectToString(playersPageColumnsToJson(visible));
}

QHash<PlayersPageColumns::Column, bool> playersPageColumnsFromString(const QString &data)
{
    return playersPageColumnsFromJson(jsonObjectFromString(data));
}

QHash<PlayersPageColumns::Column, bool> playersPageColumnsFromJson(const QJsonObject &jsonObject)
{
    QHash<PlayersPageColumns::Column, bool> columnsData;
    for (PlayersPageColumns::Column column : PlayersPageColumns::ALL_COLUMNS) {
        QJsonValue jsonValue = jsonObject.value(PlayersPageColumns::COLUMN_ID.value(column));
        columnsData.insert(column, jsonValue.type() == QJsonValue::Bool
                                   ? jsonValue.toBool()
                                   : PlayersPageColumns::DEFAULT_VISIBILITY.value(column));
    }
    return columnsData;
}

}
