/* Copyright (c) 2018-2019 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes
#include "ResourceFinder.h"
#include "StringTools.h"

// Qt includes
#include <QDebug>
#include <QFileInfo>
#include <QStandardPaths>
#include <QCoreApplication>
#include <QUrl>
#include <QDir>

namespace
{
const QLatin1String SLASH("/");
}

ResourceFinder::ResourceFinder(QObject *parent, StringTools *stringTools)
    : QObject(parent)
{
    QVector<QString> possibleLocations;

    // Add the binary's directory
    possibleLocations.append(QCoreApplication::applicationDirPath() + SLASH);

    // Add system-specific possible resource locations
#ifdef Q_OS_WIN
    possibleLocations.append(QCoreApplication::applicationDirPath() + QLatin1String("/res/"));
#elif defined(Q_OS_MACOS)
    possibleLocations.append(QCoreApplication::applicationDirPath()
                             + QLatin1String("/../Resources/"));
#else
    // Get a list of possible locations for our external files
    const QLatin1String documentationPath("doc/muckturnier/");
    const QLatin1String iconsPath        ("icons/hicolor/scalable/apps/");
    const QLatin1String resourcesPath    ("muckturnier/");

    // Add a possibly prefixed directory tree
    const QUrl prefixed = QUrl(QCoreApplication::applicationDirPath() + SLASH).resolved(
                               QUrl(QStringLiteral("../share/")));
    possibleLocations.append(prefixed.toString() + documentationPath);
    possibleLocations.append(prefixed.toString() + iconsPath);
    possibleLocations.append(prefixed.toString() + resourcesPath);

    // Add system-wide locations
    for (const QString &path
             : QStandardPaths::standardLocations(QStandardPaths::GenericDataLocation)) {

        possibleLocations.append(path + SLASH + documentationPath);
        possibleLocations.append(path + SLASH + iconsPath);
        possibleLocations.append(path + SLASH + resourcesPath);
    }
#endif

    // Assemble the list
    for (const QString location : possibleLocations) {
        QDir dir(location);
        if (dir.exists()) {
            stringTools->appendIfNotExists(m_possibleLocations, location);
        }
    }
}

QString ResourceFinder::find(const QString &file)
{
    // Check if we already have it
    if (m_path.contains(file)) {
        return m_path.value(file);
    }

    // ... then we check all other possible data locations
    for (const QString &path : m_possibleLocations) {
        const QString filePath = path + file;
        if (QFileInfo::exists(filePath)) {
            m_path[file] = filePath;
            return filePath;
        }
    }

    // This will also return QString() if nothing is found
    return m_path[file];
}
