/* Copyright (c) 2018-2019 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes
#include "SearchEngine.h"
#include "StringTools.h"

// Qt includes
#include <QRegularExpression>
#include <QDebug>

namespace
{

const QVector<QChar> CHARS_TO_REPLACE {
    QChar(0x00e4), // ä
    QChar(0x00f6), // ö
    QChar(0x00fc), // ü
    QChar(0x00df), // ß
    QChar(0x00E6), // æ
    QChar(0x0153), // œ
    QChar(0x00F8)  // ø
};

const QVector<QString> CHAR_REPLACEMENTS {
    QStringLiteral("ae"),
    QStringLiteral("oe"),
    QStringLiteral("ue"),
    QStringLiteral("ss"),
    QStringLiteral("ae"),
    QStringLiteral("oe"),
    QStringLiteral("oe")
};

const QLatin1Char SPACE(' ');
const QLatin1String REGEXP_DOTPLUS(".+");

}

SearchEngine::SearchEngine(QObject *parent, StringTools *stringTools)
    : QObject(parent),
      m_stringTools(stringTools),
      m_variantsCache {
          { SearchType::Default,                 &m_defaultVariants },
          { SearchType::DrawnPairsScore,         &m_drawnPairsScoreVariants },
          { SearchType::DrawnPairsPreAssignment, &m_drawnPairsPreAssignmentVariants }
      }
{
}

QVector<QString> SearchEngine::searchVariants(const QString &text)
{
    QVector<QString> variants;

    // First add all variants of the lower-case search term as-is ...
    const QString lower = m_stringTools->toLower(text);
    addAllVariants(lower, variants);

    // ... then add all versions of the search term sorted by words
    QStringList words = lower.split(SPACE);
    m_stringTools->sortList(words);
    const QString sorted = words.join(SPACE);
    addAllVariants(sorted, variants);

    return variants;
}

void SearchEngine::addAllVariants(const QString &searchTerm, QVector<QString> &variantsList)
{
    m_stringTools->appendIfNotExists(variantsList, searchTerm);
    const QString diacriticsStripped = stripDiacritis(searchTerm);
    m_stringTools->appendIfNotExists(variantsList, diacriticsStripped);
    m_stringTools->appendIfNotExists(variantsList, replaceUmlauts(diacriticsStripped));
    const QString lowerReplaced = replaceUmlauts(searchTerm);
    m_stringTools->appendIfNotExists(variantsList, lowerReplaced);
    m_stringTools->appendIfNotExists(variantsList, stripDiacritis(lowerReplaced));
}

QString SearchEngine::replaceUmlauts(QString text) const
{
    for (int i = 0; i < CHARS_TO_REPLACE.size(); i++) {
        text.replace(CHARS_TO_REPLACE.at(i), CHAR_REPLACEMENTS.at(i));
    }
    return text;
}

QString SearchEngine::stripDiacritis(const QString &text) const
{
    // Here, we strip out all diacritical marks and accents by first separating the combinig
    // characters and then removing them, cf. http://en.wikipedia.org/wiki/Combining_character
    QString replaced = text.normalized(QString::NormalizationForm_D);
    for (int i = 0x0300; i <= 0x036F; i++) {
        replaced.remove(QChar(i));
    }
    return replaced;
}

bool SearchEngine::checkMatch(const QString &text, SearchType searchType) const
{
    return checkMatch(text, *m_variantsCache.value(searchType));
}

bool SearchEngine::checkMatch(const QString &text,
                              const QHash<QString, QVector<QString>> &variantsList) const
{
    QRegularExpression re;
    for (const QString &searchTermRegularExpression : m_searchTermRegularExpressions) {
        re.setPattern(searchTermRegularExpression);
        for (const QString &targetVariant : variantsList.value(text)) {
            if (re.match(targetVariant).hasMatch()) {
                return true;
            }
        }
    }
    return false;
}

void SearchEngine::clearSearchCache(SearchType searchType)
{
    (*m_variantsCache.value(searchType)).clear();
}

void SearchEngine::updateSearchCache(const QString &name, SearchType searchType)
{
    (*m_variantsCache.value(searchType)).insert(name, searchVariants(name));
}

void SearchEngine::setSearchTerm(const QString &text)
{
    m_searchTermRegularExpressions.clear();
    for (const QString &variant : searchVariants(text.simplified())) {
        QStringList words = variant.split(SPACE);
        for (QString &word : words) {
            word = QRegularExpression::escape(word);
        }
        m_searchTermRegularExpressions.append(words.join(REGEXP_DOTPLUS));
     }
}

QHash<QString, QStringList> SearchEngine::checkListForPossibleDuplicates(const QStringList &list)
{
    QHash<QString, QStringList> possibleDuplicates;
    const int count = list.count();

    if (count == 1) {
        return possibleDuplicates;
    }

    emit progress(-1);
    QHash<QString, QVector<QString>> variantsCache;
    for (QString name : list) {
        variantsCache.insert(name, searchVariants(name.simplified()));
    }
    emit progress(0);

    // We actually do have to compare each entry to each other, because if we "only" do an upper
    // triangular matrix comparison, we lose potentional duplicates of entries that are on a higher
    // position in the list, because they aren't compared to the lower indices anymore.

    const int allComparisons = count * count - count;
    int currentComparison = 0;
    int lastPercent = -1;
    int percent;

    for (int searchIndex = 0; searchIndex < count; searchIndex++) {
        setSearchTerm(list.at(searchIndex));

        for (int compareIndex = 0; compareIndex < count; compareIndex++) {
            if (searchIndex == compareIndex) {
                continue;
            }

            if (checkMatch(list.at(compareIndex).simplified(), variantsCache)) {
                possibleDuplicates[list.at(searchIndex)].append(list.at(compareIndex));
            }

            percent = ++currentComparison * 100 / allComparisons;
            if (percent > lastPercent) {
                lastPercent = percent;
                emit progress(percent);
            }
        }
    }

    return possibleDuplicates;
}
