/* Copyright (c) 2019 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes
#include "SelectionPopup.h"

// Qt includes
#include <QDebug>
#include <QVBoxLayout>
#include <QCursor>
#include <QScrollArea>
#include <QScrollBar>
#include <QApplication>
#include <QTimer>
#include <QScreen>

SelectionPopup::SelectionPopup(QWidget *parent) : QDialog(parent)
{
    setWindowFlags(Qt::FramelessWindowHint | Qt::Popup);
    setStyleSheet(QStringLiteral(
        "QDialog { border: 1px solid gray; } "
        "QScrollArea { border: none; } "
        "QPushButton { border: none; border-radius: 4px; padding:0.5em; } "
        "QPushButton:hover { border: 1px solid palette(highlight); } "
    ));

    QVBoxLayout *layout = new QVBoxLayout(this);
    layout->setSpacing(0);
    layout->setMargin(1); // We need this for the QDialog's gray border

    m_scrollArea = new QScrollArea(this);
    m_scrollArea->setHorizontalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
    layout->addWidget(m_scrollArea);
}

void SelectionPopup::setScrollWidget(QWidget *widget)
{
    widget->setStyleSheet(QStringLiteral("background-color: white;"));
    m_scrollArea->setWidget(widget);
}

void SelectionPopup::show(QPoint position)
{
    if (position == QPoint(-1, -1)) {
        position = QCursor::pos();
    }

    const QRect availableGeometry = QGuiApplication::primaryScreen()->availableGeometry();
    const int availableHeight = availableGeometry.height() - position.y();
    const int availableWidth = availableGeometry.width() - position.x();

    const QSize widgetSize = m_scrollArea->widget()->size();
    const int widgetHeight = widgetSize.height() + 2; // Two pixels for the border
    const int widgetWidth = widgetSize.width(); // Who appararenly aren't necessary here

    if (availableWidth < widgetWidth) {
        if (availableGeometry.width() >= widgetWidth) {
            position.setX(position.x() - (widgetWidth - availableWidth));
        } else {
            position.setX(0);
        }
    }

    int targetHeight = -1;

    if (widgetHeight <= availableHeight) {
        targetHeight = widgetHeight;
    } else {
        if (availableGeometry.height() >= widgetHeight) {
            position.setY(position.y() - (widgetHeight - availableHeight));
            targetHeight = widgetHeight;
        } else {
            position.setY(0);
            targetHeight = availableGeometry.height();
        }
    }

    resize(m_scrollArea->widget()->width(), targetHeight);
    move(position);

    QDialog::show();
    QTimer::singleShot(0, this, &SelectionPopup::checkForScrollBar);
}

void SelectionPopup::checkForScrollBar()
{
    if (m_scrollArea->verticalScrollBar()->isVisible()) {
        resize(m_scrollArea->widget()->width() + m_scrollArea->verticalScrollBar()->width(),
               height());
    } else {
        resize(m_scrollArea->widget()->width(), height());
    }
}
