/* Copyright (c) 2019 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes
#include "Settings.h"
#include "defaultValues.h"
#include "JsonHelper.h"

// Qt includes
#include <QDebug>
#include <QAction>
#include <QJsonObject>
#include <QJsonArray>
#include <QStandardPaths>
#include <QDir>

namespace
{

// Options check state group
const QLatin1String SAVE_ON_CLOSE("save_on_close/");
const QString SAVE_ON_CLOSE_MAINWINDOW_GEOMETRY
    = SAVE_ON_CLOSE + QLatin1String("mainwindow_geometry");
const QString SAVE_ON_CLOSE_LAST_DB_PATH
    = SAVE_ON_CLOSE + QLatin1String("last_db_path");

// Session data
const QString SESSION_DATA = QStringLiteral("session_data/");
const QString SESSION_DATA_MAINWINDOW_GEOMETRY
    = SESSION_DATA + QLatin1String("mainwindow_geometry");
const QString SESSION_DATA_LAST_DB_PATH
    = SESSION_DATA + QLatin1String("last_db_path");

// Strings
const QLatin1String STRINGS("strings/");
const QString STRINGS_BOOGER_SYMBOL = STRINGS + QLatin1String("booger_symbol");
const QString STRINGS_NAMES_SEPARATOR = STRINGS + QLatin1String("names_separator");

// Tournament template

constexpr int VERSION_TEMPLATE_TOURNAMENT = 2;
const QString TOURNAMENT_TEMPLATE = QStringLiteral("tournament_template");
const QString TOURNAMENT_TEMPLATE_VERSION
    = TOURNAMENT_TEMPLATE + QLatin1String("/version");
const QString TOURNAMENT_TEMPLATE_DB_FILE_NAME_TEMPLATE
    = TOURNAMENT_TEMPLATE + QLatin1String("/db_file_name_template");
const QString TOURNAMENT_TEMPLATE_TOURNAMENT_MODE
    = TOURNAMENT_TEMPLATE + QLatin1String("/tournament_mode");
const QString TOURNAMENT_TEMPLATE_BOOGER_SCORE
    = TOURNAMENT_TEMPLATE + QLatin1String("/booger_score");
const QString TOURNAMENT_TEMPLATE_BOOGERS_PER_ROUND
    = TOURNAMENT_TEMPLATE + QLatin1String("/boogers_per_round");
const QString TOURNAMENT_TEMPLATE_AUTO_SELECT_PAIRS
    = TOURNAMENT_TEMPLATE + QLatin1String("/auto_select_pairs");
const QString TOURNAMENT_TEMPLATE_NO_TABLE_NUMBERS
    = TOURNAMENT_TEMPLATE + QLatin1String("/no_table_numbers");

// Markers template

constexpr int VERSION_TEMPLATE_MARKERS = 2;

const QString MARKERS_TEMPLATE = QStringLiteral("markers_template");

const QString MARKERS_TEMPLATE_DEFAULT_MARKER = MARKERS_TEMPLATE + QLatin1String("/default_marker");
const QString MARKERS_TEMPLATE_ABSENT_MARKER  = MARKERS_TEMPLATE + QLatin1String("/absent_marker");

const QString MARKERS_TEMPLATE_VERSION = MARKERS_TEMPLATE + QLatin1String("/version");
const QString MARKERS_TEMPLATE_MARKERS = MARKERS_TEMPLATE + QLatin1String("/markers");
const QLatin1String MARKERS_TEMPLATE_MARKERS_JSON_NAME   ("n");
const QLatin1String MARKERS_TEMPLATE_MARKERS_JSON_COLOR  ("c");
const QLatin1String MARKERS_TEMPLATE_MARKERS_JSON_SORTING("s");

// Singles Management

const QString SINGLES_MANAGEMENT_TEMPLATE = QStringLiteral("singles_management_template");

const QString SINGLES_MANAGEMENT_TEMPLATE_VERSION
    = SINGLES_MANAGEMENT_TEMPLATE + QLatin1String("/version");
const QString SINGLES_MANAGEMENT_TEMPLATE_DATA
    = SINGLES_MANAGEMENT_TEMPLATE + QLatin1String("/data");

// Network settings

const QString NETWORK = QStringLiteral("network");

const QString NETWORK_PREFER_IPV6       = NETWORK + QLatin1String("/prefer_ipv6");

const QString NETWORK_SERVER_PORT       = NETWORK + QLatin1String("/server_port");
const QString NETWORK_CONNECT_TIMEOUT   = NETWORK + QLatin1String("/connect_timeout");
const QString NETWORK_REQUEST_TIMEOUT   = NETWORK + QLatin1String("/request_timeout");

const QString NETWORK_DISCOVER_PORT     = NETWORK + QLatin1String("/discover_port");
const QString NETWORK_DISCOVER_TIMEOUT  = NETWORK + QLatin1String("/discover_timeout");
const QString NETWORK_DISCOVER_ATTEMPTS = NETWORK + QLatin1String("/discover_attempts");

// Players page
const QLatin1String PLAYERS_PAGE("players_page/");
const QString PLAYERS_PAGE_COLUMNS = PLAYERS_PAGE + QLatin1String("columns");

}

Settings::Settings(QObject *parent) : QSettings(QStringLiteral("muckturnier.org"),
                                                QStringLiteral("muckturnier"), parent)
{
}

// Main window geometry

bool Settings::saveMainWindowGeometryEnabled() const
{
    const QVariant data = value(SAVE_ON_CLOSE_MAINWINDOW_GEOMETRY);
    return data.isValid() ? data.toBool() : true;
}

void Settings::enableSaveMainWindowGeometry(bool state)
{
    setValue(SAVE_ON_CLOSE_MAINWINDOW_GEOMETRY, state);
    if (! state) {
        remove(SESSION_DATA_MAINWINDOW_GEOMETRY);
    }
}

void Settings::saveMainWindowGeometry(const QByteArray &data)
{
    setValue(SESSION_DATA_MAINWINDOW_GEOMETRY, data);
}

QByteArray Settings::mainWindowGeometry() const
{
    return value(SESSION_DATA_MAINWINDOW_GEOMETRY, QByteArray()).toByteArray();
}

void Settings::saveTournamentTemplate(const TournamentTemplate &tournamentTemplate)
{
    setValue(TOURNAMENT_TEMPLATE_VERSION, VERSION_TEMPLATE_TOURNAMENT);
    setValue(TOURNAMENT_TEMPLATE_TOURNAMENT_MODE,
             Muckturnier::TOURNAMENT_MODE_MAP.value(tournamentTemplate.tournamentMode));
    setValue(TOURNAMENT_TEMPLATE_BOOGER_SCORE, tournamentTemplate.boogerScore);
    setValue(TOURNAMENT_TEMPLATE_BOOGERS_PER_ROUND, tournamentTemplate.boogersPerRound);
    setValue(TOURNAMENT_TEMPLATE_AUTO_SELECT_PAIRS, tournamentTemplate.autoSelectPairs);
    setValue(TOURNAMENT_TEMPLATE_NO_TABLE_NUMBERS, tournamentTemplate.noTableNumbers);
}

// Last database path

bool Settings::saveLastDbPathEnabled() const
{
    const QVariant data = value(SAVE_ON_CLOSE_LAST_DB_PATH);
    return data.isValid() ? data.toBool() : true;
}

void Settings::enableSaveLastDbPath(bool state)
{
    setValue(SAVE_ON_CLOSE_LAST_DB_PATH, state);
    if (! state) {
        remove(SESSION_DATA_LAST_DB_PATH);
    }
}

void Settings::saveLastDbPath(const QString &dbPath)
{
    setValue(SESSION_DATA_LAST_DB_PATH, dbPath);
}

QString Settings::lastDbPath() const
{
    const QString documentsLocation
        = QStandardPaths::writableLocation(QStandardPaths::DocumentsLocation);
    const QString path = value(SESSION_DATA_LAST_DB_PATH, documentsLocation).toString();
    return QDir(path).exists() ? path : documentsLocation;
}

// Tournament template

Settings::TournamentTemplate Settings::defaultTournamentTemplate() const
{
    return TournamentTemplate { Muckturnier::TournamentMode::FixedPairs,
                                DEFAULT_BOOGER_SCORE,
                                DEFAULT_BOOGERS_PER_ROUND,
                                true,
                                false };
}

Settings::TournamentTemplate Settings::tournamentTemplate() const
{
    const int version = value(TOURNAMENT_TEMPLATE_VERSION, -1).toInt();
    Muckturnier::TournamentMode tournamentMode;

    // Get the tournament mode
    if (version == VERSION_TEMPLATE_TOURNAMENT) {
        // Check for a correct value of TOURNAMENT_TEMPLATE_TOURNAMENT_MODE
        const QString tournamentModeString
            = value(TOURNAMENT_TEMPLATE_TOURNAMENT_MODE,
                    Muckturnier::TOURNAMENT_MODE_FIXED_PAIRS).toString();
        if (! Muckturnier::TOURNAMENT_MODE_MAP.values().contains(tournamentModeString)) {
            return defaultTournamentTemplate();
        } else {
            tournamentMode = Muckturnier::TOURNAMENT_MODE_MAP.key(tournamentModeString);
        }
    } else if (version == 1) {
        // Read the old bool "pair_mode" value
        tournamentMode = value(TOURNAMENT_TEMPLATE + QLatin1String("/pair_mode"), true).toBool()
                             ? Muckturnier::TournamentMode::FixedPairs
                             : Muckturnier::TournamentMode::SinglePlayers;
    } else {
        return defaultTournamentTemplate();
    }

    // Check for correct booger score and boogers per round
    const int boogerScore = value(TOURNAMENT_TEMPLATE_BOOGER_SCORE, DEFAULT_BOOGER_SCORE).toInt();
    const int boogersPerRound = value(TOURNAMENT_TEMPLATE_BOOGERS_PER_ROUND,
                                      DEFAULT_BOOGERS_PER_ROUND).toInt();
    if (boogerScore < 1 || boogersPerRound < 1) {
        return defaultTournamentTemplate();
    }

    // Assemble the template and return it
    bool autoSelectPairs;
    if (tournamentMode == Muckturnier::TournamentMode::FixedPairs) {
        autoSelectPairs = value(TOURNAMENT_TEMPLATE_AUTO_SELECT_PAIRS, true).toBool();
    } else {
        autoSelectPairs = false;
    }

    return TournamentTemplate { tournamentMode,
                                boogerScore,
                                boogersPerRound,
                                autoSelectPairs,
                                value(TOURNAMENT_TEMPLATE_NO_TABLE_NUMBERS, false).toBool() };
}

void Settings::saveDbFileNameTemplate(const QString &text)
{
    setValue(TOURNAMENT_TEMPLATE_DB_FILE_NAME_TEMPLATE, text);
}

// Database file name template

QString Settings::dbFileNameTemplate() const
{
    return value(TOURNAMENT_TEMPLATE_DB_FILE_NAME_TEMPLATE,
                 QStringLiteral(DEFAULT_DB_FILE_NAME_TEMPLATE)).toString();
}

void Settings::saveMarkersTemplate(const QVector<Muckturnier::Marker> &markers)
{
    QJsonArray json;
    for (const Muckturnier::Marker &marker : markers) {
        json.append(QJsonObject {
            { MARKERS_TEMPLATE_MARKERS_JSON_NAME,    marker.name },
            { MARKERS_TEMPLATE_MARKERS_JSON_COLOR,   QString::number(marker.color.rgb()) },
            { MARKERS_TEMPLATE_MARKERS_JSON_SORTING, static_cast<int>(marker.sorting) } });
    }

    setValue(MARKERS_TEMPLATE_VERSION, VERSION_TEMPLATE_MARKERS);
    setValue(MARKERS_TEMPLATE_MARKERS, JsonHelper::jsonArrayToString(json));
}

// Markers template

bool Settings::parseMarkerData(int &id, const QJsonValue &data,
                               QVector<Muckturnier::Marker> &markers) const
{
    Muckturnier::Marker marker;
    const QJsonObject markerData = data.toObject();
    QJsonValue value;

    value = markerData.value(MARKERS_TEMPLATE_MARKERS_JSON_NAME);
    if (value.type() != QJsonValue::String) {
        return false;
    } else {
        marker.name = value.toString();
    }

    value = markerData.value(MARKERS_TEMPLATE_MARKERS_JSON_COLOR);
    if (value.type() != QJsonValue::String) {
        return false;
    } else {
        marker.color = QColor::fromRgb(value.toString().toUInt());
    }

    value = markerData.value(MARKERS_TEMPLATE_MARKERS_JSON_SORTING);
    if (value.type() != QJsonValue::Double) {
        return false;
    } else {
        marker.sorting = static_cast<Muckturnier::MarkerSorting>(value.toInt());
    }

    if (marker.name.isEmpty()) {
        marker.id = 0;
    } else {
        marker.id = ++id;
    }
    markers.append(marker);

    return true;
}

void Settings::appendMarker(const QString &name, const QColor &color,
                            Muckturnier::MarkerSorting sorting, int id,
                            QVector<Muckturnier::Marker> &markers) const
{
    Muckturnier::Marker marker;
    marker.name = name;
    marker.color = color;
    marker.sorting = sorting;
    marker.id = id;
    markers.append(marker);
}

QVector<Muckturnier::Marker> Settings::defaultMarkersTemplate() const
{
    QVector<Muckturnier::Marker> markers;
    markers.reserve(4);
    appendMarker(tr("Rot"),  QColor(255, 0,   0,   255),
                 Muckturnier::MarkerSorting::NewBlock, 1, markers);
    appendMarker(tr("Blau"), QColor(0,   0,   255, 255),
                 Muckturnier::MarkerSorting::NewBlock, 2, markers);
    appendMarker(QString(),  QColor(0,   0,   0,   255),
                 Muckturnier::MarkerSorting::NewBlock, 0, markers);
    appendMarker(tr("Grün"), QColor(0,   128, 0,   255),
                 Muckturnier::MarkerSorting::Merged,   3, markers);
    return markers;
}

QVector<Muckturnier::Marker> Settings::markersTemplate() const
{
    const int version = value(MARKERS_TEMPLATE_VERSION, 0).toInt();

    QJsonArray data;
    if (version == 2) {
        data = JsonHelper::jsonArrayFromString(value(MARKERS_TEMPLATE_MARKERS,
                                                     QString()).toString());
    } else if (version == 1) {
        data = JsonHelper::jsonObjectFromString(QString::fromUtf8(
            value(QStringLiteral("markers_template/data"), QByteArray()).toByteArray())).value(
                QStringLiteral("d")).toArray();
    } else {
        return defaultMarkersTemplate();
    }

    if (data.isEmpty()) {
        return defaultMarkersTemplate();
    }

    QVector<Muckturnier::Marker> markers;

    bool templateParsed = true;
    int id = 0;

    for (const QJsonValue &value : data) {
        if (! parseMarkerData(id, value, markers)) {
            templateParsed = false;
            break;
        }
    }

    // Be sure to have an "(unmarked)" marker with ID 0 and no name
    if (templateParsed) {
        templateParsed = false;
        for (const Muckturnier::Marker &marker : markers) {
            if (marker.id == 0 && marker.name.isEmpty()) {
                templateParsed = true;
                break;
            }
        }
    }

    if (! templateParsed) {
        return defaultMarkersTemplate();
    }

    return markers;
}

// Singles Management

void Settings::saveSinglesManagementTemplate(const Muckturnier::SinglesManagement &data)
{
    setValue(SINGLES_MANAGEMENT_TEMPLATE_VERSION, Muckturnier::SINGLES_MANAGEMENT_VERSION);
    setValue(SINGLES_MANAGEMENT_TEMPLATE_DATA,    JsonHelper::singlesManagementToString(data));
}

Muckturnier::SinglesManagement Settings::defaultSinglesManagementTemplate()
{
    Muckturnier::SinglesManagement data = Muckturnier::DEFAULT_SINGLES_MANAGEMENT;
    data.conditionString = namesSeparator().simplified();
    return data;
}

Muckturnier::SinglesManagement Settings::singlesManagementTemplate()
{
    if (value(SINGLES_MANAGEMENT_TEMPLATE_VERSION, 0).toInt()
        != Muckturnier::SINGLES_MANAGEMENT_VERSION) {

        return defaultSinglesManagementTemplate();
    }

    return JsonHelper::singlesManagementFromString(value(SINGLES_MANAGEMENT_TEMPLATE_DATA,
                                                         QString()).toString());
}

// Strings

void Settings::saveBoogerSymbol(const QString &text)
{
    setValue(STRINGS_BOOGER_SYMBOL, text);
}

QString Settings::boogerSymbol() const
{
    QString boogerSymbol = value(STRINGS_BOOGER_SYMBOL,
                                 QStringLiteral(DEFAULT_BOOGER_SYMBOL)).toString();
    if (boogerSymbol.left(1).simplified().isEmpty()) {
        return QStringLiteral(DEFAULT_BOOGER_SYMBOL);
    } else {
        return boogerSymbol;
    }
}

void Settings::saveNamesSeparator(const QString &text)
{
    setValue(STRINGS_NAMES_SEPARATOR, text);
}

QString Settings::namesSeparator() const
{
    QString namesSeparator = value(STRINGS_NAMES_SEPARATOR,
                                   QStringLiteral(DEFAULT_NAMES_SEPARATOR)).toString();
    return namesSeparator.remove(QLatin1Char('\t'));
}

void Settings::saveDefaultMarker(int id)
{
    setValue(MARKERS_TEMPLATE_DEFAULT_MARKER, id);
}

int Settings::defaultMarker() const
{
    return value(MARKERS_TEMPLATE_DEFAULT_MARKER, 0).toInt();
}

void Settings::saveAbsentMarker(int id)
{
    setValue(MARKERS_TEMPLATE_ABSENT_MARKER, id);
}

int Settings::absentMarker() const
{
    return value(MARKERS_TEMPLATE_ABSENT_MARKER, -1).toInt();
}

// Network

void Settings::saveServerPort(int port)
{
    setValue(NETWORK_SERVER_PORT, port);
}

int Settings::serverPort() const
{
    const int port = value(NETWORK_SERVER_PORT, DEFAULT_SERVER_PORT).toInt();
    return (port >= 1024 && port <= 65535) ? port : DEFAULT_SERVER_PORT;
}

void Settings::saveConnectTimeout(int timeout)
{
    setValue(NETWORK_CONNECT_TIMEOUT, timeout);
}

int Settings::connectTimeout() const
{
    return value(NETWORK_CONNECT_TIMEOUT, DEFAULT_CONNECT_TIMEOUT).toInt();
}

void Settings::saveRequestTimeout(int timeout)
{
    setValue(NETWORK_REQUEST_TIMEOUT, timeout);
}

int Settings::requestTimeout() const
{
    return value(NETWORK_REQUEST_TIMEOUT, DEFAULT_REQUEST_TIMEOUT).toInt();
}

void Settings::saveDiscoverPort(int port)
{
    setValue(NETWORK_DISCOVER_PORT, port);
}

int Settings::discoverPort() const
{
    const int port = value(NETWORK_DISCOVER_PORT, DEFAULT_DISCOVER_PORT).toInt();
    return (port >= 1024 && port <= 65535) ? port : DEFAULT_DISCOVER_PORT;
}

void Settings::saveDiscoverTimeout(int timeout)
{
    setValue(NETWORK_DISCOVER_TIMEOUT, timeout);
}

int Settings::discoverTimeout() const
{
    return value(NETWORK_DISCOVER_TIMEOUT, DEFAULT_DISCOVER_TIMEOUT).toInt();
}

void Settings::saveDiscoverAttempts(int attempts)
{
    setValue(NETWORK_DISCOVER_ATTEMPTS, attempts);
}

int Settings::discoverAttempts() const
{
    return value(NETWORK_DISCOVER_ATTEMPTS, DEFAULT_DISCOVER_ATTEMPTS).toInt();
}

void Settings::savePreferIpv6(bool state)
{
    setValue(NETWORK_PREFER_IPV6, state);
}

bool Settings::preferIpv6() const
{
    return value(NETWORK_PREFER_IPV6, false).toBool();
}

void Settings::savePlayersPageColumns(const QHash<PlayersPageColumns::Column, bool> &visible)
{
    setValue(PLAYERS_PAGE_COLUMNS, JsonHelper::playersPageColumnsToString(visible));
}

QHash<PlayersPageColumns::Column, bool> Settings::playersPageColumns() const
{
    return JsonHelper::playersPageColumnsFromString(
        value(PLAYERS_PAGE_COLUMNS, QByteArray()).toString());
}
