/* Copyright (c) 2019 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes
#include "TournamentSettings.h"
#include "Settings.h"
#include "JsonHelper.h"
#include "version.h"

// Qt includes
#include <QDebug>
#include <QJsonObject>

namespace
{

// JSON Keys for all settings

const QString MUCKTURNIER_VERSION        = QStringLiteral("muckturnier version");

const QString TOURNAMENT_MODE            = QStringLiteral("tournament mode");

const QString BOOGER_SCORE               = QStringLiteral("booger score");
const QString BOOGERS_PER_ROUND          = QStringLiteral("boogers per round");

const QString NO_TABLE_NUMBES            = QStringLiteral("no table numbers");
const QString AUTOSELECT_PAIRS           = QStringLiteral("autoselect pairs");
const QString INCLUDE_OPPONENT_GOALS     = QStringLiteral("include opponent goals");

const QString DEFAULT_MARKER             = QStringLiteral("default marker");
const QString ABSENT_MARKER              = QStringLiteral("absent marker");
const QString SINGLES_MANAGEMENT_VERSION = QStringLiteral("singles management version");
const QString SINGLES_MANAGEMENT         = QStringLiteral("singles management");

const QString PLAYERS_PAGE_COLUMNS       = QStringLiteral("players page columns");

}

TournamentSettings::TournamentSettings(QObject *parent, Settings *settings)
    : QObject(parent), m_settings(settings)
{
}

void TournamentSettings::initialize()
{
    // Here, we set the default values from the settings
    // which are not set by the "new tournament" dialog.
    m_muckturnierVersion   = QLatin1String(MT_VERSION);
    m_includeOpponentGoals = false;
    m_defaultMarker        = m_settings->defaultMarker();
    m_absentMarker         = m_settings->absentMarker();
    m_singlesManagement    = m_settings->singlesManagementTemplate();
    m_playersPageColumns   = m_settings->playersPageColumns();
}

bool TournamentSettings::load(const QString &data)
{
    const QJsonObject json = JsonHelper::jsonObjectFromString(data);

    if (   ! json.contains(MUCKTURNIER_VERSION)
        || ! json.contains(TOURNAMENT_MODE)
        || ! json.contains(BOOGER_SCORE)
        || ! json.contains(BOOGERS_PER_ROUND)
        || ! json.contains(NO_TABLE_NUMBES)
        || ! json.contains(AUTOSELECT_PAIRS)
        || ! json.contains(INCLUDE_OPPONENT_GOALS)
        || ! json.contains(DEFAULT_MARKER)
        || ! json.contains(ABSENT_MARKER)
        || ! json.contains(SINGLES_MANAGEMENT)
        || ! json.contains(SINGLES_MANAGEMENT_VERSION)
        || ! json.contains(PLAYERS_PAGE_COLUMNS)) {

        return false;
    }

    m_muckturnierVersion   = json.value(MUCKTURNIER_VERSION).toString();
    m_tournamentMode       = Muckturnier::TOURNAMENT_MODE_MAP.key(
                                 json.value(TOURNAMENT_MODE).toString());
    m_boogerScore          = json.value(BOOGER_SCORE).toInt();
    m_boogersPerRound      = json.value(BOOGERS_PER_ROUND).toInt();

    m_noTableNumbers       = json.value(NO_TABLE_NUMBES).toBool();
    m_autoSelectPairs      = json.value(AUTOSELECT_PAIRS).toBool();
    m_includeOpponentGoals = json.value(INCLUDE_OPPONENT_GOALS).toBool();

    m_defaultMarker        = json.value(DEFAULT_MARKER).toInt();
    m_absentMarker         = json.value(ABSENT_MARKER).toInt();

    // As long as the singles management representation is not changed,
    // we can ignore the version and simply parse what we have
    m_singlesManagement    = JsonHelper::singlesManagementFromJson(
                                 json.value(SINGLES_MANAGEMENT).toObject());

    m_playersPageColumns   = JsonHelper::playersPageColumnsFromJson(
                                 json.value(PLAYERS_PAGE_COLUMNS).toObject());

    return true;
}

QString TournamentSettings::toString() const
{
    return JsonHelper::jsonObjectToString(QJsonObject {
        { MUCKTURNIER_VERSION,        m_muckturnierVersion },
        { TOURNAMENT_MODE,            Muckturnier::TOURNAMENT_MODE_MAP.value(m_tournamentMode) },
        { BOOGER_SCORE,               m_boogerScore },
        { BOOGERS_PER_ROUND,          m_boogersPerRound },
        { NO_TABLE_NUMBES,            m_noTableNumbers },
        { AUTOSELECT_PAIRS,           m_autoSelectPairs },
        { INCLUDE_OPPONENT_GOALS,     m_includeOpponentGoals },
        { DEFAULT_MARKER,             m_defaultMarker },
        { ABSENT_MARKER,              m_absentMarker },
        { SINGLES_MANAGEMENT_VERSION, Muckturnier::SINGLES_MANAGEMENT_VERSION },
        { SINGLES_MANAGEMENT,         JsonHelper::singlesManagementToJson(m_singlesManagement) },
        { PLAYERS_PAGE_COLUMNS,       JsonHelper::playersPageColumnsToJson(m_playersPageColumns) }
    });
}

void TournamentSettings::setMuckturnierVersion(const QString &text)
{
    m_muckturnierVersion = text;
}

const QString &TournamentSettings::muckturnierVersion() const
{
    return m_muckturnierVersion;
}

void TournamentSettings::setTournamentMode(Muckturnier::TournamentMode tournamentMode)
{
    m_tournamentMode = tournamentMode;
}

Muckturnier::TournamentMode TournamentSettings::tournamentMode() const
{
    return m_tournamentMode;
}

bool TournamentSettings::isPairMode() const
{
    return m_tournamentMode == Muckturnier::TournamentMode::FixedPairs;
}

bool TournamentSettings::isSinglePlayerMode() const
{
    return m_tournamentMode == Muckturnier::TournamentMode::SinglePlayers;
}

void TournamentSettings::setBoogerScore(int score)
{
    m_boogerScore = score;
}

int TournamentSettings::boogerScore() const
{
    return m_boogerScore;
}

void TournamentSettings::setBoogersPerRound(int count)
{
    m_boogersPerRound = count;
}

int TournamentSettings::boogersPerRound() const
{
    return m_boogersPerRound;
}

void TournamentSettings::setNoTableNumbers(bool state)
{
    m_noTableNumbers = state;
}

bool TournamentSettings::noTableNumbers() const
{
    return m_noTableNumbers;
}

void TournamentSettings::setAutoSelectPairs(bool state)
{
    m_autoSelectPairs = state;
}

bool TournamentSettings::autoSelectPairs() const
{
    return m_autoSelectPairs;
}

void TournamentSettings::setIncludeOpponentGoals(bool state)
{
    m_includeOpponentGoals = state;
}

bool TournamentSettings::includeOpponentGoals() const
{
    return m_includeOpponentGoals;
}

void TournamentSettings::setDefaultMarker(int id)
{
    m_defaultMarker = id;
}

int TournamentSettings::defaultMarker() const
{
    return m_defaultMarker;
}

void TournamentSettings::setAbsentMarker(int id)
{
    m_absentMarker = id;
}

int TournamentSettings::absentMarker() const
{
    return m_absentMarker;
}

void TournamentSettings::setSinglesManagement(const Muckturnier::SinglesManagement &data)
{
    m_singlesManagement = data;
}

const Muckturnier::SinglesManagement &TournamentSettings::singlesManagement() const
{
    return m_singlesManagement;
}

void TournamentSettings::setPlayersPageColumns(const QHash<PlayersPageColumns::Column, bool> &data)
{
    m_playersPageColumns = data;
}

const QHash<PlayersPageColumns::Column, bool> &TournamentSettings::playersPageColumns() const
{
    return m_playersPageColumns;
}
